#!/usr/bin/perl

# This script walks the configuration tree and populates all
# internal nodes with a new configuration which contains the
# common part of all config files in the leaf nodes.

use strict;
use Cwd;
use File::Spec;
use Algorithm::Diff;   # for debian/ubuntu in libalgorithm-diff-perl

# Change this if new configuration files are added
my @configfiles = 
	qw(
		defconfig
		feature.config
		hardco.cfg
		mbus_atomic.def
		mbus_igd_artype.def
		mbus_igd.def
		project.list
		slock.txt
	);

# Update this if needed
my $rejectpattern = "generic_sdk|tools";

# Remeber where we start, so we can print filenames
# relative to here
my $startdir = cwd();
print "Root: $startdir\n";

# Recurse into the directory structure
recursive_descend();

sub recursive_descend {
	my $level = shift;
	my @content = `ls --group-directories-first -1`;
	my $internal_node = 0;
	my @subdirs = ();
	
	# First process our subdirectories, they might not have
	# config files yet if they are not leaf nodes
	foreach my $thing (@content) {
		chomp $thing;
		next if $thing =~ /$rejectpattern/;	
		if (-d $thing) {			
			# if there are subdirectories, 
			# descend first..
			my $mydir = cwd();
			chdir $thing;
			recursive_descend();
			chdir $mydir;
			# .. and construct a config here
			# based on what's in them
			$internal_node = 1;
			push @subdirs, $thing;
		}
	}
	
	if ($internal_node) {
		# construct the configs by looking
		# at the subdirs
		construct_configs(@subdirs);
	}
};

sub construct_configs() {	
	my @subdirs = @_;

	# make filenames understandable		
	my $rel_path = File::Spec->abs2rel(cwd(), $startdir);	
	if ($rel_path eq "") {
		$rel_path = '.';
	}

	# process each know config file in time
	foreach my $cfgfile (@configfiles) {
		my @leafconfs = ();
	
		if (!@subdirs) {
			die "No subdirectories";
		} else {
			print "Project $rel_path has subdirs:@subdirs\n";
		}
		
		# read in the config file into an array, and store
		# a reference to this array in @leafconfs
		foreach my $subdir (@subdirs) {
			my $success = open FILE, "$subdir/$cfgfile";
			if ($success) {				
				print "Reading in $rel_path/$subdir/$cfgfile\n";
				my @lines = <FILE>;
				push @leafconfs, \@lines;
				close FILE;
			}
		}

		# process the config files, and get back the contents
		# of the common part
		if (@leafconfs) {
			my @outfile = construct_lcs(@leafconfs);
							
			print "Writing out $rel_path/$cfgfile\n";
			open FILE, ">", "$cfgfile";
			print FILE @outfile;
			close FILE;			
		}
	}
}

# Construct the Longest Common Subsequence (LCS) of a set of files.
sub construct_lcs() {	
	# early exit for the trivial case of 1 file
	if (@_ == 1) {
		my $lcsref = shift @_;
		return @$lcsref;
	}

	my @leafconfs = @_;
	my $bestlcs_size = -1;
	my @bestlcs = ();

	print "Optimizing...";

	# for progress display, disable line buffering
	select STDOUT;
	$| = 1;

	# Constructing the LCS of an arbitrary number of files
	# is an NP-hard problem.
	# Take a reasonable number of random permuations to
	# get a good approximation.
	my $iterations = 1 + ((@leafconfs * @leafconfs) / 3);

	foreach (1..$iterations) {			
		my @tmplist = @leafconfs;
		# leafconfs is an array of array references
		# get the first one
		my $lcsref = shift @tmplist;
		my @lcs = @$lcsref;
	
		# Take the first config as the reference for the LCS,
		# then compute the LCS of this with each following config
		# file. This does not guarantee finding the global optimum.
		# Example: We might be greedy if we find 2 files with a lot of 
		# matching things, and drop stuff to get the longest match.
  		# The matching parts of those 2 files may not occur anywhere 
		# else, but the stuff we dropped might.
		foreach my $ref (@tmplist) {
			@lcs = Algorithm::Diff::LCS(\@lcs, $ref);
			print '.';
		}
		print 'o';

		my $lcs_size = @lcs;

		# is this LCS longer than our previous try?
		if ($lcs_size > $bestlcs_size) {
			if ($bestlcs_size > 0) {
				print "\nOptimized from $bestlcs_size to $lcs_size lines\n";
			}
			$bestlcs_size = $lcs_size;
			@bestlcs = @lcs;
		}

		# randomly permute the config list (and try again)
		fisher_yates_shuffle(\@leafconfs);
	}	

	# enable line buffer
	$| = 0;

	print "\n";

	return @bestlcs;
}

# fisher_yates_shuffle(\@array): 
# generate a random permutation of @array in place
sub fisher_yates_shuffle {
	my $deck = shift; 
	my $i = @$deck;	
	while (--$i) {
		my $j = int rand ($i+1);
		@$deck[$i,$j] = @$deck[$j,$i];
	}
}
