#!/usr/bin/perl
###############################################################################
#
# Tool for comparing the versions of components in the project list
# with the latest available version in perforce.
# Email <david.cluytens_ext@softathome.com>
#
###############################################################################
use strict;
use Getopt::Long;
use Cwd;
use File::Basename;

our $cplst          = {};
our $work_dir       = getcwd;
our $menuconfig     = 0;
my $cp_file         = "$work_dir/component.lst";
my $help            = 0;
my $proj_path;
my @files           = ();
my $staging_dir     = $work_dir . "/output/staging";
$ENV{'STAGINGDIR'}  = $staging_dir;

GetOptions('help'           => sub { $help = 1 },
           'path=s'         => \$proj_path,
           'menuconfig'     => sub { $menuconfig = 1 }
            ) or printSyntax();

if (defined($proj_path)) {
    @files = `find $proj_path -name project.list`;
}

if ($help) {
    printSyntax();
} elsif ($proj_path ne "") {
    @files = `find $proj_path -name project.list`;
	syncKernels($cp_file, @files);
	foreach my $proj_file (@files) {
	    doProjFile($cp_file, $proj_file);    
	}
} else {
    printSyntax();
}
exit 0;

sub doProjFile {
    my $cp_file     = shift;
    my $proj_file   = shift;

    print "Reading " . trim($proj_file) . "\n";
    parseComponentLst($cp_file);
    parseConfig($proj_file);
}

sub printSyntax {
    print "update_proj.pl   --path  <project_path>\
\t--path        proj path\n\n";
    exit 1;
}

sub trim {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

sub parseComponentLst {
    my  $file   = shift;
    our $cplst;
    my  $cp;
    my  $location;
    my  $server;

    my @lines = readFile($file);
    foreach my $line (@lines) {
        if ($line =~ /^(.*)=(.*):(.*)$/) {
            $cp         = $1;
            $server     = $2;
            $location   = $3;
            $cplst->{$cp}->{"server"}   = $server;
            $cplst->{$cp}->{"location"} = $location;
        }
    }
}

sub getComponentServer {
    my $cp = shift;
    return $cplst->{$cp}->{"server"};
}

sub getComponentLocation {
    my $cp = shift;
    return $cplst->{$cp}->{"location"};
}

sub parseConfig {
    my $fh;
    my $file = shift;
    my @lines = readFile($file);
    foreach my $line (@lines) {
        parseLine($file, $line);
    }
}

sub p4GetComponent {
    my $server      = shift;
    my $location    = shift;
    my $p4location;

    if ($server eq "w1") {
        $p4location = "//depot${location}/...";
		print "Syncing $p4location\n";
        system("p4 -p sahwp4s01.be.softathome.com:1666 sync -q -f ${p4location}");
    } else {
        $p4location = "/${location}/...";
		print "Syncing $p4location\n";
        system("p4 -p sahwp4s01.be.softathome.com:1999 sync -q -f ${p4location}");
    }
}

sub syncKernels {
	my $cp_file = shift;
	my $projs = shift;

    parseComponentLst($cp_file);
	my %kernels = (); # hash

	foreach my $proj (@files) {
		my @lines = readFile($proj);
		
		foreach my $line (@lines) {
			if ($line =~ /^sah_kernel=(.*)$/) {
				$kernels{ $1 } = '1';
			}
		}
	}

	for my $kernel ( keys %kernels ) {
		my $server = getComponentServer("sah_kernel");
		my $location = getComponentLocation("sah_kernel") . "/" . $kernel;
		p4GetComponent($server, $location);
	}
}

sub parseLine {
    our $configs;
    our $menuconfig;
    our $work_dir;
    my $file    = shift;
    my $line    = shift;
    my $cp      = "sah_kernel";
    my $version;
    my $location;
    my $server;
    my $res;

    system("cd $work_dir");
    if ($line =~ /^sah_kernel=(.*)$/) {
        $version    = $1; 
        $server     = getComponentServer($cp);
        $location   = getComponentLocation($cp) . "/" . $version;
        if ($server ne "" && $location ne "" ) {
            my $config_dir = dirname($file);
            my $ws_dir = `p4 client -o | awk '/^Root:.*\$/ {print(\$2)}'`;
            $ws_dir = trim($ws_dir);
            my $kernel_dir = $ws_dir . $location;
            my $config_kernel = $config_dir . "/kernel.config";
            my $kernel_kernel = $kernel_dir . "/.config";
           
            print "cd $work_dir; $work_dir/build.sh $config_dir nosync nocleanstaging nobuild\n";
            system("cd $work_dir; $work_dir/build.sh $config_dir nosync nocleanstaging nobuild");

            #print "Sourcing Makefile.global\n";
            #system("source Makefile.global");

            print "cp --no-preserve=mode $config_kernel $kernel_kernel\n";
            system("cp --no-preserve=mode $config_kernel $kernel_kernel");


            if ($menuconfig) {
                print "Doing make -f Component.mk menuconfig\n";
                system("make -C $kernel_dir -f Component.mk menuconfig");
				my $ret = $? >> 8;
				if ($ret != 0) {
					print "Error when handling $kernel_dir\n";
					exit(1);
				}
            } else {
                print "Doing make -f Component.mk oldconfig\n";
                system("make -C $kernel_dir -f Component.mk oldconfig");
				my $ret = $? >> 8;
				if ($ret != 0) {
					print "Error when handling $kernel_dir\n";
					exit(1);
				}
            }

	    chdir($work_dir);

	    # ignore it if only the date changed
            system("tail -n +5 $kernel_kernel > /tmp/diff1");
	    system("tail -n +5 $config_kernel > /tmp/diff2");
	    my $diffret = `diff /tmp/diff1 /tmp/diff2`;

	    if( $diffret != '' ) {
            	print "Opening file for edit in perforce $config_kernel\n";
            	system("p4 edit $config_kernel");

            	print "cp $kernel_kernel $config_kernel\n";
            	system("cp $kernel_kernel $config_kernel");
	    }

        }
    } 
}

sub readFile {
    my $name = shift;
    open FILE, $name or die "File $name " . $!; 
    my @lines = <FILE>;
    close FILE;
    return @lines;
}

