#!/usr/bin/perl
###############################################################################
#
# Tool for comparing the versions of components in the project list
# with the latest available version in perforce.
# Email <david.cluytens_ext@softathome.com>
#
###############################################################################
use strict;
use Getopt::Long;
use Term::ReadKey;

our $cplst      = {};
our $chlist     = 0;
my $cp_file     = "component.lst";
my $help        = 0;
my $proj_file;
my $proj_path;
our @files      = ();
our $upgrade    = 0;

GetOptions('help'           => sub { $help = 1 },
           'file=s'         => \$proj_file,
           'path=s'         => \$proj_path,
           'upgrade'        => sub { $upgrade = 1 },
           'upgrade_all'    => sub { $upgrade = 2 }
            ) or printSyntax();

if (defined($proj_path)) {
    @files = `find $proj_path -name project.list`;
}

if ($help) {
    printSyntax();
} elsif (defined($proj_file)) {
    doProjFile($cp_file, $proj_file);    
} elsif ($proj_path ne "") {
    @files = `find $proj_path -name project.list`;
    foreach $proj_file (@files) {
        doProjFile($cp_file, $proj_file);    
    }
} else {
    printSyntax();
}
exit 0;

sub doProjFile {
    my $cp_file     = shift;
    my $proj_file   = shift;

    print "Comparing " . trim($proj_file) . "\n";
    ReadMode('cbreak');
    parseComponentLst($cp_file);
    parseConfig($proj_file);
    ReadMode('normal');
}

sub printSyntax {
    print "update_proj.pl   --file  <project.list>\
\t--file        project.list\
\t--path        proj path\
\t--upgrade     upgrade only the project.list specified with --file\
\t--upgrade_all upgrade all the project.list(s) located under --path\n\
Example\
\t
\ttools/update_proj.pl --file ~/ws_w1/config/MAIN/sah/generic/askey-sg230/sah_multi_sip_normal/project.list --upgrade_all --path ~/ws_w1/config/MAIN/sah/generic/\
\tCompare askey-sg230/sah_multi_sip_normal/project.list with p4 and apply all the changes to the project.list(s) under  ~/ws_w1/config/MAIN/sah/generic/\n\n";
    exit 1;
}

sub trim {
    my $string = shift;
    $string =~ s/^\s+//;
    $string =~ s/\s+$//;
    return $string;
}

sub parseComponentLst {
    my  $file   = shift;
    our $cplst;
    my  $cp;
    my  $location;
    my  $server;

    my @lines = readFile($file);
    foreach my $line (@lines) {
        if ($line =~ /^(.*)=(.*):(.*)$/) {
            $cp         = $1;
            $server     = $2;
            $location   = $3;
            $cplst->{$cp}->{"server"}   = $server;
            $cplst->{$cp}->{"location"} = $location;
        }
    }
}

sub getComponentServer {
    my $cp = shift;
    return $cplst->{$cp}->{"server"};
}

sub getComponentLocation {
    my $cp = shift;
    return $cplst->{$cp}->{"location"};
}

sub parseConfig {
    my $fh;
    my $file = shift;
    my @lines = readFile($file);
    foreach my $line (@lines) {
        parseLine($file, $line);
    }
}

sub p4GetLatestComponent {
    my $server      = shift;
    my $location    = shift;
    my $fh;
    my $p4location;
    my $lver        = "";
    if ($server eq "w1") {
        $p4location = "//depot${location}/REL/";
        open($fh, "p4 -p sahwp4s01.be.softathome.com:1666 dirs ${p4location}* |");
    } else {
        $p4location = "/${location}/REL/";
        open($fh, "p4 -p sahwp4s01.be.softathome.com:1999 dirs ${p4location}* |");
    }
    while (!eof($fh)) {
        my $line = <$fh>;
        my $len  = length $p4location;
        my $cver = substr $line, $len;
        $cver = trim($cver);
        if ($cver =~ /^[\d]+-[\d]+-[\d]+_V[\.\d]+$/) {
            if ($cver gt $lver) {
                $lver = $cver;
            }
        } else {
            #print "Parsing invalid current version ${p4location}${cver}\n";
        }
    }
    close $fh;
    return $lver;
}

sub parseLine {
    our $configs;
    my $file = shift;
    my $line = shift;
    my $cp;
    my $path;
    my $version;
    my $location;
    my $server;
    my $lver;
    my $cur;
    my $latest;
    my $p   = 1;
    my $ch_info = "";
    if ($line =~ /^(.*)=(.*)\/REL\/(.*)$/) {
        $cp         = $1;
        $path       = $2;
        $version    = $3;
        $server     = getComponentServer($cp);
        $location   = getComponentLocation($cp) . "/" . $path;
        if ($server ne "" && $location ne "" ) {
            $lver       = p4GetLatestComponent($server, $location);
            $cur        = "$cp=$path/REL/$version";
            $latest     = "$cp=$path/REL/$lver";
            if ($cur ne $latest) {
                $p = 0;
                $ch_info = getChangeLists($server, $location . "/REL/" . $lver); 
                printLine($file, $cur, $latest, $ch_info);
            }
        }
    } elsif ($line =~ /^(.*)=(.*)$/) {
        $cp         = $1;
        $version    = $2; 
        $server     = getComponentServer($cp);
        $location   = getComponentLocation($cp);
        if ($server ne "" && $location ne "" ) {
            $lver       = p4GetLatestComponent($server, $location);
            $cur        = "$cp=$version";
            $latest     = "$cp=REL/$lver";
            if ($cur ne $latest) {
                $p = 0;
                $ch_info = getChangeLists($server, $location . "/REL/" . $lver); 
                printLine($file, $cur, $latest, $ch_info);
            }
        }
    } 
    if ($p) {
        #print $line
    }
}

sub printLine {
    our $upgrade;
    our $chlist;
    my $proj_file   = shift;
    my $cur         = shift;
    my $latest      = shift;
    my $ch_info     = shift;
    my $char;
    print "C $cur\nL $latest\n";
    if ($upgrade) {
        while(1) {
            print "Update component (l/y/n)? ";
            $char = ReadKey(0);
            print "$char\n";
            if ($char eq "y" || $char eq "Y") {
                $chlist = createChangeList($ch_info);
                if ($upgrade == 1) {
                    updateComponent($chlist, $proj_file, $cur, $latest, $ch_info);
                } elsif ($upgrade == 2) {
                    updateComponents($chlist, $cur, $latest, $ch_info);
                }
                return;
            } elsif ($char eq "n" || $char eq "N") {
                return;
            } elsif ($char eq "l" || $char eq "L") {
                print "Changelist:\n$ch_info\n";
            }
        }
    }
}

sub readFile {
    my $name = shift;
    open FILE, $name or die $!; 
    my @lines = <FILE>;
    close FILE;
    return @lines;
}

sub updateComponents {
    my $chlist  = shift;
    my $from    = shift;
    my $to      = shift;
    my $ch_info = shift;
    our @files;
    foreach my $proj_file (@files) {
        updateComponent($chlist, $proj_file, $from, $to, $ch_info);
    }
}

sub updateComponent {
    my $chlist  = shift;
    my $name    = shift;
    my $from    = shift;
    my $to      = shift;
    my $ch_info = shift;

    chomp $name;
    print "Scanning $name\n";
    my $touched = 0;

#read in file
    my @lines = readFile($name);

# check if we (need to) replace something
    my @newfile = (); 
    foreach my $line (@lines) {
        my $oldline = $line;
        $line =~ s#$from#$to#g;
        if ($line ne $oldline) {
            print "$name : $from -> $to\n";
            $touched = 1;
        }   
        push(@newfile, $line);
    }   

# if we edit, p4 edit and rewrite
    if ($touched) {
#print ">>>>Editing $name\n";
        system "p4 edit -c $chlist $name";
        open FILE, "+>", $name or die $!; 
        print FILE @newfile;
        close FILE;
    } 
}

sub getChangeLists {
    my $server      = shift;
    my $location    = shift;
    my $fh;
    my $p4location;
    my $ch_info;
    if ($server eq "w1") {
        $p4location = "//depot${location}/...";
        open($fh, "p4 -p sahwp4s01.be.softathome.com:1666 changes -l ${p4location} |");
    } else {
        $p4location = "/${location}/...";
        open($fh, "p4 -p sahwp4s01.be.softathome.com:1999 changes -l ${p4location} |");
    }
    while (!eof($fh)) {
        my $line = <$fh>;
        $line = trim($line);
        if ($line =~ /^Change \d+ on \d+\/\d+\/\d+ by \w+@\w+$/) {
            next;
        } elsif ($line =~ /\w+.*/) {
            $ch_info = $ch_info . "\t" . $line . "\n";
        }
    }
    close $fh;
    return $ch_info;
}

sub changeListIsPending {
    my $ch      = shift;
    my @spec    = `p4 change -o $ch`;

    foreach my $line (@spec) {
        if ($line =~ /^Status:\s+pending.*$/) {
            return 1;
        }
    }
    return 0;
}

sub createChangeList {
    our $chlist;
    my $description = shift;
    my @spec        = ();
    my $ch_spec     = "";
    my $pending     = changeListIsPending($chlist); 

    if ($pending) { 
        @spec = `p4 change -o $chlist`;
    } else {
        @spec = `p4 change -o`;
    }

    foreach my $line (@spec) {
        if ($pending) { 
            $line =~ s#^Description:$#Description:\n$description#g;
        } else {
            $line =~ s#^\s+<enter description here>$#$description#g;
        }
        $ch_spec = $ch_spec . $line;
    }
    $chlist = `echo "$ch_spec" | p4 change -i | cut -d ' ' -f 2`;
    chomp $chlist;
    return $chlist; 
}
