
/*
 * ============================================================================
 *
 * Copyright : (c) 2009 SAGEM SA - Site Velizy - URD2
 *
 * The information and source code contained herein is the exclusive
 * property of SAGEM and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the company.
 *
 * ============================================================================
 */

/**
 * \file bsp_gpio.c Bsp GPIO management.
 *
 * This file contain all Bsp Gpio functions.
 */

/*
 * Includes
 * ~~~~~~~~
 */

#include "bsp_pri.h"

/*
 * Functions
 * ~~~~~~~~~
 */

/** Write word in register. */
static inline void bsp_gpio_write_regw(unsigned short val, unsigned long addr)
{
	/* Lock */
	spin_lock(&bsp_base.Board.Cur.gpio_lock);

	/* Set register */
	*(volatile u16 *)(addr) = val;

	/* Unlock */
	spin_unlock(&bsp_base.Board.Cur.gpio_lock);
}

/** Read Word in register. */
static inline unsigned short bsp_gpio_read_regw(unsigned long addr)
{
	unsigned short _res;

	/* Lock */
	spin_lock(&bsp_base.Board.Cur.gpio_lock);

	/* Read register */
	_res = *(volatile u16 *)(addr);

	/* Unlock */
	spin_unlock(&bsp_base.Board.Cur.gpio_lock);

	/* Out */
	return(_res);
}

/** Initialize gpio. */
void bsp_gpio_initialize(void)
{
	int _indiceName, _indiceGpio;

	/* Initialize current reference. */
	bsp_base.Board.Cur.pInfo = &(bsp_base.Board.All[bsp_base.Board.Cur.boardInfo]);

	/* Initialize current gpio name access. */
	{
		/* All Gpio name. */
		for (_indiceName = 0; _indiceName < BSP_Gpio_name_max; _indiceName++) {
			/* Set to NULL */
			bsp_base.Board.Cur.pGpioName[_indiceName] = (pBsp_board_gpio_s)0;
			
			/* Check reserved Gpio */
			if (_indiceName == BSP_Gpio_name_reserved) {
#ifdef BSP_DEBUG
				/* If bsp */
				if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
					printk(	"%s: NameAcces: indiceName(%u) -> Gpio reserved\n",
							__FUNCTION__,
							_indiceName);
				}
#endif
				/* Continue. */
				continue;
			}

			/* All Gpio. */
			for (_indiceGpio = 0; _indiceGpio < BSP_GPIO_MAX; _indiceGpio++) {
				if (bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].name == _indiceName) {
					/* Set */
					bsp_base.Board.Cur.pGpioName[_indiceName] = &(bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio]);
#ifdef BSP_DEBUG
					/* If bsp */
					if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
						printk(	"%s: NameAcces: indiceName(%u) indiceGpio(%u) -> pGpio(%p)\n",
								__FUNCTION__,
								_indiceName,
								_indiceGpio,
								bsp_base.Board.Cur.pGpioName[_indiceName]);
					}
#endif
				}
			}
		}
	}

	/* Initialize Gpio. */
	{
		u16 _dir1Flgs, _dir2Flgs, _intSet1Flgs, _intSet2Flgs, _polSet1Flgs, _polSet2Flgs;
		u8 _nbInt1 = 0, _nbInt2 = 0;

		/* Load current flags. */
		_dir1Flgs =		bsp_gpio_read_regw(BSP_GPIO_DIR1);
		_dir2Flgs =		bsp_gpio_read_regw(BSP_GPIO_DIR2);

		/* Load current int Set. */
		_intSet1Flgs =		bsp_gpio_read_regw(BSP_GPIO_INT_SET1);
		_intSet2Flgs =		bsp_gpio_read_regw(BSP_GPIO_INT_SET2);

		/* Load current pol Set. */
		_polSet1Flgs =		bsp_gpio_read_regw(BSP_GPIO_POL_SET1);
		_polSet2Flgs =		bsp_gpio_read_regw(BSP_GPIO_POL_SET2);

#ifdef BSP_DEBUG
		/* If bsp */
		if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
			printk(	"%s: InitDir: curDir1(%x) curDir2(%u)\n",
				__FUNCTION__,
				_dir1,
				_dir2);
		}
#endif
	
		/* All Gpio. */
		for ( _indiceGpio = 0 ; _indiceGpio < BSP_GPIO_MAX ; _indiceGpio++) {
			/* If not reserved */
			if (bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].name != BSP_Gpio_name_reserved) {
				/* Direction Part */
				{
					/* Output */
					if (bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].flags & BSP_Gpio_flags_dir_out) {
#ifdef BSP_DEBUG
						/* If bsp */
						if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
							printk(	"%s: InitDir: indiceGpio(%u) indiceName(%u) -> Output\n",
								__FUNCTION__,
								_indiceGpio,
								bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].name);
						}
#endif

						/* Set direction */
						if (_indiceGpio < 16) {
							_dir1Flgs |= 1 << _indiceGpio;
						}
						else {
							_dir2Flgs |= 1 << (_indiceGpio - 16);
						}
					}
	
					/* Input */
					if (bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].flags & BSP_Gpio_flags_dir_in) {
#ifdef BSP_DEBUG
						/* If bsp */
						if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
							printk(	"%s: InitDir: indiceGpio(%u) indiceName(%u) -> Input\n",
								__FUNCTION__,
								_indiceGpio,
								bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].name);
						}
#endif

						if (_indiceGpio < 16) {
							_dir1Flgs &= ~(1 << _indiceGpio);
						}
						else {
							_dir2Flgs &= ~(1 << (_indiceGpio - 16));
						}
					}
				}

				/* Polarity Part only dir in */
				if (bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].flags & BSP_Gpio_flags_dir_in) {
					/* Actif low */
					if (bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].flags & BSP_Gpio_flags_active_low) {
#ifdef BSP_DEBUG
						/* If bsp */
						if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
							printk(	"%s: InitPol: indiceGpio(%u) indiceName(%u) -> Low\n",
								__FUNCTION__,
								_indiceGpio,
								bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].name);
						}
#endif

						/* Set direction */
						if (_indiceGpio < 16) {
							/* Set */
							_polSet1Flgs |= 1 << _indiceGpio;
						}
						else {
							/* Set */
							_polSet2Flgs |= 1 << (_indiceGpio - 16);
						}
					}
				}

				/* Interruption part */
				{
					/* Interrup */
					if (bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].flags & BSP_Gpio_flags_intr) {
#ifdef BSP_DEBUG
						/* If bsp */
						if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
							printk(	"%s: InitInt: indiceGpio(%u) indiceName(%u) -> Int Set\n",
								__FUNCTION__,
								_indiceGpio,
								bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].name);
						}
#endif

						/* Set direction */
						if (_indiceGpio < 16) {
							/* Set */
							_intSet1Flgs |= 1 << _indiceGpio;

							/* Active */
							_nbInt1++;
						}
						else {
							/* Set */
							_intSet2Flgs |= 1 << (_indiceGpio - 16);

							/* Active */
							_nbInt2++;
						}
					}
				}
			}
		}
	
		/* Save new flags. */
		bsp_gpio_write_regw(	_dir1Flgs,
							BSP_GPIO_DIR1);
		bsp_gpio_write_regw(	_dir2Flgs,
							BSP_GPIO_DIR2);

		/* Save new pol set. */
		bsp_gpio_write_regw(	_polSet1Flgs,
							BSP_GPIO_POL_SET1);
		bsp_gpio_write_regw(	_polSet2Flgs,
							BSP_GPIO_POL_SET2);

		/* Save new int set. */
		bsp_gpio_write_regw(	_intSet1Flgs,
							BSP_GPIO_INT_SET1);
		bsp_gpio_write_regw(	_intSet2Flgs,
							BSP_GPIO_INT_SET2);

		
		/* Gpio 1 irq */
		if (_nbInt1) {
			/* Request Irq */
			if (request_irq(	GPIO_INT,
							bsp_gpio_irq,
							IRQF_DISABLED,
							"gpio1",
							(void *)&(bsp_base.Board.GpioIntr.low)) == 0) {
			}
		}

		/* Gpio 2 irq */
		if (_nbInt2) {
			/* Request Irq */
			if (request_irq(	GPIO2_INT,
							bsp_gpio_irq,
							IRQF_DISABLED,
							"gpio2",
							(void *)&(bsp_base.Board.GpioIntr.high)) == 0) {
			}
		}

#ifdef BSP_DEBUG
		/* If bsp */
		if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
			printk(	"%s: InitDir: newDir1(%x) newDir2(%u)\n",
				__FUNCTION__,
				_dir1,
				_dir2);
		}
#endif
	}

	/* Ok */
	return;
}

/** Get GPIO num */
static unsigned int bsp_gpio_get_num(bsp_gpio_name_e _name)
{
	/* Gpio reserved */
	if (_name == BSP_Gpio_name_reserved) {
#ifdef BSP_DEBUG
		/* If bsp */
		if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
			printk(	"%s: WARNING -> Reserved GPIO\n",
					__FUNCTION__);
		}
#endif

		return(-1);
	}

	/* If Gpio name Access null */
	if (bsp_base.Board.Cur.pGpioName[_name] == (pBsp_board_gpio_s)0) {
#ifdef BSP_DEBUG
		/* If bsp */
		if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
			printk(	"%s: WARNING -> Unreachable GPIO\n",
					__FUNCTION__);
		}
#endif

		return(-1);
	}

	/* Out */
	return(bsp_base.Board.Cur.pGpioName[_name] - &(bsp_base.Board.Cur.pInfo->Gpio[0]));
}

/** Set GPIO state */
void  bsp_gpio_set_state(	bsp_gpio_name_e	_name,
						bsp_gpio_state_e	_state)
{
	unsigned int	_indiceGpio	= bsp_gpio_get_num(_name);

	/* Check num */
	if (_indiceGpio == -1) {
		/* Out */
		return;
	}

	{
		/* Set */
		switch (_state) {
			case BSP_Gpio_state_set:
			{
				if (_indiceGpio < 16) {
#ifdef BSP_DEBUG
					/* If bsp */
					if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
						printk(	"%s: Set -> nameGpio(%u)indiceGpio(%u) -> < 16\n",
								__FUNCTION__,
								_name,
								_indiceGpio);
					}
#endif
	
					/* Set Gpio */
					bsp_gpio_write_regw((u16)(1 << _indiceGpio), BSP_GPIO_FLAG_SET1);
				}
				else {
#ifdef BSP_DEBUG
					/* If bsp */
					if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
						printk(	"%s: Set -> nameGpio(%u) indiceGpio(%u) -> > 16\n",
								__FUNCTION__,
								_name,
								_indiceGpio);
					}
#endif

					/* Set Gpio */
					bsp_gpio_write_regw((u16)(1 << (_indiceGpio - 16)), BSP_GPIO_FLAG_SET2);
				}
			}
			break;

			case BSP_Gpio_state_unset:
			{
				if (_indiceGpio < 16) {
#ifdef BSP_DEBUG
					/* If bsp */
					if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
						printk(	"%s: Unset -> nameGpio(%u) indiceGpio(%u) -> < 16\n",
								__FUNCTION__,
								_name,
								_indiceGpio);
					}
#endif

					/* Set Gpio */
					bsp_gpio_write_regw((u16)(1 << _indiceGpio), BSP_GPIO_FLAG_CLEAR1);
				}
				else {
#ifdef BSP_DEBUG
					/* If bsp */
					if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
						printk(	"%s: Unset -> nameGpio(%u) indiceGpio(%u) -> > 16\n",
								__FUNCTION__,
								_name,
								_indiceGpio);
					}
#endif

					/* Set Gpio */
					bsp_gpio_write_regw((u16)(1<<(_indiceGpio - 16)), BSP_GPIO_FLAG_CLEAR2);
				}
			}
			break;

			default:
			{
				printk(	"%s: UNKNOWN -> nameGpio(%u) gpioState(%u)",
						__FUNCTION__,
						_name,
						_state);
			}
			break;
		}
	}

	return;
}
EXPORT_SYMBOL(bsp_gpio_set_state);

/** Get GPIO state */
bsp_gpio_state_e bsp_gpio_get_state(bsp_gpio_name_e _name)
{
	unsigned char	_gpioState	= BSP_Gpio_state_failure;
	unsigned int	_indiceGpio	= bsp_gpio_get_num(_name);

	/* Check num */
	if (_indiceGpio != -1) {
		/* Low part */
		if (_indiceGpio < 16) {
#ifdef BSP_DEBUG
			/* If bsp */
			if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
				printk(	"%s: nameGpio(%u) indiceGpio(%u) -> <16\n",
						__FUNCTION__,
						_name,
						_indiceGpio);
			}
#endif

			/* Read */
			if (bsp_gpio_read_regw(BSP_GPIO_FLAG_SET1) & (1 << _indiceGpio)) {
				/* Set */
				_gpioState = BSP_Gpio_state_set;
			}
			else {
				/* Unset */
				_gpioState = BSP_Gpio_state_unset;
			}
		}
		/* High part */
		else {
#ifdef BSP_DEBUG
			/* If bsp */
			if (bsp_base.debug & BSP_FLAGS_DEBUG_GPIO) {
				printk(	"%s: nameGpio(%u) indiceGpio(%u) -> >16\n",
						__FUNCTION__,
						_name,
						_indiceGpio);
			}
#endif

			/* Read */
			if (bsp_gpio_read_regw(BSP_GPIO_FLAG_SET2) & (1 << (_indiceGpio - 16))) {
				/* Set */
				_gpioState = BSP_Gpio_state_set;
			}
			else {
				/* Unset */
				_gpioState = BSP_Gpio_state_unset;
			}
		}
	}

	/* Error */
	return(_gpioState);
} 
EXPORT_SYMBOL(bsp_gpio_get_state);

/** sysctl bsp_gpio_maps_proc.
 *
 *   \par Description
 *	<b>This function allow to set connexion tampon.</b>\n
 *
 *	@param _pTable:	Sysctl table reference.
 *	@param _write:		If write.
 *	@param _pFile:		File reference.
 *	@param _pBuffer:	Buffer reference.
 *	@param _pLen:		Length reference.
 *	@param _pPpos:		Position reference.
 *
 *	@return proc_dostring return.
 */
int bsp_gpio_maps_proc(	ctl_table		*_pTable,
					int			_write,
					struct file	*_pFile,
					void __user	*_pBuffer,
					size_t		*_pLen,
					loff_t		*_pPpos)
{
	char	*_pTampon		= (char *)_pTable->data;
	u16	_dir1Flags, _dir2Flags, _int1Set, _int2Set, _polSet1, _polSet2, _edgeSet1, _edgeSet2;

	/* Load current state. */
	_dir1Flags = bsp_gpio_read_regw(BSP_GPIO_DIR1);
	_dir2Flags = bsp_gpio_read_regw(BSP_GPIO_DIR2);

	/* Load current Int Set. */
	_int1Set = bsp_gpio_read_regw(BSP_GPIO_INT_SET1);
	_int2Set = bsp_gpio_read_regw(BSP_GPIO_INT_SET2);

	/* Load current pol Set. */
	_polSet1 = bsp_gpio_read_regw(BSP_GPIO_POL_SET1);
	_polSet2 = bsp_gpio_read_regw(BSP_GPIO_POL_SET2);

	/* Load current edge Set. */
	_edgeSet1 = bsp_gpio_read_regw(BSP_GPIO_EDGE_SET1);
	_edgeSet2 = bsp_gpio_read_regw(BSP_GPIO_EDGE_SET2);

	sprintf(	_pTampon,
			"\n\n\n");

	/* Next */
	_pTampon += strlen(_pTampon);

	/* Retrieve tampon */
	if ((_pTampon = (char *)_pTable->data) != (char *)0) {
		bsp_gpio_name_e	_nameGpio;
		unsigned int		_indiceGpio;

		sprintf(	_pTampon,

				"| Nb Gpio |          Name          |            Config                       |             Maps             |\n");

		/* Next */
		_pTampon += strlen(_pTampon);

		sprintf(	_pTampon,
				"|         |                        | Nb Name | Used as | Dir | Active | Intr | Dir | Intr | Pol | Edg | Val |\n");

		/* Next */
		_pTampon += strlen(_pTampon);

		/* All Gpio number */
		for (_indiceGpio = 0; _indiceGpio < BSP_GPIO_MAX; _indiceGpio++) {
			/* Number Gpio */
			sprintf(	_pTampon,
					"|   %2u    |",
					_indiceGpio);

			/* Next */
			_pTampon += strlen(_pTampon);

			/* Retrive name gpio */
			_nameGpio = bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].name;

			/* Name */
			sprintf(	_pTampon,
					"%s|",
					bsp_base.Board.pGpioNameTxt[_nameGpio]);

			/* Next */
			_pTampon += strlen(_pTampon);

			/* Config */
			{
				/* If gpio does not exist */
				if (bsp_base.Board.Cur.pGpioName[_nameGpio] == (pBsp_board_gpio_s)0) {
					sprintf(	_pTampon,
							"    -    |    -    |  -  |   -    |  -   |  -  |  -   |  -  |  -  |  -  |\n");
	
					/* Next */
					_pTampon += strlen(_pTampon);
	
					/* Continue */
					continue;
				}
	
				/* Number Name */
				sprintf(	_pTampon,
						"    %02u   |",
						_nameGpio);
	
				/* Next */
				_pTampon += strlen(_pTampon);

				/* Used as part */
				{
					/* If Gpio */
					if (bsp_base.Board.Cur.pGpioName[_nameGpio]->flags & BSP_Gpio_flags_used_as_gpio) {
						sprintf(	_pTampon,
								"   Gpio  |");
					}
					/* If config */
					else if (bsp_base.Board.Cur.pGpioName[_nameGpio]->flags & BSP_Gpio_flags_used_as_config) {
						sprintf(	_pTampon,
								"  Config |");
					}
					/* Spi cs */
					else if (bsp_base.Board.Cur.pGpioName[_nameGpio]->flags & BSP_Gpio_flags_used_as_spi_cs) {
						sprintf(	_pTampon,
								"  Spi cs |");
					}
					else {
						sprintf(	_pTampon,
								"    -    |");
					}
				}
	
				/* Next */
				_pTampon += strlen(_pTampon);
	
				/* Dir part */
				{
					/* If Gpio */
					if (bsp_base.Board.Cur.pGpioName[_nameGpio]->flags & BSP_Gpio_flags_dir_in) {
						sprintf(	_pTampon,
								"  In |");
					}
					/* If config */
					else if (bsp_base.Board.Cur.pGpioName[_nameGpio]->flags & BSP_Gpio_flags_dir_out) {
						sprintf(	_pTampon,
								" Out |");
					}
					else {
						sprintf(	_pTampon,
								"  -  |");
					}
				}
	
				/* Next */
				_pTampon += strlen(_pTampon);
	
				/* Active part */
				{
					/* If Gpio */
					if (bsp_base.Board.Cur.pGpioName[_nameGpio]->flags & BSP_Gpio_flags_active_low) {
						sprintf(	_pTampon,
								"  Low   |");
					}
					/* If config */
					else if (bsp_base.Board.Cur.pGpioName[_nameGpio]->flags & BSP_Gpio_flags_active_high) {
						sprintf(	_pTampon,
								"  High  |");
					}
					else {
						sprintf(	_pTampon,
								"   -    |");
					}
				}
	
				/* Next */
				_pTampon += strlen(_pTampon);
	
				/* Intr part */
				{
					/* If Gpio */
					if (bsp_base.Board.Cur.pGpioName[_nameGpio]->flags & BSP_Gpio_flags_intr) {
						sprintf(	_pTampon,
								"  1   |");
					}
					else {
						sprintf(	_pTampon,
								"  0   |");
					}
				}
	
				/* Next */
				_pTampon += strlen(_pTampon);
			}
	
			/* Maps */
			{
				int	_numGpio = bsp_gpio_get_num(_nameGpio);
		
				/* Dir */
				{
					/* Set direction */
					if (_numGpio < 16) {
						if (_dir1Flags & (1 << _numGpio)) {
							sprintf(	_pTampon,
									" Out |");
						}
						else {
							sprintf(	_pTampon,
									"  In |");
						}
					}
					else {
						if (_dir2Flags & (1 << (_numGpio - 16))) {
							sprintf(	_pTampon,
									" Out |");
						}
						else {
							sprintf(	_pTampon,
									"  In |");
						}
					}
				}
	
				/* Next */
				_pTampon += strlen(_pTampon);
	
				/* Intr */
				{
					/* Set direction */
					if (_numGpio < 16) {
						if (_int1Set & (1 << _numGpio)) {
							sprintf(	_pTampon,
									"  1   |");
						}
						else {
							sprintf(	_pTampon,
									"  0   |");
						}
					}
					else {
						if (_int2Set & (1 << (_numGpio - 16))) {
							sprintf(	_pTampon,
									"  1   |");
						}
						else {
							sprintf(	_pTampon,
									"  0   |");
						}
					}
				}

				/* Next */
				_pTampon += strlen(_pTampon);
	
				/* Polarity part */
				{
					/* Set direction */
					if (_numGpio < 16) {
						/* If dir out */
						if (_dir1Flags & (1 << _numGpio)) {
							sprintf(	_pTampon,
									"  -  |");
						}
						else {
							if (_polSet1 & (1 << _numGpio)) {
								sprintf(	_pTampon,
										"  1  |");
							}
							else {
								sprintf(	_pTampon,
										"  0  |");
							}
						}
					}
					else {
						/* If dir out */
						if (_dir2Flags & (1 << (_numGpio - 16))) {
							sprintf(	_pTampon,
									"  -  |");
						}
						else {
							if (_polSet2 & (1 << (_numGpio - 16))) {
								sprintf(	_pTampon,
										"  1  |");
							}
							else {
								sprintf(	_pTampon,
										"  0  |");
							}
						}
					}
				}

				/* Next */
				_pTampon += strlen(_pTampon);
	
				/* Edge part */
				{
					/* Set direction */
					if (_numGpio < 16) {
						/* If dir out */
						if (_dir1Flags & (1 << _numGpio)) {
							sprintf(	_pTampon,
									"  -  |");
						}
						else {
							if (_edgeSet1 & (1 << _numGpio)) {
								sprintf(	_pTampon,
										"  1  |");
							}
							else {
								sprintf(	_pTampon,
										"  0  |");
							}
						}
					}
					else {
						/* If dir out */
						if (_dir2Flags & (1 << (_numGpio - 16))) {
							sprintf(	_pTampon,
									"  -  |");
						}
						else {
							if (_edgeSet2 & (1 << (_numGpio - 16))) {
								sprintf(	_pTampon,
										"  1  |");
							}
							else {
								sprintf(	_pTampon,
										"  0  |");
							}
						}
					}
				}

				/* Next */
				_pTampon += strlen(_pTampon);
	
				/* Val */
				{
					/* Get Value */
					switch (bsp_gpio_get_state(_nameGpio)) {
						/* Set */
						case BSP_Gpio_state_set:
						{
							sprintf(	_pTampon,
									"  1  |\n");
						}
						break;
	
						/* Unset */
						case BSP_Gpio_state_unset:
						{
							sprintf(	_pTampon,
									"  0  |\n");
						}
						break;
	
						/* Else */
						default:
						{
							sprintf(	_pTampon,
									"  -  |\n");
						}
						break;
					}
				}

				/* Next */
				_pTampon += strlen(_pTampon);
			}
		}
	}
	else {
		/* Print */
		sprintf(	_pTampon,
				"Process Error\n");

		/* Next */
		_pTampon += strlen(_pTampon);
	}

	sprintf(	_pTampon,
			"\n");

	/* Next */
	_pTampon += strlen(_pTampon);

	/* Call proc_dostring */
	return(proc_dostring(	_pTable,
						_write,
						_pFile,
						_pBuffer,
						_pLen,
						_pPpos));
}

/* Dying Gasp support */
void (*bsp_dying_gasp_fp)(void);
EXPORT_SYMBOL(bsp_dying_gasp_fp);

/* Interrupt */
irqreturn_t bsp_gpio_irq(	int			_irq,
						void			*_pParam,
		       			struct pt_regs *_pRegs)
{
	bsp_gpio_intr_e	_gpioIntr	= *((bsp_gpio_intr_e *)_pParam);
	int				_indiceGpio, _nameGpio;
	int				_min, _max;
	u16				_flagsSet1, _flagsSet2, _intSet1, _intSet2;

	/* Load current Flags Set. */
	_flagsSet1 = bsp_gpio_read_regw(BSP_GPIO_FLAG_SET1);
	_flagsSet2 = bsp_gpio_read_regw(BSP_GPIO_FLAG_SET2);

	/* Load current int Set. */
	_intSet1 = bsp_gpio_read_regw(BSP_GPIO_INT_SET1);
	_intSet2 = bsp_gpio_read_regw(BSP_GPIO_INT_SET2);

	/* Gpio 0 to 15 */
	switch (_gpioIntr) {
		/* Gpio 0 to 15 */
		case BSP_Gpio_intr_0_to_15:
		{
			/* Set min/max */
			_min	= 0;
			_max	= 15;
		}
		break;

		/* Gpio 16 to 31 */
		case BSP_Gpio_intr_16_to_31:
		{
			/* Set min/max */
			_min	= 16;
			_max	= 31;
            
            /* Send Dying Gasp message */            
            (*bsp_dying_gasp_fp)();
            
		}
		break;

		default:
		{
			printk(	"%s: Error -> Type of gpioIntr(%u)\n",
					__FUNCTION__,
					_gpioIntr);

			return(IRQ_HANDLED);
		}
		break;
	}

	/* All Gpio. */
	for (_indiceGpio = _min; _indiceGpio <= _max; _indiceGpio++) {
		/* Retrieve name gpio */
		_nameGpio = bsp_base.Board.Cur.pInfo->Gpio[_indiceGpio].name;

		/* Check */
		if (_indiceGpio < 16) {
			/* If int and flag */
			if ( (_intSet1 & (1 << _indiceGpio)) && (_flagsSet1 & ( 1 << _indiceGpio)) ) {
#warning "Not implemeted yet !";
			}
		}
		else {
			/* If int and flag */
			if ( (_intSet2 & (1 << (_indiceGpio - 16))) && (_flagsSet2 & ( 1 << (_indiceGpio - 16))) ) {
#warning "Not implemeted yet !";
			}
		}
	}

	return(IRQ_HANDLED);
}


// vim:ts=5:
