/*
<:copyright-gpl 
 Copyright 2002 Broadcom Corp. All Rights Reserved. 
 
 This program is free software; you can distribute it and/or modify it 
 under the terms of the GNU General Public License (Version 2) as 
 published by the Free Software Foundation. 
 
 This program is distributed in the hope it will be useful, but WITHOUT 
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
 for more details. 
 
 You should have received a copy of the GNU General Public License along 
 with this program; if not, write to the Free Software Foundation, Inc., 
 59 Temple Place - Suite 330, Boston MA 02111-1307, USA. 
:>
*/

#include <linux/init.h>
#include <linux/fs.h>
#include <linux/capability.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/netdevice.h>

#include "led_local.h"

// uncomment // for debug led
//#define	DEBUG_LED
#define k100ms		(HZ/10)	// ~100 ms
#define kFastBlinkCount	0	// ~100ms
#define kSlowBlinkCount	5	// ~600ms
#ifndef FALSE
#define FALSE		0
#endif /* FALSE */
#ifndef TRUE 
#define TRUE		1
#endif /* TRUE */

static int g_timer_on = 0;
static struct timer_list g_led_timer;
static spinlock_t g_blink_list_lock = SPIN_LOCK_UNLOCKED;
static LIST_HEAD(g_blink_list);
static kled_generic_t g_generic_leds[kLedEnd];

//**************************************************************************************
// LED operations
//**************************************************************************************
static inline pkled_generic_t 
_kled_get_generic_object(BOARD_LED_NAME led)
{
	if ((unsigned int)led < kLedEnd)
		return &g_generic_leds[(unsigned int)led];
	else return NULL;
}

void
kled_init_state(BOARD_LED_NAME led, BOARD_LED_STATE state)
{
	pkled_generic_t pled = _kled_get_generic_object(led);

	if (!pled) return;

	pled->state = pled->state_saved = state;
	pled->name = led;
	pled->blink_count_down = 0;
	pled->blink_mode = 0;
	INIT_LIST_HEAD(&pled->l_list);
}

void
_kled_set_offfail(BOARD_LED_NAME led)
{
	return;
	//TODO:
}

void
_kled_set_off(BOARD_LED_NAME led)
{
	return;
	//TODO:
}

static inline BOARD_LED_STATE
_kled_toggle(pkled_generic_t pled)
{
	if (pled->led_on) kled_hw_set_off(pled->name);
	else kled_hw_set_on(pled->name);

	pled->led_on = (pled->led_on ? 0 : 1);

	return (pled->led_on ? kLedStateOn : kLedStateOff);
}

static void
_kled_add_timer(void)
{
	if (g_timer_on) return;

	g_led_timer.expires = jiffies + k100ms;        // timer expires in ~100ms
	add_timer(&g_led_timer);
	g_timer_on = 1;
}

// led timer expire kicks in about ~100ms and perform the led operation according to the ledState and
// restart the timer according to ledState
static void
_kled_check_timer(void)
{
	struct list_head *head, *temp;
	int run_timer = 0;

	g_timer_on = 0;

	spin_lock(&g_blink_list_lock);

	list_for_each_safe(head, temp, &g_blink_list) {
		pkled_generic_t pCurLed = list_entry(head, kled_generic_t, l_list);

		switch (pCurLed->state) {

			case kLedStateOff:
			case kLedStateOn:
			case kLedStateFail:
				pCurLed->blink_mode = 0;
				list_del_init(&pCurLed->l_list);
				break;
			
			case kLedStateBlinkOnce:
				_kled_toggle(pCurLed);
				pCurLed->blink_count_down = 0;                      // reset to 0
				pCurLed->state = pCurLed->state_saved;
				if ((kLedStateSlowBlinkContinues == pCurLed->state) || (kLedStateFastBlinkContinues == pCurLed->state))
					run_timer = 1;  // start timer if in blinkContinues stats
				break;

			case kLedStateSlowBlinkContinues:
				if (pCurLed->blink_count_down-- == 0) {
					pCurLed->blink_count_down = kSlowBlinkCount;
					_kled_toggle(pCurLed);
				}
				run_timer = 1;
				break;

			case kLedStateFastBlinkContinues:
				if (pCurLed->blink_count_down-- == 0) {
					pCurLed->blink_count_down = kFastBlinkCount;
					_kled_toggle(pCurLed);
				}
				run_timer = 1;
				break;

			case kLedStateBlinkContinues:
				if (pCurLed->blink_count_down-- == 0) {
					pCurLed->blink_count_down = (pCurLed->led_on? pCurLed->blink_freq_off : pCurLed->blink_freq_on);
					_kled_toggle(pCurLed);
				}
				run_timer = 1;
				break;

			default:
				printk("Invalid state = %d for led: %d\n", pCurLed->state, pCurLed->name);
				pCurLed->blink_mode = 0;
				list_del_init(&pCurLed->l_list);
		}
	}
	
	spin_unlock(&g_blink_list_lock);

	if (run_timer) _kled_add_timer();
}

static void
_kled_start_timer(void)
{
	if (g_timer_on) return;

#ifdef DEBUG_LED
	printk("led: add_timer\n");
#endif

	init_timer(&g_led_timer);
	g_led_timer.function = (void*)_kled_check_timer;
	g_led_timer.expires = jiffies + k100ms;        // timer expires in ~100ms
	add_timer(&g_led_timer);

	g_timer_on = 1;
}

void
kled_set_state(BOARD_LED_NAME led, BOARD_LED_STATE state, int freq_on, int freq_off)
{
	pkled_generic_t pled = _kled_get_generic_object(led);
	int run_timer = 0;
	
	if (NULL == pled) return;

	switch (state) {
		case kLedStateOn:
			kled_hw_set_on(pled->name);
			pled->state = pled->state_saved = kLedStateOn;
			pled->led_on = 1;
			break;
		
		case kLedStateOff:
			kled_hw_set_off(pled->name);
			pled->state = pled->state_saved = kLedStateOff;
			pled->led_on = 0;
			break;

		case kLedStateFail:
			kled_hw_set_onfail(pled->name);
			pled->state = pled->state_saved = kLedStateOff;
			break;

		case kLedStateBlinkOnce:
			if ((kLedStateSlowBlinkContinues == pled->state_saved) || (kLedStateFastBlinkContinues == pled->state_saved))
				;
			else {
				if (pled->blink_count_down == 0) { // skip the call if it is 1
					_kled_toggle(pled);
					pled->blink_count_down = 1;  // it will be reset to 0 when timer expires
					pled->state = kLedStateBlinkOnce;
					run_timer = 1;
				}
			}
			break;
		
		case kLedStateSlowBlinkContinues:
			pled->blink_count_down = kSlowBlinkCount;
			pled->state = pled->state_saved =  kLedStateSlowBlinkContinues;
			run_timer = 1;
			break;

		case kLedStateFastBlinkContinues:
			pled->blink_count_down = kFastBlinkCount;
			pled->state = pled->state_saved =  kLedStateFastBlinkContinues;
			run_timer = 1;
			break;

		case kLedStateBlinkContinues:
			pled->blink_freq_off = freq_off;
			pled->blink_freq_on  = freq_on;
			pled->led_on = 0;
			pled->blink_count_down = pled->blink_freq_off;
			pled->state = pled->state_saved = kLedStateBlinkContinues;
			run_timer = 1;
			break;

		default:
			printk("invalid led state.\n");
			break;
	}

	if (run_timer) {
		spin_lock_bh(&g_blink_list_lock);
		if (!pled->blink_mode) {
			pled->blink_mode = 1;
			list_add(&pled->l_list, &g_blink_list);
		}
		spin_unlock_bh(&g_blink_list_lock);

		_kled_start_timer();
	}
}

void kled_get_state(BOARD_LED_NAME led, BOARD_LED_STATE *state)
{
	pkled_generic_t pled = _kled_get_generic_object(led);
	
	
	if (NULL == pled) return;
	*state = pled->state_saved;
	
}

int __init
kled_init(void)
{
	kled_hw_init();

	return 0;
}

