/*
 * (c) 2010 SoftAtHome
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/jiffies.h>
#include <linux/debugfs.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/sahprofiler.h>
#include <linux/seq_file.h>

#if defined(CONFIG_FUSIV_VX160)
#define SAH_PRF_DIVIDER 100		/* Divide 10ns values by 100 for us. */
#elif defined(CONFIG_FUSIV_VX185)
static __u32 divider = 1;
#define SAH_PRF_DIVIDER (divider)
#else
#define SAH_PRF_DIVIDER 1
#endif


struct sah_prf_id_info {
	const char *name;
        unsigned long current_ts;       /* Timestamp on entry. */
};

struct sah_prf {
        unsigned long count;
        unsigned long min;
        unsigned long max;
        unsigned long sum;
};

struct sah_prf_interval {
        unsigned long seconds;
	unsigned long ts;
        struct sah_prf buf[SAH_PRF_MAXIDS];
};

u32 sah_prf_run = 0;
EXPORT_SYMBOL(sah_prf_run);

u32 sah_prf_in_ksoftirqd = 0;
EXPORT_SYMBOL(sah_prf_in_ksoftirqd);

sah_prf_id_t sah_prf_ksoftirqd_id;
EXPORT_SYMBOL(sah_prf_ksoftirqd_id);

static struct sah_prf_id_info sah_prf_ids[SAH_PRF_MAXIDS];
static struct sah_prf_interval sah_prf_buf[SAH_PRF_INVERVALS];
static struct sah_prf_interval *sah_prf_current_interval = sah_prf_buf;
static struct sah_prf_interval *sah_prf_tail_interval = sah_prf_buf;
static DEFINE_SPINLOCK(sah_prf_id_lock);
static DEFINE_SPINLOCK(sah_prf_lock);
static struct dentry *root, *run, *clear, *dump, *list, *buf, *counter;
/* static struct dentry *ids; */

/* sah_prf_lock must be held when calling this. */
static unsigned long sah_prf_get_ts(void);
static void sah_prf_init_timer(void);

sah_prf_id_t sah_prf_register(const char *name)
{
	static sah_prf_id_t nextid = 0;
	unsigned long flags;
	sah_prf_id_t id;

	if (nextid >= (SAH_PRF_MAXIDS - 1))
		return SAH_PRF_ID_INVALID;
	spin_lock_irqsave(&sah_prf_id_lock, flags);
	id = nextid++;
	spin_unlock_irqrestore(&sah_prf_id_lock, flags);
	sah_prf_ids[id].name = name;

	return id;
}
EXPORT_SYMBOL(sah_prf_register);

void sah_prf_dump(void)
{
	if (in_interrupt())
		return;
	while (sah_prf_tail_interval != sah_prf_current_interval) {
		int i;

		printk("sah_profiler: interval %8lus %10s %10s %10s %10s %10s\n",
			sah_prf_tail_interval->seconds,
			"count", "min", "avg", "max", "sum");
		for (i = 0; i < SAH_PRF_MAXIDS && sah_prf_ids[i].name; i++)
			if (sah_prf_tail_interval->buf[i].count != 0)
				printk("%-32s %10lu %10lu %10lu %10lu %10lu\n",
				sah_prf_ids[i].name,
				sah_prf_tail_interval->buf[i].count,
				sah_prf_tail_interval->buf[i].min / SAH_PRF_DIVIDER,
				sah_prf_tail_interval->buf[i].sum / (SAH_PRF_DIVIDER * sah_prf_tail_interval->buf[i].count),
				sah_prf_tail_interval->buf[i].max / SAH_PRF_DIVIDER,
				sah_prf_tail_interval->buf[i].sum / SAH_PRF_DIVIDER);
		sah_prf_tail_interval++;
		if (sah_prf_tail_interval >= &sah_prf_buf[SAH_PRF_INVERVALS])
			sah_prf_tail_interval = sah_prf_buf;
		msleep_interruptible(40);
		if (signal_pending(current))
			break;
	}
}
EXPORT_SYMBOL(sah_prf_dump);

/* One second has passed. */
/* (Not taking into account jiffies wrapping...) */
/* sah_prf_lock must be held. */
static void sah_prf_next_interval(unsigned long seconds, unsigned long ts)
{
	int i;

	sah_prf_current_interval++;
	if (sah_prf_current_interval >= &sah_prf_buf[SAH_PRF_INVERVALS])
		sah_prf_current_interval = sah_prf_buf;
	sah_prf_current_interval->seconds = seconds;
	sah_prf_current_interval->ts = ts;
	for (i = 0; i < SAH_PRF_MAXIDS && sah_prf_ids[i].name; i++)
		memset(&sah_prf_current_interval->buf[i], 0, sizeof(struct sah_prf));
}

void __sah_prf_enter(sah_prf_id_t id)
{
	unsigned long flags;

	if (unlikely(id == SAH_PRF_ID_INVALID))
		return;
	if (unlikely(sah_prf_ids[id].name == NULL))
		return;
	spin_lock_irqsave(&sah_prf_lock, flags);
	sah_prf_ids[id].current_ts = sah_prf_get_ts();
	spin_unlock_irqrestore(&sah_prf_lock, flags);
}
EXPORT_SYMBOL(__sah_prf_enter);

void __sah_prf_leave(sah_prf_id_t id)
{
	unsigned long flags;
	unsigned long now, duration;
	unsigned long nowseconds;
	int next_interval;

	if (unlikely(id == SAH_PRF_ID_INVALID))
		return;
	if (unlikely(sah_prf_ids[id].name == NULL))
		return;
	nowseconds = jiffies / CONFIG_HZ;
	next_interval = (nowseconds != sah_prf_current_interval->seconds);
	spin_lock_irqsave(&sah_prf_lock, flags);
	now = sah_prf_get_ts();
	if (unlikely(next_interval))
		sah_prf_next_interval(nowseconds, now);
	sah_prf_current_interval->buf[id].count++;
	if (unlikely(sah_prf_ids[id].current_ts > now))
		duration = (0xffffffffUL - sah_prf_ids[id].current_ts) + now + 1;
	else
		duration = now - sah_prf_ids[id].current_ts;
	if (duration > sah_prf_current_interval->buf[id].max)
		sah_prf_current_interval->buf[id].max = duration;
	if (sah_prf_current_interval->buf[id].min == 0 ||
	    duration < sah_prf_current_interval->buf[id].min)
		sah_prf_current_interval->buf[id].min = duration;
	sah_prf_current_interval->buf[id].sum += duration;
	spin_unlock_irqrestore(&sah_prf_lock, flags);
}
EXPORT_SYMBOL(__sah_prf_leave);

void sah_prf_clear(void)
{
	unsigned long flags;

	sah_prf_run = 0;
	spin_lock_irqsave(&sah_prf_lock, flags);
	sah_prf_current_interval = sah_prf_buf;
	sah_prf_tail_interval = sah_prf_buf;
	spin_unlock_irqrestore(&sah_prf_lock, flags);
	memset(sah_prf_buf, 0, sizeof(sah_prf_buf));
}
EXPORT_SYMBOL(sah_prf_clear);

static int dummy_open(struct inode *inode, struct file *file)
{
	return 0;
}

static int dummy_release(struct inode *inode, struct file *file)
{
	return 0;
}

static ssize_t clearops_write(struct file *file, const char __user *buf,
			      size_t size, loff_t *off)
{
	sah_prf_clear();
	return size;
}

static ssize_t listops_read(struct seq_file *seq, void *offset)
{
	unsigned int i;

	(void) offset;

	for (i = 0; i < SAH_PRF_MAXIDS && sah_prf_ids[i].name; i++)
		seq_printf(seq, "%-32s %10lu\n", sah_prf_ids[i].name, sah_prf_ids[i].current_ts);
	return 0;
}

static int listops_open(struct inode *inode, struct file *file)
{
	(void) inode;

	return single_open(file, listops_read, NULL);
}

static ssize_t dumpops_write(struct file *file, const char __user *buf,
			     size_t size, loff_t *off)
{
	sah_prf_dump();
	return size;
}

static ssize_t counterops_read(struct file *file, char __user *userbuf, size_t size, loff_t *off)
{
	char buf[16];
	size_t len;

	len = snprintf(buf, sizeof(buf), "%lu\n", sah_prf_get_ts());
	if (len > size)
		return 0;
	if (copy_to_user(userbuf, buf, len))
		return -EFAULT;
	*off += len;
	return len;
}

static struct file_operations listops = {
	.owner = THIS_MODULE,
	.open = listops_open,
	.llseek = seq_lseek,
	.read = seq_read,
	.release = single_release,
};

static struct file_operations dumpops = {
	.owner = THIS_MODULE,
	.open = dummy_open,
	.write = dumpops_write,
	.release = dummy_release,
};

static struct file_operations clearops = {
	.owner = THIS_MODULE,
	.open = dummy_open,
	.write = clearops_write,
	.release = dummy_release,
};

static struct file_operations counterops = {
	.owner = THIS_MODULE,
	.open = dummy_open,
	.read = counterops_read,
	.release = dummy_release,
};

static struct debugfs_blob_wrapper buf_blob = {
	.data = sah_prf_buf,
	.size = sizeof(sah_prf_buf),
};

static int __init sah_prf_init(void)
{
	sah_prf_init_timer();

	/* Setup debugfs interface. */
	root = debugfs_create_dir("sah_prf", NULL);
	/* ids = debugfs_create_dir("id", root); */
	counter = debugfs_create_file("counter", 0444, root, NULL, &counterops);
	clear = debugfs_create_file("clear", 0644, root, NULL, &clearops);
	dump = debugfs_create_file("dump", 0644, root, NULL, &dumpops);
	list = debugfs_create_file("list", 0444, root, NULL, &listops);
	run = debugfs_create_bool("run", 0644, root, &sah_prf_run);
	buf = debugfs_create_blob("buf", 0444, root, &buf_blob);

	sah_prf_clear();
	sah_prf_ksoftirqd_id = sah_prf_register("ksoftirqd");

	return 0;
}
module_init(sah_prf_init);

#if defined(CONFIG_FUSIV_VX160)
#include <timer.h>
#define CPE_TIMER_BASE_ADDRESS  0x19070000

static volatile struct timer_regs *tregs =
	(volatile struct timer_regs *)KSEG1ADDR(CPE_TIMER_BASE_ADDRESS);

static unsigned long sah_prf_get_ts(void)
{
	__u32 ts;
	static __u32 previous = 0;

	tregs->status1 = 0x0800;
	ts = tregs->counter_lo1 | (tregs->counter_hi1 << 16);
	tregs->status1 = 0x0400;
	previous += ts;

	return previous;
}

static void sah_prf_init_timer(void)
{
	unsigned long period;
	unsigned long width;
	unsigned long flags;
	period = 0xffffffff;
	width = period - 2;

	printk(KERN_DEBUG "%s: initialising timer1. width = %lu, period = %lu\n",
		__func__, width, period);
	printk(KERN_DEBUG "%s: %p W status1:    0x%04x\n", __func__, &tregs->status1, 0x0800);
	printk(KERN_DEBUG "%s: %p W config1:    0x%04x\n", __func__, &tregs->config1, 0x000d);
	printk(KERN_DEBUG "%s: %p W period_hi1: 0x%04lx\n", __func__, &tregs->period_hi1, period >> 16);
	printk(KERN_DEBUG "%s: %p W period_lo1: 0x%04lx\n", __func__, &tregs->period_lo1, period & 0xffff);
	printk(KERN_DEBUG "%s: %p W width_hi1:  0x%04lx\n", __func__, &tregs->width_hi1, width >> 16);
	printk(KERN_DEBUG "%s: %p W width_lo1:  0x%04lx\n", __func__, &tregs->width_lo1, width & 0xffff);
	printk(KERN_DEBUG "%s: %p W status1:    0x%04x\n", __func__, &tregs->status1, 0x0800);

	local_irq_save(flags);

	/* Disable timer. */
	tregs->status1 = 0x0800;

	/*
	 * This configures the timer as follows:
	 * PWM_OUT Mode
	 * Positive Active Pulse
	 * Count to end of Period
	 * Interrupt Request Disable
	*/
	tregs->config1 = 0x000d;

	/* Write top half of period register */
	tregs->period_hi1 = period >> 16;

	/* Write bottom half of period register */
	tregs->period_lo1 = period & 0xffff;

	/* Write top half of width register */
	tregs->width_hi1 = width >> 16;

	/* Write bottom half of width register */
	tregs->width_lo1 = width & 0xffff;

	/* Set TIMEN0 and clear any lingering interrupts */
	tregs->status1 = 0x0400;

	local_irq_restore(flags);

	printk(KERN_DEBUG "%s: %p R status1:    0x%04x\n", __func__, &tregs->status1, tregs->status1);
}

#elif defined(CONFIG_FUSIV_VX185)
#include <vx185_clocks.h>
#define CPE_TIMER_BASE_ADDRESS  0x19050000
struct timer_regs {
	__u32	topcount0;		/* 00 */
	__u32	control0;		/* 04 */
	__u32	topcount1;		/* 08 */
	__u32	control1;		/* 0c */
	__u32	topcount2;		/* 10 */
	__u32	control2;		/* 14 */
	__u32	topcount3;		/* 18 */
	__u32	control3;		/* 1c */
	__u32	prescaler0;		/* 20 */
	__u32	prescaler1;		/* 24 */
	__u32	control_wdt;		/* 28 */
	__u32	pet_wdt;		/* 2c */
	__u32	topcount_wdt;		/* 30 */
	__u32	wdt_value;		/* 34 */
	__u32	int_control;		/* 38 */
	__u32	int_status;		/* 3c */
};
extern struct clock_values fusiv_718x_clks;

static volatile struct timer_regs *tregs =
	(volatile struct timer_regs *)KSEG1ADDR(CPE_TIMER_BASE_ADDRESS);

static unsigned long sah_prf_get_ts(void)
{
	__u32 ts;
	__u32 period;
	unsigned long flags;
	static __u32 previous = 0;

	/* Divide 250MHz ~ 4ns values by 250 for us. */
	divider = fusiv_718x_clks.sys_clk_val / 1000000;

	/* Just count down from anywhere. */
	period = 0xffffffff;

	local_irq_save(flags);
	/* Disable timer */
	tregs->control3 = 4;

	/* Read timer */
	ts = tregs->topcount3;

	/* Write period */
	tregs->topcount3 = period;

	/* Enable timer */
	tregs->control3 = 6;

	local_irq_restore(flags);

	previous += (period - ts);

	return previous;
}

static void sah_prf_init_timer(void)
{
	__u32 period;
	unsigned long flags;

	/* Vx185 system clock 250 MHz
	 * period = 2*250*1000*1000 ~ 2s */
	/* period = fusiv_718x_clks.sys_clk_val * 2; */
	period = 0xffffffff;

	printk(KERN_DEBUG "%s: initialising timer3.period = %lu\n",
		__func__, (unsigned long)period);
	printk(KERN_DEBUG "%s: %p W control 3:   0x%08x\n", __func__, &tregs->control3, 4);
	printk(KERN_DEBUG "%s: %p W top count 3: 0x%08x\n", __func__, &tregs->topcount3, period);
	printk(KERN_DEBUG "%s: %p W control 3:   0x%08x\n", __func__, &tregs->control3, 6);

	local_irq_save(flags);

	/*
	 * This configures the timer as follows:
	 * Periodic Mode
	 * 1ms resolution
	 * GPIO value has no effect on timer
	 * Count to zero and wrap
	 * Interrupt Request Disable
	*/

	/* Disable timer */
	tregs->control3 = 4;

	/* Write period */
	tregs->topcount3 = period;

	/* Enable timer */
	tregs->control3 = 6;

	local_irq_restore(flags);

	printk(KERN_DEBUG "%s: %p R control3:    0x%08x\n", __func__, &tregs->control3, tregs->control3);
}


#else
#include <linux/timex.h>
static unsigned long sah_prf_get_ts(void)
{
	return (unsigned long)get_cycles();
}

static void sah_prf_init_timer(void)
{
}
#endif

MODULE_LICENSE("GPL");
