#include <string.h>
#include <stdlib.h>
#include "pcb_dnsres.h"
#include <debug/sahtrace.h>
#include <pcb/utils.h>
#include <pcb/common.h>
#include "resolv.h"


struct _pcb_dnsres {
    struct dnsres _resp;
    connection_info_t *con;
    char *retrypolicy;
    pcb_timer_t *timer;

    int retry;
    int res_options;
    char *hostname;
    char *servname;
    struct addrinfo hints;
    void (*getaddrinfo_cb)(struct addrinfo *, int, void *);
    void (*getaddrinfo_cb_ext)(struct _pcb_dnsres *,struct addrinfo_ext *, int, void *);

    struct in_addr addr;
    int len;
    int af;
    unsigned int ttl;
    void (*gethostbyaddr_cb)(struct dnsres_hostent *, int, void *);

    void *arg;
    void *handle;

    bool   result_present;
    struct dnsres_hostent *result_he;
    struct addrinfo_ext *result_ai;
    int result_res;
    void * result_arg;
};

typedef struct _hba_result_t {
    struct dnsres_hostent *he;
    int error;
    int done;
} hba_result_t;

typedef struct _ai_result_t {
    struct addrinfo_ext *ai;
    int error;
    int done;
} ai_result_t;

enum {ONCE_ERROR_INTERVAL,
      NONE_ERROR_INTERVAL,
      DEFAULT_ERROR_INTERVAL};

static int pcb_dnsres_retryDetectType(char *typetoken) {
    int type = DEFAULT_ERROR_INTERVAL;
    if(strcmp(typetoken,"ONCE")==0) {
        type = ONCE_ERROR_INTERVAL;
    } else if(strcmp(typetoken,"NONE")==0) {
        type = NONE_ERROR_INTERVAL;
    }
    return type;
}

static int pcb_dnsres_nextRetryTime(const char *retrypolicy, int retry) {
    char *token=NULL;
    int rstep = 0;
    int result = RES_TIMEOUT;
    int type = DEFAULT_ERROR_INTERVAL;
    SAH_TRACEZ_INFO("DNSRES","getting retry time for policy %s (%d)",retrypolicy,retry);
    if(retrypolicy!=NULL) {
        char *completestring = strdup(retrypolicy);
        char *tmp = completestring;

        //detect the type
        token = strsep(&tmp, ":");
        if(token != NULL) {
            type = pcb_dnsres_retryDetectType(token);
            if (type == NONE_ERROR_INTERVAL) {
                if(retry==0) {
                    if(strlen(retrypolicy)>strlen("NONE:")) {
                        result = atoi(tmp);
                    }
                    free(completestring);
                    SAH_TRACEZ_INFO("DNSRES","result: %d",result);
                    return result;
                } else {
                    free(completestring);
                    SAH_TRACEZ_INFO("DNSRES","result: %d",-1);
                    return -1;
                }
            } else if(type != DEFAULT_ERROR_INTERVAL) {
                while((token = strsep(&tmp, ","))!=NULL && strlen(token)!=0) {
                    if(rstep==retry) {                    // in case we are in the specified retry step, use the step value
                        result = atoi(token);
                        break;
                    }
                    rstep++;
                }
                if(token==NULL && type==ONCE_ERROR_INTERVAL) {// we passed the last token, 
                    if(retry==rstep) { // last retry
                        free(completestring);
                        SAH_TRACEZ_INFO("DNSRES","result: %d",-1);
                        return 10; // max 10 seconds timeout
                    } else { // no more retries
                        free(completestring);
                        SAH_TRACEZ_INFO("DNSRES","result: %d",-1);
                        return -1; 
                    }
                }
            }

        }
        free(completestring);
    }

    if(type == DEFAULT_ERROR_INTERVAL) {
        result = RES_TIMEOUT;
    }
    SAH_TRACEZ_INFO("DNSRES","result: %d",result);
    return result;
}

static bool pcb_dnsres_retry(pcb_dnsres_t *item, int retrytime, void *cb) {
    if(item->timer) pcb_timer_destroy(item->timer);
    item->timer = NULL;
    item->retry += 1;
    item->timer = pcb_timer_create();
    if(item->timer==NULL) {
        SAH_TRACEZ_ERROR("DNSRES","Could not create timer");
        return false;
    }
    if(pcb_timer_setHandler(item->timer, cb)==false) {
        SAH_TRACEZ_ERROR("DNSRES","Could not set timer read callback");
        return false;
    }
    if( pcb_timer_setUserData(item->timer, item)==false) {
        SAH_TRACEZ_ERROR("DNSRES","Could not set timer user data");
        return false;
    }
    if(pcb_timer_start(item->timer, retrytime*1000)==false) {
        SAH_TRACEZ_ERROR("DNSRES","Could not start timer");
        return false;
    }
    dnsres_setTimeout(&item->_resp, retrytime);
    return true;
}


static pcb_dnsres_t *pcb_dnsres_init(connection_info_t *con, const char *retrypolicy, void *arg, unsigned int res_options,const  char *hostname, const char *servname, struct addrinfo *hints,
                                      void (*gai_cb)(struct addrinfo *, int, void *), void (*gai_cb_ext)(pcb_dnsres_t *request, struct addrinfo_ext *, int, void *),const void *addr, int len,
                                     int af, void (*hba_cb)(struct dnsres_hostent *, int, void *)) {
    pcb_dnsres_t *item = NULL;

    if(con==NULL)  {
        SAH_TRACE_ERROR("Not a valid con");
        return NULL;
    }

    item = calloc(1,sizeof(pcb_dnsres_t));
    if(item==NULL)  {
        SAH_TRACE_ERROR("Could not allocate memory");
        return NULL;
    }

    if(retrypolicy==NULL) {
        item->retrypolicy = strdup("ONCE:8,16,32,64,4");
    }
    else {
        item->retrypolicy = strdup(retrypolicy);
    }
    item->con = con;
    item->arg = arg;
    item->res_options=res_options;

    if(hostname) item->hostname = strdup(hostname);
    if(servname) item->servname = strdup(servname);
    if(addr)     memcpy(&(item->addr),addr,sizeof(struct in_addr));
    if(hints)    memcpy(&(item->hints),hints,sizeof(const struct addrinfo));
    if(gai_cb) item->getaddrinfo_cb = gai_cb;
    if(gai_cb_ext) item->getaddrinfo_cb_ext = gai_cb_ext;
    item->retry = 0;
    item->len = len;
    item->af = af;
    if(hba_cb) item->gethostbyaddr_cb = hba_cb;
    dnsres_init(con, &item->_resp);
    if(res_options & PCB_DNSRES_USE_TCP) {
        SAH_TRACE_INFO("USING TCP");

        item->_resp.options = RES_USEVC;
    } else {
        item->_resp.options &= ~RES_USEVC;
    }

    item->result_present = false;
    item->result_he = NULL;
    item->result_ai = NULL;
    item->result_res = 0;
    item->result_arg = NULL;

    return item;
}

static void pcb_dnsres_cleanup(pcb_dnsres_t *item) {
    if(item==NULL) return;

    if(item->retrypolicy) {
        free(item->retrypolicy);
        item->retrypolicy = NULL;
    }
    if(item->servname) {
        free(item->servname);
        item->servname = NULL;
    }
    if(item->hostname) {
        free(item->hostname);
        item->hostname = NULL;
    }
    if(item->timer) {
        pcb_timer_stop(item->timer);
        pcb_timer_destroy(item->timer);
        item->timer = NULL;
    }
    item->con = NULL;
    free(item);
}


static void pcb_dnsres_gethostbyaddrcb(struct dnsres_hostent *he, int res, void *arg) {
    pcb_dnsres_t *item = (pcb_dnsres_t *)arg;
    int retrytime = pcb_dnsres_nextRetryTime(item->retrypolicy,item->retry);

    // did the call succeed or was it the last retry?
    if(he!=NULL || retrytime==-1) {
        if(item->res_options&PCB_DNSRES_LAST_RETRY_ALWAYS_TIMEOUT) {
            // last call, but we have to wait untill the last retry timeout:
            // store the result and make sure that the timer callback will return this
            item->result_present = true;
            item->result_he = he;
            item->result_res = res;
            item->result_arg = arg;
        } else {
            //last call or call succeeded, call the callback and clean up
            item->gethostbyaddr_cb(he, res, item->arg);
            pcb_dnsres_cleanup(item);
            return;
        }
    }
    // there was an error, but it was not the last try: do nothing and wait until the timercb is called
    item->handle=NULL;
    return;
}

static void pcb_dnsres_gethostbyaddrtimercb(pcb_timer_t *timer, void *userdata) {
    pcb_dnsres_t *item = (pcb_dnsres_t *)userdata;
    if(item->handle) {
        dnsres_cancel_hbn_lookup(&item->_resp,item->handle);
        item->handle=NULL;
    }

    if(item->result_present) {
        item->gethostbyaddr_cb(item->result_he, item->result_res, item->result_arg);
        pcb_dnsres_cleanup(item);
        return;
    }

    //start the timeout timer
    int retrytime = pcb_dnsres_nextRetryTime(item->retrypolicy,item->retry);
    if(retrytime!=-1) {
        SAH_TRACEZ_INFO("DNSRES","Starting retry timer with %d seconds",retrytime);
        if(pcb_dnsres_retry(item, retrytime, pcb_dnsres_gethostbyaddrtimercb)==false) {
            goto cleanup;
        }
        item->handle = dnsres_gethostbyaddr(&item->_resp,(const char *) &item->addr, item->len, item->af, pcb_dnsres_gethostbyaddrcb, (void *)item);
        return;
    }

cleanup:
    item->gethostbyaddr_cb(NULL, DNSRES_SERVFAIL, item->arg);
    pcb_dnsres_cleanup(item);
    return;
}


pcb_dnsres_t *pcb_dnsres_gethostbyaddr(connection_info_t *con,unsigned int res_options,const char *retrypolicy,const void *addr, int len, int af,void (*cb)(struct dnsres_hostent *, int, void *),void *arg) {
    pcb_dnsres_t *item = NULL;
    if(con==NULL)  {
        SAH_TRACE_ERROR("Not a valid pcb connection");
        return NULL;
    }
    if(cb==NULL)  {
        SAH_TRACE_ERROR("Not a valid callback pointer");
        return NULL;
    }
    if(addr==NULL)  {
        SAH_TRACE_ERROR("Not a valid callback pointer");
        return NULL;
    }
    item = pcb_dnsres_init(con, retrypolicy,arg,res_options,NULL,NULL,NULL,NULL,NULL,addr,len,af,cb);
    if(item==NULL)  {
        SAH_TRACE_ERROR("Out of memory");
        return NULL;
    }
    int retrytime = pcb_dnsres_nextRetryTime(item->retrypolicy,item->retry);
    if(retrytime==-1) retrytime = DEFAULT_TIMEOUT;
    SAH_TRACEZ_INFO("DNSRES","Start retry timer with %d seconds",retrytime);
    if(pcb_dnsres_retry(item, retrytime, pcb_dnsres_gethostbyaddrtimercb)==false) {
        goto cleanup;
    }
    item->handle = dnsres_gethostbyaddr(&item->_resp,addr,len,af,pcb_dnsres_gethostbyaddrcb, (void *)item);
    return item;

cleanup:
    pcb_dnsres_cleanup(item);
    return NULL;
}


static void pcb_dnsres_getaddrinfocb(struct addrinfo_ext *ai, int res, void *arg) {
    pcb_dnsres_t *item = (pcb_dnsres_t *)arg;
    SAH_TRACEZ_INFO("DNSRES","pcb_dnsres_getaddrinfocb");
    int retrytime = pcb_dnsres_nextRetryTime(item->retrypolicy,item->retry);

    // did the call succeed or was it the last retry or do we need to stop on NXDOMAIN
    if(ai!=NULL ||
       retrytime==-1 ||
       (res==DNSRES_HOST_NOT_FOUND&&item->res_options & PCB_DNSRES_STOP_RETRY_ON_NXDOMAIN)) {
        if(item->res_options&PCB_DNSRES_LAST_RETRY_ALWAYS_TIMEOUT) {
            // last call, but we have to wait untill the last retry timeout:
            // store the result and make sure that the timer callback will return this
            item->result_present = true;
            item->result_ai = ai;
            item->result_res = res;
            item->result_arg = arg;
        } else {
            //last call or call succeeded, call the callback and clean up
            SAH_TRACEZ_INFO("DNSRES","call callback and clean up");
            if(item->getaddrinfo_cb) item->getaddrinfo_cb((struct addrinfo *)ai, res, item->arg);
            if(item->getaddrinfo_cb_ext) item->getaddrinfo_cb_ext(item, ai, res, item->arg);
            pcb_dnsres_cleanup(item);
            return;
        }
    }
    // there was an error: do nothing and wait until the timercb is called
    item->handle=NULL;
}

static void pcb_dnsres_getaddrinfotimercb(pcb_timer_t *timer, void *userdata) {
    pcb_dnsres_t *item = (pcb_dnsres_t *)userdata;
    SAH_TRACEZ_INFO("DNSRES","pcb_dnsres_getaddrinfotimercb");
    if(item->handle) {
        dnsres_cancel_gai_lookup(&item->_resp,item->handle);
        item->handle=NULL;
    }

    if(item->result_present) {
        if(item->getaddrinfo_cb) item->getaddrinfo_cb((struct addrinfo *)item->result_ai, item->result_res, item->result_arg);
        if(item->getaddrinfo_cb_ext) item->getaddrinfo_cb_ext(item,item->result_ai, item->result_res, item->result_arg);
        pcb_dnsres_cleanup(item);
        return;
    }

    int retrytime = pcb_dnsres_nextRetryTime(item->retrypolicy,item->retry);
    if(retrytime!=-1) {
        SAH_TRACEZ_INFO("DNSRES","start retry timer with %d seconds",retrytime);
        if(pcb_dnsres_retry(item, retrytime, pcb_dnsres_getaddrinfotimercb)==false) {
            goto cleanup;
        }
        item->handle = dnsres_getaddrinfo(&item->_resp, item->hostname, item->servname, &(item->hints), pcb_dnsres_getaddrinfocb, (void *)item);
        return;
    }

cleanup:
    if(item->getaddrinfo_cb) item->getaddrinfo_cb(NULL, DNSRES_SERVFAIL, item->arg);
    if(item->getaddrinfo_cb_ext) item->getaddrinfo_cb_ext(item,NULL, DNSRES_SERVFAIL, item->arg);
    pcb_dnsres_cleanup(item);
    return;
}

pcb_dnsres_t *pcb_dnsres_getaddrinfo(connection_info_t *con, unsigned int res_options, const char *retrypolicy, const char *hostname, const char *servname, struct addrinfo *hints, void (*cb)(struct addrinfo *, int, void *),void *arg) {
    pcb_dnsres_t *item = NULL;
    if(con==NULL)  {
        SAH_TRACE_ERROR("Not a valid pcb connection");
        return NULL;
    }
    if(cb==NULL)  {
        SAH_TRACE_ERROR("Not a valid callback pointer");
        return NULL;
    }
    item = pcb_dnsres_init(con,retrypolicy,arg,res_options,hostname,servname,hints,cb,NULL,NULL,0,0,NULL);
    if(item==NULL)  {
        SAH_TRACE_ERROR("Out of memory");
        return NULL;
    }


    //start the timeout timer
    int retrytime = pcb_dnsres_nextRetryTime(item->retrypolicy,item->retry);
    if(retrytime==-1) retrytime = DEFAULT_TIMEOUT;
    SAH_TRACEZ_INFO("DNSRES","Start retry timer with %d seconds",retrytime);

    if(pcb_dnsres_retry(item, retrytime, pcb_dnsres_getaddrinfotimercb)==false) {
        goto cleanup;
    }
    item->handle = dnsres_getaddrinfo(&item->_resp, hostname, servname, hints, pcb_dnsres_getaddrinfocb, (void *)item);
    return item;

cleanup:
    pcb_dnsres_cleanup(item);
    return NULL;
}

pcb_dnsres_t *pcb_dnsres_getaddrinfo_ext(connection_info_t *con,
                                   unsigned int res_options,
                                   const char *retry_policy,
                                   const char *hostname,
                                   const char *servname,
                                   struct addrinfo *hints,
                                   void (*cb)(pcb_dnsres_t *request, struct addrinfo_ext *ai, int error, void *arg),
                                   void *arg) {
    pcb_dnsres_t *item = NULL;
    if(con==NULL)  {
        SAH_TRACE_ERROR("Not a valid pcb connection");
        return NULL;
    }
    if(cb==NULL)  {
        SAH_TRACE_ERROR("Not a valid callback pointer");
        return NULL;
    }
    item = pcb_dnsres_init(con,retry_policy,arg,res_options,hostname,servname,hints,NULL,cb,NULL,0,0,NULL);
    if(item==NULL)  {
        SAH_TRACE_ERROR("Out of memory");
        return NULL;
    }

    //start the timeout timer
    int retrytime = pcb_dnsres_nextRetryTime(item->retrypolicy,item->retry);
    if(retrytime==-1) retrytime = DEFAULT_TIMEOUT;
    SAH_TRACEZ_INFO("DNSRES","Start retry timer with %d seconds",retrytime);

    if(pcb_dnsres_retry(item, retrytime, pcb_dnsres_getaddrinfotimercb)==false) {
        goto cleanup;
    }
    item->handle = dnsres_getaddrinfo(&item->_resp, hostname, servname, hints, pcb_dnsres_getaddrinfocb, (void *)item);
    return item;

cleanup:
    pcb_dnsres_cleanup(item);
    return NULL;
}


uint32_t pcb_dnsres_getTTL(pcb_dnsres_t *request) {
    if(request) {
        return request->ttl;
    }
    return 0;
}


void pcb_dnsres_cancelRequest(pcb_dnsres_t *request) {
    if (request&&request->result_present==false) {
        if(request->getaddrinfo_cb || request->getaddrinfo_cb_ext) {
            dnsres_cancel_gai_lookup(&request->_resp,request->handle);
        } else {
            dnsres_cancel_hbn_lookup(&request->_resp,request->handle);
        }
        pcb_dnsres_cleanup(request);
    }
}


static void pcb_dnsres_gethostbyaddr_b_cb(struct dnsres_hostent *he, int res, void *arg) {
    hba_result_t *heres = (hba_result_t *)arg;
    int i;
    if(he) {
        heres->he = calloc(1,sizeof(struct dnsres_hostent));
        memcpy(heres->he,he,sizeof(struct dnsres_hostent));
        if(he->h_name)  {
            heres->he->h_name = strdup(he->h_name);
        }
        if(he->h_aliases) {
            // count the number of entries
            for(i=0;he->h_aliases[i]!=NULL;i++) {;}

            // allocate the array
            heres->he->h_aliases = calloc(i+1,sizeof(char *));

            // allocate the entries
            for(i=0;he->h_aliases[i]!=NULL;i++) {heres->he->h_aliases[i] = strdup(he->h_aliases[i]);}
        }
        if(he->h_addr_list) {
            // count the number of entries
            for(i=0;he->h_addr_list[i]!=NULL;i++) {;}

            // allocate the array
            heres->he->h_addr_list = calloc(i+1,sizeof(char *));

            // allocate the entries
            for(i=0;he->h_addr_list[i]!=NULL;i++) {heres->he->h_addr_list[i] = strdup(he->h_addr_list[i]);}
        }
    }
    heres->error = res;
    heres->done = 1;
}

void pcb_dnsres_gethostbyaddr_b_free_hostent(struct dnsres_hostent *he) {
    int i = 0;
    if(he==NULL) {
        return;
    }
    if(he->h_name) free(he->h_name);
    if(he->h_aliases) {
        // free the entries
        for(i=0;he->h_aliases[i]!=NULL;i++) {free(he->h_aliases[i]);}

        // free the array
        free(he->h_aliases);
    }
    if(he->h_addr_list) {
        // free the entries
        for(i=0;he->h_addr_list[i]!=NULL;i++) {free(he->h_addr_list[i]);}

        // free the array
        free(he->h_addr_list);
    }
    free(he);
}

int pcb_dnsres_gethostbyaddr_b(unsigned int res_options,
                               const char *retry_policy,
                               const void *addr,
                               int len,
                               int af,
                               struct dnsres_hostent **he) {

    connection_list_t connections;
    hba_result_t resp;
    connection_info_t *con = NULL;
    resp.he = NULL;
    resp.error = 0;
    resp.done = 0;

    con = connection_create("DNSRES");
    if(con==NULL) {
        SAH_TRACEZ_ERROR("DNSRES","unable to initialize connection");
        goto error;
    }

    if (!llist_initialize(&connections)) {
        SAH_TRACEZ_ERROR("DNSRES","unable to initialize connection list %s", error_string(pcb_error));
        goto error;
    }
    llist_append(&connections, connection_getIterator(con));

    if(pcb_dnsres_gethostbyaddr(con,res_options,retry_policy,addr,len,af,pcb_dnsres_gethostbyaddr_b_cb,(void *)&resp)==NULL) {
        SAH_TRACEZ_INFO("DNSRES","pcb_dnsres_gethostbyaddr failed");
        goto error;
    }

    //simulation of pcb event loop
    while (1) {
        /* handle connection events */
        if (connection_waitForEvents(&connections, NULL) < 0) {
            SAH_TRACE_FATAL("select failed: %m");
            continue;
        }
        connection_handleEvents(&connections);
        if(resp.done!=0) {
            break;
        }
    }

    *he = resp.he;

    connection_destroy(con);
    return resp.error;

error:
    connection_destroy(con);
    return -1;
}



static void pcb_dnsres_getaddrinfo_b_cb(pcb_dnsres_t *dnsres, struct addrinfo_ext *ai, int res, void *arg) {
    SAH_TRACEZ_INFO("DNSRES","pcb_dnsres_getaddrinfo_b_cb called for %p",dnsres);
    ai_result_t *aires = (ai_result_t *)arg;
    aires->ai = ai;
    aires->error = res;
    aires->done = 1;
}


//note: res has to be freed
int pcb_dnsres_getaddrinfo_b(unsigned int res_options,
                             const char *retry_policy,
                             const char *hostname,
                             const char *servname,
                             struct addrinfo *hints,
                             struct addrinfo **res) {
    connection_list_t connections;
    ai_result_t resp;
    resp.ai = NULL;
    resp.error = 0;
    resp.done = 0;

    connection_info_t *con = connection_create("DNSRES");
    if(con==NULL) {
        SAH_TRACEZ_ERROR("DNSRES","unable to initialize connection");
        goto error;
    }

    if (!llist_initialize(&connections)) {
        SAH_TRACEZ_ERROR("DNSRES","unable to initialize connection list %s", error_string(pcb_error));
        goto error;
    }
    llist_append(&connections, connection_getIterator(con));

    if(pcb_dnsres_getaddrinfo_ext(con, res_options, retry_policy, hostname, servname, hints, pcb_dnsres_getaddrinfo_b_cb,(void *)&resp)==NULL) {
        SAH_TRACEZ_INFO("DNSRES","pcb_dnsres_getaddrinfo failed");
        goto error;
    }

    //simulation of pcb event loop
    while (1) {
        /* handle connection events */
        if (connection_waitForEvents(&connections, NULL) < 0) {
            SAH_TRACE_FATAL("select failed: %m");
            continue;
        }
        connection_handleEvents(&connections);
        if(resp.done!=0) {
            break;
        }
    }

    *res = (struct addrinfo *)resp.ai;

    connection_destroy(con);
    return resp.error;

error:
    return -1;
}


void pcb_dnsres_free_addrinfo_ext(struct addrinfo_ext *ai) {
    if(ai) {
        freeaddrinfo((struct addrinfo *) ai);
    }
}




