/*
 * Copyright (c) 2011 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 * Author : Dries De Winter (dries.dewinter@softathome.com)
 *
 */
#ifndef __MAFIA_H__
#define __MAFIA_H__

/**
@defgroup mafia Debugging Facility "MAFIA"
@{

@brief
<b>MAFIA</b> is a library that is mainly intented to facilitate debugging on the target. A quick feature summary:
- It facilitates <b>application self debugging</b> by providing a MIPS implementation for the glibc- and architecture-specific
  function backtrace(). Additionally it comes with extra backtrace()-related utilities.
- It facilitates investigating <b>unexpected terminations</b>. The typical example is an application crash because of a segmentation
  fault or a floating point exception for instance. Another possible cause of an unexpected termination is when a (third party)
  libary calls exit() under some erronous conditions.
- It facilitates investigating <b>application freezes</b>, i.e. the application is still running but it's doesn't seem to respond
  anymore. Perhaps it's stuck in an ever lasting syscall or it is a deadlocked by mutual dependant semaphores.
- It facilitates <b>infinite loop</b> debugging. Seeming to be frozen but taking 100% CPU and never getting out are the typical
  symptoms of an infinite loop. With MAFIA, you can easily get a backtrace of such process and so locate the infinite loop quickly.
- It facilitates <b>heap usage analysis</b> and <b>memory leak detection</b> by intercepting calls to malloc(), calloc(), realloc()
  and free(), store call stack information and do some extra bookkeeping. It also support <b>invalid/double free() detection</b>.

@details
@section environment Applicable environmental variables
The behavior of the library can be affected at application startup by setting some environmental variables and it can not be
modified anymore later on. The applicable environmental variables are:
	- <b>HEAPDEBUG</b> Set to <b>y</b> or <b>1</b> to intercept calls to malloc(), calloc(), realloc() and free(). If this is not
	  explicitly set, heap analysis, memory leak detection and invalid/double free detection is not supported. The default is
	  chosen to be <b>n</b>, because this feature comes with a big performance and memory penalty. If this option is enabled, 
	  the heap is dumped on the output implicitly at application exit.
	  If the application (and all involved libraries) clean up their resources nicely, the heap should be empty.
	  This gives a view on memory leaks, however, MAFIA does not distinguish still reachable memory from leaked memory like
	  valgrind does.
	- <b>EXITBACKTRACE</b> Indicates whether a backtrace should be printed when exit() is called. Possibilities are:
		- <b>never</b> never print a backtrace on exit(). This is the default.
		- <b>always</b> always print a backtrace on exit(). Notice that this includes the case where the application exits because
		  it returns from main().
		- <b>nonzero</b> only print a backtrace if the exit code of the application is non-zero.
	- <b>MAFIASINK</b> A custom file can be specified by its path to redirect debug prints to. The default is stderr, a dup()-ed
	  stderr to be more precise to circumvent that daemon() cuts MAFIA's freedom of speech.

Instead of setting these environmental variables manually, you can alternatively use command line options of the mafia tool.
See also @ref linking.

@section signal Signals
The library implements most of its features by installing appropriate signal handlers. MAFIA catches the following signals:
	- <b>SIGSEGV</b> (Segmentation violation), <b>SIGBUS</b> (Bus error), <b>SIGILL</b> (Illegal instruction), <b>SIGFPE</b>
	  (Floating point exception) and <b>SIGABRT</b> (Abort). If one of these signals is raised, a backtrace is printed, more
	  precisely a backtrace of the context that raised the signal, rather than a backtrace of the signal handling context itself.
	  See also backtrace_signal().
	- <b>SIGRTMAX</b>. Print a backtrace without terminating the process. The exact number of this signal depends on the
	  architecture. On x86 it's 64 and on MIPS it's 127.
	- <b>SIGRTMAX-1</b>. Dump a heap snapshot. All currently allocated blocks are grouped together per matching backtrace and for
	  each occuring backtrace, the payload (amount of memory allocated, excluding bookkeeping or fragmentation overhead) and
	  the number of blocks is printed, followed by the backtrace. You can also dump the heap explicitly in the application code
	  itself using heap_dump(). Notice that this only works if heap debugging is enabled.
	- <b>SIGRTMAX-2</b>. Build static memory maps. When printing a backtrace, instruction addresses are translated into
	  (<file> + <offset>)-pairs. This is implemented by parsing /proc/self/maps. However, when printing the backtrace of a
	  malloc'ed block, it's possible that the backtrace contains instructions that are not mapped in memory anymore, in which
	  case the address can not be mapped to a file. This happens for applications with a plugin infrastructure: code loaded
	  with dlopen() might be unloaded explicitly with dlclose(). A very common example of this is code loaded by the odl-parser
	  of pcb (typically inside a pcb_plugin container). Therefore you can load a static table of memory maps at a well chosen
	  point in time, so it can be used later on when resolving backtraces. If there is such a static map table, it is used
	  by heap_dump() to resolve addresses. You can also build static maps explicitly in the application code itself
	  using backtrace_loadstaticmaps().

@section output Backtrace output format
The typical format of a backtrace entry is:

<code><address> = <file> + <offset> = <symbol> + <offset></code>

So instruction addresses are mapped onto (<file> + <offset>)-pairs which are then mapped onto
(<symbol> + <offset>)-pairs. Mapping offsets relative to a file to offsets relative to a symbol is implemented by
reading the ELF information in the file, more specifically the symbol table and if that is not present, the dynamic symbol table.
If no matching symbol can be found (which is typically the case for static functions if the binary was stripped or compiled with
-s), you get an entry of the shorter format

<code><address> = <file> + <offset></code>

However, if the unstripped binary is available elsewhere, e.g. in the stagingdir of the build that produced the backtrace,
you can still post-process the backtrace with <b>crashwizard</b>. It will translate entries of the second format into entries of the
first format if possible.

If the instruction address can not even be mapped on a file, you get the even shorter format

<code><address></code>

Entries in this format are not post-processable. Instruction addresses that can not be mapped to files are a typical symptom of
code that was unloaded explicitly by dlclose(). You can avoid this by loading static maps at the right time, see
backtrace_loadstaticmaps().

@section threads With or wihout multi-threading support

In order to be thread-safe, heap bookkeeping and access to static memory maps is protected with mutexes. Locking a mutex is not
async-signal-safe, so the implementation of SIGRTMAX-1 and SIGRTMAX-2 can not be done inside signal handler context. The solution
for this issue is to implement these signals in a separate signal handling thread, implicitly created by the library at startup.

However this solution introduces two new problems:

Problem 1: Creating a thread can only happen in applications that are compiled and linked with -pthread, which is typically
not the case for single threaded applications. Therefore one build produces two versions of libmafia: one without support for
multi-threading and one with support for multi-threading. The version with multi-threading support has a separate signal handling
thread like described above. In the version without support for multi-threading, taking/releasing the mutexes is left
out at compile time. This way, the handlers for SIGRTMAX-1 and SIGRTMAX-2 become async-signal-safe and so they can be implemented
entirely in signal handling context.

Problem 2: The extra thread requires a fixed amount of extra system resources per process, which is intolerable if all applications
link with this library operationally. Therefore, the separate signal handling thread is only started if heap debugging is enabled.
This implies that SIGRTMAX-1 and SIGRTMAX-2 are not catched/handled in that case.
Notice that if heap debugging is disabled, handling SIGRTMAX-1 is useless anyway and heap debugging is also the main use case
for SIGRTMAX-2. Furthermore, static maps can still be loaded by the application itself by calling backtrace_loadstaticmaps().

@section linking Linking with MAFIA

Applications may link with MAFIA at compile time by appending <b>-lmafia_nothreads</b> or <b>-lmafia_withthreads</b> to the linker
command, or by adding package <b>mafia_nothreads</b> or <b>mafia_withthreads</b> to the pkg-config command (e.g. add
<b>`pkg-config --libs mafia_nothreads`</b> to the linker command).

The library can also be preloaded at runtime by putting <b>LD_PRELOAD=libmafia_nothreads.so</b> or
<b>LD_PRELOAD=libmafia_withthreads.so</b> in the application environment, or alternatively using the MAFIA's command line tool
called <b>mafia</b>. If you put "mafia" in front of the sartup command of your application, your application will be started with
libmafia_nothreads.so preloaded. The tool has some additional command line options to prefer the multi-threaded implementation over
the single-threaded implementation and to set the applicable environmental settings. Check the usage message with 
<b>mafia --help</b>.
Example: check the multi-threaded application called "myapp" for memory leaks with "mafia --withtreads --heapdebug myapp".

MAFIA is designed to be linked with operationally, i.e. not only for special development builds. If loaded with default 
configuration (more specifically with heap debugging disabled), its overhead in terms of memory and performance is negligible. 

@section portability Portability
	- Concerning backtrace(): If not compiled for MIPS, MAFIA relies on the presence of an implementation of the function 
	  backtrace() in the standard C-library. So MAFIA's portability is limited to MIPS systems + glibc-based systems with 
	  backtrace()-implementation (glibc does not contain a backtrace()-implementation for all architectures).
	- Concerning async-signal-safety: backtrace() is assumed to be async-signal-safe. MAFIA's own implementation for MIPS is, but
	  glibc's implementation typically isn't, which makes MAFIA unreliable on non-MIPS systems. However, it has been tested on 
	  x86 successfully.

@section acronym Where does the name "MAFIA" come from?
<b>MAFIA</b> is an acronym for:
	- <b>M</b>emory leak,
	- <b>A</b>bort,
	- <b>F</b>reeze and
	- <b>I</b>nfinite loop
	- <b>A</b>lert.
*/

#include <mafia/backtrace.h>
#include <mafia/heap.h>

/**
@brief
Get MAFIA's output file descriptor

@details
This function returns the file descriptor that is used to send debug prints to. So unless if the debug prints are redirected
by a non-empty MAFIASINK (@ref environment), this returns a copy of the original stderr.

@return a file descriptor
*/

int mafia_sink();

/**
@}
*/

#endif
