/*
 * Copyright (c) 2011 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 * Author : Dries De Winter (dries.dewinter@softathome.com)
 *
 */
#ifndef __MAFIA_BACKTRACE_H__
#define __MAFIA_BACKTRACE_H__

/**
@ingroup mafia
@{
*/

#define BACKTRACE_FORMAT_CRASHWIZARD 0 /**< Use output format understood by crashwizard when printing a backtrace. This is the default output format. See also @ref output.*/

#define BACKTRACE_FORMAT_GLIBC       1 /**< Use glibc's output format when printing a backtrace.*/

#define BACKTRACE_USESTATICMAPS      2 /**< Use static memory maps if available when resolving a backtrace.*/

/**
@brief
Get a backtrace of the calling context.

@details
Get a backtrace of the calling context, in the array pointed by <i>buffer</i>. The first item in the buffer will be the address of
the first instruction after the call to backtrace(). The rest of the buffer will be filled with return addresses formed by
unwinding the call stack. At most <i>size</i> addresses are stored in <i>buffer</i>, but less is possible if the bottom of the call
stack is reached sooner than the end of the buffer.

This function is specified by glibc and it's architecture dependent. This library has an implementation for MIPS, so it gives
limited portability to backtrace()-calling code, assuming that you're either compiling for MIPS, or either compiling against a glibc
version with a working backtrace()-implementation.

General note:

Inlined functions, tail-call optimizations and other compiler optimizations generally cause a backtrace to be incomplete.

Note regarding the MIPS implementation specifically:

Because on MIPS it's impossible to unwind the call stack using the stack itself only, the implementation does some code
scanning to find the size of the current stack frame and the place where the return address is stored on the stack, making some
assumptions on how a function looks like. This is extremely fragile, dangerous in fact. The most notable invalid assumption is that
functions only modify the stack pointer twice: decrease it once in the beginning of the function and increase it again with the same
amount at the end of the function. This amount is assumed to be the size of the stack frame. If the stack pointer is modified
multiple times in a function, the implementation just gives up unwinding resulting in an incomplete backtrace. Notice that if the
implementation concludes the wrong stack frame size, the return address would probably be invalid so code scanning would
dereference an invalid address, so it's better to give up early than to make a good guess and perhaps head for a crash.
There is still room to improve the quality of produced backtraces, but this empirical approach will never lead to 100% correctness.

@param buffer An array of void pointers, to be filled with return addresses.
@param size The size of the buffer (number of elements, not number of bytes).

@return The effective number of addresses written into buffer.
*/
int backtrace(void **buffer, int size);

/**
@brief
Translate a backtrace into symbolic representation and print to a file descriptor.

@details
Given the set of addresses returned by backtrace() in <i>buffer</i>, this function translates the addresses into strings that
describe the addresses symbolically. The <i>size</i> argument specifies  the number  of  addresses  in  buffer.  The symbolic
representation of each address consists of the function name (if this can be  determined), a hexadecimal offset into the function,
and the actual return address (in hexadecimal). Each string is written, one per line, to the file descriptor <i>fd</i>.

Like backtrace(), this function is specified by glibc. This library contains an implementation which is only compiled for MIPS.
The implementation in this library calls backtrace_dump() with attributes = BACKTRACE_FORMAT_GLIBC while on other
architectures an implementation is assumed to be present in the C library.

@param buffer An array of return addresses, as returned by backtrace().
@param size The number of addresses in <i>buffer</i>.
@param fd The file descriptor to write the symbolic representations to.
*/
void backtrace_symbols_fd(void *const *buffer, int size, int fd);


/**
@brief
Get a backtrace of the context that raised a signal.

@details
Get a backtrace like backtrace() does, except that this function is intended to be used in signal handling context to get a
backtrace of the context that raised the signal, rather than a backtrace of the signal handling context itself.

This library contains a MIPS specific implementation. <i>uctx</i> should be the void pointer that's passed as third argument
to the signal handler that's installed with sigaction() (with SA_SIGINFO set). It's casted to a ucontext_t which is an
architecture dependent structure.

If not compiled for MIPS, backtrace_signal() simply calls backtrace(), so it will always have the signal handling context on top
of the stack. This is not ideal but it gives at least some information, however, since glibc's backtrace() implementation is
typically not async-signal-safe, getting backtraces from a signal handling context on other architectures but MIPS can be
considered as unreliable.

@param buffer An array of void pointers, to be filled with return addresses.
@param size The size of the buffer (number of elements, not number of bytes).
@param uctx A pointer to a ucontext_t, should be the third argument that's passed to the signal handler.

@return The effective number of addresses written into buffer.
*/
int backtrace_signal(void **buffer, int size, const void *uctx);

/**
@brief
Translate a backtrace into symbolic representation and print to a file descriptor - extended.

@details
Given the set of addresses returned by backtrace() in <i>buffer</i>, this function translates the addresses into strings that
describe the addresses symbolically. The <i>size</i> argument specifies  the number  of  addresses  in  buffer.
Each string is written, one per line, to the file descriptor <i>fd</i>.
The symbolic representation of each address depends on the <i>attributes</i> argument.

@param buffer An array of return addresses, as returned by backtrace().
@param size The number of addresses in <i>buffer</i>.
@param fd The file descriptor to write the symbolic representations to.
@param attributes A bitwise OR of zero or more of the following:
	- BACKTRACE_FORMAT_CRASHWIZARD
	  Use output format understood by crashwizard when printing a backtrace. This is the default output format.
	  See also @ref output.
	- BACKTRACE_FORMAT_GLIBC
	  Use glibc's output format when printing a backtrace.
	- BACKTRACE_USESTATICMAPS
	  Use static memory maps if available when resolving a backtrace.
*/
void backtrace_dump(void *const *buffer, int size, int fd, int attributes);

/**
@brief
Build static memory maps.

@details
This function parses /proc/self/maps and puts the parsed information in static memory. Later on backtraces can be translated
into symbolic representation using these static maps by specifying the attribute BACKTRACE_USESTATICMAPS when calling
backtrace_dump().

The capacity of this functionality is very limited because the total static memory of libmafia should fit in 1 page to
limit the per-process overhead. Up to 48 executable segments can be stored and the filename length is limited to 28 characters,
so it's perfectly possbile that filenames are truncated. However, if a filename is too long, it's truncated at the head rather than
the tail and crashwizard and/or runtime symbol resolving will generally speaking still work as expected.
*/
void backtrace_loadstaticmaps(void);

/**
@brief
Flush static memory maps.

@details
This function erases the information stored by backtrace_loadstaticmaps().
*/
void backtrace_unloadstaticmaps(void);

/**
@brief
Translate an instruction address into a (filename + offset)-pair.

@details
Given one addresses returned by backtrace(), this function translates the address into a filename and an offset. The filename
shall be stored in <i>buffer</i>. <i>size</i> specifies the size of the buffer. This function will never write more than
<i>size</i> bytes into <i>buffer</i> while guaranteeing that the the string held by <i>buffer</i> is null-terminated
(unless if <i>size</i> is zero). The offset will be stored into <i>fileoffset</i> if it's not NULL.

@param addr The instruction address to be translated.
@param buffer Buffer to store the filename in.
@param size Size of the buffer.
@param fileoffset Pointer to write the offset to.
@param attributes A bitwise OR of zero or more of the following:
	- BACKTRACE_USESTATICMAPS
	  Use static memory maps if available to translate the address.
*/
int backtrace_file(void *addr, char *buffer, int size, unsigned long *fileoffset, int attributes);


/**
@brief
Translate a (file + offset)-pair into a (symbol + offset)-pair.

@details
Given an open file descriptor for an ELF-file and an offset into that file, this function finds the corresponding
symbol if possible and calculates the offset to that symbol.

Currently only 32-bit binaries are supported.

It has limited cross-platform support: this function is able to interpret big endian binaries on little endian systems and
vice versa.

@param elffd The file descriptor referring to an ELF-file.
@param fileoffset The offset into the ELF-file to be looked up.
@param buffer Buffer to store the symbol name in.
@param size Size of the buffer.
@param symoffset Pointer to write the offset relative to the symbol to.
*/
int backtrace_sym(int elffd, unsigned long fileoffset, char *buffer, int size, unsigned long *symoffset);

/**
@}
*/

#endif
