/*
 * Copyright (c) 2011 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 * Author : Dries De Winter (dries.dewinter@softathome.com)
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <elf.h>
#include <arpa/inet.h>
#ifdef WITHTHREADS
#include <pthread.h>
#endif

#include "mafia/backtrace.h"
#include "common.h"

#ifdef MIPS

#include <signal.h>

// Hackish redefinition of ucontext_t to avoid kernel / toolchain header mismatch
typedef struct uctx {
	unsigned long flags;
	struct uctx *link;
	stack_t stack;
	struct sigcontext sigctx;
	 __sigset_t uc_sigmask;
} uctx_t;

#define abs(s) ((s) < 0 ? -(s) : (s))

static int unwind(void **buffer, int size, unsigned long *pc, unsigned long *sp, unsigned long *ra) {
	unsigned long *addr;
	int ra_offset, stack_size, depth=0, start;
	if (!pc && ra) {
		buffer[depth++] = NULL;
		pc = ra;
		ra = NULL;
	}
	for (; depth < size && pc; depth++) {
		buffer[depth] = pc;
		ra_offset = stack_size = start = 0;
		for (addr = pc; (!ra_offset || !stack_size) && !start; addr--) {
			switch (*addr & 0xffff0000) {
			case 0x27bd0000:
				if (ra_offset) {
					stack_size = abs((short)(*addr & 0xffff));
				} else {  // modifying the stack pointer in the middle of a function -> get out of here!
					if (ra && depth + 1 < size)
						buffer[++depth] = ra;
					return depth + 1;
				}
				break;
			case 0xafbf0000:
				ra_offset = (short)(*addr & 0xffff);
				break;
			case 0x3c1c0000:
				start = 1;
				break;
			default:
				break;
			}
		}
		if (ra_offset)
			pc = *(unsigned long **)((unsigned long)sp + ra_offset);
		else if (ra)
			pc = ra;
		else
			break;
		ra = NULL;
		sp = (unsigned long *)((unsigned long)sp + stack_size);
	}
	return depth;
}

int backtrace(void **buffer, int size) {
	unsigned long *addr, *ra, *sp;
	int stack_size;

	// get current $ra and $sp
	__asm__ __volatile__ (
				"       move    %0, $ra\n"
				"       move    %1, $sp\n"
				: "=r"(ra), "=r"(sp)
				);

	// scanning to find the size of the current stack frame
	stack_size = 0;

	for (addr = (unsigned long *)backtrace; !stack_size; ++addr) {
		if ((*addr & 0xffff0000) == 0x27bd0000)
			stack_size = abs((short)(*addr & 0xffff));
		else if (*addr == 0x03e00008)
			break;
	}
	sp = (unsigned long *)((unsigned long)sp + stack_size);

	return unwind(buffer, size, ra, sp, NULL);
}

int backtrace_signal(void **buffer, int size, const void *uctx) {
	uctx_t *ctx = (uctx_t *)uctx;
	unsigned long *pc, *sp, *ra;

	pc = (unsigned long *)(unsigned long)ctx->sigctx.sc_pc;
	sp = (unsigned long *)(unsigned long)ctx->sigctx.sc_regs[29];
	ra = (unsigned long *)(unsigned long)ctx->sigctx.sc_regs[31];

	return unwind(buffer, size, pc, sp, ra);
}

#else

int backtrace_signal(void **buffer, int size, const void *uctx) {
	(void)uctx;
	return backtrace(buffer, size);
}

#endif

static map_t staticmaps[STACKMAPSIZE];
static char staticmapcaptions[STACKMAPSIZE][CAPTIONSIZE];
static int staticmapcount = 0;

#ifdef WITHTHREADS

static pthread_mutex_t staticmapsmutex = PTHREAD_MUTEX_INITIALIZER;

static void staticmapslock() {
	pthread_mutex_lock(&staticmapsmutex);
}

static void staticmapsunlock() {
	pthread_mutex_unlock(&staticmapsmutex);
}

#else
#define staticmapslock()
#define staticmapsunlock()
#endif

static ssize_t readat(int fd, void *ptr, size_t size, off_t pos) {
	ssize_t n, ret = 0;
	if (lseek(fd, pos, SEEK_SET) == -1)
		return -1;
	while (size > 0 && (n = read(fd, ptr, size)) > 0) {
		size -= n; ptr += n; ret += n;
	}
	return ret;
}

static char *strnchr(char *s, int c, int len) {
	for (; *s && *s != c && len; s++, len--);
	return len && *s ? s : NULL;
}

static int openmaps(map_t *maps, int *size) {
	char line[BUFSIZE], *end, *lineptr;
	int count = 0, pos = 0;
	int fd = open("/proc/self/maps", O_RDONLY);
	while (count < *size && readat(fd, line, BUFSIZE, pos) > (int)sizeof(void *) * 6 + 25) {
		end = strnchr(line, '\n', BUFSIZE);
		if (!end)
			break;
		maps[count].start = (void *)strtoul(line, &lineptr, 16);
		maps[count].end = (void *)strtoul(lineptr + 1, &lineptr, 16);
		maps[count].offset = strtoul(lineptr + 6, NULL, 16);
		maps[count].pathpos = pos + sizeof(void *) * 6 + 25;
		maps[count].file = -1;
		pos += end - line + 1;
		if (*(lineptr + 3) != 'x')
			continue;
		count++;
	}
	*size = count;
	return fd;
}

static void closemaps(int fd, map_t *maps, int size) {
	int i;
	for (i = 0; i < size; i++) {
		if (maps[i].file > 0)
			close(maps[i].file);
	}
	if (fd > 0)
		close(fd);
}

static int mappath(int fd, map_t *map, char *buf, int size) {
	if (!size || map->pathpos <= 0)
		return 0;
	int ret = readat(fd, buf, size, map->pathpos);
	if (ret <= 0)
		return -1;
	char *end = strnchr(buf, '\n', ret);
	if (!end)
		return -1;
	*end = '\0';
	return end - buf;
}

static int mapfile(int fd, map_t *map) {
	if (map->file <= 0) {
		char buf[BUFSIZE];
		mappath(fd, map, buf, BUFSIZE);
		map->file = open(buf, O_RDONLY);
	}
	return map->file;
}

static int mapcaption(int fd, map_t *map, char *buf, int size) {
	char path[BUFSIZE];
	int len = mappath(fd, map, path, BUFSIZE);
	if (len < 0) {
		return -1;
	} else if (len < size - 4) {
		strcpy(buf, path);
		return len;
	} else if (size > 3) {
		strcpy(buf, "...");
		strcat(buf, path + len - (size - 4));
		return size - 1;
	} else if (size > 0) {
		strncpy(buf, "...", size - 1);
		buf[size - 1] = '\0';
		return size - 1;
	} else {
		return size;
	}
}

static map_t *maplookup(map_t *maps, int size, void *addr) {
	for (; size; maps++, size--)
		if (maps->start <= addr && maps->end > addr)
			return maps;
	return NULL;
}

static void mapswap(void *data, size_t size) {
	char *cdata  = (char *)data;
	char tmp;
	size_t i;
	for (i=0; i < size/2; i++) {
		tmp = cdata[i];
		cdata[i] = cdata[size-i-1];
		cdata[size-i-1] = tmp;
	}
}

static int mapsym32(int file, unsigned long addr, char *buffer, int size, unsigned long *offset, int needswap) {
	// Read file header
	Elf32_Ehdr ehdr;
	if (readat(file, &ehdr, sizeof(Elf32_Ehdr), 0) != sizeof(Elf32_Ehdr))
		return -1;
		
	if (needswap) {
		mapswap(&ehdr.e_phnum, sizeof(Elf32_Half));
		mapswap(&ehdr.e_phoff, sizeof(Elf32_Off));
		mapswap(&ehdr.e_shoff, sizeof(Elf32_Off));
		mapswap(&ehdr.e_shstrndx, sizeof(Elf32_Half));
		mapswap(&ehdr.e_shnum, sizeof(Elf32_Half));
	}
	
	// Find program header for segment holding addr
	int i;
	for (i = 0; i < ehdr.e_phnum; i++) {

		// Get i'th program header
		Elf32_Phdr phdr;
		if (readat(file, &phdr, sizeof(Elf32_Phdr), ehdr.e_phoff + i * sizeof(Elf32_Phdr)) != sizeof(Elf32_Phdr))
			continue;
		
		if (needswap) {
			mapswap(&phdr.p_type, sizeof(Elf32_Word));
			mapswap(&phdr.p_flags, sizeof(Elf32_Word));
			mapswap(&phdr.p_offset, sizeof(Elf32_Off));
			mapswap(&phdr.p_filesz, sizeof(Elf32_Xword));
			mapswap(&phdr.p_vaddr, sizeof(Elf32_Addr));
		}
			
		if (phdr.p_type != PT_LOAD || !(phdr.p_flags & PF_X))
			continue; // only considere loadable and executable segments;

		if (phdr.p_offset > addr || phdr.p_offset + phdr.p_filesz <= addr)
			continue; // it's not a match

		addr += phdr.p_vaddr;
		break;
	}

	// Get section with section names
	Elf32_Shdr shstrhdr;
	if (readat(file, &shstrhdr, sizeof(Elf32_Shdr), ehdr.e_shoff + ehdr.e_shstrndx * sizeof(Elf32_Shdr)) != sizeof(Elf32_Shdr))
		return -1;

	if (needswap) {
		mapswap(&shstrhdr.sh_size, sizeof(Elf32_Word));
		mapswap(&shstrhdr.sh_offset, sizeof(Elf32_Off));
	}

	// Get symbol table and symbol table entry strings
	Elf32_Shdr strhdr, symhdr;
	memset(&strhdr, 0, sizeof(Elf32_Shdr)); // not needed to initialize but tell that to the compiler
	memset(&symhdr, 0, sizeof(Elf32_Shdr));
	int havestrhdr = 0, havesymhdr = 0;
	for (i = 0; i < ehdr.e_shnum; i++) {

		// Get i'th section header
		Elf32_Shdr shdr;
		if (readat(file, &shdr, sizeof(Elf32_Shdr), ehdr.e_shoff + i * sizeof(Elf32_Shdr)) != sizeof(Elf32_Shdr))
			continue;

		if (needswap) {
			mapswap(&shdr.sh_name, sizeof(Elf32_Word));
			mapswap(&shdr.sh_size, sizeof(Elf32_Word));
			mapswap(&shdr.sh_entsize, sizeof(Elf32_Word));
			mapswap(&shdr.sh_offset, sizeof(Elf32_Off));
		}

		// Get section name
		char sname[9];
		sname[8] = '\0';
		if (shdr.sh_name >= shstrhdr.sh_size)
			continue;
		if (readat(file, sname, 8, shstrhdr.sh_offset + shdr.sh_name) <= 0)
			continue;

		// Check for .dynsym and .dynstr, optionally overruled by .symtab and .strtab
		if (!strcmp(sname, ".dynstr") && !havestrhdr) {
			memcpy(&strhdr, &shdr, sizeof(Elf32_Shdr));
			havestrhdr = 1;
		} else if (!strcmp(sname, ".dynsym") && !havesymhdr) {
			memcpy(&symhdr, &shdr, sizeof(Elf32_Shdr));
			havesymhdr = 1;
		} else if (!strcmp(sname, ".strtab")) {
			memcpy(&strhdr, &shdr, sizeof(Elf32_Shdr));
			havestrhdr = 1;
		} else if (!strcmp(sname, ".symtab")) {
			memcpy(&symhdr, &shdr, sizeof(Elf32_Shdr));
			havesymhdr = 1;
		}
	}
	if (!havestrhdr || !havesymhdr)
		return -1;

	// Find matching symbol
	for (i = 0; i < (int)(symhdr.sh_size / symhdr.sh_entsize); i++) {
	
		Elf32_Sym sym;
		if (readat(file, &sym, sizeof(Elf32_Sym), symhdr.sh_offset + i * sizeof(Elf32_Sym)) != sizeof(Elf32_Sym))
			continue;
			
		if (needswap) {
			mapswap(&sym.st_value, sizeof(Elf32_Addr));
			mapswap(&sym.st_size, sizeof(Elf32_Word));
			mapswap(&sym.st_name, sizeof(Elf32_Word));
		}
			
		if (sym.st_value <= addr && sym.st_value + sym.st_size > addr) {
			if (size <= 0)
				return size;
			int ret = readat(file, buffer, size, strhdr.sh_offset + sym.st_name);
			if (ret == size)
				buffer[--ret] = '\0';
			if (offset)
				*offset = addr - sym.st_value;
			return ret;
		}
	}
	return -1;
}

static int mapsym64(int file, unsigned long addr, char *buffer, int size, unsigned long *offset, int needswap) {
	// Read file header
	Elf64_Ehdr ehdr;
	if (readat(file, &ehdr, sizeof(Elf64_Ehdr), 0) != sizeof(Elf64_Ehdr))
		return -1;
		
	if (needswap) {
		mapswap(&ehdr.e_phnum, sizeof(Elf64_Half));
		mapswap(&ehdr.e_phoff, sizeof(Elf64_Off));
		mapswap(&ehdr.e_shoff, sizeof(Elf64_Off));
		mapswap(&ehdr.e_shstrndx, sizeof(Elf64_Half));
		mapswap(&ehdr.e_shnum, sizeof(Elf64_Half));
	}
	
	// Find program header for segment holding addr
	int i;
	for (i = 0; i < ehdr.e_phnum; i++) {

		// Get i'th program header
		Elf64_Phdr phdr;
		if (readat(file, &phdr, sizeof(Elf64_Phdr), ehdr.e_phoff + i * sizeof(Elf64_Phdr)) != sizeof(Elf64_Phdr))
			continue;
		
		if (needswap) {
			mapswap(&phdr.p_type, sizeof(Elf64_Word));
			mapswap(&phdr.p_flags, sizeof(Elf64_Word));
			mapswap(&phdr.p_offset, sizeof(Elf64_Off));
			mapswap(&phdr.p_filesz, sizeof(Elf64_Xword));
			mapswap(&phdr.p_vaddr, sizeof(Elf64_Addr));
		}
			
		if (phdr.p_type != PT_LOAD || !(phdr.p_flags & PF_X))
			continue; // only considere loadable and executable segments;

		if (phdr.p_offset > addr || phdr.p_offset + phdr.p_filesz <= addr)
			continue; // it's not a match

		addr += phdr.p_vaddr;
		break;
	}

	// Get section with section names
	Elf64_Shdr shstrhdr;
	if (readat(file, &shstrhdr, sizeof(Elf64_Shdr), ehdr.e_shoff + ehdr.e_shstrndx * sizeof(Elf64_Shdr)) != sizeof(Elf64_Shdr))
		return -1;

	if (needswap) {
		mapswap(&shstrhdr.sh_size, sizeof(Elf64_Xword));
		mapswap(&shstrhdr.sh_offset, sizeof(Elf64_Off));
	}

	// Get symbol table and symbol table entry strings
	Elf64_Shdr strhdr, symhdr;
	memset(&strhdr, 0, sizeof(Elf32_Shdr)); // not needed to initialize but tell that to the compiler
	memset(&symhdr, 0, sizeof(Elf32_Shdr));
	int havestrhdr = 0, havesymhdr = 0;
	for (i = 0; i < ehdr.e_shnum; i++) {

		// Get i'th section header
		Elf64_Shdr shdr;
		if (readat(file, &shdr, sizeof(Elf64_Shdr), ehdr.e_shoff + i * sizeof(Elf64_Shdr)) != sizeof(Elf64_Shdr))
			continue;

		if (needswap) {
			mapswap(&shdr.sh_name, sizeof(Elf64_Word));
			mapswap(&shdr.sh_size, sizeof(Elf64_Xword));
			mapswap(&shdr.sh_entsize, sizeof(Elf64_Xword));
			mapswap(&shdr.sh_offset, sizeof(Elf64_Off));
		}

		// Get section name
		char sname[9];
		sname[8] = '\0';
		if (shdr.sh_name >= shstrhdr.sh_size)
			continue;
		if (readat(file, sname, 8, shstrhdr.sh_offset + shdr.sh_name) <= 0)
			continue;

		// Check for .dynsym and .dynstr, optionally overruled by .symtab and .strtab
		if (!strcmp(sname, ".dynstr") && !havestrhdr) {
			memcpy(&strhdr, &shdr, sizeof(Elf64_Shdr));
			havestrhdr = 1;
		} else if (!strcmp(sname, ".dynsym") && !havesymhdr) {
			memcpy(&symhdr, &shdr, sizeof(Elf64_Shdr));
			havesymhdr = 1;
		} else if (!strcmp(sname, ".strtab")) {
			memcpy(&strhdr, &shdr, sizeof(Elf64_Shdr));
			havestrhdr = 1;
		} else if (!strcmp(sname, ".symtab")) {
			memcpy(&symhdr, &shdr, sizeof(Elf64_Shdr));
			havesymhdr = 1;
		}
	}
	if (!havestrhdr || !havesymhdr)
		return -1;

	// Find matching symbol
	for (i = 0; i < (int)(symhdr.sh_size / symhdr.sh_entsize); i++) {
	
		Elf64_Sym sym;
		if (readat(file, &sym, sizeof(Elf64_Sym), symhdr.sh_offset + i * sizeof(Elf64_Sym)) != sizeof(Elf64_Sym))
			continue;
			
		if (needswap) {
			mapswap(&sym.st_value, sizeof(Elf64_Addr));
			mapswap(&sym.st_size, sizeof(Elf64_Xword));
			mapswap(&sym.st_name, sizeof(Elf64_Word));
		}
			
		if (sym.st_value <= addr && sym.st_value + sym.st_size > addr) {
			if (size <= 0)
				return size;
			int ret = readat(file, buffer, size, strhdr.sh_offset + sym.st_name);
			if (ret == size)
				buffer[--ret] = '\0';
			if (offset)
				*offset = addr - sym.st_value;
			return ret;
		}
	}
	return -1;
}

static int mapsym(int file, unsigned long addr, char *buffer, int size, unsigned long *offset) {
	// Read file header
	Elf32_Ehdr ehdr;
	if (readat(file, &ehdr, sizeof(Elf32_Ehdr), 0) != sizeof(Elf32_Ehdr))
		return -1;
	if (memcmp(ehdr.e_ident, ELFMAG, SELFMAG))
		return -1;

	// Little endian / Big endian mismatch ?
	static union { char c[4]; uint32_t i; } lsbtest = { .i=1 };
	int needswap = ((ehdr.e_ident[EI_DATA] == ELFDATA2LSB) != lsbtest.c[0]);
	
	if (ehdr.e_ident[EI_CLASS] == ELFCLASS32)
		return mapsym32(file, addr, buffer, size, offset, needswap);
	else if (ehdr.e_ident[EI_CLASS] == ELFCLASS64)
		return mapsym64(file, addr, buffer, size, offset, needswap);
	else
		return -1;
}

int backtrace_prologue(map_t **maps, int *mapssize, int *usestaticmaps) {
	if (*usestaticmaps) {
		staticmapslock();
		*usestaticmaps = staticmapcount;
		*maps = staticmaps;
		if (!*usestaticmaps) {
			staticmapsunlock();
		}
	}
	if (!*usestaticmaps)
		return openmaps(*maps, mapssize);
	return -1;
}

void backtrace_epilogue(map_t *maps, int mapssize, int usestaticmaps, int mapsfd) {
	if (usestaticmaps)
		staticmapsunlock();
	else
		closemaps(mapsfd, maps, mapssize);
}

void backtrace_dumppriv(void *const *buffer, int size, int fd, int attributes, map_t *maps, int mapssize, int usestaticmaps, int mapsfd) {
	int i;
	for (i = 0; i < size && !(i && !buffer[i]); i++) {
		if (!(attributes & BACKTRACE_FORMAT_GLIBC))
			mafia_fdprintf(fd, "        0x%0*lx", 2 * sizeof(unsigned long), (unsigned long)buffer[i]);
		map_t *map = maplookup(maps, mapssize, buffer[i]);
		if (map) {
			char buf[BUFSIZE], *path = NULL;
			unsigned long addr = (unsigned long)(buffer[i] - map->start + map->offset);
			if (usestaticmaps) {
				path = staticmapcaptions[map - maps];
			} else {
				if (mappath(mapsfd, map, buf, BUFSIZE) > 0)
					path = buf;
			}
			if (path) {
				if (attributes & BACKTRACE_FORMAT_GLIBC)
					mafia_fdprintf(fd, "%s", path);
				else
					mafia_fdprintf(fd, " = %s + 0x%lx", path, addr);
				if (mapsym(mapfile(mapsfd, map), addr, buf, BUFSIZE, &addr) > 0) {
					if (attributes & BACKTRACE_FORMAT_GLIBC)
						mafia_fdprintf(fd, "(%s+0x%lx)", buf, addr);
					else
						mafia_fdprintf(fd, " = %s + 0x%lx", buf, addr);
				}
			}
		}
		if (attributes & BACKTRACE_FORMAT_GLIBC)
			mafia_fdprintf(fd, "[0x%0*lx]", 2 * sizeof(unsigned long), (unsigned long)buffer[i]);
		mafia_fdprintf(fd, "\n");
	}
}

void backtrace_dump(void *const *buffer, int size, int fd, int attributes) {
	map_t mymaps[STACKMAPSIZE], *maps = mymaps;
	int mapssize = STACKMAPSIZE, usestaticmaps = (attributes & BACKTRACE_USESTATICMAPS);
	int mapsfd = backtrace_prologue(&maps, &mapssize, &usestaticmaps);
	backtrace_dumppriv(buffer, size, fd, attributes, maps, mapssize, usestaticmaps, mapsfd);
	backtrace_epilogue(maps, mapssize, usestaticmaps, mapsfd);
}

void backtrace_loadstaticmaps(void) {
	staticmapslock();
	int count = staticmapcount;
	staticmapcount = 0;
	closemaps(-1, staticmaps, count);
	count = STATICMAPSIZE;
	int i, fd = openmaps(staticmaps, &count);
	for (i = 0; i < count; i++) {
		mapfile(fd, &staticmaps[i]);
		mapcaption(fd, &staticmaps[i], staticmapcaptions[i], CAPTIONSIZE);
	}
	staticmapcount = count;
	close(fd);
	staticmapsunlock();
}

void backtrace_unloadstaticmaps(void) {
	staticmapslock();
	closemaps(-1, staticmaps, staticmapcount);
	staticmapsunlock();
}

int backtrace_file(void *addr, char *buffer, int size, unsigned long *fileoffset, int attributes) {
	map_t mymaps[STACKMAPSIZE], *maps = mymaps;
	int mapssize = STACKMAPSIZE, usestaticmaps = (attributes & BACKTRACE_USESTATICMAPS);
	int mapsfd = backtrace_prologue(&maps, &mapssize, &usestaticmaps);
	map_t *map = maplookup(maps, mapssize, addr);
	int ret = mapcaption(mapsfd, map, buffer, size);
	if (fileoffset)
		*fileoffset = (unsigned long)(addr - map->start + map->offset);
	backtrace_epilogue(maps, mapssize, usestaticmaps, mapsfd);
	return ret;
}

int backtrace_sym(int elffd, unsigned long fileoffset, char *buffer, int size, unsigned long *symoffset) {
	return mapsym(elffd, fileoffset, buffer, size, symoffset);
}

#ifdef MIPS

void backtrace_symbols_fd(void *const *buffer, int size, int fd) {
	backtrace_dump(buffer, size, fd, BACKTRACE_FORMAT_GLIBC);
}

#endif

#ifdef WITHTHREADS
static void backtrace_forkprepare() {
	pthread_mutex_lock(&staticmapsmutex);
}

static void backtrace_forkparent() {
	pthread_mutex_unlock(&staticmapsmutex);
}

static void backtrace_forkchild() {
	pthread_mutex_init(&staticmapsmutex, NULL);
}

__attribute__((constructor))
static void init() {
	pthread_atfork(backtrace_forkprepare, backtrace_forkparent, backtrace_forkchild);
}
#endif
