/*
 * Copyright (c) 2011 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 * Author : Dries De Winter (dries.dewinter@softathome.com)
 *
 */
#include <string.h>
#ifdef WITHTHREADS
#include <pthread.h>
#endif

#include "mafia.h"
#include "common.h"


/*
 * calloc/malloc/realloc/free abstraction macros
 * The backend implementation of calloc/malloc/realloc/free is referred to internally as ~_impl.
 * The exported wrappers implemented here are referred to internally as ~_wrapper
 */
#define calloc_impl dlcalloc
#define malloc_impl dlmalloc
#define free_impl dlfree
#define realloc_impl dlrealloc

#define calloc_wrapper calloc
#define malloc_wrapper malloc
#define free_wrapper free
#define realloc_wrapper realloc

void *calloc_impl(size_t nmemb, size_t size);
void *malloc_impl(size_t size);
void free_impl(void *ptr);
void *realloc_impl(void *ptr, size_t size);

extern int heap_debugging_disabled;

/*
 * backtrace() is called from within malloc(). So if backtrace() calls malloc() itself for whatever reason,
 * we get stuck in an infinite recurse loop.
 * Our own backtrace() implementation for MIPS does not use malloc, so leave this mess behind in that case.
 * The backtrace() implementation of glibc for Intel x86 calls malloc, so AVOID RECURSION!
 */
#ifndef MIPS
__thread int heaprecurse = 0;
static int heap_enter() {
	if (heaprecurse) {
		//mafia_fdprintf(mafia_sink(), "Warning: recursive malloc/calloc/free/realloc detected\n");
		return 0;
	}
	return ++heaprecurse;
}
static void heap_leave() {
	--heaprecurse;
}
#else
#define heap_enter() 1
#define heap_leave()
#endif


/*
 * All memory blocks allocated on the heap are put in "envelopes" for bookkeeping.
 * Definition of envelope and utility methods below
 */
typedef struct heapgroup heapgroup_t;
typedef struct envelope envelope_t;

struct heapgroup {
	envelope_t *prv;
	envelope_t *nxt;
	int counter;
	size_t payload;
};

struct envelope {
	envelope_t *prv;
	envelope_t *nxt;
	unsigned long magic;
	size_t payload;
	void *bt[HEAPBTSIZE];
	heapgroup_t grp;
	char data[0];
};

static struct {
	envelope_t *last;
	envelope_t *first;
} _heap = { (envelope_t *)&_heap, (envelope_t *)&_heap };

static envelope_t *heap = (envelope_t *)&_heap;

#ifdef WITHTHREADS
static pthread_mutex_t heapmutex = PTHREAD_MUTEX_INITIALIZER;

static void heap_lock() {
	pthread_mutex_lock(&heapmutex);
}

static void heap_unlock() {
	pthread_mutex_unlock(&heapmutex);
}
#else
#define heap_lock()
#define heap_unlock()
#endif

static void _heap_link(envelope_t *envelope) {
	envelope->nxt = heap;
	envelope->prv = heap->prv;
	envelope->nxt->prv = envelope->prv->nxt = envelope;
}

static void _heap_unlink(envelope_t *envelope) {
	envelope->prv->nxt = envelope->nxt;
	envelope->nxt->prv = envelope->prv;
	envelope->prv = envelope->nxt = NULL;
	envelope->magic = 0x44445701;
}

static void heap_link(envelope_t *envelope) {
	heap_lock();
	_heap_link(envelope);
	heap_unlock();
}

static void heap_unlink(envelope_t *envelope) {
	heap_lock();
	_heap_unlink(envelope);
	heap_unlock();
}


/*
 * Public functions: wrappers for malloc, calloc, free and realloc
 */
void *malloc_wrapper(size_t size) {
	if (heap_debugging_disabled)
		return malloc_impl(size);
	if (!heap_enter())
		return malloc_impl(size);
	envelope_t *envelope = malloc_impl(sizeof(envelope_t) + size);
	if (!envelope)
		goto leave;
	envelope->magic = 0x44445700;
	envelope->payload = size;
	memset(envelope->bt, 0, HEAPBTSIZE * sizeof(void *));
	backtrace(envelope->bt, HEAPBTSIZE);
	heap_link(envelope);
leave:
	heap_leave();
	return envelope?envelope->data:NULL;
}

void *calloc_wrapper(size_t nmemb, size_t size) {
	if (heap_debugging_disabled)
		return calloc_impl(nmemb, size);
	if (!heap_enter())
		return calloc_impl(nmemb, size);
	envelope_t *envelope = malloc_impl(sizeof(envelope_t) + nmemb * size);
	if (!envelope)
		goto leave;
	envelope->magic = 0x44445700;
	envelope->payload = nmemb * size;
	memset(envelope->bt, 0, HEAPBTSIZE * sizeof(void *));
	backtrace(envelope->bt, HEAPBTSIZE);
	heap_link(envelope);
	memset(envelope->data, 0, nmemb * size);
leave:
	heap_leave();
	return envelope?envelope->data:NULL;
}

static const char *check_envelope(envelope_t *envelope) {
	if (envelope->prv && envelope->nxt && envelope->magic == 0x44445700)
		return NULL;
	if (envelope->magic == 0x44445701)
		return "Double";
	else
		return "Invalid";
}

static void report_error(const char *adj, envelope_t *envelope, void *bt[], int btlen) {
	mafia_fdprintf(mafia_sink(), "%s free(%p) detected at\n", adj, envelope->data);
	backtrace_dump(bt, btlen, mafia_sink(), 0);
	if (!strcmp(adj, "Double")) {
		mafia_fdprintf(mafia_sink(), "%p was already freed by\n", envelope->data);
		backtrace_dump(envelope->bt, HEAPBTSIZE, mafia_sink(), BACKTRACE_USESTATICMAPS);
	}
	mafia_fdprintf(mafia_sink(), "\n");
	fsync(mafia_sink());
}

void free_wrapper(void *ptr) {
	if (heap_debugging_disabled) {
		free_impl(ptr);
		return;
	}
	if (!heap_enter()) {
		free_impl(ptr);
		return;
	}
	if (!ptr)
		goto leave;
	envelope_t *envelope = (envelope_t *)(ptr - sizeof(envelope_t));
	const char *adj;
	if ((adj = check_envelope(envelope))) {
		void *bt[REGULARBTSIZE];
		report_error(adj, envelope, bt, backtrace(bt, REGULARBTSIZE));
		goto leave;
	}
	heap_unlink(envelope);
	memset(envelope->bt, 0, HEAPBTSIZE * sizeof(void *));
	backtrace(envelope->bt, HEAPBTSIZE);
	free_impl(envelope);
leave:
	heap_leave();
}

void *realloc_wrapper(void *ptr, size_t size) {
	if (heap_debugging_disabled)
		return realloc_impl(ptr, size);
	if (!heap_enter())
		return realloc_impl(ptr, size);
	envelope_t *envelope = ptr ? (envelope_t *)(ptr - sizeof(envelope_t)) : NULL;
	if (envelope) {
		const char *adj;
		if ((adj = check_envelope(envelope))) {
			void *bt[REGULARBTSIZE];
			report_error(adj, envelope, bt, backtrace(bt, REGULARBTSIZE));
			envelope = NULL;
		} else {
			heap_unlink(envelope);
			memset(envelope->bt, 0, HEAPBTSIZE * sizeof(void *));
			backtrace(envelope->bt, HEAPBTSIZE);
		}
	}
	envelope = realloc_impl(envelope, sizeof(envelope_t) + size);
	if (!envelope)
		goto leave;
	envelope->magic = 0x44445700;
	envelope->payload = size;
	memset(envelope->bt, 0, HEAPBTSIZE * sizeof(void *));
	backtrace(envelope->bt, HEAPBTSIZE);
	heap_link(envelope);
leave:
	heap_leave();
	return envelope->data;
}



/*
 * Dump the heap on stderr:
 * 1) Summarize the heap: group blocks together based on backtrace information and make subtotals.
 * 2) Sort the summary so biggest blocks occur at the end of the output
 * 3) Import memory maps for address resolving
 * 4) Loop over summary, resolve symbols if possible and print on stderr
 */
static void heap_sortgroups(envelope_t *first, envelope_t *last) {
	if (first == last) // zero or one element -> nothing to be sorted
		return;
	envelope_t *pivot = first; // summary is distributed randomly -> using first element as pivot is OK.
	envelope_t *cur, *nxt = first->grp.nxt, *end = last;
	first = last = NULL;
	do {
		cur = nxt;
		nxt = cur->grp.nxt;
		if (cur->grp.payload < pivot->grp.payload) {
			cur->grp.prv->grp.nxt = cur->grp.nxt;
			cur->grp.nxt->grp.prv = cur->grp.prv;
			cur->grp.prv = pivot->grp.prv;
			cur->grp.nxt = pivot;
			cur->grp.prv->grp.nxt = cur->grp.nxt->grp.prv = cur;
			if (!first)
				first = cur;
		} else {
			last = cur;
		}
	} while (cur != end);
	if (first)
		heap_sortgroups(first, pivot->grp.prv);
	if (last)
		heap_sortgroups(pivot->grp.nxt, last);
}

void heap_dump(int fd, int attributes) {
	size_t payload = 0;
	int counter = 0;
	envelope_t _summary = { .grp = { .prv = &_summary, .nxt = &_summary } }, *summary = &_summary, *envelope, *group;

	heap_lock();
	// summarize the heap: calculate subtotal per occurring backtrace.
	for (envelope = heap; (envelope = envelope->nxt) != heap; ) {
		for (group = summary; (group = group->grp.nxt) != summary; ) {
			if (!memcmp(group->bt, envelope->bt, HEAPBTSIZE * sizeof(void *)))
				break;
		}
		if (group == summary) {
			group = envelope;
			memset(&group->grp, 0, sizeof(heapgroup_t));
			group->grp.prv = summary->grp.prv;
			group->grp.nxt = summary;
			group->grp.prv->grp.nxt = group->grp.nxt->grp.prv = group;
		}
		group->grp.counter++;
		group->grp.payload += envelope->payload;
		counter++;
		payload += envelope->payload;
	}

	// quickly sort the summary using quicksort
	heap_sortgroups(summary->grp.nxt, summary->grp.prv);

	// build temporary maps or use static maps
	map_t mymaps[STACKMAPSIZE], *maps = mymaps;
	int mapssize = STACKMAPSIZE, usestaticmaps = (attributes & BACKTRACE_USESTATICMAPS);
	int mapsfd = backtrace_prologue(&maps, &mapssize, &usestaticmaps);

	// publish on stderr
	mafia_fdprintf(fd, "\n");
	for (group = summary; (group = group->grp.nxt) != summary; ) {
		mafia_fdprintf(fd, "%d bytes (%d.%02d%% of total heap) allocated in %d blocks by\n", group->grp.payload,
				group->grp.payload * 100 / payload, group->grp.payload * 10000 / payload % 100, group->grp.counter);
		backtrace_dumppriv(group->bt, HEAPBTSIZE, fd, attributes, maps, mapssize, usestaticmaps, mapsfd);
		mafia_fdprintf(fd, "\n");
	}
	mafia_fdprintf(fd, "Total: %d bytes allocated in %d blocks\n\n", payload, counter);

	backtrace_epilogue(maps, mapssize, usestaticmaps, mapsfd);

	heap_unlock();
}

void heap_count(size_t *payload, int *blocks) {
	heap_lock();
	envelope_t *envelope;
	for (envelope = heap; (envelope = envelope->nxt) != heap; ) {
		if (blocks)
			(*blocks)++;
		if (payload)
			(*payload) += envelope->payload;
	}
	heap_unlock();
}


#ifdef WITHTHREADS
static void heap_forkprepare() {
	pthread_mutex_lock(&heapmutex);
}

static void heap_forkparent() {
	pthread_mutex_unlock(&heapmutex);
}

static void heap_forkchild() {
	pthread_mutex_init(&heapmutex, NULL);
}

__attribute__((constructor))
static void init() {
	pthread_atfork(heap_forkprepare, heap_forkparent, heap_forkchild);
}
#endif
