/*
 * Copyright (c) 2011 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 * Author : Dries De Winter (dries.dewinter@softathome.com)
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#ifdef WITHTHREADS
#include <pthread.h>
#endif

#include "mafia.h"
#include "common.h"

static int sink = -1;

#ifdef HEAP_DISABLED
void heap_dump(int fd, int attributes) {
	(void)attributes;
	mafia_fdprintf(fd, "        No heap information available - heap debugging is disabled at compile time.\n");
	mafia_fdprintf(fd, "        Enable heap debugging by setting CONFIG_SAH_LIB_MAFIA_HEAP=y in the config\n\n");
}

void heap_count(size_t *payload, int *blocks) {
	*payload = 0;
	*blocks = 0;
}
#endif

int heap_debugging_disabled = 1;

void heap_enable_debugging() {
	heap_debugging_disabled = 0;
}

int heap_debugging_enabled() {
	return !heap_debugging_disabled;
}

typedef enum exitbacktrace {
	exitbacktrace_never,
	exitbacktrace_always,
	exitbacktrace_nonzero,
} exitbacktrace_t;

static exitbacktrace_t exitbacktrace = exitbacktrace_never;

void mafia_fdprintf(int fd, const char *fmt, ...) {
	va_list args;
	va_start(args, fmt);
	char buf[BUFSIZE];
	vsnprintf(buf, BUFSIZE, fmt, args);
	va_end(args);
	if (write(fd, buf, strlen(buf)) < 0)
		return;
}

static void mafia_signalhandler(int signum, siginfo_t *info, void *ptr) {
	void *bt[REGULARBTSIZE];
	const char *codestr;
	mafia_fdprintf(sink, "[%d] ", getpid());
	if (signum == SIGRTMAX)
		mafia_fdprintf(sink, "Backtrace requested by user:\n");
	else switch (signum) {
	case SIGSEGV:
		switch (info->si_code) {
		case SEGV_MAPERR: codestr = "Address not mapped"; break;
		case SEGV_ACCERR: codestr = "Permission denied"; break;
		default: codestr = "Unknown reason"; break;
		}
		mafia_fdprintf(sink, "Segmentation fault for address %p: %s\n", info->si_addr, codestr);
		break;
	case SIGFPE:
		switch (info->si_code) {
		case FPE_INTDIV: codestr = "Integer divide by zero"; break;
		case FPE_INTOVF: codestr = "Integer overflow"; break;
		case FPE_FLTDIV: codestr = "Floating-point divide by zero"; break;
		case FPE_FLTOVF: codestr = "Floating-point overflow"; break;
		case FPE_FLTUND: codestr = "Floating-point underflow"; break;
		case FPE_FLTRES: codestr = "Floating-point inexact result"; break;
		case FPE_FLTINV: codestr = "Floating-point invalid operation"; break;
		case FPE_FLTSUB: codestr = "Subscript out of range"; break;
		default: codestr = "Unknown reason"; break;
		}
		mafia_fdprintf(sink, "Floating point exception: %s\n", codestr);
		break;
	case SIGBUS:
		switch (info->si_code) {
		case BUS_ADRALN: codestr = "Invalid address alignment"; break;
		case BUS_ADRERR: codestr = "Nonexistent physical address"; break;
		case BUS_OBJERR: codestr = "Object-specific hardware error"; break;
		default: codestr = "Unknown reason"; break;
		}
		mafia_fdprintf(sink, "Bus error for address %p: %s\n", info->si_addr, codestr);
		break;
	case SIGILL:
		switch (info->si_code) {
		case ILL_ILLOPC: codestr = "Illegal opcode"; break;
		case ILL_ILLOPN: codestr = "Illegal operand"; break;
		case ILL_ILLADR: codestr = "Illegal addressing mode"; break;
		case ILL_ILLTRP: codestr = "Illegal trap"; break;
		case ILL_PRVOPC: codestr = "Privileged opcode"; break;
		case ILL_PRVREG: codestr = "Privileged register"; break;
		case ILL_COPROC: codestr = "Coprocessor error"; break;
		case ILL_BADSTK: codestr = "Internal stack error"; break;
		default: codestr = "Unknown reason"; break;
		}
		mafia_fdprintf(sink, "Illegal instruction: %s\n", codestr);
		break;
	case SIGABRT:
		mafia_fdprintf(sink, "Process aborted:\n");
		break;
	}
	backtrace_dump(bt, backtrace_signal(bt, REGULARBTSIZE, ptr), sink, 0);
	mafia_fdprintf(sink, "\n");
	fsync(sink);
	if (signum != SIGRTMAX) {
		exitbacktrace = exitbacktrace_never;
		exit(-1);
	}
}

static void mafia_sigrt1handler(int signum) {
	(void)signum;
	if (heap_debugging_enabled()) {
		mafia_fdprintf(sink, "Heap snapshot requested by user:\n");
		heap_dump(sink, BACKTRACE_USESTATICMAPS);
	} else {
		mafia_fdprintf(sink, "        No heap information available - heap debugging is disabled.\n");
		mafia_fdprintf(sink, "        Enable heap debugging by putting HEAPDEBUG=1 in the environment\n\n");
	}
	fsync(sink);
}

static void mafia_sigrt2handler(int signum) {
	(void)signum;
	backtrace_loadstaticmaps();
}

/*
 * If WITHTHREADS is set, a seperate thread is started that handles SIGRTMAX-1 and SIGRTMAX-2:
 */
#ifdef WITHTHREADS

static pthread_t thread;

static int signalpipe[2];

static void *mafia_run(void *arg) {
	sigset_t *sigset = (sigset_t *)arg;
	pthread_sigmask(SIG_UNBLOCK, sigset, NULL);
	int signum;
	for (;;) {
		ssize_t n = read(signalpipe[0], &signum, sizeof(int));
		if (n == -1)
			continue; // assume interrupted by signal
		else if (n != sizeof(int))
			break; // short byte count -> out of sync, assume this never happens.
		if (signum == SIGRTMAX - 1)
			mafia_sigrt1handler(signum);
		else if (signum == SIGRTMAX - 2)
			mafia_sigrt2handler(signum);
	}
	return NULL;
}

static void mafia_signalpusher(int signum) {
	if (write(signalpipe[1], &signum, sizeof(int)) < 0) {
		mafia_fdprintf(sink, "write() failed: %m\n");
	}
}
#endif

static void mafia_exithook(int status, void *arg) {
	(void)arg;
	switch (exitbacktrace) {
	case exitbacktrace_never:
		break;
	case exitbacktrace_nonzero:
		if (!status)
			break;
	case exitbacktrace_always:
		{
			void *bt[REGULARBTSIZE];
			mafia_fdprintf(sink, "Exit with code %d at\n", status);
			backtrace_dump(bt, backtrace(bt, REGULARBTSIZE), sink, 0);
			mafia_fdprintf(sink, "\n");
		}
	}
}

__attribute__((constructor))
static void mafia_init() {
	const char *heapdebug = getenv("HEAPDEBUG");
	if (heapdebug && (*heapdebug == '1' || *heapdebug == 'y' || *heapdebug == 'Y'))
		heap_enable_debugging();

	const char *mafiasink = getenv("MAFIASINK");
	if (mafiasink && *mafiasink) {
		sink = open(mafiasink, O_WRONLY | O_CREAT | O_APPEND, S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);
	} else {
		sink = dup(fileno(stderr));
	}

	struct sigaction action = {.sa_sigaction=mafia_signalhandler, .sa_flags=SA_SIGINFO};
	sigaction(SIGRTMAX, &action, NULL);
	sigaction(SIGSEGV, &action, NULL);
	sigaction(SIGFPE, &action, NULL);
	sigaction(SIGBUS, &action, NULL);
	sigaction(SIGILL, &action, NULL);
	sigaction(SIGABRT, &action, NULL);

	if (heap_debugging_enabled()) {
		action.sa_flags = 0;
#ifdef NOTHREADS
		action.sa_handler = mafia_sigrt1handler;
		sigaction(SIGRTMAX-1, &action, NULL);
		action.sa_handler = mafia_sigrt2handler;
		sigaction(SIGRTMAX-2, &action, NULL);
#else
		action.sa_handler = mafia_signalpusher;
		sigaction(SIGRTMAX-1, &action, NULL);
		sigaction(SIGRTMAX-2, &action, NULL);
		static sigset_t sigset;
		sigemptyset(&sigset);
		sigaddset(&sigset, SIGRTMAX-1);
		sigaddset(&sigset, SIGRTMAX-2);
		if (pipe(signalpipe) < 0)
			mafia_fdprintf(sink, "pipe() failed: %m\n");
		pthread_sigmask(SIG_BLOCK, &sigset, NULL);
		pthread_create(&thread, NULL, mafia_run, &sigset);
		// XXX After fork, signal handling thread is lost. Create new signal handling thread in child callback of pthread_atfork? It doesn't seem to work.
#endif
	}

	const char *exitbacktracestr = getenv("EXITBACKTRACE")?:"";
	exitbacktrace =
			!strcasecmp(exitbacktracestr, "never") ? exitbacktrace_never :
			!strcasecmp(exitbacktracestr, "always") ? exitbacktrace_always :
			!strcasecmp(exitbacktracestr, "nonzero") ? exitbacktrace_nonzero : /* default */exitbacktrace_never;
	on_exit(mafia_exithook, NULL);
}

__attribute__((destructor))
static void mafia_cleanup() {
#ifdef WITHTHREADS
	pthread_cancel(thread);
	//pthread_join(thread); // deadlocks on the board.
	close(signalpipe[0]);
	close(signalpipe[1]);
#endif
	/*
	 * Dump the heap at process exit. It should be empty. However, we can not distinguish between leaked and still reachable memory
	 */
	if (heap_debugging_enabled()) {
		mafia_fdprintf(sink, "Heap snapshot at exit:\n");
		heap_dump(sink, BACKTRACE_USESTATICMAPS);
	}
	backtrace_unloadstaticmaps();
}

int mafia_sink() {
	return sink;
}



