/*
 * Copyright (c) 2010 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 */
#ifndef __NEMO_CLIENT_DHCP_H__
#define __NEMO_CLIENT_DHCP_H__

/**
@ingroup client
@defgroup dhcp dhcp.h - DHCP API
@{

@brief
DHCP API is an extension on top of the Common API (see @ref common_api) to facilitate out-of-process access to the DHCP API MIB.
*/

#include <stdbool.h>
#include <stdint.h>

#include <pcb/utils.h>

#include <nemo/common_api.h>

/**
@brief
Finds and parses the option value given by 'type' and 'tag' of any bound DHCP client Intf occurring in the traverse tree.

@param intf Starting point of the traverse tree. Default is "lo".
@param type The type of option to be queried. Possibilities are:
                - "req": consult the ReqOption table of DHCPv4 clients.
                - "sent": consult the SentOption table of DHCPv4 clients.
                - "req6": consult the ReceivedOption table of DHCPv6 clients.
                - "sent6": consult the SentOption table of DHCPv6 clients.
            This argument is mandatory.
@param tag The option tag.
@param traverse Traverse mode to build the traverse tree. Default is "down".

@return
A structured variant holding the option value. Its exact structure depends on the type and tag arguments: each option
has its particular encoding scheme. Example: if the type equals "req" and the tag equals 3 (IP Routers Option), the return value is
a variant list. Each item in the list is a string representing an IPv4-Address. Supported options for DHCPv4 are:
1-77, 90, 120, 121, 124 and 125 and for DHCPv6: 1-9, 11-34, 36-48, 51, 58 and 64. If an unsupported option tag is given as input
argument, a default encoding scheme is applied. This encoding scheme returns an ASCII-string if the option value contains valid
ASCII-characters only or the raw hexbinary prefixed with "0x" otherwise.

@warning
The returned variant is allocated on the heap, so it MUST be cleaned up AND freed by
caller of the function with variant_cleanup(), followed by free().
*/
variant_t *nemo_getDHCPOption(const char *intf, const char *type, uint16_t tag, const char *traverse);

/**
@brief
nemo_openQuery() wrapper for querying the function nemo_getDHCPOption().

@param intf see nemo_openQuery() and/or nemo_getDHCPOption()
@param subscriber see nemo_openQuery()
@param type see nemo_getDHCPOption()
@param tag see nemo_getDHCPOption()
@param traverse see nemo_getDHCPOption()
@param handler see nemo_openQuery()
@param userdata see nemo_openQuery()
@return see nemo_openQuery()
*/
nemo_query_t *nemo_openQuery_getDHCPOption(const char *intf, const char *subscriber, const char *type, uint16_t tag,
                                           const char *traverse, nemo_result_handler_t handler, void *userdata);

/**
@}
*/

#endif

