/*
 * Copyright (c) 2011 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 */
#ifndef __NEMO_CLIENT_NETDEV_H__
#define __NEMO_CLIENT_NETDEV_H__

/**
@ingroup client
@defgroup netdev netdev.h - NetDev API
@{

@brief
NetDev API is an extension on top of the Common API (see @ref common_api) to facilitate out-of-process access to the NetDev API MIB.
*/

#include <stdbool.h>
#include <stdint.h>

#include <pcb/utils.h>
#include <pcb/core.h>

#include <nemo/common_api.h>

/**
@brief
Get all active Addr objects (Status is bound or dynamic) that occur in the traverse tree.

@param intf Starting point of the traverse tree. Default is "lo".
@param flag Only Addrs for which this flag expression evaluates to true are returned.
            An Addr's flag set is defined by the union of the Flags parameter, the Scope parameter and the address family
            ("ipv4" or "ipv6"), so you could call nemo_getAddrs(NULL, "ipv6 && global", "all") to get all global IPv6 addresses
            for example. Default is empty.
@param traverse Traverse mode to build the traverse tree. Default is "down".
@return A list of object maps. An object map is a map that maps parameter names to parameter values.

@warning
The returned variant_list is allocated on the heap, so it MUST be cleaned up AND freed by
caller of the function with variant_list_cleanup(), followed by free().
*/
variant_list_t *nemo_getAddrs(const char *intf, const char *flag, const char *traverse);

/**
@brief
Get the first active Addr (Status is bound or dynamic) that occurs in the traverse tree.

@param intf Starting point of the traverse tree. Default is "lo".
@param flag Only Addrs for which this flag expression evaluates to true are considered.
            An Addr's flag set is defined by the union of the Flags parameter, the Scope parameter and the address family
            ("ipv4" or "ipv6"), so you could call nemo_luckyAddr(NULL, "ipv6 && global", "all") to get any global IPv6
            address for example. Default is empty.
@param traverse Traverse mode to build the traverse tree. Default is "down".
@return An object map. An object map is a map that maps parameter names to parameter values.

@warning
The returned variant_map is allocated on the heap, so it MUST be cleaned up AND freed by
caller of the function with variant_map_cleanup(), followed by free().
*/
variant_map_t *nemo_luckyAddr(const char *intf, const char *flag, const char *traverse);

/**
@brief
Get the only the Address parameter of the first active Addr (Status is bound or dynamic) that occurs in the traverse tree.

@param intf Starting point of the traverse tree. Default is "lo".
@param flag Only Addrs for which this flag expression evaluates to true are considered.
            An Addr's flag set is defined by the union of the Flags parameter, the Scope parameter and the address family
            ("ipv4" or "ipv6"), so you could call nemo_luckyAddrAddress(NULL, "ipv6 && global", "all") to get any global IPv6
            address for example. Default is empty.
@param traverse Traverse mode to build the traverse tree. Default is "down".
@return An IP-address in standard notation.

@warning
The returned string is allocated on the heap, so it MUST be freed by
caller of the function with free().
*/
char *nemo_luckyAddrAddress(const char *intf, const char *flag, const char *traverse);

/**
@brief
nemo_openQuery() wrapper for querying the function nemo_getAddrs().

@details
nemo_openQuery() wrapper for querying the function nemo_getAddrs().

@param intf see nemo_openQuery() and/or nemo_getAddrs()
@param subscriber see nemo_openQuery()
@param flag see nemo_getAddrs()
@param traverse see nemo_getAddrs()
@param handler see nemo_openQuery()
@param userdata see nemo_openQuery()
@return see nemo_openQuery()
*/
nemo_query_t *nemo_openQuery_getAddrs(const char *intf, const char *subscriber, const char *flag,
                                      const char *traverse, nemo_result_handler_t handler, void *userdata);

/**
@brief
nemo_openQuery() wrapper for querying the function nemo_luckyAddr().

@param intf see nemo_openQuery() and/or nemo_luckyAddr()
@param subscriber see nemo_openQuery()
@param flag see nemo_luckyAddr()
@param traverse see nemo_luckyAddr()
@param handler see nemo_openQuery()
@param userdata see nemo_openQuery()
@return see nemo_openQuery()
*/
nemo_query_t *nemo_openQuery_luckyAddr(const char *intf, const char *subscriber, const char *flag,
                                       const char *traverse, nemo_result_handler_t handler, void *userdata);

/**
@brief
nemo_openQuery() wrapper for querying the function nemo_luckyAddrAddress().

@param intf see nemo_openQuery() and/or nemo_luckyAddrAddress()
@param subscriber see nemo_openQuery()
@param flag see nemo_luckyAddrAddress()
@param traverse see nemo_luckyAddrAddress()
@param handler see nemo_openQuery()
@param userdata see nemo_openQuery()
@return see nemo_openQuery()
*/
nemo_query_t *nemo_openQuery_luckyAddrAddress(const char *intf, const char *subscriber, const char *flag,
                                              const char *traverse, nemo_result_handler_t handler, void *userdata);

/**
@}
*/

#endif
