/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <debug/sahtrace.h>

#include <pcb/common/error.h>

#if defined(TLS_AVAILABLE)
__thread uint32_t pcb_error = pcb_ok;
#else
uint32_t pcb_error = pcb_ok;
#endif

/**
@file
@brief
Implementation of error functions
*/

static const char *pcb_common_errors[] = {
    "Common error",
    "Invalid function parameter",
    "Out of memory",
    "Unknown system error",
};

static const char *pcb_utils_errors[] = {
    "utils error",
    "Can not insert an element in a list before the beginning of the list",
    "Can not insert an element in a list after the end of the list",
    "Can not remove begin or end of the list",
    "Ourside the bounderaries of a string or a list",
    "Substring not found",
    "The list is empty",
    "Invalid variant type specified",
    "Can not convert the variant to the requested type",
    "Element not found in the list",
};

static const char *pcb_core_errors[] = {
    "Core error",
    "Object or parameter name can not be empty",
    "Operation not allowed on this kind of object",
    "Object or parameter is in the wrong state for the operation",
    "Object has no parent (is root object)",
    "Object is not a template object",
    "Maximum possible instances of the object is reached",
    "Object was the last in the instance list",
    "Object is not an instance",
    "Parameter was the last in the parameter list",
    "Object or parameter not found",
    "Object can not be committed as long as it's parent is not committed",
    "Object or parameter has no unique name within it's containing object",
    "Object or parameter has no valid name",
    "Parameter type is not known",
    "Mandatory parameters or key parameters can not be empty",
    "Parameter value is not in the list of possible values",
    "Parameter value is not in the specified value range",
    "String parameter's string length is not in the specified range",
    "Parameter value is below the specified minimum",
    "Parameter value is above the specified maximum",
    "Validator type is not matching the validation function",
    "The parameter or validator values can not be converted to the correct type",
    "Object or parameter is read-only",
    "Invalid notification type",
    "Object is of the wrong kind for this operation",
    "Error last function",
    "Function can not be executed for the specified object",
    "Function is not implemented",
    "Operation is canceled",
    "Invalid parameter value",
    "Function execution failed",
    "Mandatory argument missing",
};

static const char *pcb_communication_errors[] = {
    "Communication error",
    "Already listening",
    "Already connected",
    "Not connected",
    "Connection is closed by peer",
    "The datamodel does not belong to a connection",
    "Invalid message header received",
    "Invalid or unknown data format",
    "Invalid or unknown request",
    "Request is not supported",
    "Invalid reply",
    "No (de)serializer available",
};

/**
@ingroup pcb_common_error
@{
*/

//---------------------------------------------------------------------------------------------
/**
@brief
Convert an error code into a human readable string.

@details
This function converts a error code into a human readable string.

@param error the error code

@return
    - a human readable string
*/
const char *error_string(uint32_t error) {
    switch(PCB_ERROR_GROUP(error)) {
        case pcb_error_group_system:
            return strerror(error);
        break;
        case pcb_error_group_common:
        if (error >= pcb_error_common_max)
        return "Unknown error in group \"common\"";
            else
            return pcb_common_errors[error & 0x0000FFFF];
        break;
        case pcb_error_group_utils:
        if (error >= pcb_error_utils_max)
        return "Unknown error in group \"utils\"";
            else
            return pcb_utils_errors[error & 0X0000FFFF];
        break;
        case pcb_error_group_core:
        if (error >= pcb_error_core_max)
        return "Unknown error in group \"core\"";
            else
            return pcb_core_errors[error & 0x0000FFFF];
        break;
        case pcb_error_group_communication:
        if (error >= pcb_error_communication_max)
        return "Unknown error in group \"communication\"";
            else
            return pcb_communication_errors[error & 0x0000FFFF];
        break;
        case pcb_error_group_getaddrinfo:
            return gai_strerror(-(error & 0x0000FFFF));
        default:
            SAH_TRACEZ_WARNING("pcb_com","Unknown error %d",error);
            return "Unknown error group";
    }
}

/**
@}
*/
