/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <stdlib.h>

#include <debug/sahtrace.h>
#include <pcb/common/error.h>
#include <pcb/utils/variant_map.h>
#include "parameter_priv.h"
#include "validator_priv.h"
#include "core_priv.h"


/**
@file
@brief
Implementation of the private parameter functions
*/

//---------------------------------------------------------------------------------------------
/**
@brief
Destroy the specified parameter

@details
This function will cleanup and free the complete parameter datastructure.


@param parameter The parameter we want to destroy

@return This function always returns true
*/
bool parameter_destroy(parameter_t *parameter) {
    // call the delete handler
    if (parameter->definition->handlers && parameter->definition->handlers->destroy) {
        parameter->definition->handlers->destroy(parameter);
    }

    parameter_def_t *def = parameter->definition;

    llist_iterator_take(&parameter->it);
    if (parameter->map) {
        free(parameter->map->parameterName);
        variant_map_cleanup(&parameter->map->parameters);
        free(parameter->map);
    }

    variant_cleanup(&parameter->value[1]);
    variant_cleanup(&parameter->value[0]);
    aclDestroy(&parameter->ACL);

    parameter->definition->refcount--;
    if (parameter->definition->refcount == 0) {
        free(parameter->definition->handlers);
        param_validator_destroy(parameter->definition->validator);
        if (parameter->definition->name != (char *)(((void *)parameter->definition) + sizeof(parameter_def_t) + sizeof(parameter_t))) {
            free(parameter->definition->name);
        }

#ifdef PCB_HELP_SUPPORT
        free(parameter->definition->description);
#endif
        free(parameter->definition);
    }

    if (parameter != (((void *)def) + sizeof(parameter_def_t))) {
        free(parameter);
    }
    return true;
}

//---------------------------------------------------------------------------------------------
/**
@brief
Convert the parameter type to a variant type

@details
This function converts a parameter type enumeration (@ref parameter_type_t) into a variant type enumeration (@ref variant_type_t)\n

@param paramType The parameter type we want to translate

@return The corresponding variant type, variant_type_unknown if a conversion is not possible.
*/
variant_type_t parameter_type_to_variant_type(parameter_type_t paramType) {
    switch(paramType) {
        case parameter_type_unknown:
            return variant_type_unknown;
        case parameter_type_string:
        case parameter_type_reference:
            return variant_type_string;
        case parameter_type_int8:
            return variant_type_int8;
        case parameter_type_int16:
            return variant_type_int16;
        case parameter_type_int32:
            return variant_type_int32;
        case parameter_type_int64:
            return variant_type_int64;
        case parameter_type_uint8:
            return variant_type_uint8;
        case parameter_type_uint16:
            return variant_type_uint16;
        case parameter_type_uint32:
            return variant_type_uint32;
        case parameter_type_uint64:
            return variant_type_uint64;
        case parameter_type_bool:
            return variant_type_bool;
        case parameter_type_date_time:
            return variant_type_date_time;
        default:
            return variant_type_unknown;
    }
}

//---------------------------------------------------------------------------------------------
/**
@brief
Get a pointer to the parameter handlers

@details
This function returns a pointer to the parameter handlers structure of this parameter

@param parameter The parameter in which we are interested

@return
    - The parameter handler function
    - NULL if an error occurred (out of memory)
*/

parameter_handlers_t *parameter_getHandlers(parameter_t *parameter) {
    if (!parameter->definition->handlers) {
        parameter->definition->handlers = (parameter_handlers_t *)calloc(1,sizeof(parameter_handlers_t));
        if (!parameter->definition->handlers) {
            pcb_error = pcb_error_no_memory;
            return NULL;
        }
    }

    return parameter->definition->handlers;
}

/**
@}
*/
