/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#if !defined(PCB_PARAMETER_PRIV_H)
#define PCB_PARAMETER_PRIV_H

#ifdef __cplusplus
extern "C"
{
#endif

#include <pcb/core/parameter.h>
#include <pcb/utils/linked_list.h>
#include <pcb/utils/string.h>
#include <pcb/utils/variant.h>
#include <pcb/core/mapping.h>

/**
@file
@brief
Header file with private parameter type definitions and private parameter functions
*/

/**
@ingroup pcb_core
@defgroup pcb_core_private_parameter Private parameters
@{

@brief
Parameter implementation

@details

This file contains the private parameter structures and pricate parameter functions.\n

*/

//---------------------------------------------------------------------------------------------
/**
@brief
Parameter handler structure

@details
This structure contains the parameter handler functions: destroy, read and write
*/
typedef struct _parameter_handler {
    parameter_destroy_handler_t destroy;           /**< Called when the object is going to be destroyed,this function must succeed, after this handler returns the object pointer is not valid anymore */
    parameter_read_handler_t read;                 /**< Called in the @ref object_update function, if this function is set the object is volatile. If the this handler fails, the update function will fail */
    parameter_write_handler_t write;               /**< Called when the object has been written (after the commit), this function must succeed */
    parameter_translate_value_handler_t translate; /**< Value translation function */
} parameter_handlers_t;

/**
@brief
Destination parameter structure
*/
struct _destination_parameter {
    object_destination_t *destination;
    char *parameterName;                         /**< The parameter name as provided by the destination */
    variant_map_t parameters;                    /**< List of parameters for the destination*/
};

typedef struct _parameter_def {
    char *name;                        /**< The name of the parameter */

    uint8_t type;                      /**< The parameter type, see @ref parameter_type_t */
    uint8_t attributes;                /**< The parameter attributes */

    parameter_handlers_t *handlers;    /**< User defined handler functions */
    parameter_validator_t *validator;  /**< validator for this parameter */

    uint32_t refcount;

#ifdef PCB_HELP_SUPPORT
    char *description;                 /**< Parameter help text */
#endif
} parameter_def_t;

//---------------------------------------------------------------------------------------------
/**
@brief
Parameter structure

@details
This parameter structure contains all information related to an instance of the parameter\n
*/
struct _parameter {
    llist_iterator_t it;               /**< The place in the owners global parameter list */
    parameter_def_t *definition;       /**< Pointer to parameter definition */
    llist_t ACL;                       /**< Access Control List */

    uint8_t state;                     /**< The parameter state, see @ref parameter_state_t */
    uint8_t currentConfig;             /**< 0 or 1, is set to the latest validated and committed value */
    variant_t value[2];                /**< Only one of the values is the actual value, the other one is used while editing */
    destination_parameter_t *map;      /**< Parameter mapping information */

    void *userData;                    /**< Pointer to some user data */
};

/**
@}
*/

bool __attribute__ ((visibility("hidden"))) parameter_destroy(parameter_t *parameter);
variant_type_t __attribute__ ((visibility("hidden"))) parameter_type_to_variant_type(parameter_type_t paramType);
parameter_handlers_t __attribute__ ((visibility("hidden")))  *parameter_getHandlers(parameter_t *parameter);

#ifdef __cplusplus
}
#endif

#endif


