#define _GNU_SOURCE
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <sys/time.h>
#include <string.h>

#include <components.h>

#include <debug/sahtrace.h>

#include <pcb/common/error.h>
#include <pcb/utils/string_list.h>
#include <pcb/pcb_client.h>
#include "pcb_main_priv.h"

#include "requesthandlers_priv.h"
#include "notification_priv.h"
#include "remote_object.h"
#include "object_priv.h"

typedef struct _request_wait {
    request_t *original;
    request_t *notification;
    request_replyHandler replyHandler;
    request_doneHandler doneHandler;
    request_cancelHandler cancelHandler;
    request_destroyHandler destroyHandler;
} request_wait_t;

static bool pcb_waitReplyHandler(request_t *req, pcb_t *pcb, peer_info_t *from, void *userdata);

static void pcb_waitRestoreRequest(request_wait_t *ctx)
{
    /* restore original request */
    ctx->original->replyHandler = ctx->replyHandler;
    ctx->original->doneHandler = ctx->doneHandler;
    ctx->original->cancelHandler = ctx->cancelHandler;
    ctx->original->destroyHandler = ctx->destroyHandler;
}

static bool pcb_waitErrorHandler(request_wait_t *ctx, pcb_t *pcb, peer_info_t *peer,
                                 uint32_t error, const char *description, const char *info)
{
    SAH_TRACEZ_INFO("pcb_wait", "error handler called");
    bool rc = true;
    reply_addError(request_reply(ctx->original), error, description, info);
    if (ctx->original->replyHandler)
        rc &= ctx->original->replyHandler(ctx->original, pcb, peer, ctx->original->userdata);
    if (ctx->original->doneHandler)
        rc &= ctx->original->doneHandler(ctx->original, pcb, peer, ctx->original->userdata);
    return rc;
}

static void pcb_waitCancelHandler(request_t *req, void *userdata __attribute__ ((unused)))
{
    SAH_TRACEZ_INFO("pcb_wait", "cancel handler called");
    request_wait_t *ctx = req->waitdata;

    pcb_waitRestoreRequest(ctx);

    /* if the notification request is running, than the original one cannot be canceled,
     * and if the original one is running the notification cannot be canceled, so we
     * only have to call the cancel handler for the original one any time */
    if (ctx->original->cancelHandler) {
        ctx->original->cancelHandler(ctx->original, ctx->original->userdata);
    }
}

static void pcb_waitDestroyHandler(request_t *req, void *userdata __attribute__ ((unused)))
{
    SAH_TRACEZ_INFO("pcb_wait", "destroy handler called for %s", request_path(req));
    request_wait_t *ctx = req->waitdata;

    pcb_waitRestoreRequest(ctx);

    /* when the original request is destroyed while the notification one is running */
    if (ctx->notification && req == ctx->original) {
        ctx->notification->destroyHandler = NULL;
        ctx->notification->cancelHandler = NULL;
        request_destroy(ctx->notification);
    }

    /* trigger the original destroy handler */
    if (ctx->original->destroyHandler) {
        ctx->original->destroyHandler(ctx->original, ctx->original->userdata);
    }

    /* remote everything now */
    ctx->original->waitdata = NULL;
    if (ctx->notification)
        ctx->notification->waitdata = NULL;
    free(ctx);
}


static bool pcb_waitNotificationHandler(request_t *req, pcb_t *pcb, peer_info_t *from, void *userdata __attribute__ ((unused)))
{
    request_wait_t *ctx = req->waitdata;
    reply_t *rp = request_reply(ctx->notification);
    reply_item_t *item;
    SAH_TRACEZ_INFO("pcb_wait", "got object notification");

    for (item = reply_firstItem(rp); item; item = reply_nextItem(item)) {
        if (reply_item_type(item) == reply_type_error) {
            pcb_waitErrorHandler(ctx, pcb, from, pcb_error, error_string(pcb_error), NULL);
            request_destroy(ctx->notification);
            ctx->notification = NULL;
            break;
        }
        
        if (reply_item_type(item) == reply_type_object) {
            object_t *object = reply_item_object(item);
            const char *name = object_name(object, request_attributes(ctx->original) &
                                           request_common_path_key_notation & 
                                           request_common_path_slash_seperator);
            if (!name)
                continue;

            const char *path = request_path(ctx->original);

            size_t len = strlen(name);
            if (strncmp(name, path, strlen(name)) == 0) {
                if (strlen(path) == len || path[len] == '.' || path[len] == '/') {
                    /* we have the object, resend the original request */
                    SAH_TRACEZ_INFO("pcb_wait", "found object %s (requested %s)", name, path);
                    request_destroy(ctx->notification);
                    ctx->notification = NULL;
                    if (!pcb_sendRequest(pcb, from, ctx->original)) {
                        pcb_waitErrorHandler(ctx, pcb, from, pcb_error, error_string(pcb_error), NULL);
                    }
                }
            }
        }
    }
    return true;
}

static bool pcb_waitReplyHandler(request_t *req, pcb_t *pcb, peer_info_t *from, void *userdata __attribute__ ((unused)))
{
    request_wait_t *ctx = req->waitdata;
    reply_t *rp = request_reply(req);
    reply_item_t *item;

    for (item = reply_firstItem(rp); item; item = reply_nextItem(item)) {
        if (reply_item_type(item) == reply_type_error && reply_item_error(item) == pcb_error_not_found)
            break;
    }
    
    pcb_waitRestoreRequest(ctx);

    /* if we have a not found error, send a notification request for the base object */
    if (item) {
        SAH_TRACEZ_INFO("pcb_wait", "object not found, waiting for it");
        ctx->notification = request_create_getObject("", 1, request_no_object_caching | request_notify_object_added);
        ctx->notification->waitdata = ctx;
        request_setReplyHandler(ctx->notification, pcb_waitNotificationHandler);
        request_setCancelHandler(ctx->notification, pcb_waitCancelHandler);
        request_setDestroyHandler(ctx->notification, pcb_waitDestroyHandler);
        if (!pcb_sendRequest(pcb, from, ctx->notification)) {
            request_destroy(ctx->notification);
            return pcb_waitErrorHandler(ctx, pcb, from, pcb_error, error_string(pcb_error), NULL);
        } else {
            /* we should destroy our request if the original one is destroyed */
            request_setDestroyHandler(ctx->original, pcb_waitDestroyHandler);
        }
        return true;
    }
    else {
        /* we have a reply, restore the request */
        SAH_TRACEZ_INFO("pcb_wait", "got reply, forwarding it to original request handlers");
        req->waitdata = NULL;
        free(ctx);
        if (req->replyHandler)
            return req->replyHandler(req, pcb, from, req->userdata);
        else
            return true;
    }
}


bool pcb_waitSendRequest(pcb_t *pcb, peer_info_t *peer, request_t *req)
{
    /* prepare our context data */
    request_wait_t *ctx = (request_wait_t *) calloc(1,sizeof(request_wait_t));
    if (!ctx) {
        pcb_error = pcb_error_no_memory;
        goto error;
    }

    ctx->original = req;
    ctx->original->waitdata = ctx;

    /* keep the original handlers */
    ctx->replyHandler = req->replyHandler;
    ctx->doneHandler = req->doneHandler;
    ctx->cancelHandler = req->cancelHandler;
    ctx->destroyHandler = req->destroyHandler;
    
    /* replace them by ours */
    req->replyHandler = pcb_waitReplyHandler;
    req->doneHandler = NULL;
    req->cancelHandler = pcb_waitCancelHandler;
    req->destroyHandler = pcb_waitDestroyHandler;
    
    /* first try the request, if we get an error with object not found,
     we will try then to wait for the object */
    if (!pcb_sendRequest(pcb, peer, req)) {
        SAH_TRACEZ_ERROR("pcb_wait", "unable to send request: %s", error_string(pcb_error));
        /* do not free the context has the cancel handler will be called */
        goto error;
    }

    return true;

 error:
    return false;
}
