/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#include <stdlib.h>

#include <debug/sahtrace.h>

#include <pcb/common/error.h>
#include <pcb/core/notification.h>

#include "request_priv.h"
#include "local_object.h"

bool reply_initialize(reply_t *reply) {
    if (!llist_initialize(&reply->items))
        return false;

    if (!llist_initialize(&reply->errors))
        return false;

    reply->completed = false;

    return true;
}

void reply_clear(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return;
    }

    reply_item_t *item = NULL;
    llist_iterator_t *it = llist_takeFirst(&reply->items);
    while(it) {
        item = llist_item_data(it,reply_item_t,it);
        switch(item->type) {
            case reply_type_object: // objects are stored in a data model, so do not clear the objects here
                free(item);
            break;
            case reply_type_notification: {
                item_notification_t *item_notification = llist_item_data(item,item_notification_t,item);
                notification_destroy(item_notification->notification);
                free(item);
            }
            break;
            case reply_type_error: {
                item_error_t *item_error = llist_item_data(item,item_error_t,item);
                // keep the errors when they are still in the error list
                if (!item_error->it.list) {
                    string_cleanup(&item_error->description);
                    string_cleanup(&item_error->info);
                    free(item);
                } else {
                    SAH_TRACEZ_INFO("pcb","Keeping error in list");
                }
            }
            break;
            case reply_type_function_return: {
                item_function_return_t *item_functionReturn = llist_item_data(item,item_function_return_t,item);
                variant_cleanup(&item_functionReturn->retval);
                llist_iterator_t *it = llist_takeFirst(&item_functionReturn->returnArguments);
                while(it) {
                    argument_value_t *val = llist_item_data(it,argument_value_t,it);
                    free(val->parameterName);
                    variant_cleanup(&val->value);
                    free(val);
                    it = llist_takeFirst(&item_functionReturn->returnArguments);
                }
                free(item);
            }
            break;
            default:
            break;
        }
        it = llist_takeFirst(&reply->items);
    }
}

void reply_errorsCleanup(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return;
    }

    SAH_TRACEZ_INFO("pcb","Cleanup error list");
    item_error_t *error = NULL;
    reply_item_t *item = NULL;
    llist_iterator_t *it = llist_takeFirst(&reply->errors);
    while(it) {
        SAH_TRACEZ_INFO("pcb","Cleanup error");
        error = llist_item_data(it,item_error_t,it);
        item = &error->item;
        string_cleanup(&error->description);
        string_cleanup(&error->info);
        llist_iterator_take(&item->it);
        free(item);
        it = llist_takeFirst(&reply->errors);
    }
}

void reply_cleanup(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return;
    }

    SAH_TRACEZ_INFO("pcb","Cleanup reply");
    reply_errorsCleanup(reply);
    reply_clear(reply);
}

request_t *reply_request(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    return llist_item_data(reply,request_t, reply);
}

uint32_t reply_itemCount(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return 0;
    }

    return llist_size(&reply->items);
}

reply_item_t *reply_firstItem(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }
    llist_iterator_t *it = llist_first(&reply->items);
    if (it)
        return  llist_item_data(it,reply_item_t,it);

    return NULL;
}

reply_item_t *reply_nextItem(reply_item_t *item) {
    if (!item) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }
    llist_iterator_t *it = llist_iterator_next(&item->it);
    if (it)
        return  llist_item_data(it,reply_item_t,it);

    return NULL;
}

bool reply_itemsAvailable(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    if (llist_isEmpty(&reply->items))
        return false;
    else
        return true;
}

void reply_setCompleted(reply_t *reply, bool completed) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return;
    }

    reply->completed = completed;
}

bool reply_isComplete(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    return reply->completed;
}

bool reply_hasErrors(reply_t *reply) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    return !llist_isEmpty(&reply->errors);
}

bool reply_addObject(reply_t *reply, object_t *object) {
    if (!reply || !object) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    request_t *req = reply_request(reply);
    if (request_attributes(req) & request_getObject_keep_hierarchy) {
        object_t *parent = local_object_parent(object);
        object_t *root = object_root(object);
        if (parent != root) {
            return true;
        }
    }

    item_object_t *item = (item_object_t *)calloc(1,sizeof(item_object_t));
    if (!item) {
        pcb_error = pcb_error_no_memory;
        return false;
    }

    llist_iterator_initialize(&item->item.it);
    item->item.type = reply_type_object;
    item->object = object;

    llist_append(&reply->items,&item->item.it);
    req->state = request_state_waiting_for_reply;
    return true;
}

bool reply_addNotification(reply_t *reply, notification_t *notification) {
    if (!reply || !notification) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    item_notification_t *item = (item_notification_t *)calloc(1,sizeof(item_notification_t));
    if (!item) {
        pcb_error = pcb_error_no_memory;
        return false;
    }

    llist_iterator_initialize(&item->item.it);
    item->item.type = reply_type_notification;
    item->notification = notification;

    llist_append(&reply->items,&item->item.it);
    return true;
}

bool reply_addError(reply_t *reply, const uint32_t error, const char *description, const char *info) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    item_error_t *item = (item_error_t *)calloc(1,sizeof(item_error_t));
    if (!item) {
        pcb_error = pcb_error_no_memory;
        return false;
    }

    SAH_TRACEZ_INFO("pcb","Adding error %d - %s - %s to error list",error,description,info);
    llist_iterator_initialize(&item->item.it);
    llist_iterator_initialize(&item->it);
    item->item.type = reply_type_error;
    item->error = error;
    if (!description) {
        string_fromChar(&item->description,error_string(error));
    } else {
        string_fromChar(&item->description,description);
    }
    string_fromChar(&item->info,info);

    llist_append(&reply->items,&item->item.it);
    llist_append(&reply->errors,&item->it);
    return true;
}

bool reply_addFunctionReturn(reply_t *reply, variant_t *retval, argument_value_list_t *returnArgs) {
    if (!reply) {
        pcb_error = pcb_error_invalid_parameter;
        return false;
    }

    item_function_return_t *item = (item_function_return_t *)calloc(1,sizeof(item_function_return_t));
    if (!item) {
        pcb_error = pcb_error_no_memory;
        return false;
    }

    if (!llist_initialize(&item->returnArguments)) {
        free(item);
        return false;
    }

    item->item.type = reply_type_function_return;
    variant_copy(&item->retval,retval);
    llist_iterator_t *it = llist_takeFirst(returnArgs);
    while(it) {
        llist_append(&item->returnArguments,it);
        it = llist_takeFirst(returnArgs);
    }

    llist_append(&reply->items,&item->item.it);
    return true;
}

reply_item_type_t reply_item_type(reply_item_t *item) {
    if (!item) {
        pcb_error = pcb_error_invalid_parameter;
        return reply_type_invalid;
    }

    return item->type;
}

object_t *reply_item_object(reply_item_t *item) {
    if (!item || item->type != reply_type_object) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    item_object_t *item_object = llist_item_data(item,item_object_t,item);
    return item_object->object;
}

notification_t *reply_item_notification(reply_item_t *item) {
    if (!item || item->type != reply_type_notification) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    item_notification_t *item_notification = llist_item_data(item,item_notification_t,item);
    return item_notification->notification;
}

uint32_t reply_item_error(reply_item_t *item) {
    if (!item || item->type != reply_type_error) {
        pcb_error = pcb_error_invalid_parameter;
        return 0;
    }

    item_error_t *item_error = llist_item_data(item,item_error_t,item);
    return item_error->error;
}

const char *reply_item_errorDescription(reply_item_t *item) {
    if (!item || item->type != reply_type_error) {
        pcb_error = pcb_error_invalid_parameter;
        return 0;
    }

    item_error_t *item_error = llist_item_data(item,item_error_t,item);
    return string_buffer(&item_error->description);
}

const char *reply_item_errorInfo(reply_item_t *item) {
    if (!item || item->type != reply_type_error) {
        pcb_error = pcb_error_invalid_parameter;
        return 0;
    }

    item_error_t *item_error = llist_item_data(item,item_error_t,item);
    return string_buffer(&item_error->info);
}

const variant_t *reply_item_returnValue(reply_item_t *item) {
    if (!item || item->type != reply_type_function_return) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    item_function_return_t *item_function_return = llist_item_data(item,item_function_return_t,item);
    return &item_function_return->retval;
}

argument_value_list_t *reply_item_returnArguments(reply_item_t *item) {
    if (!item || item->type != reply_type_function_return) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    item_function_return_t *item_function_return = llist_item_data(item,item_function_return_t,item);
    return &item_function_return->returnArguments;
}
