/****************************************************************************
**
** Copyright (C) 2011 SoftAtHome. All rights reserved.
**
** SoftAtHome reserves all rights not expressly granted herein.
**
** - DISCLAIMER OF WARRANTY -
**
** THIS FILE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
** EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
** PURPOSE.
**
** THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOURCE
** CODE IS WITH YOU. SHOULD THE SOURCE CODE PROVE DEFECTIVE, YOU
** ASSUME THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
**
** - LIMITATION OF LIABILITY -
**
** IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN
** WRITING WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MODIFIES
** AND/OR DISTRIBUTES THE SOURCE CODE, BE LIABLE TO YOU FOR DAMAGES,
** INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES
** ARISING OUT OF THE USE OR INABILITY TO USE THE SOURCE CODE
** (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
** INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
** OF THE SOURCE CODE TO OPERATE WITH ANY OTHER PROGRAM), EVEN IF SUCH
** HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
** DAMAGES.
**
****************************************************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>

#include <debug/sahtrace.h>

#include <pcb/common/error.h>
#include <pcb/utils/uri.h>

/**
@file
@brief
Implementation of the uri functions
*/

/**
@ingroup pcb_utils_uri
@brief
Parse an uri string

@details
This functions parses an uri string.
Call @ref uri_destroy to free up the memory used by the structure.
Never free the structure using <b>free</b>

@param URI String containing the uri

@return
    - A pointer to the uri_t structure containing the parsed uri.
    - NULL when an error occured (invalid uri)
*/
uri_t *uri_parse(const char *URI) {
    uri_t *uri = (uri_t *)calloc(1,sizeof(uri_t));
    if (!uri) {
        goto error;
    }

    uri->scheme = strdup(URI);
    if (!uri->scheme) {
        goto error_free_uri;
    }

    // get scheme
    char *scheme_end = strstr(uri->scheme,"://");
    if (!scheme_end) {
        goto error_free_scheme;
    }
    scheme_end[0] = 0;

    // get authority
    char *authority = scheme_end + 3;
    if (!*authority) {
        goto error_free_scheme;
    }

    uri->host = authority;
    char *tmp = authority;
    bool curly_braces = false;
    bool braces = false;
    for(; *tmp && !uri->path; tmp++) {
        switch(*tmp) {
            case '{':
            case '}':
                if (!braces)
                    curly_braces = !curly_braces;
            break;
            case '[':
            case ']':
                if (!curly_braces)
                    braces = !braces;
            break;
            case ':':
                if (!braces && !curly_braces) {
                    *tmp = 0;
                    uri->port = tmp + 1;
                }
            break;
            case '@':
                if (!braces && !curly_braces) {
                    *tmp = 0;
                    uri->user = uri->host;
                    uri->password = uri->port;
                    uri->port = NULL;
                    uri->host = tmp + 1;
                }
            break;
            case '/':
                if (!braces && !curly_braces) {
                    *tmp = 0;
                    uri->path = tmp + 1;
                }
            break;
            default:
            break;
        }
    }

    if (uri->host && uri->host[0] == '{') {
        uri->host++;
        uri->host[strlen(uri->host) - 1] = 0;
    }
    if (uri->port && uri->port[0] == '{') {
        uri->port++;
        uri->port[strlen(uri->port) - 1] = 0;
    }
    if (uri->host && uri->host[0] == '[') {
        uri->host++;
        uri->host[strlen(uri->host) - 1] = 0;
    }
    if (uri->port && uri->port[0] == '[') {
        uri->port++;
        uri->port[strlen(uri->port) - 1] = 0;
    }

    if (uri->path && (uri->parameters = strchr(uri->path, '?'))) {
        *uri->parameters = '\0';
        uri->parameters ++;
        uri->parametersLength = strlen(uri->parameters);
        for (tmp = uri->parameters; tmp < uri->parameters + uri->parametersLength; tmp++) {
            if (*tmp == '&') {
                *tmp = '\0';
            }
        }
    }

    SAH_TRACEZ_INFO("pcb","scheme = %s",uri->scheme?uri->scheme:"");
    SAH_TRACEZ_INFO("pcb","host = %s",uri->host?uri->host:"");
    SAH_TRACEZ_INFO("pcb","port = %s",uri->port?uri->port:"");
    SAH_TRACEZ_INFO("pcb","user = %s",uri->user?uri->user:"");
    SAH_TRACEZ_INFO("pcb","password = %s",uri->password?uri->password:"");
    SAH_TRACEZ_INFO("pcb","path = %s",uri->path?uri->path:"");

    return uri;

error_free_scheme:
    free(uri->scheme);
error_free_uri:
    free(uri);
error:
    return NULL;
}

/**
@ingroup pcb_utils_uri
@brief
Free the uri_t structure

@details
Free up the memory used by the structure.

@param URI Pointer to the uri_t structure
*/
void uri_destroy(uri_t *URI) {
    if (!URI) {
        pcb_error = pcb_error_invalid_parameter;
        return;
    }

    free(URI->scheme);
    free(URI);
}

/**
@ingroup pcb_utils_uri
@brief
Fetch the value of a parameter from the uri

@details
Fetch the value of a parameter from the uri string using a parameter name.

@param URI Pointer to the uri_t structure
@param name The name of the parameter

@return
    - The value
    - NULL if the parameter is not found
*/
const char *uri_getParameter(const uri_t *URI, const char *name)
{
    if (!URI || !name || !*name) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }

    int len = strlen(name);
    char *p;
    for (p = URI->parameters; p < URI->parameters + URI->parametersLength; p += strlen(p) + 1) {
        if (strncmp(p, name, len) == 0 && *(p + len) == '=') {
            return p + len + 1;
        }
    }
    return NULL;
}

/**
@ingroup pcb_utils_uri
@brief
Build a uri string.

@details
Generate an uri string from the given uri_t structure.
The returned string must be freed.

@param URI Pointer to the uri_t structure

@return
    - String representing the uri
*/
char *uri_string(const uri_t *URI) {
    if (!URI) {
        pcb_error = pcb_error_invalid_parameter;
        return NULL;
    }
    char *c;
    if (URI->parameters) {
        for (c = URI->parameters; c < URI->parameters + URI->parametersLength; c++) {
            if (*c == '\0')
                *c = '&';
        }
    }
    char *fulluri = NULL;
    if (asprintf(&fulluri,"%s://%s%s%s%s{%s}%s%s%s%s%s%s%s",URI->scheme,
                                          (URI->user && *(URI->user))?URI->user:"",
                                          (URI->user && *(URI->user))?":":"",
                                          (URI->password && *(URI->password))?URI->password:"",
                                          (URI->user && *(URI->user))?"@":"",
                                          URI->host,
                                          (URI->port && *(URI->port))?":{":"",
                                          (URI->port && *(URI->port))?URI->port:"",
                                          (URI->port && *(URI->port))?"}":"",
                                          (URI->path && *(URI->path))?"/":"",
                                          (URI->path && *(URI->path))?URI->path:"",
                                          (URI->parameters && *(URI->parameters))?"?":"",
                                          (URI->parameters && *(URI->parameters))?URI->parameters:"") == -1) {
        SAH_TRACEZ_ERROR("pcb","Failed to allocate memory");
        return NULL;
    }
    if (URI->parameters) {
        for (c = URI->parameters; c < URI->parameters + URI->parametersLength; c++) {
            if (*c == '&')
                *c = '\0';
        }
    }
    return fulluri;
}
