/* -*- c-basic-offset: 4 -*- */
#include <stdio.h>
#include <stdlib.h>

#include <pcb/utils/circular_buffer.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unittests/unittest.h>

#define LOOPS 100

int test_open_close() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "r");
    T_NOT_NULL(stream);
    T_SUCCEED(fclose(stream));

cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_rw() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];

    T_EQUAL(int, fwrite("123456", sizeof(char), 7, stream), 7);
    fflush(stream);
    T_EQUAL(int, fread(buffer, sizeof(char), 7, stream), 7);
    T_EQUAL(int, strncmp(buffer, "123456", 7), 0);
    T_SUCCEED(fprintf(stream, "123456"));
    fclose(stream);

cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_write_off_by_one_bug() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    char out[1024];
    T_VERIFY(circbuf_initialize(&cb, 3));
    FILE *stream = circbuf_fopen(&cb, "w+");
    T_SUCCEED(feof(stream));
    char *str = (char *) "123";
    fwrite(str, sizeof(char), 3, stream);
    fflush(stream);

    fwrite(str, sizeof(char), 3, stream);
    fflush(stream);

    ssize_t size = circbuf_availableForRead(&cb);

    T_EQUAL(int, size, 6);
    strncpy(out, (char *) circbuf_readBuffer(&cb), 6);
    circbuf_markAsRead(&cb, 6);
    T_EQUAL(int, strncmp(out, "123123", 6), 0);

    fclose(stream);

cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_reset() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "w+");

    T_EQUAL(int, fwrite("123456", sizeof(char), 7, stream), 7);
    fflush(stream);
    circbuf_reset(&cb,0);
    T_EQUAL(int, feof(stream), 0);
    fclose(stream);

cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}


int test_rw_over_boundary() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];
    int i = 0;
    char *str[2] = { (char *) "123456", (char *) "abcdef" };
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_EQUAL(int, fwrite(test, sizeof(char), 7, stream), 7);
        fflush(stream);
        T_EQUAL(int, fread(buffer, sizeof(char), 7, stream), 7);
        T_EQUAL(int, strncmp(buffer, test, 7), 0);
    }
    fclose(stream);
cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_rw_cornercase() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];
    int i = 0;
    char *str[2] = { (char *) "12345", (char *) "abcde" };
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_EQUAL(int,fwrite(test, sizeof(char), 5, stream), 5);
        fflush(stream);
        T_EQUAL(int,fread(buffer, sizeof(char), 5, stream), 5);
        T_EQUAL(int,strncmp(buffer, test, 5), 0);
    }
    fclose(stream);
cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_direct_api() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "w+");

    T_NOT_NULL(stream);
    T_VERIFY(circbuf_stream(&cb) == stream);
    T_VERIFY(circbuf_size(&cb) == 100);
    T_VERIFY(circbuf_availableForRead(&cb) == 0);

    int i = 0;
    char *str[2] = { (char *) "123456", (char *) "abcdef" };
    ssize_t toRead = 0;
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_VERIFY(fwrite(test, sizeof(char), 7, stream) == 7);
        fflush(stream);
        toRead = circbuf_availableForRead(&cb);
        T_VERIFY(strncmp((char *) circbuf_readBuffer(&cb), test, toRead) == 0);
        T_VERIFY(circbuf_markAsRead(&cb, toRead) == toRead);
        if (toRead < 7) {
            T_VERIFY(strncmp((char *) circbuf_readBuffer(&cb), test + toRead, 7 - toRead) == 0);
            T_VERIFY(circbuf_markAsRead(&cb, 7 - toRead) == 7 - toRead);
        }
    }
    fclose(stream);
cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_eof() {
    T_TESTCASE_BEGIN();
    char buffer[256];
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "w+");
    T_VERIFY(feof(stream));
    T_VERIFY(fwrite("123456", sizeof(char), 7, stream) == 7);
    fflush(stream);
    T_VERIFY(!feof(stream));
    T_VERIFY(fread(buffer, sizeof(char), 10, stream) == 7);
    T_VERIFY(feof(stream));
    T_VERIFY(fread(buffer, sizeof(char), 10, stream) == 0);
    T_VERIFY(feof(stream));
    fclose(stream);
cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_rw_increase_size() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];
    int i = 0;
    char *str[2] = { (char *) "123456", (char *) "abcdef" };
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_VERIFY(fwrite(test, sizeof(char), 7, stream) == 7);
    fflush(stream);
    }
    for (i = 0; i < LOOPS; i++) {
        char *test = str[i%2];
        T_VERIFY(fread(buffer, sizeof(char), 7, stream) == 7);
        T_VERIFY(strncmp(buffer, test, 7) == 0);
    }

    fclose(stream);
cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_resize_with_read_after_write() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    T_VERIFY(circbuf_initialize(&cb, 100));
    FILE *stream = circbuf_fopen(&cb, "w+");
    char buffer[256];

    /* fill the buffer */
    T_VERIFY(fwrite("0123456789", sizeof(char), 10, stream) == 10);
    fflush(stream);
    /* read a bit */
    T_VERIFY(fread(buffer, sizeof(char), 7, stream) == 7);

    /* trigger a resize */
    T_VERIFY(fwrite("0123456789", sizeof(char), 5, stream) == 5);
    fflush(stream);
    T_VERIFY(fwrite("0123456789", sizeof(char), 5, stream) == 5);
    fflush(stream);
    fclose(stream);
cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

int test_under_constraint() {
    T_TESTCASE_BEGIN();
    int limit = 5 * 1024 * 1024, oldLimit;
    int i;
    struct rlimit l;
    if (getenv("VALGRIND_OPTS") != NULL) {
        fprintf(stderr, "WARN   : PCBTestCircBuf::test_under_constraint() skipped, valgrind detected\n");
    T_TESTCASE_END();
    }

    T_VERIFY(getrlimit(RLIMIT_AS, &l) == 0);
    oldLimit = l.rlim_cur;
    l.rlim_cur = limit;
    T_VERIFY(setrlimit(RLIMIT_AS, &l) == 0);

    circbuf_t cb;
    T_VERIFY(!circbuf_initialize(&cb, limit * 2));
    T_VERIFY(errno == ENOMEM);

    circbuf_initialize(&cb, 100);
    FILE *stream = circbuf_fopen(&cb, "w+");
    T_VERIFY(stream);

    for (i = 0; i < limit; i++) {
        errno = 0;
        if (fwrite("0123456789", sizeof(char), 10, stream) < 10)
            break;
        errno = 0;
        fflush(stream);
        if (errno) {
            break;
        }
        T_VERIFY(errno == 0);
    }
    T_VERIFY(errno == ENOMEM);

    fclose(stream);
cleanup:
    circbuf_cleanup(&cb);

    l.rlim_cur = oldLimit;
    T_VERIFY(setrlimit(RLIMIT_AS, &l) == 0);
    T_TESTCASE_END();
}

int test_seek() {
    T_TESTCASE_BEGIN();
    circbuf_t cb;
    circbuf_initialize(&cb, 30);
    FILE *stream = circbuf_fopen(&cb, "w+");
    char result[100];

    /* simple case */
    char *str = (char *) "foo is to be replaced by bar";
    int len = strlen(str);

    T_VERIFY(fwrite(str, sizeof(char), len, stream) == len);
    T_VERIFY(fseek(stream, 0 - len, SEEK_CUR) == 0);
    T_VERIFY(fwrite("bar", sizeof(char), 3, stream) == 3);
    T_VERIFY(fseek(stream, len - 3, SEEK_CUR) == 0);
    T_VERIFY(fflush(stream) == 0);
    T_VERIFY(fread(result, sizeof(char), len, stream) == len);
    T_VERIFY(strncmp(result, "bar is to be replaced by bar", len) == 0);

    /*  cross boundaries */
    int i;
    for (i = 0; i < LOOPS; i++) {
        T_VERIFY(fwrite(str, sizeof(char), len, stream) == len);
        T_VERIFY(fseek(stream, 0 - len, SEEK_CUR) == 0);
        T_VERIFY(fwrite("bar", sizeof(char), 3, stream) == 3);
        T_VERIFY(fseek(stream, len - 3, SEEK_CUR) == 0);
        T_VERIFY(fflush(stream) == 0);
        T_VERIFY(fread(result, sizeof(char), len, stream) == len);
        T_VERIFY(strncmp(result, "bar is to be replaced by bar", len) == 0);
    }

cleanup:
    circbuf_cleanup(&cb);
    T_TESTCASE_END();
}

T_TESTCASES_BEGIN()
    T_TESTCASE(test_open_close, "Test opening and closing of a circular buffer")
    T_TESTCASE(test_rw, "Test read and write operation on a circular buffer")
    T_TESTCASE(test_write_off_by_one_bug, "Test the fix for the write off by one bug")
    T_TESTCASE(test_reset, "Test reset of circular buffer")
    T_TESTCASE(test_rw_over_boundary, "Test read and write operations over buffer boundary")
    T_TESTCASE(test_rw_cornercase, "Test special cases")
    T_TESTCASE(test_direct_api, "Test circular buffer direct C API")
    T_TESTCASE(test_eof, "Test eof conditions")
    T_TESTCASE(test_rw_increase_size, "Test automatic increasing of buffer size")
    T_TESTCASE(test_resize_with_read_after_write, "Test automatic increasing of buffer size after reading")
    T_TESTCASE(test_under_constraint, "Test circular buffer under memory constraints (can not be run when valgrind is active)")
    T_TESTCASE(test_seek, "Test circular buffer seek functionality")
T_TESTCASES_END()
