#define _XOPEN_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unittests/unittest.h>

#include <debug/sahtrace.h>

#include <pcb/utils/connection.h>
#include <pcb/utils/peer.h>
#include <pcb/utils/timer.h>

static int accepted_count = 0;
static int rejected_count = 0;
static int connected_count = 0;
static int close_count = 0;
static int destroy_count = 0;

static int peer_read_count = 0;
static int peer_write_done_count = 0;

static bool test_peer_read(peer_info_t *peer) {
    FILE *in = peer_inStream(peer);

    int bytes = 0;
    int c = 0;
    while((c = fgetc(in)) != EOF) {
        printf("%c",c);
        peer_read_count--;
        bytes++;
    }

    if (!bytes)
        return false;
    else
        return true;
}

static bool test_peer_write(peer_info_t *peer) {
    FILE *out = peer_outStream(peer,peer_stream_direct_io);
    fprintf(out,"TEST DATA");

    return true;
}

static bool test_peer_write_done(peer_info_t *peer) {
    (void)peer;
    peer_write_done_count++;

    return true;
}

static bool test_peer_close(peer_info_t *peer) {
    (void)peer;
    close_count++;
    return true;
}

static bool test_peer_destroy(peer_info_t *peer) {
    (void)peer;
    destroy_count++;
    return true;
}

static bool connection_accepted(peer_info_t *peer) {
    (void)peer;
    SAH_TRACE_INFO("Connection accept");
    accepted_count++;

    peer_setEventHandler(peer,peer_event_read,peer_handler_read_always,test_peer_read);
    return true;
}

static bool connection_connected(peer_info_t *peer) {
    (void)peer;
    SAH_TRACE_INFO("Connection connected");
    connected_count++;

    return true;
}

static bool connection_rejected(peer_info_t *peer) {
    (void)peer;
    SAH_TRACE_INFO("Connection reject");
    rejected_count++;

    return false;
}

int initialize_unit_test() {
    T_TESTCASE_BEGIN();

    T_VERIFY(connection_initLibrary());

cleanup:
    T_TESTCASE_END();
}

int cleanup_unit_test() {
    connection_exitLibrary();
    return 0;
}

int test_connection_read_write_ipc() {
    T_TESTCASE_BEGIN();

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    accepted_count = 0;
    rejected_count = 0;
    connected_count = 0;
    destroy_count = 0;

    connection_info_t *connection = connection_create("test");
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));

    T_VERIFY(connection != NULL);
    T_VERIFY(strcmp(connection_name(connection),"test") == 0);
    T_VERIFY(connection_setEventHandler(connection,peer_event_accepted,connection_accepted));
    T_VERIFY(connection_setEventHandler(connection,peer_event_connected,connection_connected));
    T_VERIFY(!connection_setEventHandler(connection,12345,NULL)); // invalid event
    peer_info_t *listen = connection_listenOnIPC(connection,"TestSocket");
    T_VERIFY(listen);
    T_VERIFY(!peer_flush(listen));
    T_VERIFY(!peer_reconnect(listen));

    peer_info_t *dest = NULL;

    dest = connection_connectToIPC(connection,"TestSocket");
    T_VERIFY(dest != NULL);
    T_VERIFY(peer_setEventHandler(dest,peer_event_destroy,peer_handler_default,test_peer_destroy));
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));
    T_VERIFY(accepted_count == 1);
    T_VERIFY(connected_count == 1);
    T_VERIFY(destroy_count == 0);
    T_VERIFY(peer_reconnect(dest));
    T_VERIFY(!peer_setStreamType(dest,peer_stream_buffered_io));

    // write 5 bytes
    FILE *out = peer_outStream(dest,peer_stream_buffered_io);
    T_VERIFY(out != NULL);
    T_VERIFY(peer_setStreamType(dest,peer_stream_buffered_io));
    peer_read_count = fprintf(out, "12345");
    fflush(out);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_EQUAL(int,peer_read_count,0);

    out = peer_outStream(dest,peer_stream_buffered_io);
    T_VERIFY(out != NULL);
    peer_read_count = fprintf(out, "67890");
    fflush(out);
    T_VERIFY(peer_flush(dest));
    T_VERIFY(peer_flush(dest));

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_EQUAL(int,peer_read_count,0);
    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_VERIFY(peer_flush(dest));

    T_VERIFY(peer_setEventHandler(dest,peer_event_write,peer_handler_default,test_peer_write));
    T_VERIFY(peer_flush(dest));
    T_VERIFY(peer_setEventHandler(dest,peer_event_write,peer_handler_write_always,test_peer_write));
    T_VERIFY(peer_flush(dest));

    T_VERIFY(peer_close(dest));
    T_VERIFY(peer_flush(dest));

    peer_destroy(dest);
    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_VERIFY(destroy_count == 1);

cleanup:
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

int test_connection_accept_ipc() {
    T_TESTCASE_BEGIN();

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    accepted_count = 0;
    rejected_count = 0;
    connected_count = 0;

    connection_info_t *connection = connection_create("test");
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));

    T_VERIFY(connection != NULL);
    T_VERIFY(strcmp(connection_name(connection),"test") == 0);
    T_VERIFY(connection_setEventHandler(connection,peer_event_accepted,connection_accepted));
    T_VERIFY(connection_setEventHandler(connection,peer_event_connected,connection_connected));

    //T_VERIFY(!connection_listenOnIPC(connection,"/var/run/test")); // must fail, need root access
    peer_info_t *listen = connection_listenOnIPC(connection,"TestSocket");
    T_VERIFY(listen);
    T_VERIFY(!connection_listenOnIPC(connection,"TestSocket"));
    T_VERIFY(peer_connection(listen) == connection);

    //T_VERIFY(!connection_connectToIPC(connection,"/var/run/test")); // must fail, need root access
    T_VERIFY(connection_connectToIPC(connection,"TestSocket"));
    T_EQUAL(int, llist_size(&connection->connections),1);
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));
    T_EQUAL(int, llist_size(&connection->connections),2);

    T_VERIFY(!peer_setEventHandler(listen,peer_event_read,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_write,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_write_done,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_close,peer_handler_default,NULL));
    T_VERIFY(!peer_inStream(listen));
    T_VERIFY(!peer_outStream(listen,peer_stream_buffered_io));

    T_VERIFY(peer_setEventHandler(listen,peer_event_accepted,0,connection_rejected));
    peer_info_t *peer =connection_connectToIPC(connection,"TestSocket");
    T_VERIFY(peer);
    T_VERIFY(!peer_isTCPSocket(peer));
    T_VERIFY(peer_isIPCSocket(peer));
    T_VERIFY(peer_setEventHandler(peer,peer_event_close,peer_handler_default,test_peer_close));
    T_VERIFY(peer_addCloseHandler(peer,test_peer_close));
    T_VERIFY(peer_setEventHandler(peer,peer_event_read,peer_handler_read_always,test_peer_read));
    T_VERIFY(!peer_setEventHandler(peer,peer_event_accepted,peer_handler_default,NULL));
    T_EQUAL(int, llist_size(&connection->connections),3);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_EQUAL(int, close_count, 1);
    T_EQUAL(int, llist_size(&connection->connections),3);

    T_VERIFY(peer_setEventHandler(listen,peer_event_accepted,0,connection_accepted));
    T_VERIFY(peer_reconnect(peer));
    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_EQUAL(int, llist_size(&connection->connections),4);

    T_VERIFY(connection_waitForEvents(&cons, &timeout) == 0);

    T_VERIFY(accepted_count == 2);
    T_VERIFY(connected_count == 3);
    T_VERIFY(rejected_count == 1);


cleanup:
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

int test_connection_read_write_tcp() {
    T_TESTCASE_BEGIN();

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    accepted_count = 0;
    rejected_count = 0;
    connected_count = 0;

    connection_info_t *connection = connection_create("test");
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));

    T_VERIFY(connection != NULL);
    T_VERIFY(strcmp(connection_name(connection),"test") == 0);
    T_VERIFY(connection_setEventHandler(connection,peer_event_accepted,connection_accepted));
    T_VERIFY(connection_setEventHandler(connection,peer_event_connected,connection_connected));
    peer_info_t *listen = connection_listenOnTCP(connection,"127.0.0.1","6431",connection_attr_default);
    T_VERIFY(listen);

    peer_info_t *dest = NULL;
    dest = connection_connectToTCP(connection,"127.0.0.1","6431",connection_attr_default);
    T_VERIFY(dest != NULL);
    T_VERIFY(peer_connection(dest) == connection);
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));
    T_VERIFY(accepted_count == 1);
    T_VERIFY(connected_count == 1);
    T_VERIFY(peer_setEventHandler(dest,peer_event_write_done,peer_handler_default,test_peer_write_done));
    T_VERIFY(peer_capabilities(dest) & peer_cap_direct_fd_access);

    // write 5 bytes
    FILE *out = peer_outStream(dest,peer_stream_buffered_io);
    T_VERIFY(out != NULL);
    peer_read_count = fprintf(out, "12345");
    fflush(out);
    T_EQUAL(int,peer_bytesQueued(dest),peer_read_count);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    T_EQUAL(int,peer_read_count,0);
    T_EQUAL(int,peer_write_done_count,1);

    out = peer_outStream(dest,peer_stream_buffered_io);
    peer_read_count = fprintf(out, "12345");
    fflush(out);
    peer_flush(dest);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    T_EQUAL(int,peer_read_count,0);
    T_EQUAL(int,peer_write_done_count,2);

    fprintf(out, "12345");

cleanup:
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

int test_connection_accept_tcp() {
    T_TESTCASE_BEGIN();

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    accepted_count = 0;
    rejected_count = 0;
    close_count = 0;
    connected_count = 0;

    connection_info_t *connection = connection_create("test");
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));

    T_VERIFY(connection != NULL);
    T_VERIFY(strcmp(connection_name(connection),"test") == 0);
    T_VERIFY(connection_setEventHandler(connection,peer_event_accepted,connection_accepted));
    T_VERIFY(connection_setEventHandler(connection,peer_event_connected,connection_connected));

    T_VERIFY(!connection_listenOnTCP(connection,"500.500.500.500","6431",connection_attr_default));
    peer_info_t *listen = connection_listenOnTCP(connection,"127.0.0.1","6431",connection_attr_default);
    T_VERIFY(listen);
    T_VERIFY(!connection_listenOnTCP(connection,"127.0.0.1","6431",connection_attr_default));
    T_VERIFY(peer_connection(listen) == connection);
    T_VERIFY(!peer_setStreamType(listen,peer_stream_direct_io));

    T_VERIFY(!peer_setEventHandler(listen,peer_event_read,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_write,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_write_done,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_close,peer_handler_default,NULL));
    T_VERIFY(!peer_inStream(listen));
    T_VERIFY(!peer_outStream(listen,peer_stream_buffered_io));

    //T_VERIFY(!connection_connectToTCP(connection,"500.500.500.500","6431",connection_attr_default));

    T_VERIFY(connection_connectToTCP(connection,"127.0.0.1","6431",connection_attr_default));
    T_EQUAL(int, llist_size(&connection->connections),1);
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));
    T_EQUAL(int, llist_size(&connection->connections),2);

    T_VERIFY(peer_setEventHandler(listen,peer_event_accepted,0,connection_rejected));
    peer_info_t *peer = connection_connectToTCP(connection,"127.0.0.1","6431",connection_attr_default);
    T_VERIFY(peer);
    T_VERIFY(peer_isTCPSocket(peer));
    T_VERIFY(!peer_isIPCSocket(peer));
    T_VERIFY(peer_setEventHandler(peer,peer_event_close,peer_handler_default,test_peer_close));
    T_VERIFY(peer_setEventHandler(peer,peer_event_read,peer_handler_read_always,test_peer_read));
    T_VERIFY(!peer_setEventHandler(peer,peer_event_accepted,peer_handler_default,NULL));
    T_EQUAL(int, llist_size(&connection->connections),3);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_EQUAL(int, close_count, 1);
    T_EQUAL(int, llist_size(&connection->connections),3);

    T_VERIFY(peer_setEventHandler(listen,peer_event_accepted,0,connection_accepted));
    T_VERIFY(peer_reconnect(peer));
    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_EQUAL(int, llist_size(&connection->connections),4);

    T_VERIFY(connection_waitForEvents(&cons, &timeout) == 0);

    T_VERIFY(accepted_count == 2);
    T_VERIFY(connected_count == 3);
    T_VERIFY(rejected_count == 1);


cleanup:
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

int test_connection_null() {
    T_TESTCASE_BEGIN();

    T_VERIFY(!connection_initialize(NULL,NULL));
    connection_cleanup(NULL);
    connection_destroy(NULL);
    T_VERIFY(!connection_listenOnIPC(NULL,NULL));
    T_VERIFY(!connection_listenOnTCP(NULL,NULL,NULL,connection_attr_default));
    T_VERIFY(!connection_connectToIPC(NULL,NULL));
    T_VERIFY(!connection_connectToTCP(NULL,NULL,NULL,connection_attr_default));
    T_VERIFY(!connection_closeAll(NULL));
    T_VERIFY(!connection_closeListenSockets(NULL));
    T_VERIFY(!connection_closeServers(NULL));
    T_VERIFY(!connection_closeClients(NULL));
    T_VERIFY(connection_waitForEvents(NULL,NULL) == -1);
    T_VERIFY(!connection_handleEvents(NULL));
    T_VERIFY(!connection_setEventHandler(NULL,0,NULL));
    T_VERIFY(!connection_setUserData(NULL,NULL));
    T_VERIFY(!connection_getUserData(NULL));
    T_VERIFY(!connection_find(NULL,123));
    T_VERIFY(!connection_setSignalEventHandler(NULL,SIGINT,NULL));
    T_VERIFY(!connection_getIterator(NULL));
    T_VERIFY(!connection_name(NULL));
    T_VERIFY(!connection_addCustom(NULL,0));

    T_VERIFY(!peer_isTCPSocket(NULL));
    T_VERIFY(!peer_isIPCSocket(NULL));
    T_VERIFY(!peer_isClientConnection(NULL));
    T_VERIFY(!peer_isServerConnection(NULL));
    T_VERIFY(!peer_isCustomFd(NULL));
    T_VERIFY(!peer_connection(NULL));
    T_VERIFY(peer_getFd(NULL) == -1);
    peer_destroy(NULL);
    T_VERIFY(!peer_close(NULL));
    T_VERIFY(!peer_reconnect(NULL));
    T_VERIFY(!peer_outStream(NULL,peer_stream_buffered_io));
    T_VERIFY(!peer_inStream(NULL));
    T_VERIFY(!peer_setEventHandler(NULL,0,0,NULL));
    T_VERIFY(!peer_setUserData(NULL,NULL));
    T_VERIFY(!peer_getUserData(NULL));
    T_VERIFY(!peer_flush(NULL));
    T_VERIFY(!peer_bytesQueued(NULL));
    T_VERIFY(!peer_isListenSocket(NULL));
    T_VERIFY(!peer_isConnected(NULL));
    T_VERIFY(!peer_capabilities(NULL));
    T_VERIFY(!peer_setStreamType(NULL,peer_stream_direct_io));
    T_VERIFY(!peer_addCloseHandler(NULL,NULL));

cleanup:
    T_TESTCASE_END();
}

int test_connection_getUserData() {
    T_TESTCASE_BEGIN();

    char data[11] = "0123456789";
    connection_info_t *connection = connection_create("test");
    T_VERIFY(connection != NULL);

    T_VERIFY(connection_setUserData(connection,data));
    T_VERIFY(connection_getUserData(connection) == data);

     peer_info_t *peer = connection_connectToTCP(connection,"127.0.0.1","6431",connection_attr_default);
     T_VERIFY(!peer);
     T_VERIFY(connection_listenOnTCP(connection,"127.0.0.1","6431",connection_attr_default));
     peer = connection_connectToTCP(connection,"127.0.0.1","6431",connection_attr_default);
     T_VERIFY(peer_setUserData(peer,data));
     T_VERIFY(peer_getUserData(peer) == data);


cleanup:
    connection_destroy(connection);
    T_TESTCASE_END();
}

int test_connection_find() {
    T_TESTCASE_BEGIN();

    connection_info_t *connection = connection_create(NULL);
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));
    T_VERIFY(connection);
    T_VERIFY(connection_listenOnTCP(connection,"127.0.0.1","6431",connection_attr_default));


    peer_info_t *peer = connection_connectToTCP(connection,"127.0.0.1","6431",connection_attr_default);
    T_VERIFY(peer);
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));

    T_VERIFY(connection_find(connection,peer_getFd(peer)) == peer);
    T_VERIFY(connection_find(connection,peer_getFd(peer) + 1) != peer);

cleanup:
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

static int signals_received = 0;

static void test_signal_raised(int signal) {
    T_TESTCASE_BEGIN();

    signals_received++;

    T_VERIFY(signal == SIGINT);
    return;

cleanup:
    return;
}

int test_signals() {
    T_TESTCASE_BEGIN();

    signals_received = 0;
    connection_info_t *connection = connection_create(NULL);
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));
    T_VERIFY(connection);

    T_VERIFY(!connection_setSignalEventHandler(connection,SIGALRM,test_signal_raised));
    T_VERIFY(connection_setSignalEventHandler(connection,SIGINT,test_signal_raised));
    T_VERIFY(connection_setSignalEventHandler(connection,SIGTERM,test_signal_raised));
    raise(SIGINT);
    T_VERIFY(connection_waitForEvents(&cons,NULL) != 0);
    T_VERIFY(connection_handleEvents(&cons));
    T_VERIFY(signals_received == 1);

    T_VERIFY(connection_setSignalEventHandler(connection,SIGINT,NULL));
    raise(SIGINT);
    T_VERIFY(connection_handleEvents(&cons));
    T_VERIFY(signals_received == 1);

cleanup:
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

static bool test_custom_write(peer_info_t *peer) {
    T_TESTCASE_BEGIN();
    char buffer[21] = "01234567890123456789";

    int bytes = write(peer_getFd(peer),buffer,21);
    T_VERIFY(bytes == 21);

    return true;

cleanup:
    return false;
}

static bool test_custom_read(peer_info_t *peer) {
    T_TESTCASE_BEGIN();

    char buffer[21];
    memset(buffer,0,21);

    read(peer_getFd(peer),buffer,21);
    printf("[%s]",buffer);

    T_VERIFY(strcmp(buffer,"01234567890123456789") == 0);
    return true;

cleanup:
    return false;
}

int test_custom() {
    T_TESTCASE_BEGIN();

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    connection_info_t *connection = connection_create(NULL);
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));
    T_VERIFY(connection);
    int fd = open("/tmp/testconnection.txt",O_CREAT|O_TRUNC|O_WRONLY|O_NONBLOCK,S_IRWXU);
    T_VERIFY(fd != -1);
    peer_info_t *custom = connection_addCustom(connection,fd);
    T_VERIFY(custom);
    T_VERIFY(!peer_isTCPSocket(custom));
    T_VERIFY(!peer_isIPCSocket(custom));
    T_VERIFY(!peer_isClientConnection(custom));
    T_VERIFY(!peer_isServerConnection(custom));
    T_VERIFY(peer_isCustomFd(custom));
    T_VERIFY(peer_setEventHandler(custom,peer_event_write,peer_handler_default,test_custom_write));

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    peer_destroy(custom);

    fd = open("/tmp/testconnection.txt",O_RDONLY|O_NONBLOCK,S_IRWXU);
    T_VERIFY(fd != -1);

    custom = connection_addCustom(connection,fd);
    T_VERIFY(custom);
    T_VERIFY(peer_setEventHandler(custom,peer_event_read,peer_handler_default,test_custom_read));
    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

cleanup:
    remove("/tmp/testconnection.txt");
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

int test_close() {
    T_TESTCASE_BEGIN();

    char data[11] = "0123456789";

    connection_info_t *connection = connection_create(NULL);
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));
    T_VERIFY(connection);

    T_VERIFY(connection_listenOnIPC(connection,"TestSocket"));

    T_VERIFY(connection_connectToIPC(connection,"TestSocket")); // +1 socket
    T_EQUAL(int, llist_size(&connection->connections),1);
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));              // +1 socket, accept
    T_EQUAL(int, llist_size(&connection->connections),2);

    T_VERIFY(connection_closeServers(connection));              // close all accepted sockets and mark them for deletion
    T_EQUAL(int, llist_size(&connection->connections),2);
    T_VERIFY(connection_connectToIPC(connection,"TestSocket")); // +1 socket
    T_EQUAL(int, llist_size(&connection->connections),3);
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));              // -1 was marked for deletion, +1 newly accepted
    T_EQUAL(int, llist_size(&connection->connections),3);

    T_VERIFY(connection_closeClients(connection));              // close all sockets created with connectTo functions (-2)
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    T_VERIFY(connection_waitForEvents(&cons, &timeout) == 0);
    T_EQUAL(int, llist_size(&connection->connections),1);

    peer_info_t *dest = connection_connectToIPC(connection,"TestSocket"); // +1 socket
    T_VERIFY(dest);

    T_EQUAL(int, llist_size(&connection->connections),2);
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));             // +1 socket (accept)
    T_EQUAL(int, llist_size(&connection->connections),3);

    // write 10 bytes
    FILE *out = peer_outStream(dest,peer_stream_buffered_io);
    T_VERIFY(out != NULL);
    peer_read_count = fprintf(out, "%s", data);
    fflush(out);
    T_EQUAL(int,peer_bytesQueued(dest),peer_read_count);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }
    T_EQUAL(int, llist_size(&connection->connections),3);

    out = peer_outStream(dest,peer_stream_direct_io);
    T_VERIFY(out != NULL);
    peer_read_count = fprintf(out, "%s", data);
    fflush(out);
    T_EQUAL(int,peer_bytesQueued(dest),0);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    T_VERIFY(connection_closeClients(connection));
    T_VERIFY(connection_closeServers(connection));
    T_VERIFY(connection_waitForEvents(&cons, &timeout) == 0);
    T_VERIFY(connection_handleEvents(&cons));
    T_EQUAL(int, llist_size(&connection->connections),0);

    dest = connection_connectToIPC(connection,"TestSocket"); // +1 socket
    T_VERIFY(connection_waitForEvents(&cons, NULL) == 1);
    T_VERIFY(connection_handleEvents(&cons));
    T_EQUAL(int, llist_size(&connection->connections),2);
    T_VERIFY(peer_close(dest));
    T_VERIFY(connection_waitForEvents(&cons, &timeout) == 0);
    T_EQUAL(int, llist_size(&connection->connections),2);

    T_VERIFY(peer_reconnect(dest));
    T_EQUAL(int, llist_size(&connection->connections),2);
    T_VERIFY(!peer_isTCPSocket(dest));
    T_VERIFY(peer_isIPCSocket(dest));
    T_VERIFY(!peer_isClientConnection(dest));
    T_VERIFY(peer_isServerConnection(dest));
    T_VERIFY(!peer_isCustomFd(dest));
    T_VERIFY(connection_waitForEvents(&cons, &timeout) == 1);
    T_VERIFY(connection_handleEvents(&cons));
    T_EQUAL(int, llist_size(&connection->connections),3);
    T_VERIFY(peer_close(dest));
    T_VERIFY(!peer_isTCPSocket(dest));
    T_VERIFY(!peer_isIPCSocket(dest));
    T_VERIFY(!peer_isClientConnection(dest));
    T_VERIFY(peer_isServerConnection(dest));
    T_VERIFY(!peer_isCustomFd(dest));

    T_VERIFY(connection_closeAll(connection));
    T_VERIFY(connection_waitForEvents(&cons, &timeout) == 0);
    T_EQUAL(int, llist_size(&connection->connections),0);
    T_EQUAL(int, llist_size(&connection->listenSockets),0);
cleanup:
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

void timer_timeout(pcb_timer_t *timer, void *userdata) {
    (void)timer;
    uint32_t *count = (uint32_t *)userdata;
    printf("T");
    (*count)++;
}

void timer_stop_action(pcb_timer_t *timer, void *userdata) {
    (void)timer;
    bool *stop = (bool *)userdata;
    printf(" - stop timers - ");
    *stop = true;
}

int test_timers() {
    T_TESTCASE_BEGIN();

    llist_t cons;
    llist_initialize(&cons);

    connection_info_t *connection = connection_create(NULL);
    T_VERIFY(connection);
    llist_append(&cons,connection_getIterator(connection));

    uint32_t count = 0;

    struct timeval timeout;
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    pcb_timer_t *timer = pcb_timer_create();
    T_VERIFY(timer);
    T_VERIFY(pcb_timer_setHandler(timer,timer_timeout));
    T_VERIFY(pcb_timer_setUserData(timer, &count));

    T_VERIFY(pcb_timer_start(timer, 100));

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    T_EQUAL(int, count, 1);

    bool stop = false;
    pcb_timer_t *timer2 = pcb_timer_create();
    T_VERIFY(timer2);
    T_VERIFY(pcb_timer_setHandler(timer2,timer_stop_action));
    T_VERIFY(pcb_timer_setUserData(timer2, &stop));
    T_VERIFY(pcb_timer_start(timer2, 4000));

    T_VERIFY(pcb_timer_setHandler(timer,timer_timeout));
    T_VERIFY(pcb_timer_setUserData(timer, &count));
    T_VERIFY(pcb_timer_getUserData(timer) == &count);
    T_VERIFY(pcb_timer_setInterval(timer,500));
    T_VERIFY(pcb_timer_start(timer, 0));

    count = 0;
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    while(connection_waitForEvents(&cons, &timeout) != 0 && !stop) {
        T_VERIFY(connection_handleEvents(&cons));
        timeout.tv_sec = 1;
        timeout.tv_usec = 0;
    }

    T_EQUAL(int, count, 8);
    pcb_timer_destroy(timer);
    pcb_timer_destroy(timer2);

cleanup:
    llist_cleanup(&cons);
    connection_destroy(connection);
    T_TESTCASE_END();
}

int test_timers_null() {
    T_TESTCASE_BEGIN();

    pcb_timer_destroy(NULL);
    T_VERIFY(!pcb_timer_setInterval(NULL,0));
    T_VERIFY(!pcb_timer_setHandler(NULL,NULL));
    T_VERIFY(!pcb_timer_setUserData(NULL,NULL));
    T_VERIFY(!pcb_timer_getUserData(NULL));
    T_VERIFY(!pcb_timer_remainingTime(NULL));
    T_VERIFY(!pcb_timer_start(NULL,0));
    T_VERIFY(!pcb_timer_stop(NULL));
    T_VERIFY(pcb_timer_getState(NULL) == pcb_timer_off);

cleanup:
    T_TESTCASE_END();
}

int test_connection_find_ipc() {
    T_TESTCASE_BEGIN();

    connection_info_t *connection = connection_create(NULL);
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));
    T_VERIFY(connection);

    T_VERIFY(connection_listenOnIPC(connection,"TestSocket"));

    peer_info_t *connected = connection_connectToIPC(connection,"TestSocket");
    T_VERIFY(connected != NULL);

    peer_info_t *found = connection_findIPC(connection,"TestSocket");
    T_VERIFY(found == connected);

    found = connection_findIPC(connection,"invalid");
    T_VERIFY(found == NULL);

cleanup:
    connection_destroy(connection);
    T_TESTCASE_END();
}

int test_connection_find_tcp() {
    T_TESTCASE_BEGIN();

    connection_info_t *connection = connection_create(NULL);
    llist_t cons;
    llist_initialize(&cons);
    llist_append(&cons,connection_getIterator(connection));
    T_VERIFY(connection);

    T_VERIFY(connection_listenOnTCP(connection,"127.0.0.1","6501",0));
    T_VERIFY(connection_listenOnTCP(connection,"::1","6501",0));

    peer_info_t *connected = connection_connectToTCP(connection,"127.0.0.1","6501",0);
    T_VERIFY(connected != NULL);

    peer_info_t *connected2 = connection_connectToTCP(connection,"::1","6501",0);
    T_VERIFY(connected2 != NULL);

    peer_info_t *found = connection_findTCP(connection,"127.0.0.1","6501");
    T_VERIFY(found == connected);

    found = connection_findTCP(connection,"127.0.0.1","6431");
    T_VERIFY(found == NULL);

    found = connection_findTCP(connection,"::1","6501");
    T_VERIFY(found == connected2);

    found = connection_findTCP(connection,"localhost","6501");
    T_VERIFY(found == connected);

    found = connection_findTCP(connection,"192.168.1.1","6501");
    T_VERIFY(found == NULL);

cleanup:
    connection_destroy(connection);
    T_TESTCASE_END();
}

#ifdef OPEN_SSL_SUPPORT
int test_connection_ssl_setup();
int test_connection_accept_tcp_ssl();
int test_connection_read_write_tcp_ssl();
int test_connection_null_ssl();
#endif

T_TESTCASES_BEGIN()
    T_TESTCASE(test_connection_read_write_ipc, "Test reading and writing to and from an ipc socket")
    T_TESTCASE(test_connection_accept_ipc, "Test accepting of incomming ipc sockets")
    T_TESTCASE(test_connection_read_write_tcp, "Test reading and writing to and from a tcp socket")
    T_TESTCASE(test_connection_accept_tcp, "Test accepting of incomming ipc sockets")
    T_TESTCASE(test_connection_null, "Test that passing  NULL pointers to connection and peer functions do not crash")
    T_TESTCASE(test_connection_getUserData, "Test getting and setting of custom data to connection and peer contexts")
    T_TESTCASE(test_connection_find, "Test find functions to retireve back a connection or peer context")
    T_TESTCASE(test_signals, "Test singal handling as an event")
    T_TESTCASE(test_custom, "Test custom file descriptors")
    T_TESTCASE(test_close, "Test closing of sockets (peers) and custom file descriptors")
    T_TESTCASE(test_timers, "Test timers and timer events (callbacks)")
    T_TESTCASE(test_timers_null, "Test that passing NULL pointers to timer functions do not crash")
    T_TESTCASE(test_connection_find_ipc, "Test finding of already open ipc sockets")
    T_TESTCASE(test_connection_find_tcp, "Test finding of already open tcp sockets")
#ifdef OPEN_SSL_SUPPORT
    T_TESTCASE(test_connection_ssl_setup, "Test setting up the ssl context")
    T_TESTCASE(test_connection_accept_tcp_ssl, "Test accepting of incomming tcp ssl sockets")
    T_TESTCASE(test_connection_read_write_tcp_ssl, "Test reading and writing from and to an ssl tcp socket")
    T_TESTCASE(test_connection_null_ssl, "Test that passing NULL pointers to timer functions do not crash")
#endif
T_TESTCASES_END()
