#define _XOPEN_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unittests/unittest.h>
#include <debug/sahtrace.h>

#include <pcb/utils/connection.h>
#include <pcb/utils/peer.h>

static int accepted_count = 0;
static int connected_count = 0;

static int peer_read_count = 0;
static int peer_write_done_count = 0;

static bool test_peer_read(peer_info_t *peer) {
    FILE *in = peer_inStream(peer);

    int bytes = 0;
    int c = 0;
    while((c = fgetc(in)) != EOF) {
        printf("%c",c);
        peer_read_count--;
        bytes++;
    }

    if (!bytes)
        return false;
    else
        return true;
}

static bool test_peer_write_done(peer_info_t *peer) {
    (void)peer;
    peer_write_done_count++;

    return true;
}

static bool connection_accepted(peer_info_t *peer) {
    SAH_TRACE_INFO("Connection accept");
    accepted_count++;

    peer_setEventHandler(peer,peer_event_read,peer_handler_read_always,test_peer_read);
    return true;
}

static bool connection_connected(peer_info_t *peer) {
    (void)peer;
    SAH_TRACE_INFO("Connection connected");
    connected_count++;

    return true;
}

int test_connection_ssl_setup() {
    T_TESTCASE_BEGIN();

    connection_info_t *connection = NULL;
    connection = connection_create("test");
    T_VERIFY(connection != NULL);
    T_VERIFY(strcmp(connection_name(connection),"test") == 0);

    T_VERIFY(!connection_setupSSL(connection, NULL, "./invalid.pem", ssl_attribute_default));
    T_VERIFY(!connection_setupSSL(connection, "./invalid.pem", NULL,  ssl_attribute_default));
    T_VERIFY(!connection_setupSSL(connection, NULL , "./srv_nokey.pem",  ssl_attribute_default));
    T_VERIFY(connection_setupSSL(connection, NULL , "./srv.pem",  ssl_attribute_default));

cleanup:
    connection_destroy(connection);
    T_TESTCASE_END();
}

long verify_server_certificate_accepted(X509 *cert, const char *host, uint32_t attributes) {
    // always accept
    (void)host;
    (void)attributes;
    if (cert)
        X509_free(cert);
    return X509_V_OK;
}

long verify_client_certificate_accepted(X509 *cert) {
    // always accept
    (void)cert;
    return X509_V_OK;
}


int test_connection_accept_tcp_ssl() {
    T_TESTCASE_BEGIN();

    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    accepted_count = 0;
    connected_count = 0;

    llist_t cons;
    llist_initialize(&cons);

    connection_info_t *server = connection_create("server");
    T_VERIFY(server != NULL);
    T_VERIFY(strcmp(connection_name(server),"server") == 0);
    llist_append(&cons,connection_getIterator(server));

    connection_info_t *client = connection_create("client");
    T_VERIFY(client != NULL);
    T_VERIFY(strcmp(connection_name(client),"client") == 0);
    llist_append(&cons,connection_getIterator(client));

    T_VERIFY(connection_setupSSL(server, NULL, "./srv.pem", ssl_attribute_default));
    T_VERIFY(connection_setupSSL(client, NULL, NULL, ssl_attribute_default | ssl_attribute_accept_self_signed | ssl_attribute_check_host_name));

    T_VERIFY(connection_setEventHandler(server,peer_event_accepted,connection_accepted));
    T_VERIFY(connection_setEventHandler(server,peer_event_connected,connection_connected));
    T_VERIFY(connection_setEventHandler(client,peer_event_connected,connection_connected));

    peer_info_t *listen = connection_listenOnTCP(server,"127.0.0.1","6431",connection_attr_ssl);
    T_VERIFY(listen);
    T_VERIFY(peer_connection(listen) == server);

    T_VERIFY(!peer_setEventHandler(listen,peer_event_read,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_write,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_write_done,peer_handler_default,NULL));
    T_VERIFY(!peer_setEventHandler(listen,peer_event_close,peer_handler_default,NULL));
    T_VERIFY(!peer_inStream(listen));
    T_VERIFY(!peer_outStream(listen,peer_stream_buffered_io));

    //T_VERIFY(!connection_connectToTCP(client,"500.500.500.500","6431",connection_attr_ssl));
    peer_info_t *peer = connection_connectToTCP(client,"127.0.0.1","6431",connection_attr_ssl);
    T_VERIFY(peer);
    T_VERIFY(peer_connection(peer) == client);
    T_EQUAL(int, llist_size(&server->connections),0);
    T_EQUAL(int, llist_size(&client->connections),1);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    // certificates exchanged but not accepted, sockets closed
    T_EQUAL(int, llist_size(&server->connections),0);
    T_EQUAL(int, llist_size(&client->connections),1);
    T_VERIFY(!peer_isConnected(peer));
    peer_destroy(peer);

    T_VERIFY(connection_setServerVerify(client,verify_server_certificate_accepted));
    T_VERIFY(connection_setClientVerify(server,verify_client_certificate_accepted));
    peer = connection_connectToTCP(client,"127.0.0.1","6431",connection_attr_ssl);
    T_VERIFY(peer);
    T_VERIFY(peer_connection(peer) == client);
    T_EQUAL(int, llist_size(&client->connections),2);

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    // certificates exchanged and accepted
    T_EQUAL(int, llist_size(&server->connections),1);
    T_EQUAL(int, llist_size(&client->connections),1);
    T_VERIFY(peer_isConnected(peer));

cleanup:
    llist_cleanup(&cons);
    connection_destroy(server);
    connection_destroy(client);
    T_TESTCASE_END();
}

int test_connection_read_write_tcp_ssl() {
    T_TESTCASE_BEGIN();

    struct timeval timeout;
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    accepted_count = 0;
    connected_count = 0;

    llist_t cons;
    llist_initialize(&cons);

    connection_info_t *server = connection_create("server");
    T_VERIFY(server != NULL);
    T_VERIFY(strcmp(connection_name(server),"server") == 0);
    llist_append(&cons,connection_getIterator(server));

    connection_info_t *client = connection_create("client");
    T_VERIFY(client != NULL);
    T_VERIFY(strcmp(connection_name(client),"client") == 0);
    llist_append(&cons,connection_getIterator(client));

    T_VERIFY(connection_setupSSL(server, NULL, "./srv.pem", ssl_attribute_default));
    T_VERIFY(connection_setupSSL(client, NULL, NULL, ssl_attribute_default | ssl_attribute_accept_self_signed));

    T_VERIFY(connection_setEventHandler(server,peer_event_accepted,connection_accepted));
    T_VERIFY(connection_setEventHandler(server,peer_event_connected,connection_connected));
    T_VERIFY(connection_setEventHandler(client,peer_event_connected,connection_connected));

    peer_info_t *listen = connection_listenOnTCP(server,"127.0.0.1","6431",connection_attr_ssl);
    T_VERIFY(listen);
    T_VERIFY(peer_connection(listen) == server);
    T_VERIFY(peer_isTCPSocket(listen));
    T_VERIFY(!peer_isIPCSocket(listen));

    T_VERIFY(connection_setServerVerify(client,verify_server_certificate_accepted));
    peer_info_t *peer = connection_connectToTCP(client,"127.0.0.1","6431",connection_attr_ssl);
    T_VERIFY(peer);
    T_VERIFY(peer_connection(peer) == client);
    T_VERIFY(peer_isTCPSocket(peer));
    T_VERIFY(!peer_isIPCSocket(peer));
    T_VERIFY(peer_capabilities(peer) & peer_cap_ssl_enabled);
    T_EQUAL(int, llist_size(&client->connections),1);

    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    // certificates exchanged and accepted
    T_EQUAL(int, llist_size(&server->connections),1);
    T_EQUAL(int, llist_size(&client->connections),1);
    T_VERIFY(peer_isConnected(peer));


    T_VERIFY(accepted_count == 1);
    T_VERIFY(connected_count == 1);

    // write 5 bytes
    FILE *out = peer_outStream(peer,peer_stream_buffered_io);
    T_VERIFY(out != NULL);
    peer_read_count = fprintf(out, "12345");
    fflush(out);
    T_EQUAL(int,peer_bytesQueued(peer),peer_read_count);
    T_VERIFY(peer_setEventHandler(peer,peer_event_write_done,peer_handler_default,test_peer_write_done));

    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    T_EQUAL(int,peer_read_count,0);
    T_EQUAL(int,peer_write_done_count,1);

    out = peer_outStream(peer,peer_stream_buffered_io);
    peer_read_count = fprintf(out, "12345");
    fflush(out);
    peer_flush(peer);

    timeout.tv_sec = 1;
    timeout.tv_usec = 0;

    while(connection_waitForEvents(&cons, &timeout) != 0) {
        T_VERIFY(connection_handleEvents(&cons));
    }

    T_EQUAL(int,peer_read_count,0);
    T_EQUAL(int,peer_write_done_count,2);

cleanup:
    llist_cleanup(&cons);
    connection_destroy(server);
    connection_destroy(client);
    T_TESTCASE_END();
}

int test_connection_null_ssl() {
    T_TESTCASE_BEGIN();

    T_VERIFY(!connection_setServerVerify(NULL,NULL));
    T_VERIFY(!connection_setClientVerify(NULL,NULL));
cleanup:
    T_TESTCASE_END();
}
