#include <stdlib.h>
#include <stdio.h>

#include <pcb/common/error.h>
#include <pcb/utils.h>
#include <pcb/core.h>
#include <unittests/unittest.h>

// create a single instance object and rollback
int test_object_create_rollback_1() {
    T_TESTCASE_BEGIN();
    const char *name = NULL;
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);
    object_t *root = object_create(datamodel_root(dm),"1234",object_attr_default);
    T_VERIFY(root == NULL);
    root = object_create(datamodel_root(dm),"ac/cd",object_attr_default);
    T_VERIFY(root == NULL);
    root = object_create(datamodel_root(dm),"ac.cd",object_attr_default);
    T_VERIFY(root == NULL);
    root = object_create(datamodel_root(dm),"$accd",object_attr_default);
    T_VERIFY(root == NULL);
    root = object_create(datamodel_root(dm),"Root",object_attr_default);
    T_VERIFY(root != NULL);
    T_VERIFY(object_state(root) == object_state_created);
    name = object_name(root,path_attr_default);
    T_VERIFY(name != NULL);
    T_VERIFY(strcmp(name,"Root") == 0);
    name = object_name(NULL,path_attr_default);
    T_VERIFY(name == NULL);
    T_VERIFY(object_hasChildren(root) == false);
    T_VERIFY(object_hasChildren(NULL) == false);
    T_VERIFY(object_childCount(root) == 0);
    T_VERIFY(object_childCount(NULL) == 0);
    T_VERIFY(object_parent(root) == datamodel_root(dm));
    T_VERIFY(object_parent(NULL) == NULL);
    T_VERIFY(object_hasInstances(root) == false);
    T_VERIFY(object_hasInstances(NULL) == false);
    T_VERIFY(object_instanceCount(root) == 0);
    T_VERIFY(object_instanceCount(NULL) == 0);
    T_VERIFY(object_instanceOf(root) == NULL);
    T_VERIFY(object_instanceOf(NULL) == NULL);
    string_t path;
    string_initialize(&path,0);
    object_path(root,&path,path_attr_default);
    T_VERIFY(string_compareChar(&path,"Root",CaseSensitive) == 0);
    string_cleanup(&path);
    object_rollback(root);
    // object should be deleted, so we can not check anything anymore
    datamodel_destroy(dm);

cleanup:
    T_TESTCASE_END();
}

// create two single instance objects and put one as a child of the other
// rollback the root object
int test_object_create_rollback_2() {
    T_TESTCASE_BEGIN();
    string_t name;
    string_initialize(&name,0);
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);
    object_t *root = object_create(datamodel_root(dm),"Root",object_attr_default);
    T_VERIFY(root != NULL);
    // no extra checks needed, these are done in test_create_object_1
    object_t *child = object_create(root,"Child",object_attr_default);
    T_VERIFY(child != NULL);
    // no extra checks needed, these are done in test_create_object_1

    // verify root
    T_VERIFY(object_hasChildren(root) == true);
    T_VERIFY(object_childCount(root) == 1);
    T_VERIFY(object_parent(root) == datamodel_root(dm));
    T_VERIFY(object_firstChild(root) == child);
    // verify child
    T_VERIFY(object_parent(child) == root);
    T_VERIFY(object_hasChildren(child) == false);
    T_VERIFY(object_childCount(child) == 0);
    T_VERIFY(object_nextSibling(child) == NULL);

    // verify path of child
    object_path(child,&name,path_attr_default);
    T_VERIFY(string_compareChar(&name,"Root.Child",CaseSensitive) == 0);
    object_path(child,&name,path_attr_slash_seperator);
    T_VERIFY(string_compareChar(&name,"Root/Child",CaseSensitive) == 0);
    object_path(child,&name,path_attr_slash_seperator | path_attr_key_notation);
    T_VERIFY(string_compareChar(&name,"Root/Child",CaseSensitive) == 0);

    object_rollback(root);
    // objects should be deleted, so we can not check anything anymore
    datamodel_destroy(dm);
    string_cleanup(&name);

cleanup:
    T_TESTCASE_END();
}

// create a a template object and 1 instance
// rollback the template object
int test_object_create_rollback_3() {
    T_TESTCASE_BEGIN();
    const char *name = NULL;
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);

    object_t *root = object_create(datamodel_root(dm),"Root",object_attr_template);
    T_VERIFY(root != NULL);
    // no extra checks needed, these are done in another test
    object_t *instance = object_createInstance(root,0xFFFFFFFF,NULL);
    T_VERIFY(instance != NULL);

    // verify the root
    T_VERIFY(object_hasChildren(root) == false);
    T_VERIFY(object_childCount(root) == 0);
    T_VERIFY(object_parent(root) == datamodel_root(dm));
    T_VERIFY(object_hasInstances(root) == true);
    T_VERIFY(object_instanceCount(root) == 1);

    // verify the instance
    T_VERIFY(object_state(instance) == object_state_created);
    name = object_name(instance,path_attr_default);
    T_VERIFY(name != NULL);
    T_VERIFY(strcmp(name,"1") == 0);
    T_VERIFY(object_hasChildren(instance) == false);
    T_VERIFY(object_childCount(instance) == 0);
    T_VERIFY(object_parent(instance) == root);
    T_VERIFY(object_hasInstances(instance) == false);
    T_VERIFY(object_instanceCount(instance) == 0);
    T_VERIFY(object_instanceOf(instance) == root);

    // verify path of instance
    string_t path;
    string_initialize(&path,0);
    object_path(instance,&path,path_attr_default);
    T_VERIFY(string_compareChar(&path,"Root.1",CaseSensitive) == 0);
    object_path(instance,&path,path_attr_slash_seperator);
    T_VERIFY(string_compareChar(&path,"Root/1",CaseSensitive) == 0);
    object_path(instance,&path,path_attr_slash_seperator | path_attr_key_notation);
    T_VERIFY(string_compareChar(&path,"Root/1",CaseSensitive) == 0);
    string_cleanup(&path);

    object_rollback(root);
    // objects should be deleted, so we can not check anything anymore
    datamodel_destroy(dm);

cleanup:
    T_TESTCASE_END();
}

// create 1 template object, which contains a single instance object
// create an instance of the template object
// rollback the template object
int test_object_create_rollback_4() {
    T_TESTCASE_BEGIN();
    const char *name = NULL;
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);

    object_t *root = object_create(datamodel_root(dm),"Root",object_attr_template);
    T_VERIFY(root != NULL);
    // no extra checks needed, these are done in another test
    object_t *child = object_create(root,"Child",object_attr_default);
    T_VERIFY(child != NULL);
    T_VERIFY(object_parent(child) == root);
    T_VERIFY(object_hasChildren(root) == true);
    T_VERIFY(object_childCount(root) == 1);

    // no extra checks needed, these are done in another test
    object_t *instance = object_createInstance(root,0xFFFFFFFF,NULL);
    T_VERIFY(instance != NULL);

    // verify the root
    T_VERIFY(object_hasChildren(root) == true);
    T_VERIFY(object_childCount(root) == 1);
    T_VERIFY(object_parent(root) == datamodel_root(dm));
    T_VERIFY(object_hasInstances(root) == true);
    T_VERIFY(object_instanceCount(root) == 1);

    // verify the instance
    T_VERIFY(object_state(instance) == object_state_created);
    name = object_name(instance,path_attr_default);
    T_VERIFY(strcmp(name,"1") == 0);
    T_VERIFY(object_hasChildren(instance) == true);
    T_VERIFY(object_childCount(instance) == 1);
    T_VERIFY(object_parent(instance) == root);
    T_VERIFY(object_hasInstances(instance) == false);
    T_VERIFY(object_instanceCount(instance) == 0);
    T_VERIFY(object_instanceOf(instance) == root);
    T_VERIFY(object_firstChild(instance) != NULL);
    T_VERIFY(object_firstChild(instance) != child);

    // verify path of instance
    string_t path;
    string_initialize(&path,0);
    object_path(instance,&path,path_attr_default);
    T_VERIFY(string_compareChar(&path,"Root.1",CaseSensitive) == 0);
    object_path(instance,&path,path_attr_slash_seperator);
    T_VERIFY(string_compareChar(&path,"Root/1",CaseSensitive) == 0);
    object_path(instance,&path,path_attr_slash_seperator | path_attr_key_notation);
    T_VERIFY(string_compareChar(&path,"Root/1",CaseSensitive) == 0);

    // verify path of child
    object_path(child,&path,path_attr_default);
    T_VERIFY(string_compareChar(&path,"Root.Child",CaseSensitive) == 0);
    object_path(child,&path,path_attr_slash_seperator);
    T_VERIFY(string_compareChar(&path,"Root/Child",CaseSensitive) == 0);
    object_path(child,&path,path_attr_slash_seperator | path_attr_key_notation);
    T_VERIFY(string_compareChar(&path,"Root/Child",CaseSensitive) == 0);

    // verify the instances child
    child =object_firstChild(instance);
    T_VERIFY(object_hasChildren(child) == false);
    T_VERIFY(object_childCount(child) == 0);
    T_VERIFY(object_parent(child) == instance);
    T_VERIFY(object_hasInstances(child) == false);
    T_VERIFY(object_instanceCount(child) == 0);
    T_VERIFY(object_instanceOf(child) == NULL);

    // verify path of instance child
    object_path(child,&path,path_attr_default);
    T_VERIFY(string_compareChar(&path,"Root.1.Child",CaseSensitive) == 0);
    object_path(child,&path,path_attr_slash_seperator);
    T_VERIFY(string_compareChar(&path,"Root/1/Child",CaseSensitive) == 0);
    object_path(child,&path,path_attr_slash_seperator | path_attr_key_notation);
    T_VERIFY(string_compareChar(&path,"Root/1/Child",CaseSensitive) == 0);
    string_cleanup(&path);

    object_rollback(root);
    // objects should be deleted, so we can not check anything anymore
    datamodel_destroy(dm);

cleanup:
    T_TESTCASE_END();
}

// create 1 template object, which contains a template object
// create an instance of the root template object,
// create an instance of the child template within the instance of the root
// rollback the template object
int test_object_create_rollback_5() {
    T_TESTCASE_BEGIN();
    string_t name;
    string_initialize(&name,0);
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);

    object_t *root = object_create(datamodel_root(dm),"Root",object_attr_template);
    T_VERIFY(root != NULL);

    object_t *child = object_create(root,"Child",object_attr_template);
    T_VERIFY(child != NULL);
    T_VERIFY(object_parent(child) == root);
    T_VERIFY(object_hasChildren(root) == true);
    T_VERIFY(object_childCount(root) == 1);

    T_VERIFY(object_hasInstances(root) == false);
    T_VERIFY(object_hasInstances(child) == false);

    object_t *instance1 = object_createInstance(root,0xFFFFFFFF,NULL);
    T_VERIFY(instance1 != NULL);
    T_VERIFY(object_instanceOf(instance1) == root);
    T_VERIFY(object_parent(instance1) == root);
    T_VERIFY(object_hasChildren(instance1) == true);

    object_t *instanceChild = object_firstChild(instance1);
    T_VERIFY(instanceChild != NULL);
    T_VERIFY(object_parent(instanceChild) == instance1);
    T_VERIFY(object_hasInstances(instanceChild) == false);
    T_VERIFY(object_hasChildren(instanceChild) == false);

    object_t *instance2 = object_createInstance(instanceChild,0xFFFFFFFF,NULL);
    T_VERIFY(instance2 != NULL);
    T_VERIFY(object_instanceOf(instance2) == instanceChild);
    T_VERIFY(object_parent(instance2) == instanceChild);
    T_VERIFY(object_hasChildren(instance2) == false);

    // verify path of instance child
    object_path(instance2,&name,path_attr_default);
    T_VERIFY(string_compareChar(&name,"Root.1.Child.1",CaseSensitive) == 0);
    object_path(instance2,&name,path_attr_slash_seperator);
    T_VERIFY(string_compareChar(&name,"Root/1/Child/1",CaseSensitive) == 0);
    object_path(instance2,&name,path_attr_slash_seperator | path_attr_key_notation);
    T_VERIFY(string_compareChar(&name,"Root/1/Child/1",CaseSensitive) == 0);

    instance2 = object_createInstance(instanceChild,0xFFFFFFFF,NULL);
    T_VERIFY(instance2 != NULL);
    T_VERIFY(object_instanceOf(instance2) == instanceChild);
    T_VERIFY(object_parent(instance2) == instanceChild);
    T_VERIFY(object_hasChildren(instance2) == false);
    T_VERIFY(object_instanceCount(instanceChild) == 2);

    object_path(instance2,&name,path_attr_default);
    T_VERIFY(string_compareChar(&name,"Root.1.Child.2",CaseSensitive) == 0);
    object_path(instance2,&name,path_attr_slash_seperator);
    T_VERIFY(string_compareChar(&name,"Root/1/Child/2",CaseSensitive) == 0);
    object_path(instance2,&name,path_attr_slash_seperator | path_attr_key_notation);
    T_VERIFY(string_compareChar(&name,"Root/1/Child/2",CaseSensitive) == 0);

    object_t *located = object_getObject(root,"1.Child.2",path_attr_default,NULL);
    T_VERIFY(located == instance2);
    located = object_getObject(root,"1/Child/2",path_attr_slash_seperator,NULL);
    T_VERIFY(located == instance2);

    object_rollback(root);
    // objects should be deleted, so we can not check anything anymore
    datamodel_destroy(dm);
    string_cleanup(&name);

cleanup:
    T_TESTCASE_END();
}

int test_object_create_commit_1() {
    T_TESTCASE_BEGIN();
    const char *name = NULL;
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);

    object_t *root = object_create(datamodel_root(dm),"Root",object_attr_default);
    T_VERIFY(root != NULL);
    T_VERIFY(object_state(root) == object_state_created);
    T_VERIFY(object_commit(root) == true);
    T_VERIFY(object_state(root) == object_state_ready);

    name = object_name(root,path_attr_default);
    T_VERIFY(strcmp(name,"Root") == 0);
    T_VERIFY(object_hasChildren(root) == false);
    T_VERIFY(object_childCount(root) == 0);
    T_VERIFY(object_parent(root) == datamodel_root(dm));
    T_VERIFY(object_hasInstances(root) == false);
    T_VERIFY(object_instanceCount(root) == 0);
    T_VERIFY(object_instanceOf(root) == NULL);

    string_t path;
    string_initialize(&path,0);
    object_path(root,&path,path_attr_default);
    T_VERIFY(string_compareChar(&path,"Root",CaseSensitive) == 0);
    string_cleanup(&path);

    T_VERIFY(object_delete(root) == true);
    T_VERIFY(object_state(root) == object_state_deleted);
    T_VERIFY(object_commit(root) == true);
    // object should be deleted, so we can not check anything anymore
    datamodel_destroy(dm);

cleanup:
    T_TESTCASE_END();
}

int test_object_create_commit_2() {
    T_TESTCASE_BEGIN();
    string_t name;
    string_initialize(&name,0);
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);

    object_t *root = object_create(datamodel_root(dm),"Root",object_attr_default);
    T_VERIFY(root != NULL);
    object_t *child = object_create(root, "Child",object_attr_default);
    T_VERIFY(child != NULL);

    T_VERIFY(object_commit(root) == true);
    T_VERIFY(object_state(child) == object_state_ready);

    // verify root
    T_VERIFY(object_hasChildren(root) == true);
    T_VERIFY(object_childCount(root) == 1);
    T_VERIFY(object_parent(root) == datamodel_root(dm));
    T_VERIFY(object_firstChild(root) == child);
    // verify child
    T_VERIFY(object_parent(child) == root);
    T_VERIFY(object_hasChildren(child) == false);
    T_VERIFY(object_childCount(child) == 0);
    T_VERIFY(object_nextSibling(child) == NULL);

    // verify path of child
    object_path(child,&name,path_attr_default);
    T_VERIFY(string_compareChar(&name,"Root.Child",CaseSensitive) == 0);
    object_path(child,&name,path_attr_slash_seperator);
    T_VERIFY(string_compareChar(&name,"Root/Child",CaseSensitive) == 0);
    object_path(child,&name,path_attr_slash_seperator | path_attr_key_notation);
    T_VERIFY(string_compareChar(&name,"Root/Child",CaseSensitive) == 0);

    T_VERIFY(object_update(root) == true);
    T_VERIFY(object_delete(root) == true);
    T_VERIFY(object_state(child) == object_state_deleted);
    T_VERIFY(object_commit(root) == true);
    // objects should be deleted, so we can not check anything anymore

    datamodel_destroy(dm);
    string_cleanup(&name);

cleanup:
    T_TESTCASE_END();
}

int test_object_iterations() {
    T_TESTCASE_BEGIN();
    string_t name;
    string_initialize(&name,64);
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);

    object_t *root = object_create(datamodel_root(dm),"Root",object_attr_default);
    T_VERIFY(root != NULL);

    object_t *child[10];
    uint32_t i = 0;
    for(i = 0; i < 10; i++) {
        string_fromUInt32(&name,i);
        string_prependChar(&name,"Child_");
        child[i] = object_create(root, string_buffer(&name),object_attr_default);
        T_VERIFY(child[i] != NULL);
    }

    object_t *test = NULL;
    i = 0;
    object_for_each_child(test, root) {
        T_VERIFY(test == child[i]);
        i++;
    }
    T_VERIFY(object_commit(root) == true);
    T_VERIFY(i == 10);
    i--;
    test = object_lastChild(root);
    T_VERIFY(test != NULL);
    while(test) {
        T_VERIFY(test == child[i]);
        T_VERIFY(object_state(child[i]) == object_state_ready);
        i--;
        test = object_prevSibling(test);
    }

    object_t *templ = object_create(root, "TemplateChild",object_attr_template);
    T_VERIFY(templ != NULL);
    object_t *instance[10];
    for(i = 0; i < 10; i++) {
        instance[i] = object_createInstance(templ,0xFFFFFFFF,NULL);
        T_VERIFY(instance[i] != NULL);
        T_VERIFY(object_instanceOf(instance[i]) == templ);
    }
    i = 0;
    object_for_each_instance(test, templ) {
        T_VERIFY(test == instance[i]);
        i++;
    }
    T_VERIFY(object_commit(templ) == true);
    T_VERIFY(i == 10);
    i--;
    test = object_lastInstance(templ);
    T_VERIFY(test != NULL);
    while(test) {
        T_VERIFY(test == instance[i]);
        T_VERIFY(object_state(instance[i]) == object_state_ready);
        i--;
        test = object_prevInstance(test);
    }

    T_VERIFY(object_delete(root) == true);
    T_VERIFY(object_commit(root) == true);
    datamodel_destroy(dm);
    string_cleanup(&name);

cleanup:
    T_TESTCASE_END();
}

int test_object_attributes() {
    T_TESTCASE_BEGIN();
    char data[10] = "1234";

    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);

    object_t *root = object_create(datamodel_root(dm),"Root", object_attr_default);
    T_VERIFY(root != NULL);
    /*
    object_t *object = object_create(root, "Test",object_attr_private |
                                            object_attr_read_only |
                                            object_attr_template |
                                            object_attr_accept_parameters |
                                            object_attr_instance |
                                            object_attr_executing_func |
                                            object_attr_remote |
                                            object_attr_has_key_params |
                                            object_attr_subtree_changed);
    T_VERIFY(object == NULL);
    */
    object_t *object = object_create(root, "Test",
                            object_attr_read_only |
                                            object_attr_template |
                                            object_attr_accept_parameters |
                                            object_attr_instance |
                                            object_attr_executing_func |
                                            object_attr_subtree_changed);
    T_EQUAL(int, object_attributes(object), (object_attr_read_only | object_attr_template | object_attr_accept_parameters));
    object_t *object2 = object_create(object, "Test", object_attr_default);
    T_VERIFY(object2 != NULL);
    T_VERIFY(object_getMaxInstances(object) == 0xFFFFFFFF);
    T_VERIFY(object_getMaxInstances(object2) == 0);
    T_VERIFY(object_setMaxInstances(object,10) == true);
    T_VERIFY(object_getMaxInstances(object) == 10);
    T_VERIFY(object_setMaxInstances(NULL,100) == false);
    T_VERIFY(object_getMaxInstances(NULL) == 0);
    T_VERIFY(object_canAcceptParameters(object) == true);
    T_VERIFY(object_canAcceptParameters(object2) == false);
    T_VERIFY(object_canAcceptParameters(NULL) == false);
    T_VERIFY(object_isRemote(object) == false);
    T_VERIFY(object_isRemote(NULL) == false);
    T_VERIFY(object_isTemplate(object) == true);
    T_VERIFY(object_isTemplate(object2) == false);
    T_VERIFY(object_isTemplate(NULL) == false);
    T_VERIFY(object_isInstance(object) == false);
    T_VERIFY(object_isInstance(NULL) == false);
    T_VERIFY(object_isReadOnly(object) == true);
    T_VERIFY(object_isReadOnly(object2) == false);
    T_VERIFY(object_isReadOnly(NULL) == false);
    object_setReadOnly(object,false);
    T_VERIFY(object_isReadOnly(object) == false);
    object_setReadOnly(object,true);
    object_setReadOnly(NULL,true);
    T_VERIFY(object_isReadOnly(object) == true);
    object_setUserData(object,(void *)data);
    T_VERIFY(object_getUserData(object) == data);
    T_VERIFY(object_getUserData(object2) == NULL);
    T_VERIFY(object_getUserData(NULL) == NULL);
    object_setUserData(object,NULL);
    object_setUserData(NULL,NULL);
    T_VERIFY(object_state(NULL) == object_state_invalid);
    datamodel_destroy(dm);

cleanup:
    T_TESTCASE_END();
}

static bool validation_fail(object_t *object, void *data) {
    (void)object;
    (void)data;
    pcb_error = 0x0100FFFF;
    return false;
}

static bool validation_ok(object_t *object, void *data) {
    (void)object;
    (void)data;
    return true;
}

int test_object_validation() {
    T_TESTCASE_BEGIN();
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);

    T_VERIFY(object_validate(NULL) == false);
    object_t *root = object_create(datamodel_root(dm),"object",object_attr_default);
    T_VERIFY(root != NULL);
    object_t *templ = object_create(root, "object",object_attr_template);
    T_VERIFY(templ != NULL);
    object_validator_t *failValidator = object_validator_create(validation_fail,NULL);
    T_VERIFY(failValidator != NULL);
    object_validator_t *okValidator = object_validator_create(validation_ok,NULL);
    T_VERIFY(okValidator != NULL);

    T_VERIFY(object_setValidator(root,failValidator) == true);
    T_VERIFY(object_validate(root) == false);

    T_VERIFY(object_setValidator(root,okValidator) == true);
    T_VERIFY(object_validate(root) == true);
    T_VERIFY(object_state(root) == object_state_created);

    T_VERIFY(object_setValidator(root,NULL) == true);
    T_VERIFY(object_setValidator(NULL,NULL) == false);

    failValidator = object_validator_create(validation_fail,NULL);
    T_VERIFY(failValidator != NULL);
    okValidator = object_validator_create(validation_ok,NULL);
    T_VERIFY(okValidator != NULL);
    T_VERIFY(object_setValidator(templ,failValidator) == true);
    T_VERIFY(object_validate(root) == false);
    T_VERIFY(object_setValidator(templ,okValidator) == true);
    T_VERIFY(object_validate(root) == true);

    T_VERIFY(object_commit(root) == true);
    object_t *instance = object_createInstance(templ,0xFFFFFFFF,NULL);
    T_VERIFY(instance != NULL);
    failValidator = object_validator_create(validation_fail,NULL);
    T_VERIFY(failValidator != NULL);
    T_VERIFY(object_setValidator(instance,failValidator) == true);
    T_VERIFY(object_validate(root) == false);
    T_VERIFY(object_setValidator(instance,NULL) == true);
    T_VERIFY(object_commit(root) == true);
    T_VERIFY(object_state(instance) == object_state_ready);

    T_VERIFY(object_delete(root) == true);
    instance = object_createInstance(templ,0xFFFFFFFF,NULL);
    T_VERIFY(instance == NULL);

    T_VERIFY(object_commit(NULL) == false);
    object_rollback(NULL);
    T_VERIFY(object_commit(root) == true);
    datamodel_destroy(dm);

cleanup:
    T_TESTCASE_END();
}

int test_object_parameters() {
    T_TESTCASE_BEGIN();
    parameter_t *parameter[10];
    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);
    object_t *root = object_create(datamodel_root(dm),"object",object_attr_default);
    T_VERIFY(root != NULL);

    parameter[0] = parameter_create(root, "Param1",parameter_type_string,parameter_attr_default);
    T_VERIFY(parameter[0] != NULL);
    parameter[1] = parameter_create(root, "Param2",parameter_type_int8,parameter_attr_default);
    T_VERIFY(parameter[1] != NULL);
    parameter[2] = parameter_create(root, "Param3",parameter_type_int16,parameter_attr_default);
    T_VERIFY(parameter[2] != NULL);
    parameter[3] = parameter_create(root, "Param4",parameter_type_int32,parameter_attr_default);
    T_VERIFY(parameter[3] != NULL);
    parameter[4] = parameter_create(root, "Param5",parameter_type_uint8,parameter_attr_default);
    T_VERIFY(parameter[4] != NULL);
    parameter[5] = parameter_create(root, "Param6",parameter_type_uint16,parameter_attr_default);
    T_VERIFY(parameter[5] != NULL);
    parameter[6] = parameter_create(root, "Param7",parameter_type_uint32,parameter_attr_default);
    T_VERIFY(parameter[6] != NULL);
    parameter[7] = parameter_create(root, "Param8",parameter_type_bool,parameter_attr_default);
    T_VERIFY(parameter[7] != NULL);
    parameter[8] = parameter_create(root, "Param9",parameter_type_date_time,parameter_attr_default);
    T_VERIFY(parameter[8] != NULL);
    parameter[9] = parameter_create(root, "Param10",parameter_type_reference,parameter_attr_default);
    T_VERIFY(parameter[9] != NULL);

    T_VERIFY(object_commit(root) == true);

    uint32_t i = 0;
    parameter_t *param = NULL;
    object_for_each_parameter(param,root) {
        T_VERIFY(param == parameter[i]);
        i++;
    }
    T_VERIFY(i == 10);
    i = 9;
    param = object_lastParameter(root);
    while(param) {
        T_VERIFY(param == parameter[i]);
        i--;
        param = object_prevParameter(param);
    }

    T_VERIFY(object_getParameter(root,"Param10") == parameter[9]);
    T_VERIFY(object_getParameter(root,NULL) == NULL);
    T_VERIFY(object_getParameter(NULL,"Param9") == NULL);
    T_VERIFY(parameter_create(root, NULL,parameter_type_string,parameter_attr_default) == false);
    T_VERIFY(parameter_create(root, "1234",parameter_type_string,parameter_attr_default) == false);
    T_VERIFY(parameter_create(root, "abc$def",parameter_type_string,parameter_attr_default) == false);
    T_VERIFY(parameter_create(root, "abc.def",parameter_type_string,parameter_attr_default) == false);
    T_VERIFY(parameter_create(root, "abc/def",parameter_type_string,parameter_attr_default) == false);
    T_VERIFY(object_firstParameter(NULL) == NULL);
    T_VERIFY(object_lastParameter(NULL) == NULL);
    T_VERIFY(object_nextParameter(NULL) == NULL);
    T_VERIFY(object_prevParameter(NULL) == NULL);
    T_VERIFY(object_hasParameters(root) == true);
    T_VERIFY(object_parameterCount(root) == 10);
    T_VERIFY(object_hasParameters(NULL) == false);
    T_VERIFY(object_parameterCount(NULL) == 0);
    T_VERIFY(object_update(root) == true);
    T_VERIFY(object_validate(root) == true);

    T_VERIFY(parameter_setFromChar(parameter[0],"1234567890") == true);
    T_VERIFY(object_state(root) == object_state_modified);
    T_VERIFY(object_validate(root) == true);
    T_VERIFY(object_state(root) == object_state_validated);
    T_VERIFY(object_commit(root) == true);
    T_VERIFY(object_state(root) == object_state_ready);

    T_VERIFY(object_delete(root) == true);
    T_VERIFY(object_commit(root) == true);
    datamodel_destroy(dm);

cleanup:
    T_TESTCASE_END();
}

int test_object_create_instances() {
    T_TESTCASE_BEGIN();

    datamodel_t *dm = datamodel_create();
    T_VERIFY(dm != NULL);
    object_t *root = object_create(datamodel_root(dm),"object",object_attr_default);
    T_VERIFY(root != NULL);

    object_t *mi = object_create(root, "Test", object_attr_template);
    T_VERIFY(mi != NULL);
    T_VERIFY(parameter_create(mi,"P1",parameter_type_string,parameter_attr_default) != NULL);
    T_VERIFY(parameter_create(mi,"P2",parameter_type_string,parameter_attr_default) != NULL);
    T_VERIFY(object_commit(root) == true);

    object_t *instance = object_createInstance(mi,0,"KEY");
    T_VERIFY(instance != NULL);
    T_VERIFY(object_instanceCount(mi) == 1);
    T_VERIFY(object_parameterCount(instance) == 2);
    T_VERIFY(object_commit(mi) == true);

    instance = object_createInstance(mi,0,"KEY");
    T_VERIFY(instance == NULL);
    T_VERIFY(object_instanceCount(mi) == 1);

    datamodel_destroy(dm);

cleanup:
    T_TESTCASE_END();
}

T_TESTCASES_BEGIN()
    T_TESTCASE(test_object_create_rollback_1, "Create a single instance object and do a rollback")
    T_TESTCASE(test_object_create_rollback_2, "create two single instance objects and put one as a child of the other, rollback the root object")
    T_TESTCASE(test_object_create_rollback_3, "create a a template object and 1 instance, rollback the template object")
    T_TESTCASE(test_object_create_rollback_4, "create 1 template object, which contains a single instance object, create an instance of the template object, rollback the template object")
    T_TESTCASE(test_object_create_rollback_5, "create 1 template object, which contains a template object, create an instance of the root template object, create an instance of the child template within the instance of the root, rollback the topmost template object")
    T_TESTCASE(test_object_create_commit_1, "Create a single instance object and do a commit")
    T_TESTCASE(test_object_create_commit_2, "create two single instance objects and put one as a child of the other, and do a commit ")
    T_TESTCASE(test_object_iterations, "Test parameter iteration and object child iterations")
    T_TESTCASE(test_object_attributes, "Testobject attributes")
    T_TESTCASE(test_object_validation, "Test object validation")
    T_TESTCASE(test_object_parameters, "Test object parameters")
    T_TESTCASE(test_object_create_instances, "Test creating of instances")
T_TESTCASES_END()
