#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>

#include <pcb/utils/tree.h>
#include <unittests/unittest.h>

typedef struct _int32_tree_item {
    uint32_t childId;
    uint32_t depth;
    tree_item_t tree_item;
} int32_tree_item_t;

static int32_tree_item_t *int32_tree_item_create(uint32_t id, uint32_t depth) {
    int32_tree_item_t *item = (int32_tree_item_t *)calloc(1,sizeof(int32_tree_item_t));
    if (item) {
        tree_item_initialize(&item->tree_item);
        item->childId = id;
        item->depth = depth;
    }
    return item;
}

static int build_tree_recursive(tree_item_t *root, uint32_t childs, uint32_t depth, uint32_t totalDepth) {
    int rc = 0;
    int32_tree_item_t *item = NULL;

    if (depth > 0) {
      for(uint32_t i = 0; i < childs; i++) {
        item = int32_tree_item_create(i,totalDepth - depth);
        T_VERIFY(item != NULL);
        T_VERIFY(tree_item_appendChild(root,&item->tree_item) == true);
        T_SUCCEED(build_tree_recursive(&item->tree_item, childs, depth - 1, totalDepth));
      }
    }

 cleanup:
    return rc;
}

static int  build_tree(tree_item_t *root, uint32_t childs, uint32_t depth) {
    return build_tree_recursive(root, childs, depth, depth);
}

static int  verify_tree_recursive(tree_item_t *root, uint32_t childs, uint32_t depth, uint32_t totalDepth) {
    int rc = 0;
    tree_item_t *child = NULL;
    int32_tree_item_t *item = NULL;

    if (depth > 0) {
      T_VERIFY(tree_item_hasChildren(root) == true);
      if (childs)
        T_VERIFY(tree_item_childCount(root) == childs);

      uint32_t index = 0;
      tree_for_each_child(child, root) {
        item = tree_item_data(child,int32_tree_item_t,tree_item);
        T_VERIFY(item != NULL);
        T_VERIFY(item->depth == totalDepth - depth);
        T_VERIFY(item->childId == index);
        T_SUCCEED(verify_tree_recursive(child,childs,depth - 1, totalDepth));
        index++;
      }
    }

 cleanup:
    return rc;
}

static int  verify_tree(tree_item_t *root, uint32_t childs, uint32_t depth) {
      return verify_tree_recursive(root, childs, depth, depth);
}


static int  destroy_tree(tree_item_t *root) {
    int rc = 0;
    tree_item_t *child = NULL;
    int32_tree_item_t *item = NULL;

    tree_for_each_child(child, root) {
      T_VERIFY(child != NULL);
      if (tree_item_hasChildren(child)==true)
        destroy_tree(child);
    }

    for(uint32_t i = tree_item_childCount(root); i > 0; i--) {
      child = tree_item_takeFirstChild(root);
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      tree_item_cleanup(child);
      free(item);
    }

 cleanup:
    return rc;
}

int  test_tree_build_and_verify() {
    T_TESTCASE_BEGIN();
    tree_t tree;

    tree_initialize(&tree);

    T_SUCCEED(build_tree(tree_root(&tree),5,1));
    T_SUCCEED(verify_tree(tree_root(&tree),5,1));
    T_SUCCEED(destroy_tree(tree_root(&tree)));

    T_SUCCEED(build_tree(tree_root(&tree),5,3));
    T_SUCCEED(verify_tree(tree_root(&tree),5,3));
    T_SUCCEED(destroy_tree(tree_root(&tree)));

 cleanup:
    tree_cleanup(&tree);
    T_TESTCASE_END();
}

#define CHILDS 5

int  test_tree_reference_functions() {
    T_TESTCASE_BEGIN();
    tree_t tree;
    tree_item_t *root = NULL;
    tree_item_t *child = NULL;
    tree_item_t *parent = NULL;
    tree_item_t *previous = NULL;
    uint32_t index = 0;

    tree_initialize(&tree);

    T_SUCCEED(build_tree(tree_root(&tree),CHILDS,1));

    root = tree_root(&tree);

    T_VERIFY(tree_item_hasChildren(root) == true);
    T_VERIFY(tree_item_childCount(root) == CHILDS);

   // test tree_item_parent()
    tree_for_each_child(child, root) {
      parent = tree_item_parent(child);
      T_VERIFY(parent == root);
    }

    // test tree_item_childAt()
    index = 0;
    tree_for_each_child(child, root) {
      T_VERIFY(child == tree_item_childAt(root,index));
      index++;
    }
    T_VERIFY(tree_item_childAt(root,CHILDS+10) == NULL);

    // test tree_item_lastChild()
    T_VERIFY(tree_item_lastChild(root) == tree_item_childAt(root,CHILDS-1));

    // test tree_item_prevSibling()
    // test tree_item_nextSibling()
    index = 0;
    tree_for_each_child(child, root) {
      if(!index) {
        T_VERIFY(tree_item_prevSibling(child) == NULL);
      }
      else {
        T_VERIFY(tree_item_prevSibling(child) == previous);
        T_VERIFY(tree_item_nextSibling(previous) == child);
      }

      if(index == (CHILDS - 1))
        T_VERIFY(tree_item_nextSibling(child) == NULL);

      previous = child;
      index++;
    }

    T_SUCCEED(destroy_tree(tree_root(&tree)));

cleanup:
    tree_cleanup(&tree);
    T_TESTCASE_END();
}

int  test_tree_insertion_functions() {
    T_TESTCASE_BEGIN();
    tree_t tree;
    tree_item_t *root = NULL;
    tree_item_t *child = NULL;
    int32_tree_item_t *item = NULL;
    int32_tree_item_t *item1 = NULL;
    int32_tree_item_t *item2 = NULL;
    uint32_t index = 0;

    tree_initialize(&tree);

    // tree_item_prependChild()
    T_SUCCEED(build_tree(tree_root(&tree),CHILDS,1));
    root = tree_root(&tree);

    item1 = int32_tree_item_create(88,0);

    T_VERIFY(tree_item_prependChild(NULL,&item1->tree_item)==false);
    T_VERIFY(tree_item_prependChild(root,NULL)==false);

    T_VERIFY(tree_item_prependChild(root,&item1->tree_item)==true);

    index = 0;
    tree_for_each_child(child, root) {
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      T_VERIFY(item->depth == 0);

      if(index == 0)
        T_VERIFY(item->childId == 88);
      else
        T_VERIFY(item->childId == index-1);

      index++;
    }

    T_SUCCEED(destroy_tree(tree_root(&tree)));

   // tree_item_insertChildAt()
    T_SUCCEED(build_tree(tree_root(&tree),CHILDS,1));
    root = tree_root(&tree);

    item1 = int32_tree_item_create(88,0);
    item2 = int32_tree_item_create(99,0);

    T_VERIFY(tree_item_insertChildAt(NULL,0,&item1->tree_item)==false);
    T_VERIFY(tree_item_insertChildAt(root,0,NULL)==false);

    T_VERIFY(tree_item_insertChildAt(root,0,&item1->tree_item)==1);
    T_VERIFY(tree_item_insertChildAt(root,CHILDS+1,&item2->tree_item)==true);

    index = 0;
    tree_for_each_child(child, root) {
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      T_VERIFY(item->depth == 0);

      if(index == 0)
        T_VERIFY(item->childId == 88);
      else if(index == CHILDS+1)
        T_VERIFY(item->childId == 99);
      else
        T_VERIFY(item->childId == index-1);

      index++;
    }

    T_SUCCEED(destroy_tree(tree_root(&tree)));

cleanup:
    tree_cleanup(&tree);
    T_TESTCASE_END();
}

int  test_tree_removal_functions() {
    T_TESTCASE_BEGIN();
    tree_t tree;
    tree_item_t *root = NULL;
    tree_item_t *child = NULL;
    int32_tree_item_t *item1 = NULL;
    int32_tree_item_t *item2 = NULL;
    int32_tree_item_t *item = NULL;
    uint32_t index = 0;

    tree_initialize(&tree);

    // tree_item_takeChild()
    T_SUCCEED(build_tree(tree_root(&tree),CHILDS,1));

    root = tree_root(&tree);

    item1 = tree_item_data(tree_item_takeChild(tree_item_firstChild(root)),int32_tree_item_t,tree_item);
    item2 = tree_item_data(tree_item_takeChild(tree_item_firstChild(root)),int32_tree_item_t,tree_item);
    T_VERIFY(item1!=NULL);
    T_VERIFY(item2!=NULL);
    T_VERIFY(item1->childId == 0);
    T_VERIFY(item2->childId == 1);

    index = 0;
    tree_for_each_child(child, root) {
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      T_VERIFY(item->depth == 0);

      T_VERIFY(item->childId == index+2);

      index++;
    }

    tree_item_cleanup(&item1->tree_item);
    tree_item_cleanup(&item2->tree_item);
    free(item1);
    free(item2);
    T_SUCCEED(destroy_tree(tree_root(&tree)));

    // tree_item_takeFirstChild()
    T_SUCCEED(build_tree(tree_root(&tree),CHILDS,1));

    root = tree_root(&tree);

    item1 = tree_item_data(tree_item_takeFirstChild(root),int32_tree_item_t,tree_item);
    item2 = tree_item_data(tree_item_takeFirstChild(root),int32_tree_item_t,tree_item);
    T_VERIFY(item1!=NULL);
    T_VERIFY(item2!=NULL);
    T_VERIFY(item1->childId == 0);
    T_VERIFY(item2->childId == 1);

    index = 0;
    tree_for_each_child(child, root) {
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      T_VERIFY(item->depth == 0);

      T_VERIFY(item->childId == index+2);

      index++;
    }

    tree_item_cleanup(&item1->tree_item);
    tree_item_cleanup(&item2->tree_item);
    free(item1);
    free(item2);
    T_SUCCEED(destroy_tree(tree_root(&tree)));

    // tree_item_takeLastChild()
    T_SUCCEED(build_tree(tree_root(&tree),CHILDS,1));

    root = tree_root(&tree);

    item1 = tree_item_data(tree_item_takeLastChild(root),int32_tree_item_t,tree_item);
    item2 = tree_item_data(tree_item_takeLastChild(root),int32_tree_item_t,tree_item);
    T_VERIFY(item1!=NULL);
    T_VERIFY(item2!=NULL);
    T_VERIFY(item1->childId == CHILDS-1);
    T_VERIFY(item2->childId == CHILDS-2);

    index = 0;
    tree_for_each_child(child, root) {
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      T_VERIFY(item->depth == 0);

      T_VERIFY(item->childId == index);

      index++;
    }

    tree_item_cleanup(&item1->tree_item);
    tree_item_cleanup(&item2->tree_item);
    free(item1);
    free(item2);

    T_SUCCEED(destroy_tree(tree_root(&tree)));

cleanup:
    tree_cleanup(&tree);
    T_TESTCASE_END();
}

int  test_tree_remove_insert_actions() {
    T_TESTCASE_BEGIN();
    tree_t tree;
    tree_item_t *root = NULL;
    tree_item_t *child = NULL;
    int32_tree_item_t *item1 = NULL;
    int32_tree_item_t *item2 = NULL;
    int32_tree_item_t *item = NULL;
    uint32_t index = 0;

    tree_initialize(&tree);

    //test: take first item for  back and attach it as the first item
    T_SUCCEED(build_tree(tree_root(&tree),CHILDS,1));
    root = tree_root(&tree);

    item1 = tree_item_data(tree_item_takeLastChild(root),int32_tree_item_t,tree_item);
    item2 = tree_item_data(tree_item_takeLastChild(root),int32_tree_item_t,tree_item);

    T_VERIFY(item1!=NULL);
    T_VERIFY(item2!=NULL);
    T_VERIFY(item1->childId == CHILDS-1);
    T_VERIFY(item2->childId == CHILDS-2);

    T_VERIFY(tree_item_prependChild(root,&item1->tree_item)==true);
    T_VERIFY(tree_item_prependChild(root,&item2->tree_item)==true);

    index = 0;
    tree_for_each_child(child, root) {
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      T_VERIFY(item->depth == 0);

      if(index == 0)
        T_VERIFY(item->childId == CHILDS-2);
      else if(index == 1)
        T_VERIFY(item->childId == CHILDS-1);
      else
        T_VERIFY(item->childId == index-2);

      index++;
    }

    T_SUCCEED(destroy_tree(tree_root(&tree)));

    //test: take the last item of the first level and and attach it as the first ellement in the first ellement of the second level
    //      take the last item of the first ellement on the second level and attach it as the first ellement of the first level
    //      second level

    T_SUCCEED(build_tree(tree_root(&tree),CHILDS,3));
    root = tree_root(&tree);

    item1 = tree_item_data(tree_item_takeLastChild(root),int32_tree_item_t,tree_item); // last item of first level
    item2 = tree_item_data(tree_item_takeLastChild(tree_item_firstChild(root)),int32_tree_item_t,tree_item); // last item of first item of second level

    T_VERIFY(item1!=NULL);
    T_VERIFY(item1->childId == CHILDS-1);
    T_VERIFY(item1->depth   == 0);

    T_VERIFY(item2!=NULL);
    T_VERIFY(item2->childId == CHILDS-1);
    T_VERIFY(item2->depth   == 1);

    T_VERIFY(tree_item_prependChild(tree_item_firstChild(root),&item1->tree_item)==true);
    T_VERIFY(tree_item_prependChild(root,&item2->tree_item)==true);

    //check level 0
    index = 0;
    tree_for_each_child(child, root) {
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      if(index == 0) {
        T_VERIFY(item->depth == 1);
        T_VERIFY(item->childId == CHILDS-1);
      }
      else {
        T_VERIFY(item->depth == 0);
        T_VERIFY(item->childId == index-1);
      }

      index++;
    }

    //check original first ellement of level 1 (= now the second one)
    index = 0;
    tree_for_each_child(child, tree_item_childAt(root,1)) {
      item = tree_item_data(child,int32_tree_item_t,tree_item);
      T_VERIFY(item != NULL);
      if(index == 0) {
        T_VERIFY(item->depth == 0);
        T_VERIFY(item->childId == CHILDS-1);
      }
      else {
        T_VERIFY(item->depth == 1);
        T_VERIFY(item->childId == index-1);
      }
      index++;
    }

    T_SUCCEED(destroy_tree(tree_root(&tree)));

cleanup:
    tree_cleanup(&tree);
    T_TESTCASE_END();
}

int  test_tree_null() {
    T_TESTCASE_BEGIN();
    tree_t tree;
    tree_item_t item;
    T_VERIFY(tree_initialize(&tree) == true);
    T_VERIFY(tree_item_initialize(&item) == true);

    T_VERIFY(tree_initialize(NULL) == false);
    T_VERIFY(tree_item_initialize(NULL) == false);
    tree_cleanup(NULL);
    tree_item_cleanup(NULL);
    T_VERIFY(tree_root(NULL) == NULL);
    T_VERIFY(tree_item_parent(NULL) == NULL);
    T_VERIFY(tree_item_firstChild(NULL) == NULL);
    T_VERIFY(tree_item_lastChild(NULL) == NULL);
    T_VERIFY(tree_item_childAt(NULL,100) == NULL);
    T_VERIFY(tree_item_nextSibling(NULL) == NULL);
    T_VERIFY(tree_item_prevSibling(NULL) == NULL);
    T_VERIFY(tree_item_appendChild(NULL,NULL) == false);
    T_VERIFY(tree_item_appendChild(&item,NULL) == false);
    T_VERIFY(tree_item_prependChild(NULL,NULL) == false);
    T_VERIFY(tree_item_prependChild(&item,NULL) == false);
    T_VERIFY(tree_item_insertChildAt(NULL,100,NULL) == false);
    T_VERIFY(tree_item_insertChildAt(&item,100,NULL) == false);
    T_VERIFY(tree_item_takeChild(NULL) == NULL);
    T_VERIFY(tree_item_takeFirstChild(NULL) == NULL);
    T_VERIFY(tree_item_takeLastChild(NULL) == NULL);
    T_VERIFY(tree_item_childCount(NULL) == 0);
    T_VERIFY(tree_item_hasChildren(NULL) == false);

cleanup:
    tree_cleanup(&tree);
    tree_item_cleanup(&item);
    T_TESTCASE_END();
}

T_TESTCASES_BEGIN()
    T_TESTCASE(test_tree_build_and_verify, "Test tree building and verify content of the tree")
    T_TESTCASE(test_tree_reference_functions, "Test functions for walking the tree")
    T_TESTCASE(test_tree_insertion_functions, "Test insertions of branches and leafs")
    T_TESTCASE(test_tree_removal_functions, "Test removal of branches and leafs")
    T_TESTCASE(test_tree_remove_insert_actions, "Test removal and insertion of branches and leafs (moving branches)")
    T_TESTCASE(test_tree_null, "Test passing null pointers to tree functions")
T_TESTCASES_END()
