#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <debug/sahtrace.h>
#include <sahtest/sahtest.h>

#include "mock.h"

void test_open_close_setup() {
    mock_reset(openlog);
    mock_reset(open);
    mock_reset(dup2);
    mock_reset(getenv);
}

void test_open_close_teardown() {
    sahTraceClose();
}

int test_open_stdout() {
    mock_add_expectation(getenv)->varname = "SAH_TRACE_ZONES";
    mock_add_expectation(getenv)->value = NULL;

    sahTraceOpen("test",TRACE_TYPE_STDOUT);
    T_CMP(sahTraceIdentifier(),"test");
    T_VERIFY(sahTraceIsOpen());

    T_VERIFY(mock_verify_expectations(getenv) == true);

    return 0;
}

int test_open_stderr() {
    mock_add_expectation(getenv)->varname = "SAH_TRACE_ZONES";
    mock_add_expectation(getenv)->value = NULL;

    sahTraceOpen("test",TRACE_TYPE_STDERR);
    T_CMP(sahTraceIdentifier(),"test");
    T_VERIFY(sahTraceIsOpen());

    T_VERIFY(mock_verify_expectations(getenv) == true);

    return 0;
}

int test_open_syslog() {
    mock_add_expectation(getenv)->varname = "SAH_TRACE_ZONES";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(openlog)->ident = "test";
    mock_add_expectation(openlog)->field = LOG_CONS;
    mock_add_expectation(openlog)->field = LOG_USER;
    
    sahTraceOpen("test",TRACE_TYPE_SYSLOG);
    T_CMP(sahTraceIdentifier(),"test");
    T_VERIFY(sahTraceIsOpen());

    T_VERIFY(open_log_hasUnexpectedInput() == false);

    T_VERIFY(mock_verify_expectations(openlog) == true);
    T_VERIFY(mock_verify_expectations(getenv) == true);

    return 0;
}

int test_open_file() {
    mock_add_expectation(getenv)->varname = "SAH_TRACE_ZONES";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "logfile.txt";
    mock_add_expectation(open)->flags = O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR;
    mock_add_expectation(open)->fd = 10;

    mock_add_expectation(dup2)->fd = 10;
    mock_add_expectation(dup2)->fd = STDERR_FILENO;
    mock_add_expectation(dup2)->fd = 11;

    sahTraceOpenFile("test","logfile.txt");
    T_CMP(sahTraceIdentifier(),"test");
    T_VERIFY(sahTraceIsOpen());

    T_VERIFY(mock_verify_expectations(open) == true);
    T_VERIFY(mock_verify_expectations(dup2) == true);
    T_VERIFY(mock_verify_expectations(getenv) == true);

    return 0;
}

int test_open_file_fails() {
    mock_add_expectation(getenv)->varname = "SAH_TRACE_ZONES";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "logfile.txt";
    mock_add_expectation(open)->flags = O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR;
    mock_add_expectation(open)->fd = -1;
    
    mock_add_expectation(dup2)->fd = STDERR_FILENO;
    mock_add_expectation(dup2)->fd = STDERR_FILENO;
    mock_add_expectation(dup2)->fd = STDERR_FILENO;

    // tracing should not be opened when opening file fails
    sahTraceOpenFile("test","logfile.txt");
    T_VERIFY(strcmp(sahTraceIdentifier(),"test") == 0);
    T_VERIFY(sahTraceIsOpen() == true);

    T_VERIFY(mock_verify_expectations(open) == true);
    T_VERIFY(mock_verify_expectations(getenv) == true);

    return 0;
}

int test_open_file_dup_fails() {
    mock_add_expectation(getenv)->varname = "SAH_TRACE_ZONES";
    mock_add_expectation(getenv)->value = NULL;

    mock_add_expectation(open)->pathname = "logfile.txt";
    mock_add_expectation(open)->flags = O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR;
    mock_add_expectation(open)->fd = 10;
    
    mock_add_expectation(dup2)->fd = 10;
    mock_add_expectation(dup2)->fd = STDERR_FILENO;
    mock_add_expectation(dup2)->fd = -1;

    // tracing should not be opened when opening file fails
    sahTraceOpenFile("test","logfile.txt");
    T_VERIFY(strcmp(sahTraceIdentifier(),"test") == 0);
    T_VERIFY(sahTraceIsOpen() == true);

    T_VERIFY(mock_verify_expectations(open) == true);
    T_VERIFY(mock_verify_expectations(getenv) == true);

    return 0;
}

int test_open_twice() {
    mock_add_expectation(getenv)->varname = "SAH_TRACE_ZONES";
    mock_add_expectation(getenv)->value = NULL;

    sahTraceOpen("test",TRACE_TYPE_STDOUT);
    T_CMP(sahTraceIdentifier(),"test");
    T_VERIFY(sahTraceIsOpen());

    sahTraceOpen("test2",TRACE_TYPE_STDOUT);
    T_CMP(sahTraceIdentifier(),"test");
    T_VERIFY(sahTraceIsOpen());

    T_VERIFY(mock_verify_expectations(getenv) == true);

    return 0;
}

int test_close_twice() {
    mock_add_expectation(getenv)->varname = "SAH_TRACE_ZONES";
    mock_add_expectation(getenv)->value = NULL;

    sahTraceOpen("test",TRACE_TYPE_STDOUT);
    T_CMP(sahTraceIdentifier(),"test");
    T_VERIFY(sahTraceIsOpen());

    sahTraceClose();
    sahTraceClose();

    T_VERIFY(mock_verify_expectations(getenv) == true);

    return 0;
}
