#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <debug/sahtrace.h>
#include <openssl/md5.h>
#include <openssl/sha.h>
#include <usermngt/usermngt.h>
#include <usermngt/auth.h>
#include "usermngt.h"

static bool usermngt_hashMD5Password(const char *password, const char *salt __attribute__ ((unused)), char hashed[USERMNGT_PASSWORD_LENGTH])
{
  int rc = false;
  char base[USERMNGT_PASSWORD_LENGTH + 1];
  unsigned char md5[MD5_DIGEST_LENGTH];

  snprintf(base, USERMNGT_PASSWORD_LENGTH, "%s", password);

  MD5_CTX md5ctx;
  if (!MD5_Init(&md5ctx)) {
    SAH_TRACE_ERROR("unable to hash password");
    goto error;
  }
  if (!MD5_Update(&md5ctx, base, strlen(base))) {
    SAH_TRACE_ERROR("unable to hash password");
    goto error_final;
  }
  
 error_final:
  if (!MD5_Final(md5, &md5ctx)) {
    SAH_TRACE_ERROR("unable to hash password");
    goto error;
  }

  /* convert to hex */
  int i;
  for (i = 0; i < MD5_DIGEST_LENGTH; i++) {
    snprintf(hashed + (i * 2), 3, "%02x", md5[i]);
  }

  rc = true;
 error:
  return rc;
}

static bool usermngt_hashSSHA256Password(const char *password, const char *salt, char hashed[USERMNGT_PASSWORD_LENGTH])
{
  int rc = false;
  char base[USERMNGT_SALT_LENGTH + USERMNGT_PASSWORD_LENGTH + 2];
  unsigned char sha[SHA256_DIGEST_LENGTH];

  snprintf(base, USERMNGT_SALT_LENGTH +  USERMNGT_PASSWORD_LENGTH, "%s|%s", salt, password);

  SHA256_CTX shactx;
  if (!SHA256_Init(&shactx)) {
    SAH_TRACE_ERROR("unable to hash password");
    goto error;
  }
  if (!SHA256_Update(&shactx, base, strlen(base))) {
    SAH_TRACE_ERROR("unable to hash password");
    goto error_final;
  }
  
 error_final:
  if (!SHA256_Final(sha, &shactx)) {
    SAH_TRACE_ERROR("unable to hash password");
    goto error;
  }

  /* convert to hex */
  int i;
  for (i = 0; i < SHA256_DIGEST_LENGTH; i++) {
    snprintf(hashed + (i * 2), 3, "%02x", sha[i]);
  }

  rc = true;
 error:
  return rc;
}

bool usermngt_hashPassword(const char *type, const char *password, const char *salt, char hashed[USERMNGT_PASSWORD_LENGTH]) 
{
  if (strcmp(type, "MD5") == 0) {
    return usermngt_hashMD5Password(password, salt, hashed);
  } else if (strcmp(type, "SSHA256") == 0) {
    return usermngt_hashSSHA256Password(password, salt, hashed);
  } else {
    return false;
  }
}

/**
 * @brief
 * Try to authenticate a user
 *
 * @details
 * Check that a username and password pair is valid. If
 * a valid pair is found, the corresponding user is returned.
 *
 * @param username the name of the user to authenticate
 * @param password the password to use for authentication
 * @return the matching user if found, NULL otherwise
 */
const usermngt_user_t *usermngt_authenticate(const char *username, const char *password)
{
  INITOK(NULL);

  if (username == NULL || *username == '\0') {
    SAH_TRACEZ_ERROR("usermngt", "invalid user name");
    errno = EINVAL;
    return NULL;
  }

  if (password == NULL || *password == '\0') {
    SAH_TRACEZ_ERROR("usermngt", "invalid user name");
    errno = EINVAL;
    return NULL;
  }

  const usermngt_user_t *user = usermngt_userFindByName(username);
  if (user == NULL)
    return NULL;

  char hash[USERMNGT_PASSWORD_LENGTH + 1];
  switch (user->passwordType) {
  case USERMNGT_PASSWD_MD5:
    if (!usermngt_hashMD5Password(password, user->salt, hash)) {
      return NULL;
    }
    break;
  case USERMNGT_PASSWD_SSHA256:
    if (!usermngt_hashSSHA256Password(password, user->salt, hash)) {
      return NULL;
    }
    break;
  case USERMNGT_PASSWD_UNKNOWN:
    return NULL;
  }

  if (strcmp(hash, user->password) != 0) {
    return NULL;
  }

  return user;
}
