/**
 * @file usermngt.c
 * @mainpage User datamodel access library
 * 
 * This library provides support for accessing the UserManagement datamodel.
 *
 * The system has many groups and users. A users belongs to many groups. A user
 * as a id, a name and an enable flag. A group as an id, a name and an enable flag.
 *
 * To fetch specific user, see
 * - usermngt_userFindByID() and
 * - usermngt_userFindByName().
 *
 * To loop over all users, see
 * - usermngt_userFirst() and
 * - usermngt_userNext().
 *
 * To fetch specific group, see
 * - usermngt_groupFindByID() and
 * - usermngt_groupFindByName().
 *
 * To loop over all groupss, see
 * - usermngt_groupFirst() and
 * - usermngt_groupNext().
 *
 * To check if a user belongs to a group, see
 * - usermngt_userIsInGroup()
 *
 * To fetch a specific group to which a user belongs to, see
 * - usermngt_userGroupFindByID() and
 * - usermngt_userGroupFindByName().
 *
 * To loop over the groups a user belongs to, see
 * - usermngt_userGroupFirst() and
 * - usermngt_userGroupNext().
 *
 * To authenticate a user, see
 * - usermngt_authenticate().
 */

#include <usermngt/usermngt.h>
#include <debug/sahtrace.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include "usermngt.h"

struct usermngt {
  int passwdfd;
  int iuser;
  int iusergroup;
  int igroup;
  usermngt_model_t *model;
  int error;
  time_t passwdChanged;
} usermngt;

int usermngt_getLastError()
{
  return usermngt.error;
}

int usermngt_usercmp(const void *a, const void *b)
{
  return ((usermngt_user_t *) a)->uid -  ((usermngt_user_t *) b)->uid;
}

int usermngt_groupcmp(const void *a, const void *b)
{
  return ((usermngt_group_t *) a)->gid -  ((usermngt_group_t *) b)->gid;
}

static __attribute__ ((constructor)) void usermngt_initialize()
{
  SAH_TRACEZ_INFO("usermngt", "initializing library");

  usermngt.passwdfd = open(USERMNGT_PASSWD_FILE, O_RDONLY);
  if (usermngt.passwdfd < 0) {
    usermngt.error = errno;
    usermngt.model = MAP_FAILED;
    SAH_TRACE_FATAL("unable to open password file %s: %m (is usermngt plugin started ?)", USERMNGT_PASSWD_FILE);
    return;
  }
  
  struct stat sb;
  fstat(usermngt.passwdfd, &sb);
  usermngt.passwdChanged = sb.st_mtime;
  usermngt.model = mmap(NULL, sizeof(usermngt_model_t), PROT_READ, MAP_PRIVATE, usermngt.passwdfd, 0);
  if (usermngt.model == MAP_FAILED) {
    usermngt.error = errno;
    SAH_TRACE_FATAL("unable to access password file %s: %m", USERMNGT_PASSWD_FILE);
    return;
  }
  usermngt.error = 0;
}

static __attribute__ ((destructor)) void usermngt_cleanup()
{
  if (usermngt.model != MAP_FAILED)
    munmap(usermngt.model, sizeof(usermngt_model_t));
  if (usermngt.passwdfd >= 0)
    close(usermngt.passwdfd);
  usermngt.error = 0;
}

void usermngt_checkChanges()
{
  struct stat sb;
  if (stat(USERMNGT_PASSWD_FILE, &sb) < 0) {
    SAH_TRACE_FATAL("unable to stat %s: %m", USERMNGT_PASSWD_FILE);
    usermngt.error = errno;
  }

  if (usermngt.passwdfd < 0 || sb.st_mtime != usermngt.passwdChanged) {
    usermngt_cleanup();
    usermngt_initialize();
  }
}


/**
 * @brief
 * Fetch a user by its name
 *
 * @details
 * This function fetches a user matching a name. The usermngt_user_t 
 * returned can then be used with the various accessors.
 *
 * If there was an error, NULL is returned and errno is set.
 * If the user cannot be found, errno will be set to ENOENT.
 *
 * The searching algorithm is linear.
 *
 * @param username the name of the user to fetch
 * @return the user matching the name, NULL if it could not be found.
 */
const usermngt_user_t *usermngt_userFindByName(const char *username)
{
  INITOK(NULL);

  if (username == NULL || *username == '\0') {
    SAH_TRACEZ_ERROR("usermngt", "invalid user name");
    errno = EINVAL;
    return NULL;
  }

  usermngt_user_t *user = NULL;
  int i;
  for (i = 0; i < usermngt.model->userCount; i++) {
    if (strncmp(usermngt.model->users[i].name, username, USERMNGT_USERNAME_LENGTH) == 0) {
      user = &(usermngt.model->users[i]);
      break;
    }
  }

  if (user == NULL)
    errno = ENOENT;

  return user;
}

/**
 * @brief
 * Fetch a user by its id
 *
 * @details
 * This function fetches a user matching an id. The usermngt_user_t 
 * returned can then be used with the various accessors.
 *
 * If there was an error, NULL is returned and errno is set.
 * If the user cannot be found, errno will be set to ENOENT.
 *
 * The searching algorithm is using a binary search.
 *
 * @param uid the user id of the user to fetch
 * @return the user matching the user id, NULL if it could not be found.
 */
const usermngt_user_t *usermngt_userFindByID(uint32_t uid)
{
  INITOK(NULL);

  if (uid == 0) {
    SAH_TRACEZ_ERROR("usermngt", "invalid user id");
    errno = EINVAL;
    return NULL;
  }

  usermngt_user_t key = { .uid = uid, 0 };
  usermngt_user_t *user = bsearch(&key, usermngt.model->users, usermngt.model->userCount,
                                  sizeof(usermngt_user_t), usermngt_usercmp);
  if (user == NULL)
    errno = ENOENT;

  return user;
}

/**
 * @brief
 * Fetch the first user in the list of user
 *
 * @details
 * This function fetches the first user in the user list. 
 *
 * If there was an error, NULL is returned and errno is set
 *
 * @return the first user, NULL if there are no users
 */
const usermngt_user_t *usermngt_userFirst()
{
  INITOK(NULL);

  if (usermngt.model->userCount == 0)
    return NULL;
  
  usermngt.iuser = 0;
  return usermngt.model->users;
}

/**
 * @brief
 * Fetch the next user in the list of user
 *
 * @details
 * This function fetches the next user in the user list. 
 *
 * If there was an error, NULL is returned and errno is set.
 *
 * @return the next user, NULL if there are no more users
 */
const usermngt_user_t *usermngt_userNext()
{
  INITOK(NULL);

  if (++usermngt.iuser == usermngt.model->userCount)
    return NULL;

  return usermngt.model->users + usermngt.iuser;
}

/**
 * @brief
 * Fetch the first group in the user's group list.
 *
 * @details
 * This function fetches the first group in the list of groups a user belongs to. 
 *
 * If there was an error, NULL is returned and errno is set
 *
 * @return the first group, NULL if this user does not belong to any group
 */
const usermngt_group_t *usermngt_userGroupFirst(const usermngt_user_t *user)
{
  INITOK(NULL);
  if (user->groupCount == 0)
    return NULL;
  
  usermngt.iusergroup = 0;
  return user->groups;
}

/**
 * @brief
 * Fetch the next user in the list of user
 *
 * @details
 * This function fetches the next user in the user list. 
 *
 * If there was an error, NULL is returned and errno is set.
 *
 * @return the user matching the user id, NULL if there are no more users
 */
const usermngt_group_t *usermngt_userGroupNext(const usermngt_user_t *user)
{
 INITOK(NULL);

  if (++usermngt.iusergroup == user->groupCount)
    return NULL;

  return user->groups + usermngt.iusergroup;
}
/**
 * @brief
 * Fetch a group from a user's group list by its name
 *
 * @details
 * This function fetches a group from the list of groups a user belongs to by 
 * matching a group name. 
 *
 * If there was an error, NULL is returned and errno is set.
 * If the group cannot be found, errno will be set to ENOENT.
 *
 * The searching algorithm is linear.
 *
 * @param user the user belonging to the group
 * @param groupname the name of the group to fetch
 * @return the group matching the group name, NULL if it could not be found.
 */
const usermngt_group_t *usermngt_userGroupFindByName(const usermngt_user_t *user, const char *groupname)
{
  INITOK(NULL);

  if (groupname == NULL || *groupname == '\0') {
    SAH_TRACEZ_ERROR("usermngt", "invalid group name");
    errno = EINVAL;
    return NULL;
  }

  const usermngt_group_t *group = NULL;
  int i;
  for (i = 0; i < user->groupCount; i++) {
    if (strncmp(user->groups[i].name, groupname, USERMNGT_GROUPNAME_LENGTH) == 0) {
      group = &(user->groups[i]);
      break;
    }
  }

  if (group == NULL)
    errno = ENOENT;

  return group;
}

/**
 * @brief
 * Fetch a group from a user's group list by its id
 *
 * @details
 * This function fetches a group from the list of groups a user belongs to
 * by matching an id.
 *
 * If there was an error, NULL is returned and errno is set.
 * If the user cannot be found, errno will be set to ENOENT.
 *
 * The searching algorithm is using a binary search.
 *
 * @param user the user belonging to the group
 * @param gid the user id of the user to fetch
 * @return the user matching the user id, NULL if it could not be found.
 */
const usermngt_group_t *usermngt_userGroupFindByID(const usermngt_user_t *user, uint32_t gid)
{
  INITOK(NULL);

  if (gid == 0) {
    SAH_TRACEZ_ERROR("usermngt", "invalid group id");
    errno = EINVAL;
    return NULL;
  }

  usermngt_group_t key = { .gid = gid, 0 };
  usermngt_group_t *group = bsearch(&key, user->groups, user->groupCount,
                                    sizeof(usermngt_group_t), usermngt_groupcmp);
  if (group == NULL)
    errno = ENOENT;

  return group;
}

/**
 * @brief
 * Fetch a group by its name
 *
 * @details
 * This function fetches a group matching a name. The usermngt_group_t 
 * returned can then be used with the various accessors.
 *
 * If there was an error, NULL is returned and errno is set.
 * If the group cannot be found, errno will be set to ENOENT.
 *
 * The searching algorithm is linear.
 *
 * @param groupname the name of the group to fetch
 * @return the group matching the name, NULL if it could not be found.
 */
const usermngt_group_t *usermngt_groupFindByName(const char *groupname)
{
  INITOK(NULL);

  if (groupname == NULL || *groupname == '\0') {
    SAH_TRACEZ_ERROR("usermngt", "invalid group name");
    errno = EINVAL;
    return NULL;
  }

  usermngt_group_t *group = NULL;
  int i;
  for (i = 0; i < usermngt.model->groupCount; i++) {
    if (strncmp(usermngt.model->groups[i].name, groupname, USERMNGT_GROUPNAME_LENGTH) == 0) {
      group = &(usermngt.model->groups[i]);
      break;
    }
  }

  if (group == NULL)
    errno = ENOENT;

  return group;

}

/**
 * @brief
 * Fetch a group by its id
 *
 * @details
 * This function fetches a group matching an id. The usermngt_group_t 
 * returned can then be used with the various accessors.
 *
 * If there was an error, NULL is returned and errno is set.
 * If the group cannot be found, errno will be set to ENOENT.
 *
 * The searching algorithm is using a binary search.
 *
 * @param gid the group id of the group to fetch
 * @return the group matching the group id, NULL if it could not be found.
 */
const usermngt_group_t *usermngt_groupFindByID(uint32_t gid)
{
  INITOK(NULL);

  if (gid == 0) {
    SAH_TRACEZ_ERROR("usermngt", "invalid group id");
    errno = EINVAL;
    return NULL;
  }

  usermngt_group_t key = { .gid = gid, 0 };
  usermngt_group_t *group = bsearch(&key, usermngt.model->groups, usermngt.model->groupCount,
                                  sizeof(usermngt_group_t), usermngt_groupcmp);
  if (group == NULL)
    errno = ENOENT;

  return group;
}


/**
 * @brief
 * Fetch the first group in the list of group
 *
 * @details
 * This function fetches the first group in the group list. 
 *
 * If there was an error, NULL is returned and errno is set
 *
 * @return the first group, NULL if there are no groups
 */
const usermngt_group_t *usermngt_groupFirst()
{
  INITOK(NULL);

  if (usermngt.model->groupCount == 0)
    return NULL;
  
  usermngt.igroup = 0;
  return usermngt.model->groups;
}

/**
 * @brief
 * Fetch the next group in the list of group
 *
 * @details
 * This function fetches the next group in the group list. 
 *
 * If there was an error, NULL is returned and errno is set.
 *
 * @return the next group, NULL if there are no more groups
 */
const usermngt_group_t *usermngt_groupNext()
{
  INITOK(NULL);

  if (++usermngt.igroup == usermngt.model->groupCount)
    return NULL;

  return usermngt.model->groups + usermngt.igroup;
}

