/* -*- c-basic-offset: 4 -*- */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <assert.h>

#include "circular_buffer.h"


inline ssize_t max(ssize_t a, ssize_t b)
{
    return (a > b ? a : b);
}

inline ssize_t min(ssize_t a, ssize_t b)
{
    return (a > b ? b : a);
}

#ifndef NDEBUG
static void circbuf_dump(struct _circbuf *cb)
{
    printf("[");
    fwrite(cb->buffer, sizeof(char), cb->size, stdout);
    printf("]\n");
    if (cb->read < cb->write)
        printf(" % *s% *s\n", cb->read - cb->buffer, "r", cb->write - cb->read, "w");
    else if (cb->read > cb->write)
        printf(" % *s% *s\n", cb->write - cb->buffer, "w", cb->read - cb->write, "r");
    else
        printf(" % *s\n", cb->write - cb->buffer, "b");
}

#define assert_circbuf(CB) circbuf_isValid(CB, __FILE__, __LINE__)
static void circbuf_isValid(struct _circbuf *cb, const char *file, int line)
{
    assert(cb->buffer);
    assert(cb->size > 0);
    assert(cb->read);
    assert(cb->toRead >= 0);
    assert(cb->write);
    assert(cb->toWrite >= 0);
    assert(cb->size >= cb->toRead + cb->toWrite);
    assert(cb->buffer + cb->size > cb->read);
    assert(cb->buffer <= cb->read);
    assert(cb->buffer + cb->size > cb->write);
    assert(cb->buffer <= cb->write);
    //circbuf_dump(cb);
}
#else
#define assert_circbuf(CB) ((void) 0)
#endif

static ssize_t circbuf_resize(struct _circbuf *cb, size_t size)
{
    assert_circbuf(cb);
 
    /* store the offset of the read/write pointer to restore them */
    off_t read = cb->read - cb->buffer;
    off_t write = cb->write - cb->buffer;

    /* we might need to move the read pointer if it is after the write*/
    int readIsAfterWrite = (cb->toRead > 0 && cb->read == cb->write) || cb->read > cb->write;

    /* allocate more memory */
    char *newBuffer = realloc(cb->buffer, cb->size + size);
    if (newBuffer == NULL)
        return -1;

    /* restore state */
    cb->buffer = newBuffer;
    cb->write = cb->buffer + write;


    /* we might need to move the read pointer if it is after the write*/
    if (readIsAfterWrite)
        cb->read = cb->buffer + read + size;
    else 
        cb->read = cb->buffer + read;

    /* we need to insert the free space at the write pointer */
    char *r = cb->buffer + cb->size - 1, *w = cb->buffer + cb->size + size -1;
    while (r > cb->write) {
        assert(r < w);
        *w-- = *r--;
    }

    /* we now have more space to write to */
    cb->size += size;
    cb->toWrite +=  size;

    assert_circbuf(cb);

    return cb->size;
}

static ssize_t circbuf_read(void *cookie, char *buffer, size_t size)
{
    struct _circbuf *cb = (struct _circbuf *) cookie;
    assert_circbuf(cb);

    /* EOF is when the readPos and writePos matches */
    if (cb->toRead == 0)
        return 0;

    ssize_t toRead = min(size, cb->toRead);
    ssize_t read = 0;
    while (read < toRead) {
        *(buffer + read) = *cb->read;
        read ++;
        cb->read ++;
        cb->toRead --;
        cb->toWrite ++;

        /* reset the read pointer if we cross the buffer boundary */
        if (cb->read >= cb->buffer + cb->size)
            cb->read = cb->buffer;
    }

    assert(read == toRead);
    assert_circbuf(cb);

    return toRead;
}

static ssize_t circbuf_write(void *cookie, const char *buffer, size_t size)
{
    struct _circbuf *cb = (struct _circbuf *) cookie;
    clearerr(cb->stream);
    errno = 0;
    
    /* check if we have enough space */
    if (size >= cb->toWrite) {
        if (circbuf_resize(cb, max(size, cb->size)) < 0)
            return -1;
    }
        
    ssize_t writen = 0;
    while (writen < size) {
        *cb->write = *(buffer + writen);
        writen ++;
        cb->write ++;
        cb->toWrite --;
        cb->toRead ++;
        /* reset the read pointer if we cross the buffer boundary */
        if (cb->write >= cb->buffer + cb->size)
            cb->write = cb->buffer;
    }
    
    assert(writen == size);
    assert_circbuf(cb);
    return size;
}

/**
 * when seeking on a circular buffer for read, the beginning and the end are
 * at the write pointer. When seeking a write buffer, the beginning and the end are 
 * at the read pointer. Moving further thant the beginning or end of the buffer WON't
 * resize it.
 */
static int circbuf_seek(void *cookie, off64_t *position, int whence)
{
    struct _circbuf *cb = (struct _circbuf *) cookie;
    char **seek = NULL, *bound = NULL;
    size_t *toSeek = NULL, *toBound = NULL;

    assert_circbuf(cb);

    /* check which pointer we have to move */
    switch (cb->seekMode) {
    case 'r':
        seek = &cb->read;
        toSeek = &cb->toRead;
        bound = cb->write;
        toBound = &cb->toWrite;
        break;
    case 'a':
    case 'w':
        seek = &cb->write;
        toSeek = &cb->toWrite;
        bound = cb->read;
        toBound = &cb->toRead;
        break;
    default:
        errno = EINVAL;
        return -1;
    }

    /* seek from ... */
    switch (whence) {
    case SEEK_SET: /* ... the beginning */
        /* set position to the beginning */
        *seek = bound;
        *toSeek = cb->size;
        *toBound = 0;
        break;

    case SEEK_END: /* ... the end */
        /* set position to the end */
        *seek = bound;
        *toSeek = 0;
        *toBound = cb->size;
        break;
    }

    int8_t sign = abs(*position) / *position;
    off64_t moved = 0;
    for (moved = 0; moved < ((off64_t) sign) * *position; moved ++) {
        *toSeek -= sign;
        *toBound += sign;
        *seek += sign;

        if (*seek < cb->buffer)
            *seek = cb->buffer + cb->size - 1;
        else if (*seek >= cb->buffer + cb->size)
            *seek = cb->buffer;

        if (*seek == bound)
            break;
    }
    assert_circbuf(cb);

    /* we now need to get the absolute position */
    *position = moved;

    return 0;
}

static int circbuf_close(void *cookie)
{
    struct _circbuf *cb = (struct _circbuf *) cookie;
    cb->stream = NULL;
    return 1;
}

bool circbuf_initialize(circbuf_t *circbuf, size_t size)
{
    struct _circbuf *cb = (struct _circbuf *) circbuf;

    cb->stream = NULL;

    cb->buffer = calloc(sizeof(char), size);
    if (cb->buffer == NULL) {
        errno = ENOMEM;
        goto error;
    }

    memset(cb->buffer, 'X', size);

    /* initialize buffer state */
    cb->read = cb->buffer;
    cb->write = cb->buffer;
    cb->size = size;
    cb->toRead = 0;
    cb->toWrite = size;

    assert_circbuf(cb);
    return true;

 error:
    return false;
}

/**
 * cleanup resources
 */
void circbuf_cleanup(circbuf_t *circbuf)
{
    struct _circbuf *cb = (struct _circbuf *) circbuf;
    if (cb->stream)
        fclose(cb->stream);
    free(cb->buffer);
}

/**
 * empty the buffer
 */
void circbuf_reset(circbuf_t *circbuf)
{
    struct _circbuf *cb = (struct _circbuf *) circbuf;
    cb->toRead = 0;
    cb->toWrite = cb->size;
    assert_circbuf(cb);
}

/**
 * return the file pointer
 */
FILE *circbuf_stream(circbuf_t *circbuf)
{
    return ((struct _circbuf *) circbuf)->stream;
}

/**
 * return the total allocated size for the buffer
 */
ssize_t circbuf_size(circbuf_t *circbuf)
{
    return ((struct _circbuf *) circbuf)->size;
}

/**
 * return the bytes available for reading as contiguous memory
 */
ssize_t circbuf_availableForRead(circbuf_t *circbuf)
{
    struct _circbuf *cb = (struct _circbuf *) circbuf;
    ssize_t leftUntilBoundary = cb->buffer + cb->size - cb->read;

    return min(leftUntilBoundary, cb->toRead);
}

/**
 * return a buffer for reading
 */
void *circbuf_readBuffer(circbuf_t *circbuf)
{
    return ((struct _circbuf *) circbuf)->read;
}

/**
 * mark x bytes as read and return the number of bytes marked as read
 */
ssize_t circbuf_markAsRead(circbuf_t *circbuf, size_t size)
{
    struct _circbuf *cb = (struct _circbuf *) circbuf;

    if (cb->toRead == 0)
        return 0;

    ssize_t toRead = min(size, cb->toRead);

    /* if the read is crossing the buffer boundary */
    ssize_t leftUntilBoundary = cb->buffer + cb->size - cb->read;
    if (toRead >= leftUntilBoundary)
        cb->read = cb->buffer + toRead - leftUntilBoundary;
    else 
        cb->read += toRead;

    cb->toRead -= toRead;
    cb->toWrite += toRead;

    assert_circbuf(cb);

    return toRead;
}

/**
 * open a stream usable with the stdio functions
 * When opening with r or r+, seeking the stream moves the read pointer.
 * When opening with w, w+, a or a+, seeking moves the write pointer.
 */
FILE *circbuf_fopen(circbuf_t *circbuf, const char *opentype)
{
    struct _circbuf *cb = (struct _circbuf *) circbuf;

    cookie_io_functions_t hooks = {
        .read  = &circbuf_read,
        .write = &circbuf_write,
        .seek  = &circbuf_seek,
        .close = &circbuf_close
    };

    if ((cb->stream = fopencookie(cb, opentype, hooks)) == NULL) 
        goto error;

    /* buffer completely so we avoid writing /reading one byte at a time
     * it is the responsability from the writer to flush the buffer*/
    if (setvbuf(cb->stream, (char *) NULL, _IOFBF, 0) != 0)
        goto error_close_cb_stream;


    /* force eof if circbuf is empty */
    if (cb->toRead == 0) {
        char c;
        fread(&c, sizeof(char), 1, cb->stream);
    }

    /* set the seek mode */
    cb->seekMode = *opentype;

    return cb->stream;

 error_close_cb_stream:
    fclose(cb->stream);
 error:
    return NULL;
}

