#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <assert.h>
#include <fcntl.h>

#include "cupsd.h"
#include "circular_buffer.h"


/**** INTERNAL ROUTINES ****/

static int ps_open_printer_fd(void)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_open_printer_fd()");
  int fd = -1;
  char device[] = "/dev/usb/lp0";

  if ((fd = open(device, O_RDWR | O_EXCL | O_NONBLOCK)) < 0 && errno == ENOENT)
     return -1;

  return (fd);
}


static int ps_open_printer()
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_open_printer()");
  struct termios opts;			/* Parallel port options */
  int device_fd = -1;

  if ((device_fd = ps_open_printer_fd()) == -1)
  {
    if (errno == EBUSY)
    {
      _cupsLangPuts(stderr,
                    _("INFO: Printer busy; Cannot print...\n"));
      return -1;
    }
    else if (errno == ENXIO || errno == EIO || errno == ENOENT ||
             errno == ENODEV)
    {
      _cupsLangPuts(stderr,
                    _("INFO: Printer not connected; Cannot print\n"));
       return -1;
    }
    else
    {
       _cupsLangPuts(stderr,
                      _("ERROR: Unable to open device file\n"));
       return -1;
    }
  }

 /*
  * Set any options provided...
  */

  tcgetattr(device_fd, &opts);

  opts.c_lflag &= ~(ICANON | ECHO | ISIG );	/* Raw mode */

  /**** No options supported yet ****/

  tcsetattr(device_fd, TCSANOW, &opts);

  return device_fd;
}

/**** printer state machine routines ****/


int ps_printer_ready(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_printer_ready()");
  if((myStream->printer_status == PRINTER_IDLE || myStream->printer_status == PRINTER_OPENED) && myStream->stream_status == STREAM_IDLE ) {
    cupsdLogMessage(CUPSD_LOG_INFO, "ps_printer_ready(): Printer is ready");
    return 1;
  }
  return 0;
}

int ps_printer_open(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_printer_open()");

  myStream->writeFD        = ps_open_printer();
  cupsdLogMessage(CUPSD_LOG_INFO, "ps_printer_open(): opened: %d",myStream->writeFD);
  if(myStream->writeFD < 0) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_printer_open(): Cannot open printer file descriptor");
    return 0;
  }
  myStream->printer_status = PRINTER_OPENED;

  return 1;
}

int ps_printer_close(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_printer_close()");


  fclose(myStream->buffer);
  circbuf_cleanup(myStream->cb);
  free(myStream->cb);

  myStream->cb = calloc(1,sizeof(circbuf_t));
  if(!myStream) {
    free(myStream);
    return NULL;
  }
  circbuf_initialize(myStream->cb, PS_BUFFER_SIZE);
  myStream->buffer = circbuf_fopen(myStream->cb, "w+");
  if(!myStream->buffer) {
    cupsdLogMessage(CUPSD_LOG_INFO, "ps_printer_close(): errror re-initializing file descriptor");
    free(myStream);
    return NULL;
  }


  if(myStream->writeFD != -1) {
    int c = close(myStream->writeFD);  
    if(c==0) {
      cupsdLogMessage(CUPSD_LOG_INFO, "ps_printer_close(): Printer closed succesfully");
    } else if (c==-1){
      cupsdLogMessage(CUPSD_LOG_INFO, "ps_printer_close(): Printer closing failed: %s", strerror(errno));
    } else {
      cupsdLogMessage(CUPSD_LOG_INFO, "ps_printer_close(): Printer close should NOT retrun this value!");
    }
    myStream->writeFD = -1;
  }
  myStream->printer_status = PRINTER_IDLE;

  return 1;
}

int ps_printer_start(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_printer_start()");
  if(myStream->printer_status != PRINTER_OPENED) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_printer_start(): Invalid printer state to start the printer");
    return 0;
  }
  myStream->printer_status = PRINTER_PRINTING;

  return 1;
}

int ps_printer_stop(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_printer_stop()");
  if(myStream->printer_status != PRINTER_PRINTING) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_printer_stop(): Invalid printer state to stop the printer");
    return 0;
  }
  myStream->printer_status = PRINTER_OPENED;

  return 1;
}

int ps_printer_error(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_printer_error()");
  if(myStream->printer_status != PRINTER_PRINTING) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_printer_error(): Invalid printer state to put the printer into error state");
    return 0;
  }
  myStream->printer_status = PRINTER_ERROR;

  return 1;
}

int ps_printer_is_printing(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_printer_is_printing()");
  if(myStream->printer_status == PRINTER_PRINTING) {
    return 1;
  }
  return 0;
}

/**** stream state machine routines ****/

int ps_stream_prepare(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_stream_prepare()");
  if(myStream->stream_status != STREAM_IDLE) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_stream_prepare(): Cannot prepare streaming because the stream status is wrong!");
    return 0;
  }
  myStream->stream_status = STREAM_STARTING;

  return 1;
}

int ps_stream_bind(print_stream* myStream, int streamFD)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_stream_bind()");
  if(myStream->stream_status != STREAM_STARTING) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_stream_bind(): Cannot bind streaming because the stream status is wrong!");
    return 0;
  }
  myStream->readFD = streamFD;
  myStream->stream_status = STREAM_STREAMING;

  return 1;
}

int ps_stream_release(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_stream_release()");
  if(myStream->stream_status != STREAM_STREAMING) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_stream_release(): Cannot release streaming because the stream status is wrong!");
    return 0;
  }

  myStream->readFD = -1;
  myStream->stream_status = STREAM_DONE;

  return 1;
}

int ps_stream_reset(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_stream_reset()");
  myStream->stream_status = STREAM_IDLE;
  myStream->readFD = -1;

  return 1;
}


/**** print_stream struct routines ****/

print_stream* ps_print_stream_init()
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_init_print_stream()");

  print_stream* myStream = calloc(1,sizeof(print_stream));
  if(!myStream)
    return NULL;

  myStream->writeFD        = -1;
  myStream->readFD         = -1;
  myStream->lastRead       = 0;
  myStream->lastWrite      = 0;
  myStream->printer_status = PRINTER_IDLE;
  myStream->stream_status  = STREAM_IDLE;

  myStream->cb = calloc(1,sizeof(circbuf_t));
  if(!myStream) {
    free(myStream);
    return NULL;
  }
  circbuf_initialize(myStream->cb, PS_BUFFER_SIZE);
  myStream->buffer = circbuf_fopen(myStream->cb, "w+");
  if(!myStream->buffer) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_init_print_stream(): Error allocating buffer");
    free(myStream);
    return NULL;
  }
  cupsdLogMessage(CUPSD_LOG_INFO, "ps_init_print_stream(): print stream struct created");
  return myStream;
}

/*void ps_print_stream_cleanup(print_stream** myStream)
{
  fclose(myStream->buffer);
  free(*myStream);
  myStream = NULL;
  cupsdLogMessage(CUPSD_LOG_INFO, "ps_init_print_stream(): print stream struct closed");
}*/

void ps_print_stream_status()
{
  char *printer_status = "";
  char *stream_status = "";

  switch (printStream->printer_status) {
    case PRINTER_IDLE:
      printer_status = "PRINTER_IDLE";      break;
    case PRINTER_OPENED:
      printer_status = "PRINTER_OPENED";    break;
    case PRINTER_PRINTING:
      printer_status = "PRINTER_PRINTING";  break;
    case PRINTER_ERROR:
      printer_status = "PRINTER_ERROR";     break;
    default:
      printer_status = "UNKNOWN STATE";     break;
  };

  switch (printStream->stream_status) {
    case STREAM_IDLE:
      stream_status = "STREAM_IDLE";       break;
    case STREAM_STARTING:
      stream_status = "STREAM_STARTING";   break;
    case STREAM_STREAMING:
      stream_status = "STREAM_STREAMING";  break;
    case STREAM_DONE:
      stream_status = "STREAM_DONE";       break;
    default:
      stream_status = "UNKNOWN STATE";     break;
  };


  cupsdLogMessage(CUPSD_LOG_DEBUG, "cupsdDoSelect: printer FD=%d (ts=%d) - status = %s",printStream->writeFD,printStream->lastWrite,printer_status);
  cupsdLogMessage(CUPSD_LOG_DEBUG, "cupsdDoSelect: stream  FD=%d (ts=%d) - status = %s",printStream->readFD,printStream->lastRead,stream_status);
  cupsdLogMessage(CUPSD_LOG_DEBUG, "cupsdDoSelect: buffer  size=%d (available for read =%d)",circbuf_size(printStream->cb),circbuf_availableForRead(printStream->cb));
}


/**** stream public routines ****/


/*
 * return value:
 *  1 = mask stream
 *  0 = do not mask stream
 */
int ps_stream_mask(print_stream* myStream, int fd)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_stream_mask()");
  int ret = 0;

  if(!myStream) {
    return ret;
  }
  
  if(myStream->readFD == fd) {
    // this is the stream file descriptor, find out if we are ready to fetch some more data

    // only fetch more data if:
    //  - the buffer contains < PS_BUFFER_FILL
    //  - PS_HTTP_INTERVAL seconds have passed, to avoid http/ipp timeouts on the http POST message
    if(circbuf_availableForRead(myStream->cb) == 0)
      ret = 0;
    else if(time(NULL) - myStream->lastRead > PS_HTTP_INTERVAL) {
      ret = 0;
    } else {
      ret = 1;
    }
  }

  return ret;
}

int ps_stream_is_starting(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_stream_is_starting");
  int ret = 0;

  if(!myStream) 
    return ret;
  if(myStream->stream_status == STREAM_STARTING) 
    ret = 1;

  return ret;
}


int ps_stream_overflow(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_stream_overflow()");
  int ret = 0;

  if(!myStream)     
    return ret;
  if(circbuf_size(myStream->cb) > PS_BUFFER_SIZE)   
    ret = 1;
  return ret;
}


/****  print public routines ****/


int ps_print_job_start(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_start()");
  // check the state machines
  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_start_print_job(): stream structure is not properly initialized!");
    return 0;
  }

  if(myStream->printer_status != PRINTER_IDLE && myStream->printer_status != PRINTER_OPENED) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_start_print_job(): Cannot start print job because the print status is wrong!");
    return 0;
  }

  if(myStream->stream_status != STREAM_IDLE) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_start_print_job(): Cannot start print job because the stream status is wrong!");
    return 0;
  }

  // open the printer when needed
  if(myStream->printer_status == PRINTER_IDLE) {
    if(!ps_printer_open(myStream)) {
      cupsdLogMessage(CUPSD_LOG_ERROR, "ps_start_print_job(): Cannot open the printer!");
      return 0;
    }
  }
  if(!ps_printer_start(myStream)) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_start_print_job(): Cannot start the printer!");
    return 0;
  }
 
  // update the stream status
  ps_stream_prepare(myStream);

  return 1;

}

int ps_print_job_get_data(print_stream* myStream, char *data, int bytes)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_get_data()");
  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_get_data(): stream structure is not properly initialized!");
    return 0;
  }

  if(fwrite(data, sizeof(char), bytes, myStream->buffer) < 0) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_get_data(): fwrite failed!");
    return 0;
  }
  fflush(myStream->buffer);

  myStream->lastRead = time (NULL);

  return 1;
}


int ps_print_job_data_available(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_data_available()");
  int ret = 0;

  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_data_available(): stream structure is not properly initialized!");
    return ret;
  }

  if(circbuf_availableForRead(myStream->cb) != 0)
    ret = 1;

  return ret;
}

int ps_print_job_put_data(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_put_data()");
  int ret = 0;
  int    selectRes = 0;
  fd_set writefs;
  struct timeval Timeout;
  Timeout.tv_usec = 100;           /* milliseconds */
  Timeout.tv_sec  = 0;          /* seconds */

  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_put_data(): stream structure is not properly initialized!");
    return ret;
  }

  if(myStream->writeFD == -1) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_put_data(): Invalid write file descriptor");
    ps_printer_error(myStream);
    return ret;
  }


  FD_ZERO(&writefs);                      /* clear writefs */
  FD_SET(myStream->writeFD, &writefs);    /* set testing for source 1 */  

  selectRes = select(myStream->writeFD + 1, NULL, &writefs, NULL, &Timeout);
  if(selectRes==-1) {
    // something went wrong?
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_put_data(): Select went wrong?");
    ps_printer_error(myStream);
  } else if(selectRes>0) {
    // ready for writing
    cupsdLogMessage(CUPSD_LOG_INFO, "ps_print_job_put_data(): Writing data to the printer");
    int writtenBytes  = write(myStream->writeFD, (char *)circbuf_readBuffer(myStream->cb), circbuf_availableForRead(myStream->cb));
    if(writtenBytes == -1) {
      // something went wrong?
      ps_printer_error(myStream);
      cupsdLogMessage(CUPSD_LOG_INFO, "ps_print_job_put_data(): Error writing the data to the printer");
    } else {
      // written some bytes, update the buffer!
      cupsdLogMessage(CUPSD_LOG_INFO, "ps_print_job_put_data(): write succesfull - writtenBytes=%d",writtenBytes);
      ret = 1;
      int check = circbuf_markAsRead(myStream->cb, writtenBytes);
      if(check != writtenBytes)
        cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_put_data(): error marking the bytes as being read!");
      myStream->lastWrite = time(NULL);
    }
  }
  else {
    // timer expired, cannot write currently, just keep all settings
    if(time(NULL) - myStream->lastWrite > PS_PRINT_ERROR_TIMEOUT) {
      cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_put_data(): print error timeout - there must be something wrong with the printer");
      // something definitely went wrong on the printer, it was impossible
      // to write data to it during PS_PRINT_ERROR seconds, abort     
      ps_printer_error(myStream);
    }
  }


  return ret;
}

int ps_print_job_finished(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_finished()");
  int ret = 0;
  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_finished(): stream structure is not properly initialized!");
    return ret;
  }

  if(myStream->stream_status == STREAM_DONE && !ps_print_job_data_available(myStream))
    ret = 1;

  return ret;
}


int ps_print_job_finish(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_finish()");
  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_finish(): stream structure is not properly initialized!");
    return 0;
  }
  
  ps_printer_stop(myStream);

  return 1;
}


int ps_print_job_error(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_error()");
  int ret = 0;
  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_error(): stream structure is not properly initialized!");
    return ret;
  }

  if(myStream->printer_status == PRINTER_ERROR)
    ret = 1;

  return ret;
}

int ps_print_job_reset(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_reset()");
  int ret = 0;
  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_error(): stream structure is not properly initialized!");
    return ret;
  }

  // we reset the stream to avoid lockups in the future (e.g. usb plugged/unplugged)
  if((myStream->printer_status == PRINTER_OPENED) && (time(NULL) - myStream->lastWrite > PS_PRINT_JOB_RESET_TIMEOUT))
    ret = 1;

  return ret;
}


int ps_print_job_interupted(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_print_job_interupted()");
  int ret = 0;
  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_print_job_error(): stream structure is not properly initialized!");
    return ret;
  }

  // if the writeFD is -1, the print job was interupted
  if(myStream->writeFD == -1)
    ret = 1;

  return ret;
}



int ps_printer_fd(print_stream* myStream)
{
  cupsdLogMessage(CUPSD_LOG_DEBUG2, "ps_printer_fd()");
  int ret = -1;
  if(!myStream) {
    cupsdLogMessage(CUPSD_LOG_ERROR, "ps_printer_fd(): stream structure is not properly initialized!");
    return ret;
  }

  return myStream->writeFD;
}





