
/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GLib Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GLib Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GLib at ftp://ftp.gtk.org/pub/gtk/.
 */

/*
 * This file contains all the defines/functions/macros needed to get
 * the ghash routines up and running. The code was extraced from different
 * source files in glib v2.20.4.
 */

#include "gmacros.h"
#include "gtypes.h"
#include "ghash.h"

#include <pthread.h>

pthread_mutex_t add_mutex = PTHREAD_MUTEX_INITIALIZER;

void g_atomic_int_add (volatile gint *atomic, gint val)
{
  pthread_mutex_lock(&add_mutex);
  *atomic += val;
  pthread_mutex_unlock(&add_mutex); 
}



gint g_atomic_int_exchange_and_add (volatile gint *atomic, gint val)
{
  gint result;
 
  pthread_mutex_lock(&add_mutex); 
  result = *atomic;
  *atomic += val;
  pthread_mutex_unlock(&add_mutex); 

  return result;
}

/**
 * g_string_hash:
 * @str: a string to hash
 *
 * Creates a hash code for @str; for use with #GHashTable.
 *
 * Returns: hash code for @str
 */
/* 31 bit hash function */
guint
g_string_hash (gconstpointer *str)
{
  const gchar *p = str;
  gsize n = strlen(str);    
  guint h = 0;

  while (n--)
    {
      h = (h << 5) - h + *p;
      p++;
    }

  return h;
}

guint g_str_hash(gconstpointer v)
{
 return g_string_hash(v);
}

/**
 * g_string_equal:
 * @v: a #GString
 * @v2: another #GString
 *
 * Compares two strings for equality, returning %TRUE if they are equal. 
 * For use with #GHashTable.
 *
 * Returns: %TRUE if they strings are the same length and contain the 
 *   same bytes
 */
gboolean
g_string_equal (gconstpointer *v,
                gconstpointer *v2)
{
  gchar *p, *q;
  gsize i = strlen(v);    

  if (i != strlen(v2))
    return FALSE;

  p = v;
  q = v2;
  while (i)
    {
      if (*p != *q)
	return FALSE;
      p++;
      q++;
      i--;
    }
  return TRUE;
}

gboolean g_str_equal (gconstpointer  v1,
                      gconstpointer  v2)
{

 return g_string_equal(v1,v2);
}


/**
 * g_direct_hash:
 * @v: a #gpointer key
 *
 * Converts a gpointer to a hash value.
 * It can be passed to g_hash_table_new() as the @hash_func parameter, 
 * when using pointers as keys in a #GHashTable.
 *
 * Returns: a hash value corresponding to the key.
 */
guint
g_direct_hash (gconstpointer v)
{
  return GPOINTER_TO_UINT (v);
}




