/*
 * Copyright (c) 2009 SoftAtHome
 *
 * The information and source code contained herein is the exclusive
 * property of SoftAtHome and may not be disclosed, examined, or
 * reproduced in whole or in part without explicit written authorization
 * of the copyright owner.
 *
 * fw_common.c - SoftAtHome firewall plug-in: common functions
 *
 */

#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <arpa/inet.h>
#include <assert.h>

#include <debug/sahtrace.h>
#include <pcb/utils.h>
#include <pcb/core.h>

#include "fw_common.h"
#include "nemo/common_api.h"

bool arg_getServiceEnableValue(uint32_t attr, argument_value_list_t *args, service_enable_t **value, const char *name){
	argument_value_t *arg = NULL;

	arg = (attr&request_function_args_by_name) ? argument_valueByName(args, name) : argument_valueFirstArgument(args);
	if (arg) {
		argument_valueTakeArgument(arg);
		*value = variant2service_enable_t(argument_value(arg));
		if (*value == NULL)
			return false;
	}
	if (arg) {
		argument_valueDestroy(arg);
	}
	return true;
}

void service_enable_t_cleanup(service_enable_t *data) {
	if (data) {
		free(data);
	}
}

service_enable_t *variant2service_enable_t(const variant_t *var) {
	variant_map_iterator_t *it = NULL;		// variant must be of type map
	if (variant_type(var) != variant_type_map) {
		goto error;
	}
	// get the map
	variant_map_t *map = variant_da_map(var);
	if (!map) {
		goto error;
	}
	service_enable_t *ret = (service_enable_t *)calloc(1,sizeof(service_enable_t));
	if (!ret) {
		goto error;
	}
	// get enableIPv4
	it = variant_map_find(map,"enableIPv4");
	if (!it) {
		goto error_free;
	}
	// remove from map file
	variant_map_iterator_take(it);
	ret->enableIPv4 = variant_bool(variant_map_iterator_data(it));
	// cleanup iterator
	variant_map_iterator_destroy(it);
	// get enableIPv6
	it = variant_map_find(map,"enableIPv6");
	if (!it) {
		goto error_free;
	}
	// remove from map file
	variant_map_iterator_take(it);
	ret->enableIPv6 = variant_bool(variant_map_iterator_data(it));
	// cleanup iterator
	variant_map_iterator_destroy(it);
	return ret;

error_free:
	free(ret);
error:
	return NULL;
}

variant_t *service_enable_t2variant(const service_enable_t *data) {
	variant_t *ret = (variant_t *)calloc(1,sizeof(variant_t));
	if (!ret) {
		goto error;
	}

	if (!data) {
		if (!variant_initialize(ret,variant_type_unknown)) {
			goto error_free;
		}
		return ret;
	} else {
		if (!variant_initialize(ret,variant_type_map)) {
			goto error_free;
		}
	}

	if (!variant_map_addBool(variant_da_map(ret),"enableIPv4",data->enableIPv4)) {
		goto error_free;
	}

	if (!variant_map_addBool(variant_da_map(ret),"enableIPv6",data->enableIPv6)) {
		goto error_free;
	}
	return ret;

error_free:
	free(ret);
error:
	return NULL;
}

bool variant_setServiceEnable(variant_t *variant, service_enable_t service_enable)
{
	if (!variant) {
		return false;
	}
	
	variant_cleanup(variant);

	if (!variant_initialize(variant,variant_type_map)) {
		SAH_TRACEZ_ERROR(FW, "Failed to initialize variant");
		return false;
	}

	if (!variant_map_addBool(variant_da_map(variant), "enableIPv4", service_enable.enableIPv4)) {
		SAH_TRACEZ_ERROR(FW, "Failed to add enableIPv4 to map");
		goto error;
	}
	
	if (!variant_map_addBool(variant_da_map(variant), "enableIPv6", service_enable.enableIPv6)) {
		SAH_TRACEZ_ERROR(FW, "Failed to add enableIPv6 to map");
		goto error;
	}

	return true;

error:
		variant_map_cleanup(variant_da_map(variant));
		return false;
}

service_result_t *variant2service_result_t(const variant_t *var) {
	variant_map_iterator_t *it = NULL;	// variant must be of type map
	if (variant_type(var) != variant_type_map) {
		goto error;
	}

	variant_map_t *map = variant_da_map(var);
	if (!map) {
		goto error;
	}
	service_result_t *ret = (service_result_t *)calloc(1,sizeof(service_result_t));
	if (!ret) {
		goto error;
	}
	// get name
	it = variant_map_find(map, "name");
	if (!it) {
		goto error_free;
	}
	variant_map_iterator_take(it);
	ret->name = variant_char(variant_map_iterator_data(it));
	variant_map_iterator_destroy(it);

	// get port 
	it = variant_map_find(map, "port");
	if (!it) {
		goto error_free;
	}
	variant_map_iterator_take(it);
	ret->port = variant_char(variant_map_iterator_data(it));
	variant_map_iterator_destroy(it);

	return ret;

error_free:
	free(ret);
error:
	return NULL;
}

/** 
 * Check if an ip address is valid, 
 * return true if valid, false if invalid
 * the empty string is also a valid value.
 * if the address equals "0.0.0.0" or "::", 
 * the address parameter is free'd and set to NULL
 */
bool checkIpAddress(char **address) {
	unsigned char buf[sizeof(struct in6_addr)] = {};
	int s = 0;
	char *addr = NULL;
	char *needle = NULL;

	if (*address == NULL) {
		return true;
	}
	if (*address[0] == '\0') {
		return true;
	}

	addr = strdup(*address);
	needle = strstr(addr, "/");
	if (needle) {
		SAH_TRACEZ_INFO(FW, "netmask detected in ip address");
		needle[0] = '\0';
	}
	
	s = inet_pton(AF_INET, addr, buf);
	if (s == 1) {
		struct in_addr *a = (struct in_addr *)buf;
		SAH_TRACEZ_INFO(FW, "Valid IPv4 address[%s]", addr);
		if (a->s_addr == 0) {
			free(*address);
			*address = NULL;
		}
		free(addr);
		return true;
	}
	else {
		s = inet_pton(AF_INET6, addr, buf);
		if (s == 1) {
			struct in6_addr *a = (struct in6_addr *)buf;
			SAH_TRACEZ_INFO(FW, "Valid IPv6 address[%s]", addr);
			if (IN6_IS_ADDR_UNSPECIFIED(a)) {
				SAH_TRACEZ_INFO(FW, "Valid ANY address[%s]", addr);
				free(*address);
				*address = NULL;
			}
			free(addr);
			return true;
		}
		else {
			SAH_TRACEZ_WARNING(FW, "Invalid IPv6 address[%s]", *address);
		}
	}
	free(addr);
	return false;
}


void start_timer(struct timeval *start) {
	assert(start);
	gettimeofday(start, NULL);
};

void stop_timer(struct timeval *start, struct timeval *res){
	assert(start && res);
	struct timeval stop_tv;
	gettimeofday(&stop_tv, NULL);
	timersub(&stop_tv, start, res);
	SAH_TRACEZ_INFO(FW, "ACTION takes [%lusec, %luusec]", res->tv_sec, res->tv_usec);
};

