/*
 * libip6t_IMARK.c    Inheritance Marking user module
 *
 * Copyright (c) 2008-2010 France Telecom - Orange Labs - All rights reserved
 * 
 * Author: Frederic Leroy <frederic1.leroy@orange-ftgroup.com>
 * 
 */

/* Shared library add-on to iptables for imark target */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <ip6tables.h>
#include <linux/netfilter_ipv6/ip6_tables.h>
#include "xt_IMARK.h"

static void imark_tg_help(void) 
{
	printf(
"IMARK target options\n"
"  --lan-to-wan		Set IMARK direction from lan to wan for these packets\n"
"  --wan-to-lan		Set IMARK direction from wan to lan \n"
"  --set-priority	Set packet priority according to DSCP precedence \n"
"\n"
);
}

static struct option opts[] = {
	{ "lan-to-wan", 0, 0, 'L' },
	{ "wan-to-lan", 0, 0, 'W' },
	{ "set-priority", 0, 0, 'P' },
	{ 0 }
};

static int
imark_tg_parse(int c, char **argv, int invert, unsigned int *flags,
               const void *entry, struct xt_entry_target **targetinfo)
{
	struct xt_imark_tginfo *dinfo
		= (struct xt_imark_tginfo *)(*targetinfo)->data;

	switch (c) {
	case 'L':
		if (*flags)
			xtables_error(PARAMETER_PROBLEM,
			           "IMARK target: Only use --lan-to-wan ONCE!");
		dinfo->flags=dinfo->flags & ~IMARK_DIR;
		*flags = 1;
		break;
	case 'W':
		if (*flags)
			xtables_error(PARAMETER_PROBLEM,
				   "IMARK target: Only use --wan-to-lan ONCE!");
		dinfo->flags=dinfo->flags | IMARK_DIR_WAN2LAN;
		*flags = 1;
		break;
	case 'P':
		dinfo->flags=dinfo->flags | IMARK_SET_PRIORITY;
		break;
	default:
		return 0;
	}
	return 1;
}

/* Prints out the targinfo. */
static void
imark_tg_print(const void *ip,
               const struct xt_entry_target *target, int numeric)
{
	const struct xt_imark_tginfo *dinfo =
		(const struct xt_imark_tginfo *)target->data;
	printf("IMARK direction ");
	if ( dinfo->flags & IMARK_DIR )
		printf("wan_to_lan");
	else
		printf("lan_to_wan");
	if ( dinfo->flags & IMARK_SET_PRIORITY )
		printf(" --set-priority");
}

static void
imark_tg_final_check	(unsigned int flags)
{
	if (!flags)
		xtables_error(PARAMETER_PROBLEM,
		           "DSCP target: Parameter --lan-to-wan or --wan-to-lan is required");
}

static struct xtables_target imark = { 
	.next		= NULL,
	.name		= "IMARK",
	.version	= XTABLES_VERSION,
	.size		= IP6T_ALIGN(sizeof(struct xt_imark_tginfo)),
	.userspacesize	= IP6T_ALIGN(sizeof(struct xt_imark_tginfo)),
	.help		= &imark_tg_help,
	.parse		= &imark_tg_parse,
	.print		= &imark_tg_print,
	.final_check	= &imark_tg_final_check,
	.extra_opts	= opts
};

void _init(void)
{
	xtables_register_target(&imark);
}
