/*
* This file adds functionality to iptable command to configure ikanos fusiv QOS modules for IPv6 (fusivTC)
*
* Copyright © Ikanos Communications 2008-2013.
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version
* 2 of the License, or (at your option) any later version.
*
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <getopt.h>


#include <xtables.h>
#include <linux/netfilter_ipv6/ip6_tables.h>
#include <linux/netfilter_ipv4/ipt_DSCP.h>
#include <fusivTC.h>
#include "dscp_helper.c"

/* Function which prints out usage message. */
static void
fusivTC_help(void)
{
printf(
"fusivTC target v%s options:\n"
" Commands \n"
"  --map           <info>                Map QoS information to the Scheduler Q\n"
"  --set-priority  <priority>            QoS Priority [0-7]\n"
"  --mark          <stamp>:<priority>    Mark QoS information \n"
"  --remark        <stamp>               Re-Mark (Convert) QoS Information in packet\n"
/*
"  --rate-control  <bool>[-<abw>-<mbw>]  Enable / Disable Rate Control, required for WRT \n\n"
*/
"  --set-dscp value                      Set DSCP field in packet header to value\n"
"                                        This value can be in decimal (ex: 32)\n"
"                                        or in hex (ex: 0x20)\n"
"  --set-dscp-class class                Set the DSCP field in packet header to\n" 
"                                        the value represented by the DiffServ\n"
"                                        class value.This class may be CS0/CS1\n"       
"                                        CS2/CS3/CS4/CS5/CS6/CS7/BE/AF11/AF12\n"
"                                        AF13/AF21/AF22/AF23/AF31/AF32/AF33\n"
"                                        AF41/AF42/AF43/EF \n"
" Command Options:\n"
"  info                                  : 'tos' / 'cos' \n"
"  stamp                                 : 'tos' / 'dscp' / 'cos' \n"
"  priority                              : Value between 0 to 7 \n"
/*
"  bool                                  : 'true' / 'false' "
"  abw                                   : Assured or Committed Rate in KBPS"
"  mbw                                   : Maximum or Peak Rate in KBPS"
*/
"\n",
XTABLES_VERSION);
}

static struct option fusivTC_opts[] = {
	{ "map", 1, 0, '1' },
	{ "set-priority", 1, 0, '2' },
	{ "mark", 1, 0, '3' },
	{ "rate-control", 1, 0, '4' },
	{ "remark", 1, 0, '5' },
        { "set-dscp", 1, 0, '6' },
        { "set-dscp-class", 1, 0, '7' },
	{ 0 }
};

/* Initialize the target. */
static void
fusivTC_init(struct xt_entry_target *t)
{
  	struct fusivtc_target_info  *target = (struct fusivtc_target_info*)t->data;
	/* Reset the target info structure */
	target->ucPriority 		= 0xFF;
	target->marking_priority	= 0xFF;
	target->marking_cos_priority            = 0xFF;
	target->markingInfo		= 0;
	target->remarkInfo	        = 0;
	target->mappingInfo             = 0;
	target->rateControl             = 0;
	target->dscp_info               = 0;
}

static void
parse_dscp(const unsigned char *s, struct fusivtc_target_info *tcinfo)
{
        unsigned int dscp;

        //if (string_to_number(s, 0, 255, &dscp) == -1)
        if (!xtables_strtoui(s, NULL, &dscp, 0, UINT8_MAX))
                xtables_error(PARAMETER_PROBLEM, "Invalid dscp `%s'\n", s);

        if (dscp > IPT_DSCP_MAX)
                xtables_error(PARAMETER_PROBLEM,
                           "DSCP `%d` out of range\n", dscp);

	tcinfo->dscp_info =  1;
        tcinfo->dinfo.dscp = (u_int8_t )dscp;
        return;
}

static void
parse_class(const unsigned char *s, struct fusivtc_target_info *tcinfo)
{
        unsigned int dscp = class_to_dscp(s);

	tcinfo->dscp_info =  1;
        /* Assign the value */
        tcinfo->dinfo.dscp = (u_int8_t)dscp;
}


/* Function which parses command options; returns true if it
   ate an option */
static int
fusivTC_parse(int c, char **argv, int invert, unsigned int *flags,
      const void *entry,
      struct xt_entry_target **target)
{
  int opt_index;

	struct fusivtc_target_info *tcinfo
		= (struct fusivtc_target_info *)(*target)->data;
	switch (c) {
		char *end;
	case '1':  /* Do Mapping */
	  if(tcinfo->mappingInfo != 0x0)
			xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
	  if(tcinfo->ucPriority != 0xFF)
			xtables_error(PARAMETER_PROBLEM, "Mapping and Policy can not be configured together");
	  if(strcasecmp(optarg,"tos") == 0)
	    tcinfo->mappingInfo =  IPQOS_TOS;
	  else if(strcasecmp(optarg,"dscp") == 0)
	    tcinfo->mappingInfo =  IPQOS_DSCP;
	  else if(strcasecmp(optarg,"cos") == 0)
	    tcinfo->mappingInfo =  IPQOS_COS;
	  else
	    xtables_error(PARAMETER_PROBLEM, "Bad Mapping value `%s'", optarg);
	  *flags |= 1;
	break;
	case '2': /* Configure the policy */
	  if(tcinfo->ucPriority != 0xFF)
      xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
	  if(tcinfo->mappingInfo != 0)
			xtables_error(PARAMETER_PROBLEM, "Mapping and Policy can not be configured together");
	  tcinfo->ucPriority = strtoul(optarg, &end, 0);
		if (*end != '\0' || end == optarg)
      xtables_error(PARAMETER_PROBLEM, "Bad Policy value `%s'", optarg);
		if(tcinfo->ucPriority > 7)
      xtables_error(PARAMETER_PROBLEM, "Bad Policy value `%s'", optarg);
    *flags |= 0x2;
		break;
	case '3': /* Configure the Markinginformation */
		if ((*flags & 0x80))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--mark can not be combined with --set-dscp!");
		if ((*flags & 0x100))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--mark can not be combined with --set-dscp-class!");
    if(strncasecmp(optarg,"tos",3) == 0)
		{
			if(tcinfo->markingInfo & IPQOS_DSCP)
				xtables_error(PARAMETER_PROBLEM,"Select either DSCP or TOS marking, not both");
			else if(tcinfo->markingInfo & IPQOS_TOS)
				xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
      tcinfo->markingInfo =  IPQOS_TOS;
			opt_index = 3;
		}
    else if(strncasecmp(optarg,"dscp",4) == 0)
		{
			if(tcinfo->markingInfo & IPQOS_TOS)
				xtables_error(PARAMETER_PROBLEM,"Select either DSCP or TOS marking, not both");
			else if(tcinfo->markingInfo & IPQOS_DSCP)
				xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
		  tcinfo->markingInfo =  IPQOS_DSCP;
			opt_index = 4;
		}
    else if(strncasecmp(optarg,"cos",3) == 0)
		{
			if(tcinfo->markingInfo & IPQOS_COS)
				xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
      tcinfo->markingInfo =  IPQOS_COS;
			opt_index = 3;
		}
    else
      xtables_error(PARAMETER_PROBLEM, "Bad Marking value `%s'", optarg);

		if(optarg[opt_index++] != ':')
			xtables_error(PARAMETER_PROBLEM, "Bad priority value '%s'", optarg);
  if(tcinfo->markingInfo & IPQOS_COS)
               {
                   tcinfo->marking_cos_priority = optarg[opt_index++] - '0';
                   if(tcinfo->marking_cos_priority > 7)
                       xtables_error(PARAMETER_PROBLEM, "Bad cos priority value `%d'", tcinfo->marking_cos_priority);
               }
               else
               {
                   tcinfo->marking_priority = optarg[opt_index++] - '0';
                   if(tcinfo->marking_priority > 7)
                       xtables_error(PARAMETER_PROBLEM, "Bad priority value `%d'", tcinfo->marking_priority);
               }

		if (optarg[opt_index] != '\0')
      xtables_error(PARAMETER_PROBLEM, "Bad priority value `%s'", optarg);
    *flags |= 0x4;
		break;
	case '5': /* Configure the Convert  */
		if ((*flags & 0x80))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--remark can not be combined with --set-dscp!");
		if ((*flags & 0x100))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--remark can not be combined with --set-dscp-class!");
    if(strcasecmp(optarg,"tos") == 0)
		{
			if(tcinfo->remarkInfo & IPQOS_DSCP)
				xtables_error(PARAMETER_PROBLEM,"Select either DSCP or TOS marking, not both");
			else if(tcinfo->remarkInfo & IPQOS_TOS)
				xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
      tcinfo->remarkInfo =  IPQOS_TOS;
		}
    else if(strcasecmp(optarg,"dscp") == 0)
		{
			if(tcinfo->remarkInfo & IPQOS_TOS)
				xtables_error(PARAMETER_PROBLEM,"Select either DSCP or TOS marking, not both");
			else if(tcinfo->remarkInfo & IPQOS_DSCP)
				xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
		  tcinfo->remarkInfo =  IPQOS_DSCP;
		}
    else if(strcasecmp(optarg,"cos") == 0)
		{
			if(tcinfo->remarkInfo & IPQOS_COS)
				xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
      tcinfo->remarkInfo =  IPQOS_COS;
		}
    else
      xtables_error(PARAMETER_PROBLEM, "Bad Mapping value `%s'", optarg);
    *flags |= 0x8;
		break;
	case '4':
    *flags |= 0x10;
		if(tcinfo->rateControl)
      xtables_error(PARAMETER_PROBLEM,"Redundent configurations invoked");
		if(strncasecmp(optarg,"true",4) == 0)
		{
			end = &optarg[4];
			if(*end++ != '-')
			  return 0;
      while(*end != '\0' && *end != '-')
			{
				if(isdigit(*end) == 0)
					return 0;
				end++;
			}
			if(*end++ != '-')
        return 0;
      while(*end != '\0')
      {
				if(isdigit(*end) == 0)
					return 0;
				end++;
			}
		  sscanf(&optarg[4],"-%d-%d",&tcinfo->confAssuredBW,&tcinfo->confMaxBW);
			if(tcinfo->confAssuredBW < 0 || tcinfo->confMaxBW < 0)
		    xtables_error(PARAMETER_PROBLEM, "Bad rate-control value `%s'", optarg);
			tcinfo->rateControl = 1;
		}
		else if(strncasecmp(optarg,"false",5) != 0)
		  xtables_error(PARAMETER_PROBLEM, "Bad rate-control value `%s'", optarg);
		break;
	case '6': /* Parse --set-dscp option */
		if ((*flags & 0x4))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--set-dscp can not be combined with --mark!");
		if ((*flags & 0x8))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--set-dscp can not be combined with --remark!");
		if ((*flags & 0x100))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--set-dscp can not be combined with --set-dscp-class!");
		parse_dscp(optarg,tcinfo);
		*flags |= 0x80;
		break;
	case '7': /* Parse --set-dscp-class option */
		if ((*flags & 0x4))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--set-dscp-class can not be combined with --mark!");
		if ((*flags & 0x8))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--set-dscp-class can not be combined with --remark!");
		if ((*flags & 0x80))
        		xtables_error(PARAMETER_PROBLEM,
                                 "--set-dscp-class can not be combined with --set-dscp!");
		parse_class(optarg,tcinfo);
		*flags |= 0x100;
		break;
	default:
	  xtables_error(PARAMETER_PROBLEM, "This feature is not available now `%s'", optarg);
		return 0;
	}
	return 1;
}

static void
fusivTC_final_check(unsigned int flags)
{
	if (!flags)
		xtables_error(PARAMETER_PROBLEM, "fusivTC target: one of the action required");
	if( (flags & 0x8) && !(flags & 0x1))
	{
		xtables_error(PARAMETER_PROBLEM, "fusivTC target: Re-Marking Requires a Mapping");
	}
}

/* Prints out the targinfo. */
static void
fusivTC_print(const void *ip,
      const struct xt_entry_target *target,
      int numeric)
{
	struct fusivtc_target_info *tcinfo = (struct fusivtc_target_info *)target->data;
	if(tcinfo->mappingInfo)
	{
		switch(tcinfo->mappingInfo)
		{
			case IPQOS_TOS:
    	  printf("map=tos");
				break;
			case IPQOS_DSCP:
			  printf("map=dscp");
				break;
			case IPQOS_COS:
			  printf("map=cos");
				break;
		}
		printf("  ");
	}
	if(tcinfo->markingInfo)
	{
		switch(tcinfo->markingInfo)
		{
			case IPQOS_TOS:
    	                               printf("mark=tos");
				       break;
			case IPQOS_DSCP:
			               printf("mark=dscp");
				       break;
			case IPQOS_COS:
			               printf("mark=cos");
				       break;
		}
		printf("  ");
	}
	if(tcinfo->ucPriority != 0xFF)
	{
		printf("priority:  %d ",tcinfo->ucPriority);
	}
	if(tcinfo->rateControl)
	{
		printf("rate-control:  true-%d-%d [STAT:  %d/%d-%d/%d-%d/%d]",
		      tcinfo->confAssuredBW,tcinfo->confMaxBW,tcinfo->greenPkts,tcinfo->greenBytes,
					tcinfo->yellowPkts,tcinfo->yellowBytes,tcinfo->redPkts,tcinfo->redBytes);
	}
	else 
		printf("rate-control:  false");
        if(tcinfo->dscp_info)
		printf("DSCP %d \n",tcinfo->dinfo.dscp);

}

/* Saves the union ip6t_targinfo in parsable form to stdout. */
static void
fusivTC_save(const void *ip, const struct xt_entry_target *target)
{
    struct fusivtc_target_info *tcinfo = (struct fusivtc_target_info *)target->data;
    if(tcinfo->mappingInfo)
    {
        switch(tcinfo->mappingInfo)
        {
            case IPQOS_TOS:
                            printf("--map tos");
                            break;
            case IPQOS_DSCP:
                            printf("--map dscp");
                            break;
            case IPQOS_COS:
                            printf("--map cos");
                            break;
        }
        printf(" ");
   }
   if(tcinfo->markingInfo)
   {
       switch(tcinfo->markingInfo)
       {
           case IPQOS_TOS:
                           printf("--mark tos:%d",tcinfo->marking_priority);
                           break;
           case IPQOS_DSCP:
                           printf("--mark dscp:%d",tcinfo->marking_priority);
                           break;
           case IPQOS_COS:
                           printf("--mark cos:%d",tcinfo->marking_priority);
                           break;
       }
       printf(" ");
   }
   if(tcinfo->remarkInfo)
   {
       switch(tcinfo->remarkInfo)
       {
          case IPQOS_TOS:
                          printf("--remark tos");
                          break;
          case IPQOS_DSCP:
                          printf("--remark dscp");
                          break;
          case IPQOS_COS:
                          printf("--remark cos");
                          break;
       }
       printf(" ");

   }
   if(tcinfo->ucPriority != 0xFF)
   {
       printf("--set-priority %d",tcinfo->ucPriority);
       printf(" ");
   }
   if(tcinfo->rateControl)
   {
       printf("--rate-control true-%d-%d ",
                   tcinfo->confAssuredBW,tcinfo->confMaxBW);
       printf(" ");
   }
   else
   {
       printf("--rate-control false");
       printf(" ");
   } 

   if(tcinfo->dscp_info)
   {
       printf("--set-dscp %d ",tcinfo->dinfo.dscp);
       printf(" ");
   }

}
static
struct xtables_target ip6t_fusivTC_target
= {
         .name   		= "fusivTC",
         .version  		= XTABLES_VERSION,
	 .family         	= NFPROTO_IPV6,
         .size  	  	= XT_ALIGN(sizeof(struct fusivtc_target_info)),
         .userspacesize  	= XT_ALIGN(sizeof(struct fusivtc_target_info)),
         .help   		= &fusivTC_help,
         .init   		= &fusivTC_init,
         .parse    		= &fusivTC_parse,
         .final_check  		= &fusivTC_final_check,
         .print    		= &fusivTC_print,
         .save   		= &fusivTC_save,
         .extra_opts 		= fusivTC_opts
};

void _init(void)
{
       xtables_register_target(&ip6t_fusivTC_target);
}


