/*
 * libipt_IMARK.c    Inheritance Marking user module
 *
 * Copyright (c) 2008-2010 France Telecom - Orange Labs - All rights reserved
 * 
 * Author: Frederic Leroy <frederic1.leroy@orange-ftgroup.com>
 * 
 */

/* Shared library add-on to iptables for imark target */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include "xt_IMARK.h"

/*#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
#define XT_ALIGN IPT_ALIGN
#endif*/

static void
imark_tg_init(struct xt_entry_target *t)
{
}

static void
imark_tg_save(const void *ip, const struct xt_entry_target *target)
{
	printf("%s not implemented\n", __FUNCTION__);
}

static void imark_tg_help(void) 
{
	printf(
"IMARK target options\n"
"  --lan-to-wan		Set IMARK direction from lan to wan for these packets\n"
"  --wan-to-lan		Set IMARK direction from wan to lan \n"
"  --set-priority	Set packet priority according to DSCP precedence \n"
"  --wait-answer	Trust dscp field, until the other side answers\n"
"\n"
);
}

static struct option opts[] = {
	{ "lan-to-wan", 0, 0, 'L' },
	{ "wan-to-lan", 0, 0, 'W' },
	{ "set-priority", 0, 0, 'P' },
	{ "wait-answer", 0, 0, 'T' },
	{ 0 }
};

static int
imark_tg_parse(int c, char **argv, int invert, unsigned int *flags, 
               const void *entry, struct xt_entry_target **targetinfo)
{
	struct xt_imark_tginfo *dinfo
		= (struct xt_imark_tginfo *)(*targetinfo)->data;

	switch (c) {
	case 'L':
		if (*flags)
			xtables_error(PARAMETER_PROBLEM,
			           "IMARK target: Only use --lan-to-wan ONCE!");
		dinfo->flags=dinfo->flags & ~IMARK_DIR;
		*flags = 1;
		break;
	case 'W':
		if (*flags)
			xtables_error(PARAMETER_PROBLEM,
				   "IMARK target: Only use --wan-to-lan ONCE!");
		dinfo->flags=dinfo->flags | IMARK_DIR_WAN2LAN;
		*flags = 1;
		break;
	case 'P':
		dinfo->flags=dinfo->flags | IMARK_SET_PRIORITY;
		break;
	case 'T':
		dinfo->flags=dinfo->flags | IMARK_WAIT_ANSWER;
		break;
	default:
		return 0;
	}

	return 1;
}

/* Prints out the targinfo. */
static void
imark_tg_print(const void *ip,
               const struct xt_entry_target *target, int numeric)
{
	const struct xt_imark_tginfo *dinfo =
		(const struct xt_imark_tginfo *)target->data;
	if ( dinfo->flags & IMARK_DIR )
		printf("--wan-to-lan");
	else
		printf("--lan-to-wan");
	if ( dinfo->flags & IMARK_SET_PRIORITY )
		printf(" --set-priority");
	if ( dinfo->flags & IMARK_WAIT_ANSWER )
		printf(" --wait-answer");
}

static void
imark_tg_final_check(unsigned int flags)
{
	if (!flags)
		xtables_error(PARAMETER_PROBLEM,
		           "IMARK: Parameter --lan-to-wan or --wan-to-lan is required");
}

static struct xtables_target imark = { 
	.next		= NULL,
	.name		= "IMARK",
	.version	= XTABLES_VERSION,
	.size		= XT_ALIGN(sizeof(struct xt_imark_tginfo)),
	.userspacesize	= XT_ALIGN(sizeof(struct xt_imark_tginfo)),
	.help		= &imark_tg_help,
	.init		= &imark_tg_init,
	.parse		= &imark_tg_parse,
	.final_check	= &imark_tg_final_check,
	.print		= &imark_tg_print,
    	.save		= &imark_tg_save,
	.extra_opts	= opts
};

void _init(void)
{
	xtables_register_target(&imark);
}
