#include "fw_wrapper.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util.h"
#include "pcb_eventloop.h"
#include "globals.h"

#include <pcb/pcb_client.h>
#include <fw/fw_client.h>

void ExpireMapping(void *input);
void phv6_expiration(void *data);


/* static variables */
static const char *s_fwObjectPath = "Firewall";
static const char *s_fwOrigin = "upnp";
static const char *s_dataIntf = "data";

static const char *s_UDPNmbr = "6";
static const char *s_TCPNmbr = "17";

static const char *s_upnpd_ipv4 = "upnpd_ipv4";
#ifdef UPNP_ENABLE_IPV6
static const char *s_upnpd_ipv6 = "upnpd_ipv6";
#endif // UPNP_ENABLE_IPV6

static volatile unsigned int s_pmIndex = 0;     // portmapping index

static const char *pmProtocolToNmbrString(const char *protocol) {
  /* for upnp only UDP and TCP are allowed */
  return (protocol && strncmp("TCP", protocol, 3)) ? s_TCPNmbr : s_UDPNmbr;
}

// portmap expiration event functions
static void pcb_plugin_portmap_eventTimerHandler(pcb_timer_t *timer, void *userdata)
{
  (void) timer;

  ExpireMapping(userdata);
}

bool FWCancelPortMapExpirationTimer(struct portMap *portMap)
{
  if (portMap && portMap->m_expirationTimer)
  {
    pcb_timer_stop(portMap->m_expirationTimer);
    pcb_timer_destroy(portMap->m_expirationTimer);
    portMap->m_expirationTimer = NULL;
    portMap->expirationTime = 0;
  }
  return true;
}

static bool FWSchedulePortMapExpirationTimer(struct portMap *portMap)
{
  bool rc = false;
  if (portMap)
  {
    FWCancelPortMapExpirationTimer(portMap);

    if (portMap->m_PortMappingLeaseDuration <= 0 ||
        portMap->m_PortMappingLeaseDuration > MAXIMUM_DURATION)
    {
      portMap->m_PortMappingLeaseDuration = MAXIMUM_DURATION;
    }

    if (NULL != (portMap->m_expirationTimer = pcb_timer_create()) &&
        pcb_timer_setHandler(portMap->m_expirationTimer, pcb_plugin_portmap_eventTimerHandler) &&
        pcb_timer_setUserData(portMap->m_expirationTimer, portMap) &&
        pcb_timer_start(portMap->m_expirationTimer, portMap->m_PortMappingLeaseDuration*1000))
    {
      portMap->expirationTime = time(NULL) + portMap->m_PortMappingLeaseDuration;

      rc = true;
    }
    else
    {
      trace(1, "%s: Unable to schedule portmapping expiration event", __FUNCTION__);
    }
  }
  return rc;
}

// pinhole expiration event functions
static void pcb_plugin_pinhole_eventTimerHandler(pcb_timer_t *timer, void *userdata)
{
  (void) timer;

  phv6_expiration(userdata);
}

bool FWCancelPinHoleExpirationTimer(struct pinholev6 *pinhole)
{
  if (pinhole && pinhole->expirationTimer)
  {
    pcb_timer_stop(pinhole->expirationTimer);
    pcb_timer_destroy(pinhole->expirationTimer);
    pinhole->expirationTimer = NULL;
  }
  return true;
}

static bool FWSchedulePinHoleExpirationTimer(struct pinholev6 *pinhole)
{
  bool rc = false;
  if (pinhole)
  {
    FWCancelPinHoleExpirationTimer(pinhole);

    if (pinhole->lease_time <= 0 ||
        pinhole->lease_time > MAXIMUM_DURATION)
    {
      pinhole->lease_time = MAXIMUM_DURATION;
    }

    if (NULL != (pinhole->expirationTimer = pcb_timer_create()) &&
        pcb_timer_setHandler(pinhole->expirationTimer, pcb_plugin_pinhole_eventTimerHandler) &&
        pcb_timer_setUserData(pinhole->expirationTimer, pinhole) &&
        pcb_timer_start(pinhole->expirationTimer, pinhole->lease_time*1000))
    {

      rc = true;
    }
    else
    {
      trace(1, "%s: Unable to schedule pinhole expiration event", __FUNCTION__);
    }
  }
  return rc;
}



bool FWSetPortMapping(struct portMap *portMap, bool newPortMap)
{
  bool rc = false;

  if (!portMap) {
    trace(1, "%s: error portMap == NULL",__FUNCTION__);
  }
  else {
    char indexBuffer[32];
    char *resultId = NULL;
    const char *protocol = pmProtocolToNmbrString(portMap->m_PortMappingProtocol);

    if (newPortMap) {
      s_pmIndex++;
      snprintf(indexBuffer, sizeof(indexBuffer)-1, "%d", s_pmIndex);
    }


    trace(3, "%s: fw_setPortForwarding: id=%s, origin=%s, sourceinterface=%s, external_port=%s, internal_port=%s, "
               "destinationIPAddress=%s, sourceprefix=%s, protocol=%s, enable=%d, persistent=%d, description=%s",
           __FUNCTION__, newPortMap? indexBuffer : portMap->m_PCBId, s_fwOrigin, s_dataIntf, portMap->m_ExternalPort,
           portMap->m_InternalPort, portMap->m_InternalClient, portMap->m_RemoteHost, protocol, portMap->m_PortMappingEnabled,
           0, portMap->m_PortMappingDescription);

    resultId = setPortForwarding( PCB_getPeer(),
                                  s_fwObjectPath,                                           // objectpath
                                  newPortMap? indexBuffer : portMap->m_PCBId,               // id
                                  s_fwOrigin,                                               // origin
                                  s_dataIntf,                                               // sourceinterface
                                  portMap->m_ExternalPort,                                  // external port
                                  portMap->m_InternalPort,                                  // internal port
                                  portMap->m_InternalClient,                                // internalclient
                                  portMap->m_RemoteHost,                                    // sourceprefix
                                  protocol,                                                 // protocol
                                  portMap->m_PortMappingEnabled ? true: false,              // enable
                                  false,                                                    // non-persistent
                                  portMap->m_PortMappingDescription);                       // description

    if (resultId && strlen(resultId)) {

      if (newPortMap) {
        strncpy(portMap->m_PCBId, resultId, sizeof(portMap->m_PCBId)-1);
      }
      rc = true;

      trace(3, "%s: portmapping with key \"%s\" created", __FUNCTION__, portMap->m_PCBId);

      // now schedule for expiration
      FWSchedulePortMapExpirationTimer(portMap);


    } else {
      trace(1, "%s: setPortForwarding returned an error", __FUNCTION__);
    }

    free(resultId);
  }

  return rc;
}

bool FWDelPortMapping(struct portMap *portMap)
{
  bool rc = false;

  if (!portMap) {
    trace(1, "%s: error portMap == NULL", __FUNCTION__);
  }
  else {

    FWCancelPortMapExpirationTimer(portMap);

    if (deletePortForwarding(PCB_getPeer(), s_fwObjectPath, portMap->m_PCBId, s_fwOrigin, "")) {
      rc = true;
      trace(3, "%s: portmapping with key \"%s\" deleted", __FUNCTION__, portMap->m_PCBId);
    } else {
      trace(1, "%s: deletePortForwarding failed",__FUNCTION__);
    }
  }

  return rc;
}

bool FWAddPinHole(struct pinholev6 *pinHole, char *internal_client, char *remote_host, char *internal_port, char *remote_port, char *protocol) {
  bool rc = false;

  if (!pinHole) {
    trace(1, "%s: error pinHole == NULL", __FUNCTION__);
  }
  else {

    char indexBuffer[32];
    char *resultId = NULL;

    snprintf(indexBuffer, sizeof(indexBuffer)-1, "%d", pinHole->unique_id);

    resultId = setPinhole(PCB_getPeer(),                             // peer_info_t *dest
                          s_fwObjectPath,                            // const char *objectPath
                          indexBuffer,                               // const char *id
                          s_fwOrigin,                                // const char *origin
                          s_dataIntf,                                // const char *sourceInterface
                          remote_port ? remote_port : "",            // const char *sourcePort
                          internal_port ? internal_port : "",        // const char *destinationPort
                          internal_client ? internal_client : "",    // const char *destinationIPAddress
                          remote_host ? remote_host : "",            // const char *sourcePrefix
                          protocol ? protocol : "",                  // const char *protocol
                          6,                                         // int32_t ipversion
                          true,                                      // bool enable
                          false,                                     // bool persistenti
                          "");                                       // const char *description


    if (resultId && strlen(resultId)) {
      strncpy(pinHole->PCBId, resultId, sizeof(pinHole->PCBId) - 1);

      FWSchedulePinHoleExpirationTimer(pinHole);

      rc = true;
      trace(3, "%s: setPinhole with key \"%s\" created", __FUNCTION__, resultId);
    } else {
      trace(1, "%s: setPinhole failed",__FUNCTION__);
    }

    free(resultId);
  }
  return rc;
}

bool FWDelPinHole(struct pinholev6 *pinHole) {
  bool rc = false;

  if (!pinHole) {
    trace(1, "%s: error pinHole == NULL", __FUNCTION__);
  }
  else {

    FWCancelPinHoleExpirationTimer(pinHole);

    if (deletePinhole(PCB_getPeer(), s_fwObjectPath, pinHole->PCBId, s_fwOrigin)) {
      rc = true;
      trace(3, "%s: pinhole with key \"%s\" deleted", __FUNCTION__, pinHole->PCBId);
    }
    else {
      trace(1, "%s: deletePinhole failed",__FUNCTION__);
    }
  }

  return rc;
}

bool FWUpdatePinHole(struct pinholev6 *pinHole)
{
  bool rc = false;

  if (pinHole)
  {
    rc = FWSchedulePinHoleExpirationTimer(pinHole);
  }

  return rc;
}

static void FWFreeService_Result(service_result_t **serviceId)
{
	if (serviceId && *serviceId) {
		free((*serviceId)->name);
		free((*serviceId)->port);
		free(*serviceId);
		*serviceId = NULL;
	}
}

bool FWOpenServicePorts()
{
  bool rc1 = false, rc2 = false;
  char portNmbrBuffer[7];
  service_result_t *serviceId = NULL;
  peer_info_t *peer = PCB_getPeer();

  if ( NULL !=  peer) {

    snprintf(portNmbrBuffer, 6, "%i", (int) UpnpGetServerPort());
    if (NULL != (serviceId = setService(peer, s_fwObjectPath, s_upnpd_ipv4, "lan", portNmbrBuffer, "6,17", 4, true))) {
      FWFreeService_Result(&serviceId);
      rc1 = true;
      trace(3, "%s: Firewall service \"%s\"created", __FUNCTION__, s_upnpd_ipv4);
    } else {
      trace(1, "%s: Firewall setService \"%s\" failed",__FUNCTION__, s_upnpd_ipv4);
    }

#ifdef UPNP_ENABLE_IPV6
    snprintf(portNmbrBuffer, 6, "%i", (int) UpnpGetServerPort6());
    if (NULL != (serviceId = setService(peer, s_fwObjectPath, s_upnpd_ipv6, "lan", portNmbrBuffer, "6,17", 6, true))) {
      FWFreeService_Result(&serviceId);
      rc2 = true;
      trace(3, "%s: Firewall service \"%s\"created", __FUNCTION__, s_upnpd_ipv6);
    } else {
      trace(1, "%s: Firewall setService \"%s\" failed",__FUNCTION__, s_upnpd_ipv6);
    }
#else
    rc2 = true;
#endif

  }
  else
  {
    trace(1, "%s: PCB system is not initialized", __FUNCTION__);
  }

  return (rc1&&rc2);
}

bool FWCloseFirewall()
{
  bool rc1 = false, rc2 = false;
  peer_info_t *peer = PCB_getPeer();

  if ( NULL != peer) {

    if (deleteService(peer, s_fwObjectPath, s_upnpd_ipv4)) {
      rc1 = true;
      trace(3, "%s: Firewall service \"%s\" deleted", __FUNCTION__, s_upnpd_ipv4);

    } else {
      trace(1, "%s: Firewall deleteService \"%s\" failed",__FUNCTION__, s_upnpd_ipv4);
    }

#ifdef UPNP_ENABLE_IPV6
    if (deleteService(peer, s_fwObjectPath, s_upnpd_ipv6)) {
      rc2 = true;
      trace(3, "%s: Firewall service \"%s\" deleted", __FUNCTION__, s_upnpd_ipv6);

    } else {
      trace(1, "%s: Firewall deleteService \"%s\" failed",__FUNCTION__, s_upnpd_ipv6);
    }
#else
    rc2 = true;
#endif
  }
  else
  {
    trace(1, "%s: PCB system is not initialized", __FUNCTION__);
  }

  return (rc1&&rc2);
}

bool FWDelAllPortMappings()
{
    if (!PCB_getPeer()) {
        trace(1, "%s: PCB system is not initialized", __FUNCTION__);
        goto error_exit;
    }

    if (deletePortForwarding(PCB_getPeer(), s_fwObjectPath, "", s_fwOrigin, "")) {
        trace(3, "%s: all portmappings with origin \"%s\" deleted", __FUNCTION__, s_fwOrigin);
    } else {
        trace(1, "%s: unable to delete all portmappings with origin \"%s\"",__FUNCTION__, s_fwOrigin);
        goto error_exit;
    }

    return true;

error_exit:
    return false;
}

bool FWDelAllPinholes()
{
    if (!PCB_getPeer()) {
        trace(1, "%s: PCB system is not initialized", __FUNCTION__);
        goto error_exit;
    }

    if (deletePinhole(PCB_getPeer(), s_fwObjectPath, "", s_fwOrigin)) {
        trace(3, "%s: all pinholes with origin \"%s\" deleted", __FUNCTION__, s_fwOrigin);
    } else {
        trace(1, "%s: unable to delete all pinholes with origin \"%s\"",__FUNCTION__, s_fwOrigin);
        goto error_exit;
    }

    return true;

error_exit:
    return false;
}

