#include "ipc_wrapper.h"
#include "fw_wrapper.h"
#include "pcb_eventloop.h"
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include "util.h"
#include <stdio.h>
#include <errno.h>
#include <pthread.h>

typedef enum {
  FUNCTION_ADD_PORTMAPPING,
  FUNCTION_DEL_PORTMAPPING,
  FUNCTION_ADD_PINHOLE,
  FUNCTION_DEL_PINHOLE,
  FUNCTION_UPD_PINHOLE,
  FUNCTION_GET_LINKLAYER_BIT_RATES,
  FUNCTION_FIREWALL_IS_ENABLED
} function_to_execute;


typedef struct _ipc_data {
  function_to_execute exec;
  bool functionResult;

  struct portMap *portMap;
  bool newPortMap;

  struct pinholev6 *pinHole;
  char *internal_client;
  char *remote_host;
  char *internal_port;
  char *remote_port;
  char *protocol;

  bool firewallIsEnabled;

  nemo_link_layer_max_bit_rates bitRates;
} ipc_data_t;

#define PIPE_READ_END         0
#define PIPE_WRITE_END        1

static ipc_data_t s_ipc_data;
static pthread_t s_parentPID = 0;

static int ipc_pcb_pipe[2] = {-1, -1};       // pcb loop is allowed to write, lib_upnp reads
static int ipc_lib_upnp_pipe[2] = {-1, -1};  // lib_upnp loop is allowed to write, pcb reads

static volatile bool s_needToStop = false;

static void set_nonblocking(int fd)
{
  int val = 0;

  if ((val = fcntl(fd, F_GETFL, 0)) == -1) {
    return;
  }
  if (!(val & O_NONBLOCK)) {
    val |= O_NONBLOCK;
    fcntl(fd, F_SETFL, val);
  }
}

bool IPC_Init()
{
  bool rc = true;
  if (pipe(ipc_pcb_pipe) < 0 ||
      pipe(ipc_lib_upnp_pipe) < 0)
  {
    rc = false;
    trace(1, "%s: Error opening pipes for IPC", __FUNCTION__);
  }
  else
  {
    // store the pcb_eventloop thread ID
    s_parentPID = pthread_self();
    trace(3, "Parent PID = %lu", s_parentPID);
    s_needToStop = false;
    set_nonblocking(ipc_lib_upnp_pipe[PIPE_READ_END]); // read end of pipe for PCB has to be non-blocking
  }

  return rc;
}

bool IPC_Exit()
{
  // make sure we are not going to deadlock
  s_needToStop = true;

  while (1 != write(ipc_pcb_pipe[PIPE_WRITE_END], "1", 1))
  {
     trace(4, "%s: write failed %s", __FUNCTION__, strerror(errno));
  }
  return true;
}

inline int IPC_GetReadSocket()
{
  return ipc_lib_upnp_pipe[PIPE_READ_END];
}

static void IPC_signalEventLoop()
{
  if (pthread_self() != s_parentPID)
  {
    if (!s_needToStop)
    {
      char buf[2];
      unsigned int cnt = 0;
      // write to socket to signal that we need the result of a function
      while (1 != write(ipc_lib_upnp_pipe[PIPE_WRITE_END], "1", 1))
      {
        trace(4, "%s: write failed %s", __FUNCTION__, strerror(errno));
        sleep(1);
        cnt++;
        if (s_needToStop || cnt == 60) {
          return;
        }
      }

      cnt = 0;
      while (1 != read(ipc_pcb_pipe[PIPE_READ_END], buf, 1))
      {
        trace(4, "%s: read failed %s", __FUNCTION__, strerror(errno));
        sleep(1);
        cnt++;
        if (s_needToStop || cnt == 60) {
          return;
        }
      }
    }
    else
    {
      trace(1, "%s: Shutting down, no longer handling IPC requests", __FUNCTION__);
    }
  }
  else
  {
    IPC_handle_request(0);
  }
}

bool IPC_request_SetPortMapping(struct portMap *portMap, bool newPortMap)
{
  memset(&s_ipc_data, 0, sizeof(s_ipc_data));
  s_ipc_data.exec = FUNCTION_ADD_PORTMAPPING;
  s_ipc_data.portMap = portMap;
  s_ipc_data.newPortMap = newPortMap;
  s_ipc_data.functionResult = false;

  trace(4, "%s: pid(%lu) - ppid(%lu)", __FUNCTION__, pthread_self(), s_parentPID);

  // if application needs to exit, return false
  IPC_signalEventLoop();

  return s_ipc_data.functionResult;
}

bool IPC_request_DelPortMapping(struct portMap *portMap)
{
  memset(&s_ipc_data, 0, sizeof(s_ipc_data));
  s_ipc_data.exec = FUNCTION_DEL_PORTMAPPING;
  s_ipc_data.portMap = portMap;
  s_ipc_data.functionResult = false;

  trace(4, "%s: pid(%lu) - ppid(%lu)", __FUNCTION__, pthread_self(), s_parentPID);

  // if application needs to exit, return false
  IPC_signalEventLoop();

  return s_ipc_data.functionResult;
}

bool IPC_request_AddPinHole(struct pinholev6 *pinHole, char *internal_client, char *remote_host, char *internal_port, char *remote_port, char *protocol)
{
  memset(&s_ipc_data, 0, sizeof(s_ipc_data));
  s_ipc_data.exec = FUNCTION_ADD_PINHOLE;
  s_ipc_data.pinHole = pinHole;
  s_ipc_data.internal_client = internal_client;
  s_ipc_data.remote_host = remote_host;
  s_ipc_data.internal_port = internal_port;
  s_ipc_data.remote_port = remote_port;
  s_ipc_data.protocol = protocol;
  s_ipc_data.functionResult = false;

  trace(4, "%s: pid(%lu) - ppid(%lu)", __FUNCTION__, pthread_self(), s_parentPID);

  // if application needs to exit, return false
  IPC_signalEventLoop();

  return s_ipc_data.functionResult;
}

bool IPC_request_DelPinHole(struct pinholev6 *pinHole)
{
  memset(&s_ipc_data, 0, sizeof(s_ipc_data));
  s_ipc_data.exec = FUNCTION_DEL_PINHOLE;
  s_ipc_data.pinHole = pinHole;
  s_ipc_data.functionResult = false;

  trace(4, "%s: pid(%lu) - ppid(%lu)", __FUNCTION__, pthread_self(), s_parentPID);

  // if application needs to exit, return false
  IPC_signalEventLoop();

  return s_ipc_data.functionResult;
}

bool IPC_request_UpdatePinHole(struct pinholev6 *pinHole)
{
  memset(&s_ipc_data, 0, sizeof(s_ipc_data));
  s_ipc_data.exec = FUNCTION_UPD_PINHOLE;
  s_ipc_data.pinHole = pinHole;

  trace(4, "%s: pid(%lu) - ppid(%lu)", __FUNCTION__, pthread_self(), s_parentPID);

  // if application needs to exit, return false
  IPC_signalEventLoop();

  return s_ipc_data.functionResult;
}

bool IPC_request_GetLinkLayerMaxBitRates(nemo_link_layer_max_bit_rates *bitRates)
{
  memset(&s_ipc_data, 0, sizeof(s_ipc_data));
  s_ipc_data.exec = FUNCTION_GET_LINKLAYER_BIT_RATES;

  // if application needs to exit, return false
  IPC_signalEventLoop();

  if (s_ipc_data.functionResult)
  {
    bitRates->wanAccessType = s_ipc_data.bitRates.wanAccessType;
    bitRates->downStreamMaxBitRate = s_ipc_data.bitRates.downStreamMaxBitRate;
    bitRates->upstreamMaxBitRate = s_ipc_data.bitRates.upstreamMaxBitRate;
    bitRates->UpstreamCurrRate = s_ipc_data.bitRates.UpstreamCurrRate;
    bitRates->DownstreamCurrRate = s_ipc_data.bitRates.DownstreamCurrRate;
  }

  return s_ipc_data.functionResult;
}

bool IPC_request_FirewallIsEnabled(bool *enabled)
{
  memset(&s_ipc_data, 0, sizeof(s_ipc_data));
  s_ipc_data.exec = FUNCTION_FIREWALL_IS_ENABLED;

  // if application needs to exit, return false
  IPC_signalEventLoop();

  if (s_ipc_data.functionResult)
  {
    *enabled = s_ipc_data.firewallIsEnabled;
  }

  return s_ipc_data.functionResult;
}

// only to be called from inside the PCB-eventloop
bool IPC_handle_request(bool signalThread)
{
  bool rc = false;

  switch (s_ipc_data.exec)
  {
    case FUNCTION_ADD_PORTMAPPING:
      rc = FWSetPortMapping(s_ipc_data.portMap, s_ipc_data.newPortMap);
      break;
    case FUNCTION_DEL_PORTMAPPING:
      rc = FWDelPortMapping(s_ipc_data.portMap);
      break;
    case FUNCTION_ADD_PINHOLE:
      rc = FWAddPinHole(s_ipc_data.pinHole, s_ipc_data.internal_client, s_ipc_data.remote_host, s_ipc_data.internal_port, s_ipc_data.remote_port, s_ipc_data.protocol);
      break;
    case FUNCTION_DEL_PINHOLE:
      rc = FWDelPinHole(s_ipc_data.pinHole);
      break;
    case FUNCTION_UPD_PINHOLE:
      rc = FWUpdatePinHole(s_ipc_data.pinHole);
      break;
    case FUNCTION_GET_LINKLAYER_BIT_RATES:
      rc = NeMoGetLinkLayerMaxBitRates(&(s_ipc_data.bitRates));
      break;
    case FUNCTION_FIREWALL_IS_ENABLED:
      s_ipc_data.firewallIsEnabled = PCB_firewallIsEnabled();
      rc = true;
      break;
    default:
      break;
  }

  s_ipc_data.functionResult = rc;

  if (signalThread)
  {
    unsigned int cnt = 0;
    while (1 != write(ipc_pcb_pipe[PIPE_WRITE_END], "1", 1))
    {
      trace(4, "%s: write failed %s", __FUNCTION__, strerror(errno));
      sleep(1);
      cnt++;
      if (s_needToStop || cnt == 60) {
          trace(4, "Stop trying: s_needToStop(%d)", s_needToStop);
          rc = false;
          goto out;
      }
    }
  }

out:
  return rc;
}
