#include <string.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>

#include <pcb/common/error.h>
#include <pcb/utils.h>
#include <pcb/core.h>
#include <pcb/pcb_client.h>

#include <nemo/client.h>

#include "util.h"
#include "pmlist.h"
#include "pinholev6.h"
#include "gatedevice.h"
#include "wanipv6fw.h"

#include "nemo_wrapper.h"
#include "ipc_wrapper.h"
#include "fw_wrapper.h"

typedef struct _delete_event_struct {
  pcb_timer_t *timer;
  char *key;
  bool isPortmapping;   // portmapping or pinhole
} delete_event_struct_t;


#define FIREWALL_PORTFORWARDING            "Firewall.PortForwarding"
#define FIREWALL_PINHOLE                   "Firewall.Pinhole"

extern ithread_mutex_t DevMutex;

plugin_config_t s_appconfig;

static volatile bool s_exit_app = false;
volatile bool g_FirewallEnabled = false;

// DSLITE overrules the WAN externalIP parameter, only the Firewall knows the CGN provided external IP address
volatile bool g_FirewallPCPEnabled = false;
char *g_FirewallPCPExternalIP = NULL;

static request_t *s_portForwardingRequest = NULL;
static request_t *s_pinHoleRequest = NULL;
static request_t *s_FirewallRequest = NULL;
static request_t *s_FirewallPCBRequest = NULL;

static nemo_query_t * s_query_logical = NULL;
static nemo_query_t * s_query_physical = NULL;

static bool s_logical_intf_enabled = false;
static bool s_physical_intf_enabled = false;

static bool PCB_initialized = true;

peer_info_t* PCB_getPeer()
{
  return s_appconfig.system_bus;
}

pcb_t* PCB_getClient()
{
  return s_appconfig.pcb;
}

// access functions: since we have a subscription in the MAIN (PCB) thread on the firewall status
bool PCB_firewallIsEnabled()
{
    return g_FirewallEnabled;
}

bool PCB_firewallPCPIsEnabled()
{
  return g_FirewallPCPEnabled;
}

const char* PCB_firewallPCPExternalIP()
{
  return g_FirewallPCPExternalIP;
}

static void wan_interface_state_handler() {
  IXML_Document *propSet = NULL;
  nemo_interface nemo_wan_interface;

  NeMoUpdateWanStatus(s_appconfig.pcb, s_appconfig.system_bus, s_physical_intf_enabled, s_logical_intf_enabled);
  NeMoGetWanStatus(&nemo_wan_interface);

  ithread_mutex_lock(&DevMutex);

  EthernetLinkStatusEventing(propSet, &nemo_wan_interface);
  ExternalIPAddressEventing(propSet, &nemo_wan_interface);
  ConnectionStatusEventing(propSet, &nemo_wan_interface);

  ithread_mutex_unlock(&DevMutex);

  ixmlDocument_free(propSet);
}

static void handler_logical(const variant_t *result, void *userdata) {
  (void)userdata;
  s_logical_intf_enabled = variant_bool(result);

  wan_interface_state_handler();
}

static void handler_physical(const variant_t *result, void *userdata) {
  (void)userdata;
  s_physical_intf_enabled = variant_bool(result);

  wan_interface_state_handler();
}

static void pcb_eventloop_deleteHandler(pcb_timer_t *timer, void *userdata)
{
  (void) timer;

  delete_event_struct_t *event = (delete_event_struct_t*) userdata;

  if (!event)
    return;

  if (EBUSY != ithread_mutex_trylock(&DevMutex))
    {
      if (event->isPortmapping)
      {
        struct portMap* portmap = pmlist_FindByPCBKey(event->key);
        if (portmap)
        {
          char DevUDN[NAME_SIZE];
          char ServiceID[NAME_SIZE];

          strncpy(DevUDN, portmap->DevUDN, sizeof(DevUDN));
          strncpy(ServiceID, portmap->ServiceID, sizeof(ServiceID));

          trace(3,"received portmap deleted event - %s", event->key? event->key : "NULL");
          if (pmlist_Delete(portmap, 0))
          {
            NotifyPortMappingNmbrOfEntriesChanged(DevUDN, ServiceID);
          }
        }
      }
      else
      {
        struct pinholev6 *pinhole = NULL;
        if(phv6_findPinholeByPCBKey(event->key, &pinhole))
        {
           trace(3,"received pinhole deleted event - %s", event->key? event->key : "NULL");
           phv6_deletePinhole(pinhole->unique_id, 0);
        }
      }

      ithread_mutex_unlock(&DevMutex);

      pcb_timer_destroy(event->timer);
      free(event->key);
      free(event);
    }
    else
    {
      // reschedule
      pcb_timer_start(event->timer, 0);
    }
}

// do not handle the delete events directly in the main_loop as we have the risk of a deadlock
bool CreateDeleteEvent(char *key, bool isPortMapping)
{
  bool rc = false;
  delete_event_struct_t *event = (delete_event_struct_t*) malloc(sizeof(delete_event_struct_t));

  if (key &&
      event &&
      NULL != (event->timer = pcb_timer_create()) &&
      NULL != (event->key = strdup(key)) &&
      pcb_timer_setHandler(event->timer, pcb_eventloop_deleteHandler) &&
      pcb_timer_setUserData(event->timer, event) &&
      pcb_timer_start(event->timer, 0))
  {
    event->isPortmapping = isPortMapping;
    rc = true;
  }
  else
  {
    free(event);
  }

  return rc;
}

static bool eventHandler(request_t *req, pcb_t *pcb, peer_info_t *from, void *userdata)
{
	notification_t *notification = NULL;
	char *key = NULL;
	reply_t *reply = request_reply(req);
	// Loop over reply items and build reply structure, call get_events_done when done
	reply_item_t *item = reply_firstItem(reply);
	while(item) {
		switch(reply_item_type(item)) {
			case reply_type_notification:
				notification = reply_item_notification(item);
				switch (notification_type(notification))
				{
					case notify_object_deleted:
            key = strstr(notification_objectPath(notification), ".upnp_");

            if (key && strlen(key)>6)
            {
              key = &(key[1]);

              if (userdata==(void*)s_portForwardingRequest)
              {
                CreateDeleteEvent(key, true);
              }
              else if (userdata==(void*)s_pinHoleRequest)
              {
                CreateDeleteEvent(key, false);
              }
            }
						break;
					default:
						break;
				}

				break;
			default: // all other items not handled
				break;
		}
		item = reply_nextItem(item);
	}

	return true;
}

inline static void pcb_updateFirewallStatus() {
    IXML_Document *propSet = NULL;

    ithread_mutex_lock(&DevMutex);
    WANIPv6FirewallStatusEventing(propSet, g_FirewallEnabled);
    ithread_mutex_unlock(&DevMutex);

    ixmlDocument_free(propSet);
}

static bool firewallStatusEventHandler(request_t *req, pcb_t *pcb, peer_info_t *from, void *userdata) {
    (void)pcb; (void)from; (void)userdata;
    reply_t *reply = request_reply(req);
    reply_item_t *item;
    object_t *object;

    if (!reply)
        return true;

    for(item = reply_firstItem(reply); item; item = reply_nextItem(item)) {
        if (reply_item_type(item) != reply_type_object)
            continue;
        object = reply_item_object(item);

        g_FirewallEnabled = object_parameterBoolValue(object, "Enable");
        pcb_updateFirewallStatus();
    }

    return true;
}

static bool firewallPCBStatusEventHandler(request_t *req, pcb_t *pcb, peer_info_t *from, void *userdata) {
  (void)pcb; (void)from; (void)userdata;
  bool update = false;
  bool FirewallPCBEnabled = false;
  const char *FirewallPCBExternalIP = NULL;
  reply_t *reply = request_reply(req);
  reply_item_t *item;
  object_t *object;

  if (!reply)
      return true;

  for(item = reply_firstItem(reply); item; item = reply_nextItem(item)) {
      if (reply_item_type(item) != reply_type_object)
          continue;
      object = reply_item_object(item);

      FirewallPCBEnabled = object_parameterBoolValue(object, "Enable");
      FirewallPCBExternalIP = object_da_parameterCharValue(object, "WANIPAddress");

      ithread_mutex_lock(&DevMutex);
      if (g_FirewallPCPEnabled != FirewallPCBEnabled) {
          g_FirewallPCPEnabled = FirewallPCBEnabled;
          update = true;
      }
      else if (g_FirewallPCPEnabled && 0 != strcmp(GET_STRING(g_FirewallPCPExternalIP), GET_STRING(FirewallPCBExternalIP))) {
        free(g_FirewallPCPExternalIP);
        g_FirewallPCPExternalIP = NULL;

        if (FirewallPCBExternalIP){
          g_FirewallPCPExternalIP = strdup(FirewallPCBExternalIP);
        }
        update = true;
      }
      ithread_mutex_unlock(&DevMutex);
  }

  if (update) {
    trace(3, "PCP Update: Firewall.PCP.Enable=%s - Firewall.PCP.WANIPAddress.=%s", FirewallPCBEnabled ? "1" : "0", GET_STRING(FirewallPCBExternalIP));
    wan_interface_state_handler();
  }

  return true;
}

static void pcb_plugin_signalHandler(int signal)
{
	(void)signal;

  trace(1,"Signal event handler called(%d)", signal);

	switch(signal)
	{
		case SIGUSR1:
      //DeleteAllPortMappings();
      //CloseFirewallv6();
			break;
		case SIGTERM:
		case SIGINT:
		case SIGUSR2:
      s_exit_app = true;
			break;
		default:
			break;
	}
}

static bool ipc_socket_read(peer_info_t *peer) {
  char buf[2];
  while (1 != read(IPC_GetReadSocket(), buf, 1))
  {
    trace(1, "%s: Error, read returned error: %s", __FUNCTION__, strerror(errno));
    sleep(1);
  }

  if (buf[0] == '1')
  {
    IPC_handle_request(1);
  }
  else
  {
    trace(1, "%s: Unknown request", __FUNCTION__);
  }

  return true;
}

static void pcb_plugin_eventloop() {
  int retval = 0;

  trace(1,"%s entered event loop",s_appconfig.name);

	llist_t cons;
	llist_initialize(&cons);
  llist_append(&cons,connection_getIterator(pcb_connection(s_appconfig.pcb)));

  while(!s_exit_app) {
		// call select
		retval = connection_waitForEvents(&cons,NULL);
		if (retval < 0) {
				// an error has happened
				trace(1,"Error %d (%s)",pcb_error,error_string(pcb_error));
				continue;
		}

		if (retval >= 0) {
				// events available (signals or fd events)
				connection_handleEvents(&cons);
		}
	}

  /* delete all portmappings and pinholes while we still get the notifications and in this wasy can inform all clients with notifications */
  //DeleteAllPortMappings();
  //CloseFirewallv6();

  IPC_Exit();

	llist_cleanup(&cons);
  trace(1,"%s leaving event loop",s_appconfig.name);
}

static bool pcb_plugin_connect_system() {
	const char *uri = getenv("PCB_SYS_BUS");
	if (uri) {
		trace(1,"Try to connect to %s",uri);
    s_appconfig.system_bus = pcb_client_connect(s_appconfig.pcb,uri);
	} else {
    if (!s_appconfig.system_bus) {
			char *address = getenv("PCB_SYSTEM");
			if (!address) {
				trace(1,"env PCB_SYSTEM not found, using default system bus");
				address = "/var/run/pcb_sys";
			}

			trace(1,"Try to connect to %s",address);
      s_appconfig.system_bus = connection_connectToIPC(pcb_connection(s_appconfig.pcb),address);
		}
	}

  if (!s_appconfig.system_bus) {
			trace(1,"Failed to connect to system bus");
      return false;
	}

  return true;
}

static bool pcb_plugin_initialize() {

    int listen = 0;
    peer_info_t *peer_event = NULL;

    connection_initLibrary();

    // create the connection object
    s_appconfig.pcb = pcb_create(s_appconfig.name,s_appconfig.plugin_argc,s_appconfig.plugin_argv);
    if (!s_appconfig.pcb) {
        trace(1,"Failed to create connection object");
        goto error;
    }

    pcb_setPluginConfig(s_appconfig.pcb,&s_appconfig);

    // set the default format
    if (!pcb_setDefaultFormat(s_appconfig.pcb,SERIALIZE_FORMAT(serialize_format_default,0,0))) {
        trace(1,"Failed to find serialization format");
        goto error_cleanup_connection;
    }

    if (!pcb_plugin_connect_system()) {
      trace(1,"pcb_plugin_connect_system failed");
      goto error_cleanup_connection;
    }

    // set the custom socket handler
    peer_event = connection_addCustom(pcb_connection(s_appconfig.pcb), IPC_GetReadSocket());
    peer_setEventHandler(peer_event,peer_event_read,peer_handler_read_always,ipc_socket_read);

    // create the listen sockets
    // IPC listen socket
    if (s_appconfig.ipc_socket_name && *s_appconfig.ipc_socket_name) {
        trace(1,"Start listening on IPC socket %s",s_appconfig.ipc_socket_name);
        if (connection_listenOnIPC(pcb_connection(s_appconfig.pcb), s_appconfig.ipc_socket_name)) {
            listen++;
        } else {
            trace(1,"Failed to listen on IPC socket %s",s_appconfig.ipc_socket_name);
        }
    }

    // TCP listen socket
    if (s_appconfig.tcp_address &&  *s_appconfig.tcp_address &&  s_appconfig.tcp_port && *s_appconfig.tcp_port) {
        trace(1,"Start listening on TCP socket %s:%s",s_appconfig.tcp_address,s_appconfig.tcp_port);
        uint32_t flags = connection_attr_default;

        if (connection_listenOnTCP(pcb_connection(s_appconfig.pcb), s_appconfig.tcp_address,s_appconfig.tcp_port,flags)) {
            listen++;
        } else {
            trace(1,"Failed to listen on TCP socket %s:%s",s_appconfig.tcp_address,s_appconfig.tcp_port);
        }
    }

    // set signal event handlers
    if (!connection_setSignalEventHandler(pcb_connection(s_appconfig.pcb),SIGINT,pcb_plugin_signalHandler))
      trace(1, "Unable to register SIGINT signal handler");

    if (!connection_setSignalEventHandler(pcb_connection(s_appconfig.pcb),SIGTERM,pcb_plugin_signalHandler))
      trace(1, "Unable to register SIGTERM signal handler");

    if (!connection_setSignalEventHandler(pcb_connection(s_appconfig.pcb),SIGUSR1,pcb_plugin_signalHandler))
      trace(1, "Unable to register SIGUSR1 signal handler");

    if (!connection_setSignalEventHandler(pcb_connection(s_appconfig.pcb),SIGUSR2,pcb_plugin_signalHandler))
      trace(1, "Unable to register SIGUSR2 signal handler");

    // set the callback handlers
    s_portForwardingRequest = request_create_getObject(FIREWALL_PORTFORWARDING, 1,  request_notify_object_deleted | request_common_path_key_notation);
    request_setReplyHandler(s_portForwardingRequest, eventHandler);
    request_setData(s_portForwardingRequest, s_portForwardingRequest);

    if (!pcb_sendRequest(s_appconfig.pcb, s_appconfig.system_bus, s_portForwardingRequest))
			trace(1,"Failed to send portForwardingRequest");

    s_pinHoleRequest = request_create_getObject(FIREWALL_PINHOLE, 1, request_notify_object_deleted | request_common_path_key_notation);
    request_setReplyHandler(s_pinHoleRequest, eventHandler);
    request_setData(s_pinHoleRequest, s_pinHoleRequest);

    if (!pcb_sendRequest(s_appconfig.pcb, s_appconfig.system_bus, s_pinHoleRequest))
			trace(1,"Failed to send pinHoleRequest");

    s_FirewallRequest = request_create_getObject("Firewall", 0, request_getObject_parameters | request_notify_values_changed | request_no_object_caching);
    request_setReplyHandler(s_FirewallRequest, firewallStatusEventHandler);
    request_setData(s_FirewallRequest, s_pinHoleRequest);

    if (!pcb_sendRequest(s_appconfig.pcb, s_appconfig.system_bus, s_FirewallRequest))
            trace(1,"Failed to send FirewallRequest");

    s_FirewallPCBRequest = request_create_getObject("Firewall.PCP", 0, request_getObject_parameters | request_notify_values_changed | request_no_object_caching);
    request_setReplyHandler(s_FirewallPCBRequest, firewallPCBStatusEventHandler);

    if (!pcb_sendRequest(s_appconfig.pcb, s_appconfig.system_bus, s_FirewallPCBRequest))
            trace(1,"Failed to send s_FirewallPCBRequest");

    // Check that at least one listen socket is available or a connection to the system bus has been made
    // otherwise just exit the application
    if (!listen && !s_appconfig.system_bus && !peer_isConnected(s_appconfig.system_bus)) {
        trace(1,"No listen sockets available or no connection to the system bus, stop");
        goto error_cleanup_connection;
    }

    trace(3,"pcb_plugin initialized");

    return true;

error_cleanup_connection:
    pcb_destroy(s_appconfig.pcb);
    s_appconfig.pcb = NULL;

error:
    trace(1,"pcb plugin (%s) failed to initialize\n",s_appconfig.name);
    return false;
}

static void pcb_plugin_cleanup() {

  if (s_portForwardingRequest) {
    request_destroy(s_portForwardingRequest);
    s_portForwardingRequest = NULL;
  }
  if (s_pinHoleRequest) {
    request_destroy(s_pinHoleRequest);
    s_pinHoleRequest = NULL;
  }
  if (s_FirewallRequest) {
      request_destroy(s_FirewallRequest);
      s_FirewallRequest = NULL;
  }
  if (s_FirewallPCBRequest) {
      request_destroy(s_FirewallPCBRequest);
      s_FirewallPCBRequest = NULL;
  }

  free(g_FirewallPCPExternalIP);
  g_FirewallPCPExternalIP = NULL;

  trace(1,"Cleanup pcb context");
  pcb_destroy(s_appconfig.pcb);

  trace(1,"Cleanup library");
  connection_exitLibrary();
}

static bool PCB_NeMoInit() {
  argument_value_list_t args;
  argument_value_list_initialize(&args);
  variant_t flag;
  variant_initialize(&flag, variant_type_string);

  nemo_client_initialize(s_appconfig.pcb, s_appconfig.system_bus);

  variant_setChar(&flag, "logical");
  argument_valueAdd(&args, "flag", &flag);
  s_query_logical = nemo_openQuery("data", "upnpd", "isUp", &args, 0, handler_logical, NULL);

  variant_setChar(&flag, "physical");
  argument_valueAdd(&args, "flag", &flag);
  s_query_physical = nemo_openQuery("data", "upnpd", "isUp", &args, 0, handler_physical, NULL);

  argument_valueClear(&args);
  variant_cleanup(&flag);

  return true;
}

static void PCB_NeMoExit() {
  nemo_closeQuery(s_query_logical);
  nemo_closeQuery(s_query_physical);

  nemo_client_cleanup();
}

bool PCB_mainLoop() {
  int rc = false;

  if (PCB_initialized) {

        // all initialization done, start the eventloop
        pcb_plugin_eventloop();

        ithread_mutex_lock(&DevMutex);
        PCB_NeMoExit();
        ithread_mutex_unlock(&DevMutex);
        FWCloseFirewall();
        FWDelAllPortMappings();
        FWDelAllPinholes();
        pcb_plugin_cleanup();
        rc = true;
    }

  return rc;
}


bool PCB_initialize() {
  int rc = false;

  s_appconfig.name = "upnp-igd-eventloop";

  if (IPC_Init() &&
      NeMoInitWrapper("data") &&
      pcb_plugin_initialize() &&
      PCB_NeMoInit())
  {
    PCB_initialized = true;
    rc = true;
  }

  return rc;
}
