#include <stdlib.h>
#include <errno.h>
#include <inttypes.h>

#include <debug/sahtrace.h>

#include <pcb/common.h>
#include <pcb/utils.h>
#include <pcb/core.h>

#include <components.h>

#ifdef CONFIG_PCB_ACL_USERMNGT
#include <usermngt/usermngt.h>
#endif

static uint32_t level = 0;

static void odl_save_add_identation(FILE *out) {
    uint32_t i = 0;
    for(i = 0; i < level + 1; i++) {
        fprintf(out,"    ");
    }
}

#ifdef CONFIG_PCB_ACL_USERMNGT
static void odl_save_acl(const llist_t *acllist,FILE *out) {
    odl_save_add_identation(out);
    fprintf(out,"acl {\n");
    ACL_t *acl = NULL;
    const char *sep = " ";
    const usermngt_user_t *user = NULL;
    const usermngt_group_t *group = NULL;

    level++;
    acl_for_each(acl,acllist) {
        odl_save_add_identation(out);
        sep = " ";
        group = NULL;
        user = NULL;

        SAH_TRACEZ_INFO("odl","Saving acl for id = %d",acl_id(acl));

        if (acl_id(acl) == UINT32_MAX) {
            fprintf(out,"world");
        } else if (acl_flags(acl) & acl_group_id) {
            group = usermngt_groupFindByID(acl_id(acl));
            fprintf(out,"group %s",usermngt_groupName(group));
        } else {
            user = usermngt_userFindByID(acl_id(acl));
            fprintf(out,"user %s",usermngt_userName(user));
        }

        if (acl_flags(acl) & acl_read) {
            fprintf(out,"%sread",sep);
            sep = ",";
        }
        if (acl_flags(acl) & acl_write) {
            fprintf(out,"%swrite",sep);
            sep = ",";
        }
        if (acl_flags(acl) & acl_execute) {
            fprintf(out,"%sexecute",sep);
            sep = ",";
        }
        if ((acl_flags(acl) & (acl_read | acl_write | acl_execute)) == 0) {
            fprintf(out,"%snone",sep);
            sep = ",";
        }
        fprintf(out,";\n");
    }
    level--;

    odl_save_add_identation(out);
    fprintf(out,"}\n");
}
#else
#define odl_save_acl(acllist, out)
#endif

static void odl_save_parent_tree(object_t *object, object_t *root, FILE *out) {
    if (object != root) {
        SAH_TRACEZ_INFO("odl","Tree up: object = %s",object_name(object,path_attr_key_notation));
        odl_save_parent_tree(object_parent(object), root, out);
    }

    if (object == root)
        return;

    if (object_isTemplate(object))
        return;

    SAH_TRACEZ_INFO("odl","Saving object: %s",object_name(object,path_attr_key_notation));
    if (object_isInstance(object)) {
        odl_save_add_identation(out);
        fprintf(out,"instance of %s(%s,\"%s\") {\n",object_name(object_parent(object),path_attr_key_notation),object_name(object,path_attr_default),object_name(object,path_attr_key_notation));
    } else {
        odl_save_add_identation(out);
        fprintf(out,"object %s {\n",object_name(object,path_attr_default));
    }

    level++;
    if (object_hasAcl(object)) {
        odl_save_acl(object_getACL(object),out);
    }

    return;
}

static void odl_save_parameter_value(parameter_t *parameter, FILE *out) {
    switch(parameter_type(parameter)) {
        case parameter_type_string:
        case parameter_type_reference: {
            string_t data;
            string_initialize(&data,0);
            string_copy(&data,variant_da_string(parameter_getValue(parameter)));
            string_replaceChar(&data,"\\","\\\\");
            string_replaceChar(&data,"\"","\\\"");
            string_replaceChar(&data,"$","\\$");
            string_replaceChar(&data,"(","\\(");
            string_replaceChar(&data,")","\\)");
            fprintf(out,"\"%s\"",string_isEmpty(&data)?"":string_buffer(&data));
            string_cleanup(&data);
        }
        break;
        case parameter_type_date_time: {
            char *value = variant_char(parameter_getValue(parameter));
            fprintf(out,"\"%s\"",value);
            free(value);
        }
        break;
        case parameter_type_bool:
            fprintf(out,"%s",variant_bool(parameter_getValue(parameter))?"true":"false");
        break;
        case parameter_type_int8:
        case parameter_type_int16:
        case parameter_type_int32:
            fprintf(out,"%d",variant_int32(parameter_getValue(parameter)));
        break;
        case parameter_type_int64:
            fprintf(out,"%"PRId64,variant_int64(parameter_getValue(parameter)));
        break;
        case parameter_type_uint8:
        case parameter_type_uint16:
        case parameter_type_uint32:
            fprintf(out,"%u",variant_uint32(parameter_getValue(parameter)));
        break;
        case parameter_type_uint64:
            fprintf(out,"%"PRIu64,variant_uint64(parameter_getValue(parameter)));
        break;
    }
}

static void odl_save_parameter(parameter_t *parameter, FILE *out) {
    odl_save_add_identation(out);
    if (parameter_hasAcl(parameter)) {
        fprintf(out,"parameter %s {\n",parameter_name(parameter));

        level++;
        odl_save_acl(parameter_getACL(parameter),out);
        // write default value
        odl_save_add_identation(out);
        fprintf(out,"default ");
        odl_save_parameter_value(parameter,out);
        fprintf(out,";\n");
        level--;

        odl_save_add_identation(out);
        fprintf(out,"}\n");
    } else {
        fprintf(out,"parameter %s=",parameter_name(parameter));
        odl_save_parameter_value(parameter,out);
        fprintf(out,";\n");
    }
}

static void odl_save(object_t *object, uint32_t depth, FILE *out) {
    bool hasBody = false;

    SAH_TRACEZ_INFO("odl","Saving object: %s (%p)",object_name(object,path_attr_key_notation),object);
    if (!object_isTemplate(object)) {
        // create object
        if (object_isInstance(object)) {
            odl_save_add_identation(out);
            fprintf(out,"persistent instance of %s(%s,\"%s\")",object_name(object_parent(object),path_attr_key_notation),object_name(object,path_attr_default),object_name(object,path_attr_key_notation));
        } else {
            odl_save_add_identation(out);
            fprintf(out,"object %s",object_name(object,path_attr_default));
        }

        if (object_hasAcl(object)) {
            fprintf(out," {\n");
            hasBody = true;
            level++;
            odl_save_acl(object_getACL(object),out);
        }

        // save persistent parameters
        parameter_t *parameter = NULL;
        object_for_each_parameter(parameter, object) {
            // do not store mapped parameters
            if (parameter_destination(parameter))
                continue;

            // only store persistent parameters
            if (!(parameter_attributes(parameter) & parameter_attr_persistent))
                continue;

            if (!hasBody) {
                fprintf(out," {\n");
                level++;
                hasBody = true;
            }

            odl_save_parameter(parameter,out);
        }

        // max depth reached, stop here
        if (depth == 0) {
            goto close_object;
        }

        // save persistent childs
        object_t *child = NULL;
        object_for_each_child(child,object) {
            // only store persistent objects
            if (!(object_attributes(child) & object_attr_persistent) && !object_isTemplate(child))
                continue;

            // skip empty template objects
            if (object_isTemplate(child) && !object_hasInstances(child))
                continue;

            if (!hasBody) {
                fprintf(out," {\n");
                level++;
                hasBody = true;
            }

            odl_save(child,(depth==UINT32_MAX)?depth:depth-1,out);
        }
    } else {
        // save persistent childs
        object_t *child = NULL;
        object_for_each_instance(child,object) {
            // only store persistent objects
            if (!(object_attributes(child) & object_attr_persistent))
                continue;

            odl_save(child,(depth==UINT32_MAX)?depth:depth-1,out);
        }
    }

close_object:
    if (hasBody) {
        level--;
        odl_save_add_identation(out);
        fprintf(out,"}\n");
    } else {
        if (!object_isTemplate(object))
            fprintf(out,";\n");
    }
}

static void odl_save_close_parent_tree(FILE *out) {
    while(level) {
        odl_save_add_identation(out);
        fprintf(out,"}\n");
        level--;
    }
}

bool odl_save_datamodel(int fd, object_t *object, uint32_t depth, const char *filename) {
    FILE *out = fdopen(fd,"w");

    if (!out) {
        SAH_TRACEZ_ERROR("odl","failed to create stream for output file");
        pcb_error = errno;
        return false;
    }

    SAH_TRACEZ_INFO("odl","Saving data model to %s (object = %p)",filename,object);
    level = 0;

    object_t *root = object_root(object);
    object_t *parent = object_parent(object);

    SAH_TRACEZ_INFO("odl","Write datamodel");
    fprintf(out,"datamodel {\n");

    // build odl file up to given object
    if (parent) {
        SAH_TRACEZ_INFO("odl","Build parent tree");
        odl_save_parent_tree(parent,root,out);
    }

    if (object == root) {
        SAH_TRACEZ_INFO("odl","Write childeren of root object");
        object_t *child = NULL;
        object_for_each_child(child,object) {
            // only store persistent objects
            if (!(object_attributes(child) & object_attr_persistent))
                continue;

            odl_save(child, (depth==UINT32_MAX)?depth:depth-1, out);
        }
    } else {
        // save the requested objects
        SAH_TRACEZ_INFO("odl","Save objects");
        odl_save(object, depth, out);
    }

    // close objects
    SAH_TRACEZ_INFO("odl","Close objects");
    odl_save_close_parent_tree(out);

    // close datamodel
    SAH_TRACEZ_INFO("odl","Close datamodel");
    fprintf(out,"}\n");

    // implementation according to http://lwn.net/Articles/457667/
    /*
     * After the fwrite call returns, the data is in libc's stdio
     * buffer (still in the application's address space).  So, the
     * next thing we want to do is flush that buffer.
     */
    if (fflush(out) != 0) {
        SAH_TRACEZ_ERROR("odl","fflush failed: %d - %s",errno,error_string(errno));
    }
    /*
     * Now the data is in the kernel's page cache.  The next steps
     * flush the page cache for this file to disk.
     */
    if (fsync(fd) < 0) {
        SAH_TRACEZ_ERROR("odl","fsync failed: %d - %s",errno,error_string(errno));
    }

    if (fclose(out) < 0) {
        SAH_TRACEZ_ERROR("odl","fclose failed: %d - %s",errno,error_string(errno));
    }

    return true;
}
