# Include this makefile to compile autoconfigured opensource packages
# you need to define the following from the calling Makefile :
#
# AUTO_ARCHIVE       : the source archive to be expanded
# AUTO_SRCPATH       : the sources path once extraction is done
# AUTO_PATCHES       : list of patches to apply to the source tree
# AUTO_RELEASE       : list of files to be copied in the release directory
# AUTO_CONF          : the configuration flags to be passed to the ./configure
# AUTO_ISAUTOCONF    : flag to say if standard autoconf is used or not, default (y)
# AUTO_RECONF        : force enable/disable reconfigure
# AUTO_BUILD_INPLACE : flag to say if inplace compilation is to be used default (n)
# AUTO_CONFINSTALL   : list of host-dependent *-config files

# Will unpack again if archive name, archive path or list of patches changes
AUTO_COOKIE_UNPACK_FILE=auto_cookie_unpack
AUTO_COOKIE_UNPACK_MD5=$(shell echo $(AUTO_ARCHIVE) $(AUTO_SRCPATH) \
                                    $(AUTO_PATCHES) \
                         | md5sum - | awk '{print $$1}')

# Will reconf again if unpacked
AUTO_COOKIE_RECONF_FILE=auto_cookie_reconf
AUTO_COOKIE_RECONF_MD5=$(shell echo $(AUTO_COOKIE_UNPACK_MD5) $(AUTO_RECONF) \
                         | md5sum - | awk '{print $$1}')

# Will configure again if unpacked or if conf changed
AUTO_COOKIE_CONF_FILE=auto_cookie_conf
AUTO_COOKIE_CONF_MD5=$(shell echo $(AUTO_COOKIE_RECONF_MD5) $(AUTO_CONF) \
                                  $(CONFIG_CROSS_COMPILE_CFLAGS) \
                                  $(CONFIG_CROSS_COMPILE_LDFLAGS) \
                       | md5sum - | awk '{print $$1}')

ifeq ($(AUTO_BUILD_INPLACE),y)
AUTO_BUILDDIR=$(AUTO_SRCPATH)
else
AUTO_BUILDDIR=build
endif

# By default patches are applied from the source path
AUTO_PATCHFROM?=$(AUTO_SRCPATH)
# By default archives are unpacked from the component path
AUTO_UNPACKFROM?=

AUTO_RECONF?=y

AUTO_CONFINSTALL?=

include $(dir $(lastword $(MAKEFILE_LIST)))/tmpdir.mk

auto_all: auto_make

auto_check:
	@# Check mandatory vars
	@[ "$(AUTO_ARCHIVE)" ]
	@[ "$(AUTO_SRCPATH)" ]

auto_unpack: auto_check
	@# Cookie does not exist, or is different
	@# Unpack
	@# Apply patches
	@# Reconfigure if patches applied
	@set -e ; \
	if    [ ! -f "$(AUTO_COOKIE_UNPACK_FILE)" ] \
	   || [    "$$(cat $(AUTO_COOKIE_UNPACK_FILE))" \
	        != "$(AUTO_COOKIE_UNPACK_MD5)" ] ; \
	then \
		[ -d "$(AUTO_SRCPATH)" ] && rm -rf "$(AUTO_SRCPATH)" ; \
		echo -e "\nUnpacking $(AUTO_ARCHIVE) in $(AUTO_SRCPATH)..." ; \
		auto_comproot="$$(pwd)" ; \
		if [ "$(AUTO_UNPACKFROM)" ] ; \
		then \
			mkdir -p "$(AUTO_UNPACKFROM)" ; \
			pushd "$(AUTO_UNPACKFROM)" 1>/dev/null ; \
		fi ; \
		case "$(AUTO_ARCHIVE)" in \
			*.zip) unzip   "$$auto_comproot/$(AUTO_ARCHIVE)" ;; \
			*)     tar -xf "$$auto_comproot/$(AUTO_ARCHIVE)" ;; \
		esac ; \
		if [ "$(AUTO_UNPACKFROM)" ] ; \
		then \
			popd 1>/dev/null ; \
		fi ; \
		if [ "$(AUTO_PATCHES)" ] ; \
		then \
			echo -e "\nPatching..." ; \
			pushd "$(AUTO_PATCHFROM)" 1>/dev/null ; \
			for file in $(AUTO_PATCHES) ; \
			do \
				echo " - $$file" ; \
				patch -p1 -g0 -i "$$auto_comproot/$$file" ; \
			done ; \
			popd 1>/dev/null ; \
		fi ; \
		echo -e "Done" ; \
		echo -n "$(AUTO_COOKIE_UNPACK_MD5)" \
		        > "$(AUTO_COOKIE_UNPACK_FILE)" ; \
	fi

auto_reconf: auto_unpack
	@set -e ; \
	if    [ ! -f "$(AUTO_COOKIE_RECONF_FILE)" ] \
	   || [    "$$(cat $(AUTO_COOKIE_RECONF_FILE))" \
	        != "$(AUTO_COOKIE_RECONF_MD5)" ] ; \
	then \
		if [ "$(AUTO_RECONF)" = "y" ] ; \
		then \
			pushd "$(AUTO_SRCPATH)" 1>/dev/null ; \
			echo -e "\nRunning autoreconf..." ; \
			autoconf_file=configure.ac; \
			[ ! -r "$$autoconf_file" ] && autoconf_file=configure.in; \
			if [ -r "$$autoconf_file" ] ; \
			then \
				AUX_DIR=$$(sed -rn -e 's/^AC_CONFIG_AUX_DIR\(\[?([^]]+)\]?\)/\1/ p' $$autoconf_file) ; \
				MACRO_DIR=$$(sed -rn -e 's/^AC_CONFIG_MACRO_DIR\(\[?([^]]+)\]?\)/\1/ p' $$autoconf_file) ; \
				# fallback to m4 directory, a better way could be to parse ACLOCAL_AMFLAGS from Makefile.am \
				[ -z "$$MACRO_DIR" ] && [ -d m4 ] && MACRO_DIR=m4 ; \
				[ -n "$$AUX_DIR" ] && rm -f $$AUX_DIR/ltmain.sh ; \
				[ -n "$$MACRO_DIR" ] && rm -f $$MACRO_DIR/libtool.m4 $$MACRO_DIR/ltversion.m4 $$MACRO_DIR/ltoptions.m4 ; \
			fi ; \
			autoreconf --force --install ; \
			echo -e "Done" ; \
			popd 1>/dev/null ; \
		fi ; \
		echo -n "$(AUTO_COOKIE_RECONF_MD5)" \
		        > "$(AUTO_COOKIE_RECONF_FILE)" ; \
	fi

AUTO_ISCROSSCOMPILE?=y

auto_conf: auto_reconf
	if    [ ! -d "./$(AUTO_BUILDDIR)" ] ; \
	then \
		mkdir ./$(AUTO_BUILDDIR) ; \
	fi
	
	@set -e ; \
	if    [ ! -f "$(AUTO_COOKIE_CONF_FILE)" ] \
	   || [    "$$(cat $(AUTO_COOKIE_CONF_FILE))" \
	        != "$(AUTO_COOKIE_CONF_MD5)" ] ; \
	then \
		pushd $(AUTO_BUILDDIR) 1>/dev/null ; \
		CONF="" ; \
		if [ "$(AUTO_ISAUTOCONF)" != "n" ] ; \
		then \
			if    [ "$(AUTO_ISCROSSCOMPILE)" = "y" ] \
			   && [ -e "../$(AUTO_SRCPATH)/config.guess" ] ; \
			then \
				BUILD="$$($(SHELL) ../$(AUTO_SRCPATH)/config.guess)" ; \
				CONF="$$CONF --build=$$BUILD" ; \
			fi ; \
			CONF="$$CONF --host=$$($(CC) -dumpmachine)" ; \
			CONF="$$CONF --prefix=/usr" ; \
			CONF="$$CONF --sysconfdir=/etc --localstatedir=/var" ; \
			CONF="$$CONF --disable-docs" ; \
		fi ; \
		CONF="$$CONF $(AUTO_CONF)" ; \
		echo ; \
		echo "Configure options: $$CONF" ; \
		echo ; \
		export CC="$(CC)" ; \
		export CXX="$(CXX)"	; \
		export CPPFLAGS="$(CPPFLAGS)" ; \
		export CFLAGS="$(CFLAGS)" ; \
		export LDFLAGS="$(LDFLAGS)" ; \
		if [ "$(AUTO_ISCROSSCOMPILE)" = "y" ] ; \
		then \
			export CPPFLAGS="$$CPPFLAGS -isystem $(STAGINGDIR)/include" ; \
			export CPPFLAGS="$$CPPFLAGS -isystem $(STAGINGDIR)/usr/include" ; \
			export LDFLAGS="$$LDFLAGS -L$(STAGINGDIR)/lib" ; \
			export LDFLAGS="$$LDFLAGS -L$(STAGINGDIR)/usr/lib" ; \
			export LDFLAGS="$$LDFLAGS -Wl,-rpath-link,$(STAGINGDIR)/lib" ; \
			export LDFLAGS="$$LDFLAGS -Wl,-rpath-link,$(STAGINGDIR)/usr/lib" ; \
		fi ; \
		export CPPFLAGS="$$CPPFLAGS "$(CONFIG_CROSS_COMPILE_CPPFLAGS) ; \
		export CFLAGS="$$CFLAGS "$(CONFIG_CROSS_COMPILE_CFLAGS) ; \
		export CXXFLAGS="$$CXXFLAGS "$(CONFIG_CROSS_COMPILE_CXXFLAGS) ; \
		export LDFLAGS="$$LDFLAGS "$(CONFIG_CROSS_COMPILE_LDFLAGS) ; \
		if [ "$(filter 1 2,$(V))" ] ; then \
		echo "Environment variables: " ; \
		echo "CC           $$CC" ; \
		echo "CXX          $$CXX" ; \
		echo "CPPFLAGS     $$CPPFLAGS" ; \
		echo "CFLAGS       $$CFLAGS" ; \
		echo "CXXFLAGS     $$CXXFLAGS" ; \
		echo "LDFLAGS      $$LDFLAGS" ; \
		echo ; \
		fi ; \
		"../$(AUTO_SRCPATH)/configure" $$CONF ; \
		popd ; \
		echo -n "$(AUTO_COOKIE_CONF_MD5)" \
		        > "$(AUTO_COOKIE_CONF_FILE)" ; \
	fi

auto_make: auto_conf
	cd ./$(AUTO_BUILDDIR) && $(MAKE) $(AUTO_MAKEOPTS)

#IDESTPREFIX:=$(shell echo $$PWD)/$(AUTO_BUILDDIR)/dest
IDESTPREFIX:=$(realpath .)/$(AUTO_BUILDDIR)/dest
TMPDIR_SRC:=$(IDESTPREFIX)/
TMPDIR_TEMPFILE:=$(IDESTPREFIX).list
TMPDIR_DST:=$(STAGINGDIR)

auto_subinstall:
	rm -rf "$(TMPDIR_SRC)"
	mkdir -p "$(TMPDIR_SRC)"
	# Install in dest
	$(MAKE) -C ./$(AUTO_BUILDDIR) install DESTDIR="$(TMPDIR_SRC)"
auto_cleaninstall: auto_subinstall
	@# Clean dest
	@# ..."*.la"
	@find "$(TMPDIR_SRC)" -name "*.la" -exec rm {} \;
	@# ...man and doc
	@rm -rf "$(TMPDIR_SRC)/share/man" \
	        "$(TMPDIR_SRC)/usr/share/man" \
	        "$(TMPDIR_SRC)/share/doc" \
	        "$(TMPDIR_SRC)/usr/share/doc" \
	        "$(TMPDIR_SRC)/share/gtk-doc" \
	        "$(TMPDIR_SRC)/usr/share/gtk-doc"
tmpdir_install: auto_cleaninstall
auto_wrappers: tmpdir_install
	# Create *-conf wrappers
	set -e ; \
	if [ "$(AUTO_CONFINSTALL)" ] ; \
	then \
		install -d $(STAGINGDIR)/buildutils/pkgconfig ; \
		for each in $(AUTO_CONFINSTALL) ; \
		do \
			OUTCONF="$(STAGINGDIR)/buildutils/pkgconfig/$$(basename $$each)" ; \
			sed -e 's|^prefix=.*$$|prefix=$(STAGINGDIR)/usr|' "$(STAGINGDIR)/$$each" > "$$OUTCONF" ; \
			chmod +x "$$OUTCONF" ; \
		done ; \
	fi
auto_install: tmpdir_install auto_wrappers

auto_release:
	set -e ; \
	for each in $(AUTO_RELEASE) ; \
	do \
		install -d "$(ROOT_FS_DIR)/$$(dirname $$each)" ; \
		cp -af "$(STAGINGDIR)"/$$each "$(ROOT_FS_DIR)/$$(dirname $$each)" ; \
	done

auto_clean:
	rm -f auto_*
	rm -rf $(AUTO_SRCPATH)
	rm -rf $(AUTO_BUILDDIR)
	rm -rf $(TMPDIR_SRC)
	rm -f $(TMPDIR_TEMPFILE)

